/*
 *      OSU Research Computing, architects and contractors      Trillium
 *
 *      VAR
 *
 *      $Log:	wavcub.c,v $
 * Revision 2.2  92/04/24  15:40:19  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  13:58:17  gdburns
 * initial version
 * 
 *
 *      Function:       Solve the one-dimensional wave equation problem.
 *
 *      Version:        C2 : CUBIX program in C
 *
 *      Reference:      Solving Problems on Concurrent Processors - Vol I
 *                      Fox, Johnson et al
 */

#include <stdio.h>
#include <typical.h>	/* for ERROR, SUCCESS	*/
#include <net.h>	/* for message header	*/
#include <t_types.h>	/* for ltotd()		*/

extern	int	errno;	/* global error variable */
/*
 * Global constants.
 */
#define MAXSIZE		500	/* Maximum number of nodal points	*/
#define INITIA		1	/* Task type 				*/
#define UPDATE		2	/* Task type				*/
#define QUIT		3	/* Task type				*/
#define	KONST		1	/* Constant used in computation		*/
#define	PI     3.14159265	/* Value of Pi				*/

extern 	double	sin();

/*
 * Local functions.
 */
void	gparam();
void	getask();
void	initze();
void	dostep();
void	outres();
void	cshift();

/*
 * Global declarations.
 */
int	npts; 		/* Number of points or size of domain	*/
int	nstep;		/* Number of time steps or iterations	*/
double	delt;		/* Time step size			*/
int	instd = 0;	/* Stdin descriptor			*/
int	outstd = 1;	/* Stdout descriptor			*/

int	nproc;		/* Total number of processors or nodes	*/
int	myproc;		/* My processor identifier		*/
int	nmin;		/* Minimum size of subdomain		*/
int	nleft;		/* nleft = modulo (npts, nproc)		*/
int	mypts;		/* number of points in my domain	*/
int	last;		/* last task selected			*/
int	task;		/* current task selected		*/

double  values[MAXSIZE+2];
double  oldval[MAXSIZE+2];
double  newval[MAXSIZE+2];

struct 	nmsg	header;	/* network message structure used in	*/
			/* message communication between 	*/
			/* host and node processes		*/

char	*malloc();

/*-----------------------------------------------------------------------*/
main()

{

/*
 * Attach to Trollius kernel.
 */
	if (kinit(0)) {
		terror("Err in kinit\n");
		kexit(1);
	}

/*
 * Set I/O to singular mode.
 */
	if (singl(instd) != SUCCESS) {
		printf("Err in singl : %d\n", errno);
		kexit(1);
	}

	if (singl(outstd) != SUCCESS) {
		printf("Err in singl : %d\n", errno);
		kexit(1);
	}

	printf("\nThis is a 1-D wave equation solver - \n\n");

/*
 * Initialize variables.
 */
	gparam();

	last = 0;
	task = 0;
/*
 * Get command option until quit.
 */
	while (task != QUIT) {
		getask();
/*		printf("task = %d\n", task); */
		last = task;
		if (task == INITIA) {
/*
 * Get parameters.
 */
			initze();
		}
		else if (task == UPDATE) {
/*
 * Update points for parameters specified.
 */
			dostep();
/*
 * Output results to host node.
 */
			outres();
		}
	}
/*
 * Detach from Trollius kernel.
 */
	kexit(0);
}

/*-----------------------------------------------------------------------
 * gparam()	Determine total number of processors in the topology.
 */

void
gparam()
{
        nproc = getnitb(); 
	myproc = getnodeid();

	if ((myproc < 0) || (myproc > nproc-1)) {
		printf("Err: Nodeid's must be 0 thru' (nproc-1)\n\n");
		kexit(1);
	}
}

/*-----------------------------------------------------------------------
 * getask()	Obtain task option.
 */

void
getask()

{
	int	reply;
	int	done = 0;
	int	out = 0;
	char	input[20];

	while (done == 0) {
		printf("\nSelect one of the following :\n");
		printf(" 1 ... to initialize a new problem\n");
		printf(" 2 ... to compute points and print results\n");
		printf(" 3 ... to terminate program\n\n");

		if ((out = tread(instd, input, 5)) < 0) {
			printf("Err in tread(reply) : %d\n", errno);
			kexit(1);
		}

		input[5] = '\0';
		sscanf(input, "%d", &reply);
/*		printf("reply = %d\n", reply);*/
		done = 1;
		switch(reply) {
		    case 1:
			task = INITIA;
			break;
		    case 2:
			task = UPDATE;
			break;
		    case 3:
			task = QUIT;
			break;
		    default:
			done = 0;
			printf("\nErr: Invalid input- please try again.\n");
			break;
		}

		if ((last == 0) && (task == UPDATE)) {
			printf("\nErr: Problem not initialized \n");
			done = 0;
		}
	}
}

/*-----------------------------------------------------------------------
 * initze()	Get parameters npts, delt.
 */

void
initze()

{
	int	i, j, k;
	double	fac, dx;
	int	out = 0;
	char	input[20];

/*
 * Read input parameters.
 */
	printf("\nEnter npts (integer) (number of points) :\n");

	if ((out = tread(instd, input, 10)) < 0) {
		printf("Err in tread(npts) : %d\n", errno);
		kexit(1);
	}

	sscanf(input, "%d", &npts);

	if (npts == 0) {
		printf("\nErr: Zero points specified \n");
		kexit(1);
	}

/*
 * Creating equal (almost) size subdomains.
 */
	nmin = npts / nproc;
	nleft = npts % nproc;

	if (nmin+1 > MAXSIZE) {
		printf("\nErr: Too many points specified.\n");
		kexit(1);
	}

	printf("\nEnter delt (float) (time step size) :\n");

	if ((out = tread(instd, input, 10)) < 0) {
		printf("Err in tread(delt) : %d\n", errno);
		kexit(1);
	}

	sscanf(input, "%lf", &delt);

	fac = 2.0 * PI;
	k = 0;

	for (i=0; i<=nproc-1; i++) {
		mypts = nmin;
                if (i < nleft) mypts = nmin + 1;

                if (myproc == i) {
/*                        printf("initial data in %d\n", myproc); */
                        for (j=1; j<=mypts; j++) {
                                dx = (double) k / (double) (npts+1);
                                values[j] = sin(fac*dx);
                                k = k+1;
/*                                printf("%d : %lf\n", myproc, values[j]); */
                        }
                        break;
                }
                else
                        k = k + mypts;
	}

        for (i=1; i<=mypts; i++) {
                oldval[i] = values[i];
        }
}

/*-----------------------------------------------------------------------
 * dostep()	Get parameter nstep.
 */

void
dostep()

{
	int	i, j;
	double	pfac;
	int	out = 0;
	char	input[20];

	printf("\nEnter (nstep) (integer) (number of time steps) :\n");

	if ((out = tread(instd, input, 10)) < 0) {
		printf("Err in tread(nstep) : %d\n", errno);
		kexit(1);
	}

	sscanf(input, "%d", &nstep);

	printf("npts  = %d \ndelt  = %lf \nnstep = %d\n\n",
        	npts, delt, nstep);

        pfac = (KONST * delt / 1.0);
        pfac = pfac * pfac;

        for (i=1; i<=nstep; i++) {

            cshift (&values[0], (myproc-1), &values[mypts], (myproc+1));

            cshift (&values[mypts+1], (myproc+1), &values[1], (myproc-1));

            for (j=1; j<=mypts; j++) {

                if ((myproc == 0) && (j == 1)) {
                    newval[j] = 0.0;
                }
                else if ((myproc == nproc-1) && (j == mypts)) {
                    newval[j] = 0.0;
                }
                else {
                    newval[j] = 2.0*values[j] - oldval[j]
                        + ( values[j+1] - 2.0*values[j] + values[j-1] )
                        * pfac;
                }
            }

            for (j=1; j<=mypts; j++) {
                oldval[j] = values[j];
                values[j] = newval[j];
            }
        }
}

/*--------------------------------------------------------------------
 * cshift()	Exchange boundary nodal values with neighbour nodes.
 */

void
cshift(inval, inindex, outval, outindex)
double	*inval;
int	inindex;
double	*outval;
int	outindex;

{

        header.nh_type = 0;
        header.nh_flags = 0;
        header.nh_length = sizeof(double);

	if ((myproc % 2) == 0) {
/* 
 * Even node index.
 * Send to outindex only if outindex in (0,nproc-1).
 * Receive from inindex only if inindex in (0,nproc-1).
 */

	    if ((outindex >= 0) && (outindex <= nproc-1)) {
		header.nh_node = outindex;
		header.nh_event = myproc * 65536 + outindex;
		header.nh_msg = (char *) outval;
/*	        printf("node %d --> %d on event %d : %lf\n", 
			myproc, outindex, header.nh_event, *outval);*/

		if (nsend(&header)) {
		    printf("Err in nsend(cshift) on node %d\n", myproc);
		    kexit(1);
		}
	    }

	    if ((inindex >= 0) && (inindex <= nproc-1)) {
		*inval = 0;
		header.nh_event = inindex * 65536 + myproc;
		header.nh_msg = (char *) inval;
/*		printf("node %d <-- %d on event %d\n", 
			myproc, inindex, header.nh_event); */
                if (nrecv(&header)) { 
                    printf("Err in nrecv(cshift) on node %d\n", myproc);    
                    kexit(1); 
                } 
	    }
	}

	else {

/*
 * Odd node index.
 * Receive from inindex only if inindex in (1,nproc).
 * Send to outindex only if outindex in (1,nproc).
 */
            if ((inindex >= 0) && (inindex <= nproc-1)) {
		*inval = 0;
		header.nh_event = inindex * 65536 + myproc;
		header.nh_msg = (char *) inval;
/*		printf("node %d <-- %d on event %d\n",
                        myproc, inindex, header.nh_event); */
                if (nrecv(&header)) {
                    printf("Err in nrecv(cshift) on node %d\n", myproc);
                    kexit(1);
                }
            }

            if ((outindex >= 0) && (outindex <= nproc-1)) {
                header.nh_node = outindex;
		header.nh_event = myproc * 65536 + outindex;
                header.nh_msg = (char *) outval;
/*              printf("node %d --> %d on event %d : %lf\n",
                        myproc, outindex, header.nh_event, *outval); */
                if (nsend(&header)) {
                    printf("Err in nsend(cshift) on node %d\n", myproc);
                    kexit(1);
                }
            }
        }
}

/*-----------------------------------------------------------------------
 * outres()	Output results.
 */

void
outres()

{

	char	*outbuf;
	char	*ptr;
	int	j, nout;

	printf("\nData points on the wave : \n\n");
/*
 * Caution: printf() instead of sprintf() in the following loop will
 * not print data of processors 0,1,2,...n-1 in the order 
 * 0,0,0,1,1,1,2,2,3,3,...(n-1),(n-1). It will print data of 
 * processors in the order 0,1,2,...(n-1),0,1,2,...(n-1),0,1..
 * Hence the use of malloc and sprintf.
 */

	if ((outbuf = malloc ((mypts * 22) + 2)) == (char *) 0) {
		printf("Err in malloc : %d\n", errno);
		kexit(1);
	}	

	ptr = outbuf;
	for (j=1; j<=mypts; j++) {
/*     		printf(outbuf, "% lf\n", values[j]); */
        	sprintf(ptr, "%20.5f \n", values[j]);
		ptr = ptr + 22;
	}
	sprintf(ptr, " \n");

/*
 * Change to multi() mode.
 */
	if (multi(outstd) != SUCCESS) {
		printf("Err in multi : %d\n", errno);
		kexit(1);
	}

/*
 * With CUBIX mode, it is ensured that the data points will be output
 * in the order of the processor identifiers.
 */
/*
 * Caution: A `printf("%s\n",outbuf)' instead of rfwrite will not
 * work. The line feeds following each data value cause only the 
 * first data values of each processor to be printed (in the order 
 * of processor numbers.
 */
	if ((nout = twrite(outstd, outbuf, mypts*22+2)) < 0) {
		printf("Err in twrite : %d\n", errno);
		kexit(1);
	}

	if (singl(outstd) != SUCCESS) {
		printf("Err in singl : %d\n", errno);
		kexit(1);
	}
}

/*=======================================================================*/

