/*
 *      OSU Research Computing, architects and contractors      Trillium
 *
 *      VAR
 *
 *      $Log:	wavfcp.c,v $
 * Revision 2.2  92/04/24  15:40:21  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  13:59:07  gdburns
 * initial version
 * 
 *
 *      Function:       Solve the one-dimensional wave equation problem.
 *
 *      Version:        C1 : Host-node programs in C
 *
 *      Reference:      Solving Problems on Concurrent Processors - Vol I
 *                      Fox, Johnson et al
 *
 *      Host Program
 */

#include <stdio.h>
#include <events.h>	/* for NOTEVENT		*/
#include <net.h>	/* for message header	*/
#include <t_types.h>	/* for ltotd()		*/

/*
 * Global constants.
 */
#define MAXSIZE		500	/* Maximum number of nodal points	*/
#define INITIA		1	/* Task type 				*/
#define UPDATE		2	/* Task type				*/
#define QUIT		3	/* Task type				*/
#define EVTASK		11	/* Event number for task		*/
#define EVNPTS		12	/* Event number for npts		*/
#define EVDELT		13	/* Event number for delt		*/
#define EVNSTEP		15	/* Event number for nstep		*/
#define EVDONE		16	/* Event number for done		*/
#define EVRES		17	/* Event number for result		*/
#define EVOUT	       100	/* Event number for out results		*/

/*
 * Local functions.
 */
void	setvar();
void	getask();
void	initcp();
void	dostep();
void	outres();

/*
 * Global declarations.
 */
int	npts; 		/* Number of points or size of domain	*/
int	nstep;		/* Number of time steps or iterations	*/
double	delt;		/* Time step size			*/

int	nproc;		/* Total number of processors or nodes	*/
int	nmin;		/* Minimum size of subdomain		*/
int	nleft;		/* nleft = modulo (npts, nproc)		*/
int	last;		/* last task selected			*/
int	task;		/* current task selected		*/

struct 	nmsg	header;	/* network message structure used in	*/
			/* message communication between 	*/
			/* host and node processes		*/

/*-----------------------------------------------------------------------*/
main()

{

/*
 * Attach to Trollius kernel.
 */
	if (kinit(0)) {
		terror("Err in kinit\n");
		kexit(1);
	}
/*
 * Initialize variables.
 */
	setvar();

	last = 0;
	task = 0;
/*
 * Get command option until quit.
 */
	while (task != QUIT) {
		getask();
		last = task;
		if (task == INITIA) {
/*
 * Get parameters.
 */
			initcp();
		}
		else if (task == UPDATE) {
/*
 * Update points for parameters specified.
 */
			dostep();
/*
 * Output results to host node.
 */
			outres();
		}
	}
/*
 * Detach from Trollius kernel.
 */
	kexit(0);
}

/*-----------------------------------------------------------------------
 * setvar()	Determine total number of processors in the topology.
 */

void
setvar()
{

	int	i, event;
/*
 * Note that there could be more than one OTB node in the system.
 */
        nproc = getnitb(); 
 
	printf("\nThis is a 1-D wave equation solver with %d processor(s).\n",
			nproc);

/*
 * Check if nodeids are 0,1,...(nproc-1), this is an assumption here.
 */
        event = 0;
	for (i=0; i<=nproc-1; i++) {
/*
 * If node i is a valid node, routing from this host node to node i
 * as destination node must be possible. This is one way of validating
 * nodeids.
 */
	    if (getroute(i, event) == NOTEVENT) {
            	printf("*** Node %d", i," does not exist - verify bnail file\n\n");
		kexit(1);
	    }
	} 
}

/*-----------------------------------------------------------------------
 * getask()	Obtain task option, send task selected to all nodes.
 */

void
getask()

{
	int	reply;
	int	done = 0;
	int	i;

	while (done == 0) {
		printf("\nSelect one of the following :\n");
		printf(" 1 ... to initialize a new problem\n");
		printf(" 2 ... to compute points and print results\n");
		printf(" 3 ... to terminate program\n\n");

		scanf("%d", &reply);

		done = 1;
		switch(reply) {
		    case 1:
			task = INITIA;
			break;
		    case 2:
			task = UPDATE;
			break;
		    case 3:
			task = QUIT;
			break;
		    default:
			done = 0;
			printf("\n*** Invalid input- please try again ***\n\n");
			break;
		}

		if ((last == 0) && (task == UPDATE)) {
			printf("\n*** Problem not initialized ***\n");
			done = 0;
		}
	}

/*
 * Send task option selected to all nodes.
 */
	header.nh_event = EVTASK;
	header.nh_type = 0;
	header.nh_length = sizeof(int);
	header.nh_flags = DINTMSG;
	header.nh_msg = (char *) &task;

	for (i=0; i<=nproc-1; i++) {
	    header.nh_node = i;

/*	    printf("sending task = %d to node %d\n", task, i);*/
	    if (nsend(&header)) {
		terror("Err in nsend to node %d (task)\n", i);
		kexit(1);
	    }
	}
}

/*-----------------------------------------------------------------------
 * initcp()	Get parameters npts, delt.
 *		Send all these to all nodes.
 */

void
initcp()

{
	int	i;

/*
 * Read input parameters.
 */
	printf("\nEnter npts (integer) (number of points) :\n");
	scanf("%d", &npts);

	if (npts == 0) {
		printf("\n*** Zero points specified ***\n");
		kexit(1);
	}

/*
 * Creating equal (almost) size subdomains.
 */
	nmin = npts / nproc;
	nleft = npts % nproc;

	if (nmin+1 > MAXSIZE) {
		printf("\n*** Too many points specified ***\n");
		kexit(1);
	}

	printf("\nEnter delt (float) (time step size) :\n");
	scanf("%lf", &delt);

/*
 * Send parameters to all nodes.
 */
        header.nh_event = EVNPTS;
        header.nh_type = 0;
        header.nh_length = sizeof(int);
        header.nh_flags = DINTMSG;
        header.nh_msg = (char *) &npts;

	for (i=0; i<=nproc-1; i++) { 
            header.nh_node = i;
/*          printf("sending npts = %d to node %d\n", npts, i);*/
	    if (nsend(&header)) {
		terror("Err in nsend(npts)\n");
		kexit(1);
	    }
	}

        header.nh_event = EVDELT;
        header.nh_type = 0;
        header.nh_length = sizeof(double);
        header.nh_flags = DDBLMSG;
        header.nh_msg = (char *) &delt;
 
	for (i=0; i<=nproc-1; i++) { 
            header.nh_node = i;
/*          printf("sending delt = %lf to node %d\n", delt, i);*/
            if (nsend(&header)) {
		terror("Err in nsend(delt)\n");
		kexit(1);
	    }
	}
}

/*-----------------------------------------------------------------------
 * dostep()	Get parameter nstep, send this to all nodes.
 */

void
dostep()

{
	int	i;

	printf("\nEnter (nstep) (integer) (number of time steps) :\n");
	scanf("%d", &nstep);
/*
 * Send nstep to nodes.
 */
        header.nh_event = EVNSTEP;
        header.nh_type = 0;
        header.nh_length = sizeof(int);
        header.nh_flags = DINTMSG;
        header.nh_msg = (char *) &nstep;

	for (i=0; i<=nproc-1; i++) {
            header.nh_node = i;
/*          printf("sending nstep = %d to node %d\n", nstep, i);*/
            if (nsend(&header)) {
		terror("Err in nsend(nstep)\n");
		kexit(1);
	    }
	}

	printf("npts  = %d \ndelt  = %lf \nnstep = %d\n\n",
        	npts, delt, nstep);

/*
 * Recieve DONE message from node process.
 */

        header.nh_event = EVDONE; 
        header.nh_type = 0;
        header.nh_length = 0;
        header.nh_flags = 0;

	for (i=0; i<=nproc-1; i++) { 
/*          printf("receive done message from node %d\n", i); */
            if (nrecv(&header)) {
		terror("Err in nrecv(done)\n"); 
		kexit(1);
	    }
	}
}

/*-----------------------------------------------------------------------
 * outres()	Get results from nodes in order and print them.
 */

void
outres()

{

	int		i, j;
	int		mypts;
	double		values[MAXSIZE+1];

/*
 * Send request message.
 * To have this application workable on both the multi-node Trollius 
 * and the Stand-alone version, the event (see below) is assigned 
 * the value (EVOUT+i) because :
 * Stand-alone Trollius does not distinguish processes waiting on 
 * the same event. In order that the Host node receive and print 
 * the final data from the pseudo-compute nodes in ascending order, 
 * we need to specify distinct event numbers for each process.
 */
        header.nh_type = 0;
        header.nh_length = 0;

	for (i=0; i<=nproc-1; i++) {
            header.nh_node = i;
            header.nh_event = EVOUT + i;
/*          printf("sending out message to node %d\n", i);*/
            if (nsend(&header)) {
		terror("Err in nsend(out)\n");
		kexit(1);
	    }
	}

/*	
 * Receive data.
 */

        header.nh_event = EVRES;
        header.nh_type = 0;
        header.nh_flags = 0;
	header.nh_msg = (char *) &values[1]; 

	printf("\nData points on the wave :\n\n");

	for (i=0; i<=nproc-1; i++) { 
            if (i < nleft) 
	    	mypts = (nmin+1);
	    else
	    	mypts = nmin;
	    header.nh_length = mypts * sizeof (double);  
/*          printf("receive data points from node %d\n", i); */
            if (nrecv(&header)) {
		terror("Err in nrecv(data)\n"); 
		kexit(1);
	    }

/*
 * Output data points.
 */

	    for (j=1; j<=mypts; j++) {
		values[j] = ttold(values[j]);
        	printf("% lf\n",values[j]);
	    }
	    printf("\n");
	}
}

/*=======================================================================*/

