/*
 *      OSU Research Computing, architects and contractors      Trillium
 *
 *      VAR
 *
 *      $Log:	wavnod.c,v $
 * Revision 2.2  92/04/24  15:40:22  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  13:59:16  gdburns
 * initial version
 * 
 *
 *      Function:       Solve the one-dimensional wave equation problem.
 *
 *      Version:        C1 : Host-node programs in C
 *
 *      Reference:      Solving Problems on Concurrent Processors - Vol I
 *                      Fox, Johnson et al
 *
 *      Node Program
 */

#include <math.h>	/* for sin function	*/
#include <net.h>	/* for network header	*/
#include <t_types.h>	/* for ttold()		*/

/*
 * Global constants.
 */
#define MAXSIZE        500     	/* Maximum number of nodal points       */
#define INITIA           1      /* Task type                            */
#define UPDATE           2      /* Task type                            */
#define QUIT             3      /* Task type                            */
#define EVTASK          11      /* Event number for task                */
#define EVNPTS          12      /* Event number for npts                */
#define EVDELT          13      /* Event number for delt                */
#define EVNSTEP         15      /* Event number for nstep               */
#define EVDONE          16      /* Event number for done                */
#define EVRES           17      /* Event number for result              */
#define EVOUT          100      /* Event number for out results         */
				/* Do not use EVOUT thru (EVOUT+nproc) 	*/
#define KONST		 1	/* Constant used in computation 	*/
#define PI	3.14159265	/* Value of Pi				*/

extern	double	sin();

/*
 * Local functions.
 */

void	gparam();
void	getask();
void	initze();
void	cshift();
void	dostep();
void	outres();

/*
 * Global declarations.
 */
int     npts;
int     nstep;
double  delt;

int	host;
int     nproc;
int     myproc;
int	task; 
int     nmin;
int     nleft;
int	mypts;
 
double	values[MAXSIZE+2]; 
double	oldval[MAXSIZE+2];
double	newval[MAXSIZE+2];

struct  nmsg    header;

/*-----------------------------------------------------------------------*/
main()

{
	if (kinit(0)) {
                terror("Err in kinit\n");
                kexit(1);
        }

	gparam();

	task = 0;
	while (task != QUIT) {
	    getask();
	    switch (task) {
		case INITIA:
		    initze();
		    break;
		case UPDATE:
		    dostep();
		    outres();
		    break;
		case QUIT:
		    break;
	    }
	}

	kexit(0);
}

/*--------------------------------------------------------------------
 * gparam()	get my processor (node) number.
 */

void
gparam()

{
	host = getorigin();
        nproc = getnitb();
        myproc = getnodeid();
        if ((myproc < 0) || (myproc > nproc-1)) {
	    printf("*** Node identifiers must be 1 thru n ***\n");
	    printf("    Terminate and rerun node program\n\n");
        }
}

/*--------------------------------------------------------------------
 * getask()	receive task from host
 */

void
getask()

{
        header.nh_event = EVTASK;
        header.nh_type = 0;
        header.nh_flags = 0;
        header.nh_length = sizeof(int);
	header.nh_msg = (char *) &task;

/*	printf("node %d waiting for task\n", myproc);*/
	if (nrecv(&header)) {
	    printf("Node %d: Err in nrecv(task)\n", myproc);
	    kexit(1);
	}
}

/*--------------------------------------------------------------------
 * initze()	receive npts, delt from host
 */

void
initze()
{

	int	i, j, k;
	double fac, dx;

        header.nh_event = EVNPTS;
        header.nh_type = 0;
        header.nh_flags = 0;
        header.nh_length = sizeof(int);
        header.nh_msg = (char *) &npts;
 
/*	printf("node %d waiting for npts\n", myproc);*/
        if (nrecv(&header)) {
            printf("Node %d: Err in nrecv(npts)\n", myproc);
            kexit(1);
        }

/*	printf("npts = %d\n", npts); */

	nmin = npts / nproc;
	nleft = npts % nproc;

	fac = 2.0 * PI;
	k = 0;

	for (i=0; i<=nproc-1; i++) {
		mypts = nmin;
		if (i < nleft) mypts = nmin + 1;

		if (myproc == i) {
/*			printf("initial data in %d\n", myproc); */
			for (j=1; j<=mypts; j++) {
				dx = (double) k / (double) (npts+1);
				values[j] = sin(fac*dx);
				k = k+1;
/*				printf("%d : %lf\n", myproc, values[j]); */
			}
			break;
		}
		else
			k = k + mypts;
	}

        header.nh_event = EVDELT;
        header.nh_type = 0;
        header.nh_flags = 0;
        header.nh_length = sizeof(double);
        header.nh_msg = (char *) &delt;
 
/*      printf("node %d waiting for delt\n", myproc);*/
        if (nrecv(&header)) {
            printf("Node %d: Err in nrecv(delt)\n", myproc);
            kexit(1);
        }

/*	printf("delt = %lf\n", delt); */

	for (i=1; i<=mypts; i++) {
 		oldval[i] = values[i];
	}
}

/*--------------------------------------------------------------------
 * cshift()	Exchange boundary nodal values with neighbour nodes.
 */

void
cshift(inval, inindex, outval, outindex)
double	*inval;
int	inindex;
double	*outval;
int	outindex;

{

        header.nh_type = 0;
        header.nh_flags = 0;
        header.nh_length = sizeof(double);

	if ((myproc % 2) == 0) {
/* 
 * Even node index.
 * Send to outindex only if outindex in (0,nproc-1).
 * Receive from inindex only if inindex in (0,nproc-1).
 */

	    if ((outindex >= 0) && (outindex <= nproc-1)) {
		header.nh_node = outindex;
		header.nh_event = myproc * 65536 + outindex;
		header.nh_msg = (char *) outval;
/*	        printf("node %d --> %d on event %d : %lf\n", 
			myproc, outindex, header.nh_event, *outval);*/

		if (nsend(&header)) {
		    printf("Err in nsend(cshift) on node %d\n", myproc);
		    kexit(1);
		}
	    }

	    if ((inindex >= 0) && (inindex <= nproc-1)) {
		*inval = 0;
		header.nh_event = inindex * 65536 + myproc;
		header.nh_msg = (char *) inval;
/*		printf("node %d <-- %d on event %d\n", 
			myproc, inindex, header.nh_event); */
                if (nrecv(&header)) { 
                    printf("Err in nrecv(cshift) on node %d\n", myproc);    
                    kexit(1); 
                } 
	    }
	}

	else {

/*
 * Odd node index.
 * Receive from inindex only if inindex in (1,nproc).
 * Send to outindex only if outindex in (1,nproc).
 */
            if ((inindex >= 0) && (inindex <= nproc-1)) {
		*inval = 0;
		header.nh_event = inindex * 65536 + myproc;
		header.nh_msg = (char *) inval;
/*		printf("node %d <-- %d on event %d\n",
                        myproc, inindex, header.nh_event); */
                if (nrecv(&header)) {
                    printf("Err in nrecv(cshift) on node %d\n", myproc);
                    kexit(1);
                }
            }

            if ((outindex >= 0) && (outindex <= nproc-1)) {
                header.nh_node = outindex;
		header.nh_event = myproc * 65536 + outindex;
                header.nh_msg = (char *) outval;
/*              printf("node %d --> %d on event %d : %lf\n",
                        myproc, outindex, header.nh_event, *outval); */
                if (nsend(&header)) {
                    printf("Err in nsend(cshift) on node %d\n", myproc);
                    kexit(1);
                }
            }
        }
}


/*--------------------------------------------------------------------
 * dostep()	Update nodal points for nstep iterations.
 */

void
dostep()

{

	int	i, j;
	double 	pfac;

        header.nh_event = EVNSTEP;
        header.nh_type = 0;
        header.nh_flags = 0;
        header.nh_length = sizeof(int);
        header.nh_msg = (char *) &nstep;

/*      printf("node %d waiting for nstep\n", myproc);*/
        if (nrecv(&header)) {
            printf("Node %d: Err in nrecv(nstep)\n", myproc);
            kexit(1);
        }
 
/*        printf("nstep = %d\n", nstep); */

	pfac = (KONST * delt / 1.0);
	pfac = pfac * pfac;

	for (i=1; i<=nstep; i++) {

	    cshift (&values[0], (myproc-1), &values[mypts], (myproc+1));

	    cshift (&values[mypts+1], (myproc+1), &values[1], (myproc-1));

	    for (j=1; j<=mypts; j++) {

	        if ((myproc == 0) && (j == 1)) {
		    newval[j] = 0.0;
		}
		else if ((myproc == nproc-1) && (j == mypts)) {
		    newval[j] = 0.0;
		}
		else {
		    newval[j] = 2.0*values[j] - oldval[j] 
      			+ ( values[j+1] - 2.0*values[j] + values[j-1] )
      			* pfac;
		}
	    } 

	    for (j=1; j<=mypts; j++) {
	        oldval[j] = values[j];
		values[j] = newval[j];
	    }
	}

	header.nh_node = host;
	header.nh_event = EVDONE;
	header.nh_type = 0;
	header.nh_length = 0;
	header.nh_flags = 0;
/*	printf("node %d sending done msg to host\n", myproc);*/

	if (nsend(&header)) {
	    printf("Node %d : Err in nsend(done)\n", myproc);
	    kexit(1);
	}
}

/*--------------------------------------------------------------------
 * outres()	Send results to host.
 */

void
outres()

{
	int	j;

/* 
 * Receive out msg from control processor.
 */

        header.nh_event = EVOUT + myproc;
        header.nh_type = 0;
        header.nh_length = 0;
        header.nh_flags = 0;

/*	printf("Node %d receive out message from host\n", myproc);*/
	if (nrecv(&header)) {
            printf("Node %d : Err in nrecv(out)\n", myproc);
	    kexit(1);
	}

/* 
 * Send data to control processor.
 */
	for (j=1; j<=mypts; j++) {
	    values[j] = ltotd(values[j]);
	}

	header.nh_node = host;
        header.nh_event = EVRES;
        header.nh_type = 0;
        header.nh_length = mypts * sizeof(double); 
        header.nh_flags = 0;
        header.nh_msg = (char *) &values[1]; 

/*	printf("Node %d sending results to host\n", myproc);*/
	if (nsend(&header)) {
            printf("Node %d : Err in nsend(results)\n", myproc);
            kexit(1);
        }
}

/*====================================================================*/


