/*
 *	OSU Research Computing					Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	MB
 *
 *	$Log:	brendad.c,v $
 * Revision 2.2  92/04/24  15:45:00  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:34:42  gdburns
 * Ohio release
 * 
 *
 *	Implementation of BRENDA via Trollius: the server.
 *
 *	by Moshe Braner,  880901, 891026, 900316
 *
 *	This version for Trollius 2.0.
 *
 *	Copyright 1988, 1989 Cornell University
 */

/*
 *	This server is mostly stateless, and thus does not
 *	require much buffering from the system.  Statelessness
 *	here means that it receives all messages on the same
 *	event and regardless of type.  If what arrives is not
 *	what was expected (as detected by the type) then it must
 *	be buffered internally and dealt with later.
 *	The only exception is in receiving the data part of a
 *	large (>MAXNMSGLEN) GLOBAL PUT.  That is because we do
 *	not send an ACK message to the client in that case:
 *	the client broadcasts the request and then the data
 *	with no pause.  The data must not be mixed up with
 *	other such data, and its size may be smaller than other
 *	such data, so we must get the correct message into the
 *	buffer we allocate upon receiving the request.  This
 *	should not cause deadlock unless many clients try to
 *	PUT large GLOBAL objects at once.  (In the case of a
 *	nonglobal large PUT, the data is sent only after an ACK
 *	is received from the deamon, thus only one client may
 *	be sending a message of type _DATA_ at a time, and there
 *	is no need to break the statelessness.)
 */

#include "brenda.h"

#include <t_types.h>
#include <terror.h>


/*
 *  Buffered incoming requests are stored in a list in
 *  this form while previous requests are being processed:
 */
struct bufd_req {
	struct bufd_req *next;
	struct nmsg msg;
};

static struct bufd_req *headreq;	/* head of that list */
static struct bufd_req *tailreq;	/* tail of that list */

/*
 *  Requests and objects are stored long-term in this form,
 *  to be matched with requests that may come in later.
 *  Lists of these structs are rooted in two hash tables,
 *  one for objects and one for requests for objects.
 */
struct queued_req {
	struct queued_req *next;
	long	node;
	long	event;
	long	request;
	long	tag;
	long	size;
	long	data;
	char	*buf;
};

/* #define DBREC */
#ifdef DBREC
/* set recptr to somewhere in high memory where it is safe (we hope) */
#define DBLOC 0x801F0000
struct dbrec {
	long	node;
	long	reqtype;
	long	tag;
	long	data;
};
struct dbrec *recptr;
#endif DBREC

static struct queued_req *obj_space[B_SLOTS];	/* the hash table itself */
static struct queued_req *req_space[B_SLOTS];

static struct queued_req *free_que;	/* list of recyclable structs */

/*
 *  To avoid data copying as much as possible, we receive
 *  requests and their data blocks into these common areas:
 */
static struct nmsg reqmsg;
static char *reqbuf;

static struct nmsg repmsg;		/* common reply-to-client header */

static long  mynode, mypid;		/* avoid calling getpid(), etc */


/* this function is called when we cannot handle an error: */
void
panic()
{
	tprint ("brendad: runtime error\n");
	_kexit (errno);
}

/*
 *  Wrappers for malloc() and free().
 */

char *
Malloc (amount)
	unsigned int amount;
{
	char *cp;

	if (amount == 0)
		return (NULL);
	cp = malloc (amount);
	if (cp == NULL)
		tprint ("brendad: out of memory\n");
	return (cp);
}

void
Free (addr)
	char *addr;
{
	if (addr == NULL)  return;
#ifdef FREEVOID
	free (addr);
#else
	if (free (addr))  panic();
#endif
}


/*
 *  Reallocate the reqbuf common buffer in the most efficient way.
 */
char *
rebuf (pbuf, size)
	char **pbuf;
	long size;
{
	char *addr;

	if (size < MAXNMSGLEN / 4) {
		/* allocate new space for the data, copy it */
		/* (hopefully this will fight memory fragmentation) */
		addr = Malloc ((unsigned int) size);
		if (addr != NULL)
			bcopy (*pbuf, addr, (int) size);
	} else {
		/* keep data where it is, but get a new buffer */
		addr = *pbuf;
		*pbuf = Malloc ((unsigned int) MAXNMSGLEN);
		if (*pbuf == NULL) {
			*pbuf = addr;
			return (NULL);
		}
		if (size <= 3 * MAXNMSGLEN / 4) {
			/* return the rest to the system */
			addr = realloc (addr, (unsigned int) size);
		}
	}
	return (addr);
}


/*
 *  Wait for and receive a message.
 *  To make this server stateless we have to accept
 *  any message.  If we are waiting for data and get
 *  a new request we have to buffer it for later.
 *  Called from main() (with NULL buf) and from do_lput().
 */
struct nmsg *
do_recv (type, buf, length)
	long type;
	char *buf;
	long length;
{
	char *maddr, *baddr;
	long mlength;
	struct bufd_req *req, *preq;
	struct nmsg *msg;

	/* anything acceptable in list of buffered requests? */

	preq = NULL;
	req = headreq;
	while (req) {				/* look at buffered reqs */
		if (req->msg.nh_type & type) {	/* acceptable right now */
			/* unlink req from list: */
			if (preq) {
				preq->next = req->next;
				if (req == tailreq)
					tailreq = preq;
			} else {
				if (req == tailreq)
					headreq = NULL;
				else
					headreq = req->next;
			}
			/* minimum length: */
			mlength = req->msg.nh_length;
			if (mlength > length) {
				mlength = length;
				req->msg.nh_length = mlength;
			}
			/* copy data if necessary: */
			if (buf != NULL && mlength != 0) {
				bcopy (req->msg.nh_msg, buf, (int) mlength);
				Free (req->msg.nh_msg);
				req->msg.nh_msg = buf;
			}
			/* return what looks like a fresh message: */
			bcopy (&req->msg, &reqmsg, sizeof (struct nmsg));
			Free ((char *) req);
			return (&reqmsg);
		}
		preq = req;
		req = req->next;
	}

	/* no, need to nrecv a brand new request */

	if (buf == NULL) {
		maddr = reqbuf;
		mlength = MAXNMSGLEN;
	} else if (length < MAXNMSGLEN) {
		maddr = reqbuf;
		mlength = MAXNMSGLEN;
	} else {
		maddr = buf;
		mlength = length;
	}

	while (1) {			/* wait until we got the right thing */

		msg = &reqmsg;
		msg->nh_event  = BRENDA;
		msg->nh_type   = 0;		/* receive any type */
		msg->nh_flags  = DINTDATA;	/* do ttol() on nh_data */
		msg->nh_length = mlength;
		msg->nh_msg    = maddr;
/*0*/		if (nrecv (msg) != 0)  panic();
		mlength = msg->nh_length;

#ifdef DBREC
		recptr->node    = ((struct brenda_req *) msg->nh_data)->node;
		recptr->reqtype = ((struct brenda_req *) msg->nh_data)->request;
		recptr->tag     = ((struct brenda_req *) msg->nh_data)->tag;
		recptr->data    = ((struct brenda_req *) msg->nh_data)->data;
		++recptr;
		recptr->node    = -1;
		recptr->reqtype = -1;
#endif

/*
 * Before bailing out, set nh_msg to NULL if nh_length is 0 or this
 * spagetti will slip off your fork.  (GDB)
 */
		if (msg->nh_type & type) {	/* acceptable right now */
			if (!mlength) msg->nh_msg = NULL;
			break;
		}

		/* insert msg into bufd list: */
		req = (struct bufd_req *) Malloc (sizeof (struct bufd_req));
		if (req == NULL)
			panic();		/* out of memory */
		if (maddr == reqbuf && mlength != 0) {
			/* get the data out of reqbuf: */
			baddr = rebuf (&reqbuf, msg->nh_length);
			if (baddr == NULL)
				panic();	/* out of memory */
		} else if (mlength != 0) {
			/* copy it out of caller's buf: */
			baddr = Malloc ((unsigned int) mlength);
			if (baddr == NULL)
				panic();	/* out of memory */
			bcopy (maddr, baddr, (int) mlength);
		} else {
			baddr = NULL;
		}
		msg->nh_msg    = baddr;
		msg->nh_length = mlength;
		bcopy (msg, &req->msg, sizeof (struct nmsg));
		if (headreq == NULL)		/* empty list */
			headreq = req;
		else
			tailreq->next = req;
		tailreq = req;
		req->next = NULL;
	}

	if (buf != NULL && maddr != buf && mlength != 0) {
		bcopy (maddr, buf, (int) mlength);
		msg->nh_msg = buf;
	}
	return (msg);
}


/*
 *	Send the client a reply.
 */
void
nreply (node, event, buf, nbytes)
	long nbytes, node, event;
	char *buf;
{
	struct nmsg *msg = &repmsg;
	msg->nh_node   = node;
	msg->nh_event  = event;
	msg->nh_type   = _REPLY_;
	msg->nh_flags  = DINTDATA;		/* do ltot() on nh_data */
	msg->nh_msg    = buf;
	msg->nh_length = nbytes;
/*1,2*/	if (nsend (msg) != 0)  panic();

#ifdef DBREC
	recptr->node    = msg->nh_node;
	recptr->reqtype = ((struct brenda_req *) msg->nh_data)->request;
	recptr->tag     = ((struct brenda_req *) msg->nh_data)->tag;
	recptr->data    = ((struct brenda_req *) msg->nh_data)->data;
	++recptr;
	recptr->node    = -1;
	recptr->reqtype = -1;
#endif

}

/*
 *	Send the client an error reply.
 */
void
erreply (node, event, tag, errcode)
	long node, event, tag;
	int errcode;
{
	struct brenda_req *reply = (struct brenda_req *) repmsg.nh_data;
	reply->request = _ERROR_;
	reply->tag     = tag;
	reply->node    = mynode;
	reply->data    = errcode;
	nreply (node, event, NULLP, 0L);
}

/* send the client the most common error reply: */
void
badreq (node, event, tag)
	long node, event, tag;
{
	erreply (node, event, tag, EBADREQUEST);
}
void
do_error (req)
	struct brenda_req *req;
{
	badreq (req->node, req->event, req->tag);
}


/*
 *	Send out data.
 */
void
dreply (node, event, buf, nbytes)
	long nbytes, node, event;
	char *buf;
{
	struct nmsg *msg = &repmsg;
	msg->nh_node   = node;
	msg->nh_event  = event;
	msg->nh_type   = _DATA_;
	msg->nh_flags  = DRAWDATA;
	msg->nh_msg    = buf;
	msg->nh_length = nbytes;
/*4*/	if (nsend (msg) != 0)  panic();
}


/*
 *	Search for a stored object that matches the request.
 *	We do not unlink the found object here since it may be a LOOK.
 *	This routine doubles for searching the list of requests.
 */
struct queued_req *
find_obj (reqtag, list, ppobj)
	long reqtag;
	struct queued_req *list, **ppobj;
{
	long objtag;
	struct queued_req *obj, *pobj;

	pobj = NULL;
	obj = list;
	while (obj) {
		objtag = obj->tag;
		if (objtag == reqtag || objtag == 0 || reqtag == 0) {
			*ppobj = pobj;
			return (obj);
		}
		pobj = obj;
		obj = obj->next;
	}
	return (NULL);	/* not found */
}

/*
 *	Search for a queued request that is satisfied by a given object.
 *	If found, unlink it, but do NOT put on free list yet.
 */
struct queued_req *
find_req (tag, slot)
	long tag;
	int  slot;
{
	struct queued_req *que, *pque;

	que = find_obj (tag, req_space[slot], &pque);
	if (que == NULL)
		return (NULL);			/* not found */
	if (pque)
		pque->next = que->next;
	else					/* at head */
		req_space[slot] = que->next;
	return (que);
}

/*
 *	Get a queued_req list node.
 *	Called by store_que() only.
 */
struct queued_req *
get_que()
{
	int	i;
	char	*addr;
	struct queued_req *que;

	if (free_que == NULL) {
		addr = Malloc (B_NBLOCK * sizeof (struct queued_req));
		if (addr == NULL)  return (NULL);
		free_que = (struct queued_req *) addr;
		for (i=0; i<B_NBLOCK-1; i++)
			free_que[i].next = &free_que[i+1];
		free_que[B_NBLOCK-1].next = NULL;
	}
	que = free_que;
	free_que = que->next;			/* unlink from free list */
	que->next = NULL;
	que->buf  = NULL;
	return (que);
}

/*
 *	Store the current request:
 *	Get a new list node, but do not link it in a list yet.
 *	Allocate data space if necessary.
 *	Return pointer to new list node.
 *	Called by do_put(), do_lput(), do_look(), do_bump().
 */
struct queued_req *
store_que (req)
	struct brenda_req *req;
{
	int	reqtype, isput;
	struct queued_req *que;

	reqtype = req->request;
	isput = (reqtype == B_PUT || reqtype == _GPUT_);

	que = get_que();

	if (que == NULL) {
		if (isput) {
			/* out of memory: drop request */
			return (NULL);
		} else {
			/* out of memory: reply(error) to this client: */
			erreply (req->node, req->event, req->tag, errno);
			return (NULL);
		}
	}

	que->node    = req->node;
	que->event   = req->event;
	que->request = reqtype;
	que->tag     = req->tag;
	que->size    = req->size;
	que->data    = req->data;

	if (isput && req->size > sizeof(long)) {
		if (req->buf == reqbuf) {
			/* need to relocate the object data: */
			req->buf = rebuf (&reqbuf, req->size);
			if (req->buf == NULL)
				return (NULL);
		}
		que->buf = req->buf;
	}

	return (que);
}

/*
 *	Link a stored object or request into a list
 *	in the hash tables for long-term residence.
 *	Called by do_put(), do_lput(), do_look(), do_bump().
 */
void
link_que (que, list)
	struct queued_req *que, **list;
{
	struct queued_req *pque;

	pque = *list;
	if (pque) {				/* link at end */
		while (pque->next)
			pque = pque->next;
		pque->next = que;
	} else {				/* first in list */
		*list = que;
	}
	que->next = NULL;
}


/************/
/*   PUT    */
/*  _GPUT_  */
/************/

void
do_put (obj)
	struct brenda_req *obj;
{
	int  slot;
	long tag, reqtype, msize;
	struct queued_req *que;
	struct brenda_req *reply = (struct brenda_req *) repmsg.nh_data;

	tag = obj->tag;
	slot = HASHSLOT(tag);

	que = find_req (tag, slot);	/* satisfies any queued request? */

	while (que) {				/* yes, fits an old request */

		que->next = free_que;		/* link onto free list */
		free_que = que;

		reqtype = que->request;

		if (obj->request == _GPUT_ && reqtype != B_LOOK) {

			/* global objects may not be grabbed */
/*1*/			badreq (que->node, que->event, tag);

		} else if (reqtype==B_BUMP) {	/* bump request pending */

			if (obj->size != sizeof(long)) {

/*1*/			    badreq (que->node, que->event, tag);

			} else {

				/* ship current contents to this client */
				reply->request = _FOUND_;
				reply->tag     = obj->tag;
				reply->data    = obj->data;
/*1*/				nreply (que->node, que->event, NULLP, 0L);

				/* bump object by amount given by client */
				obj->data += que->size;
			}

		} else {

			/* ship this fresh object to the queued client */
			if (obj->buf == NULL) {
				reply->data = obj->data;
				msize = 0;
			} else {
				msize = que->size;
				if (msize > obj->size)
					msize = obj->size;
			}
			reply->tag     = obj->tag;
			reply->request = _FOUND_;
/*1*/			nreply (que->node, que->event, obj->buf, msize);
		}

		/* if it's been grabbed then nobody else gets it */
		if (reqtype == B_GRAB && obj->request != _GPUT_)
			return;

		/* satisfies any additional queued requests? */
		que = find_req (tag, slot);
	}

	/* store the object since it was not grabbed */
	que = store_que (obj);
	if (que == NULL)
		return;
	link_que (que, &obj_space[slot]);
}


/***********/
/* _LPUT_  */
/* _LGPUT_ */
/***********/

/*
 *  Receive a global-data message from the client.
 *  This is the only exception to statelessness in this server.
 */
void
recv_gdata (event, buf, nbytes)
	long event;
	char *buf;
	long nbytes;
{
	struct nmsg msg;

	msg.nh_event  = event;
	msg.nh_type   = _DATA_;
	msg.nh_flags  = DRAWDATA;
	msg.nh_length = nbytes;
	msg.nh_msg    = buf;
/*3*/	if (nrecv (&msg) != 0)  panic();
}

/* link a list of satisfied requests onto free_que */
void
free_list (head)
	struct queued_req *head;
{
	struct queued_req *que;

	if (head == NULL)  return;
	que = head;
	while (que->next)  que = que->next;
	que->next = free_que;
	free_que = head;
}

/* relink a list of unsatisfied requests back into the queue */
void
re_link (head, slot)
	struct queued_req *head;
	int slot;
{
	struct queued_req *pque;

	if (head == NULL)  return;
	pque = req_space[slot];
	if (pque) {				/* link at end */
		while (pque->next)
			pque = pque->next;
		pque->next = head;
	} else {				/* link onto empty list */
		req_space[slot] = head;
	}
}

void
do_lput (obj)
	struct brenda_req *obj;
{
	int s, global, reqtype, recv_it, keep_it, slot;
	long size, msize, tag, event;
	struct nmsg *dmsg;
	struct queued_req *que, *pque, *head;
	struct brenda_req *reply = (struct brenda_req *) repmsg.nh_data;

	size = obj->size;
	tag  = obj->tag;
	slot = HASHSLOT(tag);
	global = (obj->request == _LGPUT_);
	reqtype = B_PUT;		/* distinct from GRAB and LOOK */
	que = find_req (tag, slot);	/* satisfies any queued request? */
	head = que;				/* first found */
	pque = NULL;				/* previous one found */
	s = 0;
	while (que) {				/* yes, fits an old request */
		++s;				/* count how many requests */
		if (pque)
			pque->next = que;	/* link requests in a list */
		que->next = NULL;
		reqtype = que->request;		/* GRAB, LOOK or BUMP */
		if (reqtype == B_GRAB)
			break;			/* only one grabs it */
		pque = que;
		que = find_req (tag, slot);	/* any other requests? */
	}

	keep_it = (global || reqtype != B_GRAB);   /* only LOOK(s), or none */
	recv_it = ((keep_it || s > 1) && size > sizeof(long));

	/* tell grab-ing (or look-ing) party(s) that an object is coming */
	for (que=head; que!=NULL; que=que->next) {
		if (global && que->request != B_LOOK) {
/*1*/			badreq (que->node, que->event, tag);
		} else if (size == sizeof(long)) {
			/* send the data itself right away: */
			reply->request = _FOUND_;
			reply->tag     = tag;
			reply->data    = obj->data;
/*1*/			nreply (que->node, que->event, NULLP, 0L);
			if (que->request == B_BUMP)
				obj->data += que->size;
		} else {
			msize = (size > que->size? que->size : size);
			if (msize <= MAXNMSGLEN)  /* single-packet object? */
				continue;	  /* then no advance message */
			reply->request = _LFOUND_;
			reply->tag     = tag;
			reply->size    = msize;
/*1*/			nreply (que->node, que->event, NULLP, 0L);
		}
	}

	s = 0;
	if (recv_it) {
		/* get space for the data */
		obj->buf = Malloc ((unsigned int) size);
		if (obj->buf == NULL) {
			s = errno;
			if (global)		/* no NACK */
				return;
		}
	} else {
		obj->buf = NULL;		/* data in obj->data */
	}

	/* if not global then tell put-ing party where to send it to */
	if (! global) {
		if (recv_it) {
			reply->node  = mynode;
			reply->size  = size;
			reply->event = BRENDA;
		} else {
			if (head->size > size)
				head->size = size;
				/* -- can't be more than we got */
			reply->node  = head->node;
			reply->event = head->event;
			reply->size  = head->size;
				/* -- may be less than offered */
		}
		reply->request = _ACK_;
		reply->tag     = tag;
		reply->data    = s;		/* nonzero means not ack! */
/*2*/		nreply (obj->node, obj->event, NULLP, 0L);
		if (s) {			/* can't do it */
			re_link (head, slot);
			/* -- return unsatisfied requests to queue */
			return;
		}
	}

	if (keep_it) {
		/* need to store object in hash table */
		if (global)		/* mark it as such */
			obj->request = _GPUT_;
		que = store_que (obj);
		if (que == NULL)	/* storage failed: drop request */
			return;
		/* (obj->buf is allocated but data is not there yet) */
		link_que (que, &obj_space[slot]);
	}

	if (! recv_it) {
		free_list (head);	/* satisfied requests to free list */
		return;			/* done, or they do it without us */
	}

	/* get the object data */
	if (global) {
		/* this is the non-stateless case: */
		event = (obj->node ^ obj->event);
/*3*/		recv_gdata (event, obj->buf, size);
	} else {
/*3*/		dmsg = do_recv ((long) _DATA_, obj->buf, size);
		if (dmsg == NULL)
			panic();
	}

	/* send copies to waiting client(s) */
	for (que=head; que!=NULL; que=que->next) {

		if (que->request == B_BUMP) {

			/* we're here 'cause size > sizeof(long), so: */
/*1*/			badreq (que->node, que->event, tag);

		} else  /* LOOK or GRAB */ {

			if (global && que->request != B_LOOK) {
				/* we already sent an error reply */
				continue;
			}
			msize = (size > que->size? que->size : size);
			if (msize > MAXNMSGLEN) {
/*4*/			    dreply (que->node, que->event, obj->buf, msize);
			} else {
			    /* there was no advance warning... */
			    reply->request = _FOUND_;
			    reply->tag     = tag;
			    reply->size    = msize;
/*1*/			    nreply (que->node, que->event, obj->buf, msize);
			}
		}
	}				/* next que */

	free_list (head);		/* satisfied requests to free list */

	if (! keep_it)			/* it's been grabbed: throw it out */
		Free (obj->buf);
}


/*************/
/*  GRAB     */
/*  _GGRAB_  */
/*  LOOK     */
/*  TRYGRAB  */
/*  TRYLOOK  */
/*************/
void
do_look (req)
	struct brenda_req *req;
{
	int slot, send_it;
	long size, reqtype, tag;
	struct queued_req *obj, *pobj, *que;
	struct brenda_req *reply = (struct brenda_req *) repmsg.nh_data;

	reqtype = req->request;
	tag   = req->tag;
	size  = req->size;
	slot = HASHSLOT(tag);

	/* search for such an object: */
	obj = find_obj (tag, obj_space[slot], &pobj);

	if (obj == NULL) {		/* not found */

		if (reqtype == B_GRAB || reqtype == B_LOOK) {

			/* it's a blocking request, queue it: */
			que = store_que (req);
			if (que == NULL)
				return;
			link_que (que, &req_space[slot]);

		} else if (reqtype == _GGRAB_) {

			/* global grab is non-blocking */
			/* only the local deamon replies: */
			if (mynode == req->node)
/*1*/			  erreply (req->node, req->event, tag, EWOULDBLOCK);

		} else  /* TRYGRAB or TRYLOOK */ {

/*1*/			erreply (req->node, req->event, tag, EWOULDBLOCK);
		}

		return;
	}

	/* object found: should we send it?  Is the request legal? */

	send_it = 1;

	if (obj->request == _GPUT_) {	/* global object found */

	    if (reqtype == B_GRAB || reqtype == B_TRYGRAB) {

		/* only a global grab is legal on a global object */
/*1*/		badreq (req->node, req->event, tag);
		return;

	    } else if (reqtype == _GGRAB_) {

		if (req->node != obj->node || req->event != obj->event) {

			/* only the originator can grab a global object */
			/* only the local deamon replies to the broadcast */
			if (mynode == req->node)
/*1*/			  badreq (req->node, req->event, tag);
			return;

		} else {

			/* originator grabs: all the deamons delete
			   the object but only the local deamon replies */
			send_it = (mynode == req->node);

		}
	    }

	} else {			/* normal object found */

		if (reqtype == _GGRAB_) {

			/* global grab allowed on global objects only */
			/* only the local deamon replies: */
			if (mynode == req->node)
/*1*/			  badreq (req->node, req->event, tag);
			return;
		}
	}


	/* can't send more than there is */
	if (size > obj->size)
		size = obj->size;

	if (send_it && size <= MAXNMSGLEN) {

		/* ship the found object to this client */
		reply->request = _FOUND_;
		reply->tag     = obj->tag;
		reply->size    = size;
		if (obj->buf == NULL) {
			reply->data = obj->data;
			size = 0;
		}
/*1*/		nreply (req->node, req->event, obj->buf, size);

	} else if (send_it) {		/* long object */

		/* tell grab-ing party that a long object is coming */
		reply->request = _LFOUND_;
		reply->tag     = obj->tag;
		reply->size    = size;
/*1*/		nreply (req->node, req->event, NULLP, 0L);

		/* send the data separately */
/*4*/		dreply (req->node, req->event, obj->buf, size);
	}

	/* if it's a grab, unlink stored object from list */
	if (reqtype==B_GRAB || reqtype==B_TRYGRAB || reqtype==_GGRAB_) {
		Free (obj->buf);
		if (pobj)
			pobj->next = obj->next;
		else					/* at head */
			obj_space[slot] = obj->next;
		obj->next = free_que;
		free_que = obj;
	}
}


/********/
/* BUMP */
/********/
void
do_bump (req)
	struct brenda_req *req;
{
	int  slot;
	long tag;
	struct queued_req *obj, *pobj, *que;
	struct brenda_req *reply = (struct brenda_req *) repmsg.nh_data;

	tag = req->tag;
	slot = HASHSLOT(tag);

	/* search for such an object: */
	obj = find_obj (tag, obj_space[slot], &pobj);

	if (obj) {			/* found */

		if (obj->request == _GPUT_) {
			/* bumping global object is verbotten */
/*1*/			badreq (req->node, req->event, tag);
			return;
		}

		if (obj->size != sizeof(long)) {
/*1*/			badreq (req->node, req->event, obj->tag);
			return;
		}

		/* ship current contents to this client */
		reply->request = _FOUND_;
		reply->tag     = obj->tag;
		reply->data    = obj->data;
/*1*/		nreply (req->node, req->event, NULLP, 0L);

		/* bump the object by the amount given by the client */
		obj->data += req->size;

	} else  /* not found */ {

		/* it's a blocking request, queue it */
		que = store_que (req);
		if (que == NULL)
			return;
		link_que (que, &req_space[slot]);
	}
}


/*************/
/* _OBJ_SUM  */
/* _OBJ_LST  */
/* _REQ_SUM  */
/* _REQ_LST  */
/*************/
void
do_report (req)
	struct brenda_req *req;
{
	int  i, count, maxcount, slot;
	long reqtype, size, osize, tag;
	char *addr;
	struct queued_req *que;
	struct obj_lst *objlst;
	struct req_lst *reqlst;
	struct brenda_req *reply = (struct brenda_req *) repmsg.nh_data;

	reqtype = req->request;
	tag  = req->tag;
	slot = HASHSLOT(tag);
	size = 0;
	objlst = NULL;
	reqlst = NULL;
	addr = NULL;
	if (reqtype == _OBJ_LST
	||  reqtype == _REQ_LST) {
		size = req->size;
		addr = Malloc ((unsigned int) size);
		if (addr == NULL) {
/*1*/			erreply (req->node, req->event, tag, errno);
			return;
		}
		if (reqtype == _OBJ_LST) {
			objlst = (struct obj_lst *) addr;
			maxcount = size / sizeof(struct obj_lst);
		} else {
			reqlst = (struct req_lst *) addr;
			maxcount = size / sizeof(struct req_lst);
		}
	}
	count = 0;
	osize = 0;
	for (i=(tag?slot:0); (tag?i==slot:i<B_SLOTS); i++) {
		if (reqtype == _OBJ_SUM
		||  reqtype == _OBJ_LST)
			que = obj_space[i];
		else
			que = req_space[i];
		while (que) {
			if (tag      == 0
			||  que->tag == 0
			||  que->tag == tag) {
				if (objlst && count < maxcount) {
					objlst->tag  = ltot (que->tag);
					objlst->size = ltot (que->size);
					++objlst;
				}
				if (reqlst && count < maxcount) {
					reqlst->node    = ltot (que->node);
					reqlst->event   = ltot (que->event);
					reqlst->request = ltot (que->request);
					reqlst->tag     = ltot (que->tag);
					reqlst->size    = ltot (que->size);
					++reqlst;
				}
				osize += que->size;
				++count;
			}
			que = que->next;
		}
	}

	reply->request = reqtype;
	reply->node = mynode;
	reply->tag  = tag;
	reply->data = count;
	reply->size = osize;
/*1*/	nreply (req->node, req->event, addr, size);
	Free (addr);
}


/************/
/* _CLEAR_  */
/************/
int
do_clear (req)
	struct brenda_req *req;
{
	int  i, count, slot;
	long tag;
	struct queued_req *obj, *pobj, *oobj, *que, *pque;
	struct brenda_req *reply = (struct brenda_req *) repmsg.nh_data;

#ifdef DBREC
	/* clear the debugging record too */
	recptr = (struct dbrec *) DBLOC;
	recptr->node    = -1;
	recptr->reqtype = -1;
#endif

	tag = req->tag;

#ifdef FREEALL
	if (tag == 0) {			/* complete reset */
		/* dirty trick: return all memory to system */
		rfreeall (LOCAL);
		_reset_malloc();
		reply->request = _CLEAR_;
		reply->node    = mynode;
		reply->tag     = 0;
		reply->data    = 0;
/*1*/		nreply (req->node, req->event, NULLP, 0L);
		return (1);			/* go start over */
	}
#else
	/* fall through, delete all if tag == 0 */
#endif

	/* selective deletion */

	slot = HASHSLOT(tag);
	count = 0;
	for (i=(tag?slot:0); (tag?i==slot:i<B_SLOTS); i++) {
		que = req_space[i];
		pque = NULL;
		while (que) {
			if (tag      == 0
			||  que->tag == 0
			||  que->tag == tag) {
				if (pque)
					pque->next = que->next;
				else
					req_space[slot] = que->next;
				que->next = free_que;
				free_que = que;
				++count;
			}
			pque = que;
			que = que->next;
		}
		obj = obj_space[i];
		pobj = NULL;
		while (obj) {
			oobj = obj;
			if (tag      == 0
			||  obj->tag == 0
			||  obj->tag == tag) {
				Free (obj->buf);
				if (pobj)
					pobj->next = obj->next;
				else
					obj_space[slot] = obj->next;
				obj->next = free_que;
				free_que = obj;
				++count;
			}
			pobj = obj;
			obj = oobj->next;
		}
	}
	reply->request = _CLEAR_;
	reply->node    = mynode;
	reply->tag     = tag;
	reply->data    = count;
/*1*/	nreply (req->node, req->event, NULLP, 0L);
	return (0);
}


/************/
/*   MAIN   */
/************/

/*
 *  Initialize the data structures:
 */
void
init()
{
	int i;

	/* initialize hash tables */
	for (i=0; i<B_SLOTS; i++) {
		obj_space[i] = NULL;
		req_space[i] = NULL;
	}
	free_que = NULL;	/* no free list nodes yet */
	headreq = NULL;
	tailreq = NULL;

	/* allocate first request buffer */
	reqbuf = Malloc ((unsigned int) MAXNMSGLEN);
	if (reqbuf == NULL)
		_kexit (errno);
}


/*
 *	**** The main program of the Brenda server ****
 */
main()
{
	long type;
	struct nmsg *msg;
	struct brenda_req *req;

	if (kinit (0))
		kexit (1);

	mynode = getnodeid();
	mypid  = getpid();

	init();

#ifdef DBREC
	/* set recptr to somewhere in high memory where it is safe (we hope) */
	recptr = (struct dbrec *) DBLOC;
	recptr->node    = -1;
	recptr->reqtype = -1;
#endif

	/*****************/
	/*   main loop   */
	/*****************/

	type = _REQ_;

	while (1) {			/* loop forever */

		/* get a new request to work on: */
/*0*/		msg = do_recv (type, NULLP, (long) MAXNMSGLEN);
		if (msg == NULL)
			break;
		req = (struct brenda_req *) msg->nh_data;
		req->buf = msg->nh_msg;

		switch ((int) req->request) {

		case B_PUT:
		case _GPUT_:
				do_put (req);
				break;
		case _LPUT_:
		case _LGPUT_:
				do_lput (req);
				break;
		case B_GRAB:
		case _GGRAB_:
		case B_LOOK:
		case B_TRYGRAB:
		case B_TRYLOOK:
				do_look (req);
				break;
		case B_BUMP:
				do_bump (req);
				break;
		case _OBJ_SUM:
		case _OBJ_LST:
		case _REQ_SUM:
		case _REQ_LST:
				do_report (req);
				break;
		case _CLEAR_:
				if (do_clear (req) != 0)  init();
				break;
		case _FREEZE_:
				type = _FROZEN_;
				break;
		case _THAW_:
				type = _REQ_;
				break;
		default:			/* not a valid request type */
				do_error (req);

		}				/* end of switch (request) */
	
	}				/* end of while (1) */

	/* this point not reached unless major error */
	panic();
}
