/*
 *      OSU Research Computing					Trillium
 *      Copyright 1992 The Ohio State University
 *      GDB
 *
 *      $Log:	xled.c,v $
 * Revision 2.2  92/04/24  16:14:49  trillium
 * Ohio Release
 * 
 *	Function:	- OTB LED simulator
 *			- based on X Windows, Athena widget set
 *			- operates as separate process to ledd
 *			- communicates via a pipe
 *			- becomes a Trollius process only if
 *			  killed by X, in order to kill the ledd parent
 */

#include <kio.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <terror.h>

#undef NULL
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Viewport.h>

#include <xled.h>

/*
 * local variables
 */
static Pixmap		win;			/* drawing window */
static GC		draw_gc, undraw_gc;	/* pen/eraser */
static GC		copy_gc;		/* adds colour */
static int		winwidth, winheight;	/* win dimensions */
static int		nodewidth;		/* node panel in pixels */
static int		minx, maxx,		/* max/min exposure */
			miny, maxy;
static int		nnodes;			/* # node panels */
static int		pnodes[MAXNODES];	/* node ids */
static int		pshadow[MAXNODES];	/* shadow regs for each node */
static int		parentid;		/* pid of ledd parent */
static int		xfd[2];			/* pipe ends */
static char		*me;			/* command line name */

static Widget		wtoplevel;
static Widget		wleds;
static Widget		wviewport;


typedef struct {
	Pixel		ap_ledfg;		/* LED panel colour */
	int		ap_ledsize;		/* 1 LED pix size on side */
} AppData, *AppDataPtr;

AppData			app_res;		/* application resources */

static XtResource	resources[] = {		/* app resource list */
	{
		XtNledForeground,
		XtCLedForeground,
		XtRPixel,
		sizeof(Pixel),
		XtOffset(AppDataPtr, ap_ledfg),
		XtRString,
		XtDefaultForeground
	},
	{
		XtNledSizeInPixels,
		XtCLedSizeInPixels,
		XtRInt,
		sizeof(int),
		XtOffset(AppDataPtr, ap_ledsize),
		XtRImmediate,
		(caddr_t) 16,
	},
};

static XrmOptionDescRec	options[] = {		/* command line options */
	{"-ledsize",	"*ledSizeInPixels",	XrmoptionSepArg,	NULL},
	{"-ledfg",	"*ledForeground",	XrmoptionSepArg,	NULL},
};

/*
 * local functions
 */
static void		Bail();
static void		display();
static void		drawled();
static void		redraw();
static void		setup();
static void		xled();

/*
 * global functions
 */
void			leinit();
void			ledisplay();
int4			lecount();

/*
 * external functions
 */
extern char		*mnemonic();

/*
 * external variables
 */
extern int		errno;
extern struct kio_t	_kio;

/*
 *	leinit
 *
 *	Function:	- initializes node dependent LED interface
 *			- X Window version
 *			- forks a new process for the X client
 *	Accepts:	- argc, argv to pass to X child
 */

void
leinit(argc, argv)

int			argc;
char			*argv[];

{
	int		pid;

	parentid = getpid();

	if (pipe(xfd) < 0) {
		perror(argv[0]);
		exit(errno);
	}

	if ((pid = fork()) < 0) {
		perror(argv[0]);
		exit(errno);
    	}

	else if (pid == 0) {		/* child */
		close(xfd[1]);
		_kio.ki_rtf &= ~RTF_KERNEL;
		xled(argc, argv);
    	}

	close(xfd[0]);
}

/*
 *	ledisplay
 *
 *	Function:	- displays new value of LEDs
 *	Accepts:	- client nodeid
 *			- new value of LEDs
 */

void
ledisplay(nodeid, shadow)

int4			nodeid;
int4			shadow;

{
	int		n;

	n = write(xfd[1], (char *) &nodeid, sizeof(int4));

	if (n < 0) {
		perror(me);
		exit(errno);
	}

	n = write(xfd[1], (char *) &shadow, sizeof(int4));

	if (n < 0) {
		perror(me);
		exit(errno);
	}
}

/*
 *	lecount
 *
 *	Returns:	- number of LEDs per node
 */

int4
lecount()

{
	return(NLEDS);
}

/*
 *	xled
 *
 *	Function:	- mainline of LED X Window process
 *			- initializes and then enters event loop
 */

static void
xled(argc, argv)

int			argc;
char			*argv[];

{
	Arg		args[8];	/* resource arguments */
	int		i;		/* favourite index */
	int		n_index;	/* node index */
	int		n_id;		/* node identifier */
	int		n_flags;	/* ignored */
	int		*p;		/* favourite ptr */
	int		defwinwidth;	/* adjusted max win width */

	String		trans =		/* event translations */
		"<Expose>:	redraw()";

	static XtActionsRec		/* action table */
			window_actions[] = {
		{"redraw", redraw},
	};

	me = argv[0];
/*
 * Parse the nodes off the command line.
 */
	if (parse_nodes(&argc, argv) < 0) {
		terror(argv[0]);
		exit(errno);
	}
/*
 * Establish the node list.
 */
	nnodes = 0;
	p = pnodes;
	getnode(&n_index, &n_id, &n_flags);

	if (n_index < 0) {
		errno = EUSAGE;
		terror(argv[0]);
		exit(errno);
	}

	do {
		if (nnodes >= MAXNODES) {
			errno = EUSAGE;
			terror(me);
			exit(errno);
		}

		*p++ = n_id;
		nnodes += 1;
		getnode(&n_index, &n_id, &n_flags);
	} while (n_index > 0);

	for (i = 0; i < nnodes; ++i) {
		pshadow[i] = 0;
	}

	wtoplevel = XtInitialize(
		me,			/* not really used */
		"XLedd",		/* application class */
		options,		/* resource mgr options */
		XtNumber(options),	/* # of RM options */
		&argc,			/* # of args */
		argv			/* command line */
	);

	if (argc > 1) {
		errno = EUSAGE;
		terror(argv[0]);
		exit(errno);
	}

	XtGetApplicationResources(wtoplevel,
		&app_res,
		resources,
		XtNumber(resources),
		NULL,
		0);

	if ((app_res.ap_ledsize < MINLEDSIZE) ||
		(app_res.ap_ledsize > MAXLEDSIZE)) {

		errno = EINVAL;
		terror(me);
		exit(errno);
	}

	setup();
/*
 * Create widgets.
 */
	defwinwidth = ((DEFWINWIDTH / nodewidth) * nodewidth) +
		app_res.ap_ledsize;
	i = 0;
	XtSetArg(args[i], XtNallowHoriz, TRUE);   i++;
	XtSetArg(args[i], XtNuseBottom, TRUE);   i++;
	XtSetArg(args[i], XtNwidth, (winwidth > defwinwidth) ?
		defwinwidth : winwidth);   i++;
	XtSetArg(args[i], XtNheight, winheight + 14);   i++;

	wviewport = XtCreateManagedWidget(
		"viewport",		/* instance name */
		viewportWidgetClass,	/* widget class */
		wtoplevel,		/* parent */
		args,			/* resources */
		i			/* # of resources */
	);

	i = 0;
	XtSetArg(args[i], XtNtranslations,
		XtParseTranslationTable(trans));   i++;
	XtSetArg(args[i], XtNwidth, winwidth);   i++;
	XtSetArg(args[i], XtNheight, winheight);   i++;

	wleds = XtCreateManagedWidget(
		"leds",			/* instance name */
		widgetClass,		/* widget class */
		wviewport,		/* parent */
		args,			/* resources */
		i			/* # of resources */
	);

	XSetIOErrorHandler(Bail);

	XtAddActions(window_actions, XtNumber(window_actions));

	XtAddInput(xfd[0], XtInputReadMask, display, NULL);
/*
 * Realize widgets.
 */
	XtRealizeWidget(wtoplevel);
/*
 * X event loop
 */
	XtMainLoop();
}

/*
 *	setup
 *
 *	Function:	- calculates sizes and draws permanent pixels
 */

static void
setup()

{
	XGCValues	values;			/* GC fields */
	int		i;			/* favourite index */
	int		n;			/* favourite node index */
	int		origin_x;		/* origin of node panel */
	int		origin_y;		/* origin of node panel */
	XSegment	segment[MAXSEGS];	/* panel grid */
	char		*label;			/* mnemonic node ID */

	nodewidth = NVERTSEGS * app_res.ap_ledsize;
	winwidth = (nnodes * nodewidth) + app_res.ap_ledsize;
	winheight = (NHORIZSEGS + 1) * app_res.ap_ledsize;

	win = XCreatePixmap(XtDisplay(wtoplevel),
		RootWindowOfScreen(XtScreen(wtoplevel)),
		winwidth, winheight, 1);

	values.foreground = 1;
	values.background = 0;
	values.line_style = LineSolid;

	draw_gc = XCreateGC(XtDisplay(wtoplevel), win,
		GCForeground | GCBackground | GCLineStyle, &values);

	values.foreground = 0;
	values.background = 1;

	undraw_gc = XCreateGC(XtDisplay(wtoplevel), win,
		GCForeground | GCBackground | GCLineStyle, &values);

	values.foreground = app_res.ap_ledfg;
	values.background = 0;

	copy_gc = XCreateGC(XtDisplay(wtoplevel),
		RootWindowOfScreen(XtScreen(wtoplevel)),
		GCForeground | GCBackground, &values);
/*
 * Initialize pixel map.
 */
	XFillRectangle(XtDisplay(wtoplevel), win, undraw_gc,
		0, 0,
		winwidth,
		winheight);

	origin_y = app_res.ap_ledsize;
/*
 * Draw one LED panel for every node on the command line.
 */
	for (n = 0; n < nnodes; n += 1) {
		origin_x = (n * nodewidth) + app_res.ap_ledsize;

		for (i = 0; i < NHORIZSEGS; i += 1) {
			segment[i].x1 = origin_x;
			segment[i].x2 = origin_x +
				(app_res.ap_ledsize * NHORIZLEDS);
			segment[i].y1 = origin_y + (app_res.ap_ledsize * i);
			segment[i].y2 = origin_y + (app_res.ap_ledsize * i);
		}

		XDrawSegments(XtDisplay(wtoplevel), win, draw_gc, segment,
			NHORIZSEGS);

		for (i = 0; i < NVERTSEGS; i += 1) {
			segment[i].x1 = origin_x + (app_res.ap_ledsize * i);
			segment[i].x2 = origin_x + (app_res.ap_ledsize * i);
			segment[i].y1 = origin_y;
			segment[i].y2 = origin_y +
				(app_res.ap_ledsize * NVERTLEDS);
		}

		XDrawSegments(XtDisplay(wtoplevel), win, draw_gc, segment,
			NVERTSEGS);
/*
 * Label node.
 * The value 4 is a total eyeball job.
 * I can't spend the rest of my life fiddling with ledd.
 */
		if (app_res.ap_ledsize >= NPIXLABEL) {
			label = mnemonic(pnodes[n]);
			XDrawString(XtDisplay(wtoplevel), win, draw_gc,
				origin_x, origin_y - 4, label, strlen(label));
		}
	}
}

/*
 *	redraw
 *
 *	Function:	- redraws exposed portion widget window
 *	Accepts:	- widget handle
 *			- expose event
 */

static void
redraw(w, event)

Widget			w;
XExposeEvent		*event;

{
	int		x, y;
	unsigned int	width, height;

	if (event) {
		x = event->x;
		y = event->y;
		width = event->width;
		height = event->height;
	} else {
		x = 0;
		y = 0;
		width = winwidth;
		height = winheight;
	}

	if (DefaultDepthOfScreen(XtScreen(w)) == 1) {
		XCopyArea(XtDisplay(w), win, XtWindow(w),
			copy_gc, x, y, width, height, x, y);
	} else {
		XCopyPlane(XtDisplay(w), win, XtWindow(w),
			copy_gc, x, y, width, height, x, y, 1);
	}
}

/*
 *	display
 *
 *	Function:	- displays changes in a shadow register
 *			- reads node ID and new shadow register
 *			  from pipe from ledd
 */

static void
display()

{
	int4		nodeid;		/* which LED panel */
	int4		shadow;		/* new shadow register */
	int4		delta;		/* change in old/new shadow */
	int4		i, j;		/* favourite indices */
	int		n;		/* #bytes read */
	XExposeEvent	fake;		/* prerequisite for redraw() */

	n = read(xfd[0], (char *) &nodeid, sizeof(int4));

	if (n < 0) {
		perror(me);
		exit(errno);
	} else if (n < sizeof(int4)) {
		exit(0);
	}

	n = read(xfd[0], (char *) &shadow, sizeof(int4));

	if (n < 0) {
		exit(errno);
	} else if (n < sizeof(int4)) {
		exit(1);
	}
/*
 * Locate shadow register.
 */
	for (i = 0; (i < nnodes) && (pnodes[i] != nodeid); ++i) {
	}

	if (i >= nnodes) {
		return;
	}

	delta = pshadow[i] ^ shadow;
	pshadow[i] = shadow;

	if (delta == 0) {
		return;
	}

	minx = winwidth;
	maxx = 0;
	miny = winheight;
	maxy = 0;

	for (j = 0; j < NLEDS; ++j) {

		if (delta & 1) {
			drawled(i, j, (shadow & 1) ? draw_gc : undraw_gc);
		}

		delta >>= 1;
		shadow >>= 1;
	}

	fake.x = minx;
	fake.y = miny;
	fake.width = maxx - minx;
	fake.height = maxy - miny;
	redraw(wleds, &fake);
}

/*
 *	drawled
 *
 *	Function:	- (un)draws one LED
 *			- leaves a white border between LED and grid
 *	Accepts:	- node index
 *			- LED number
 *			- GC
 */

static void
drawled(inode, iled, gc)

int4			inode;
int			iled;
GC			gc;

{
	int		x, y;		/* origin of LED */
	register int	ledsize;	/* fast temporary */

	ledsize = app_res.ap_ledsize;

	x = (inode * nodewidth) + ledsize + ((iled / NVERTLEDS) * ledsize);
	y = ledsize + ((iled % NVERTLEDS) * ledsize);

	XFillRectangle(XtDisplay(wtoplevel), win, gc,
		x + 2,
		y + 2,
		ledsize - 3,
		ledsize - 3);

	x += 2;
	y += 2;

	if (x < minx) minx = x;
	if (y < miny) miny = y;

	x += ledsize - 3;
	y += ledsize - 3;

	if (x > maxx) maxx = x;
	if (y > maxy) maxy = y;
}

/*
 *	Bail
 *
 *	Function:	- kills ledd parent and exits
 */

static void
Bail()

{
	if (kinit(0) < 0) {
		terror(me);
		exit(errno);
	}

	if (kdoom(parentid, SIGUDIE) < 0) {
		terror(me);
		kexit(errno);
	}

	(void) close(xfd[0]);
	kexit(0);
}
