/*
 *      OSU Research Computing					Trillium
 *      Copyright 1992 The Ohio State University
 *      RBD, GDB
 *
 *      $Log:	xmtv.c,v $
 * Revision 2.2  92/04/24  16:15:33  trillium
 * Ohio Release
 * 
 *	Function:	- OTB graphics plotting tool
 *			- based on X Windows, Athena widget set
 *			- operates as separate process to mtvd
 *			- communicates via a pipe
 *			- becomes a Trollius process only if
 *			  killed by X, in order to kill the mtvd parent
 */

#include <stdio.h>
#include <kio.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <terror.h>
#include <tvreq.h>

#undef NULL
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Viewport.h>

#include <xmtv.h>

/*
 * local variables
 */
static Pixmap		win;			/* drawing window */
static GC		pen;			/* copying pen */
static int		winxsize, winysize;	/* win dimensions */
static int		windepth;		/* win depth */
static int		parentid;		/* pid of mtvd parent */
static int		t2xfd[2];		/* Trollius -> X pipe */
static int		x2tfd[2];		/* X -> Trollius pipe */
static char		inbuf[MAXNMSGLEN];	/* input packet buffer */
static char		*me;			/* command line name */

static int		ncolours;		/* size of colour table */
static unsigned long	*colours;		/* colour table */

static Widget		wtoplevel;
static Widget		wviewport;
static Widget		wdraw;


typedef struct {
	int		ap_xsize;		/* screen horizontal size */
	int		ap_ysize;		/* screen vertical size */
} AppData, *AppDataPtr;

AppData			app_res;		/* application resources */

static XtResource	resources[] = {		/* app resource list */
	{
		XtNmtvXSizeInPixels,
		XtCMtvXSizeInPixels,
		XtRInt,
		sizeof(int),
		XtOffset(AppDataPtr, ap_xsize),
		XtRImmediate,
		(caddr_t) DEFXSIZE
	},
	{
		XtNmtvYSizeInPixels,
		XtCMtvYSizeInPixels,
		XtRInt,
		sizeof(int),
		XtOffset(AppDataPtr, ap_ysize),
		XtRImmediate,
		(caddr_t) DEFYSIZE,
	},
};

static XrmOptionDescRec	options[] = {		/* command line options */
	{"-xsize",	"*mtvXSizeInPixels",	XrmoptionSepArg,	NULL},
	{"-ysize",	"*mtvYSizeInPixels",	XrmoptionSepArg,	NULL},
};

/*
 * external functions
 */
extern char		*malloc();
extern void		exit();

/*
 * local functions
 */
static void		Bail();
static void		do_clr();
static void		do_rect();
static void		do_snake();
static void		do_state();
static void		do_text();
static void		expose_rect();
static void		plotchar();
static void		readback();
static void		readpipe();
static void		redraw();
static void		setpen();
static void		setup();
static void		setup_colour();
static void		trequest();
static void		writeback();
static void		writepipe();
static void		xmtv();
static unsigned long	name2colour();

/*
 * global functions
 */
int			tv_init();
int			tv_clr();
int			tv_rect();
int			tv_snake();
int			tv_text();

/*
 * external variables
 */
extern int		errno;
extern struct kio_t	_kio;

/*
 *	tv_init
 *
 *	Function:	- initializes mtvd graphics interface
 *			- X Window version
 *			- forks a new process for the X client
 *
 *	Accepts:	- argc, argv to pass to X child
 *
 *	Returns:	- 0 or aborts
 */
int
tv_init(argc, argv)

int			argc;
char			*argv[];

{
	int		pid;

	parentid = getpid();

	if ((pipe(t2xfd) < 0) || (pipe(x2tfd) < 0)) {
		perror(argv[0]);
		exit(errno);
	}

	if ((pid = fork()) < 0) {
		perror(argv[0]);
		exit(errno);
    	}
	else if (pid == 0) {		/* child */
		(void) close(t2xfd[1]);
		(void) close(x2tfd[0]);
		xmtv(argc, argv);
    	}

	else {
		(void) close(t2xfd[0]);
		(void) close(x2tfd[1]);
	}

	return(0);
}

/*
 *	tv_clr
 *
 *	Function:	- clear entire screen to black
 *			- xmtv version
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_clr()

{
	int		req;

	req = TVQCLR;
	writepipe((char *) &req, sizeof(int));
	return(0);
}

/*
 *	tv_rect
 *
 *	Function:	- plot a multicoloured rectangle
 *			- xmtv version
 *
 *	Accepts:	- upper left corner x-ordinate
 *			- upper left corner y-ordinate
 *			- rectangle horizontal size (width)
 *			- rectangle vertical size (height)
 *			- array of pixel colours
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_rect(x, y, xsize, ysize, pixcol)

int4			x,
			y,
			xsize,
			ysize;
char			*pixcol;

{
	int		req;

	req = TVQRECT;
	writepipe((char *) &req, sizeof(int));
	writepipe((char *) &x, sizeof(int4));
	writepipe((char *) &y, sizeof(int4));
	writepipe((char *) &xsize, sizeof(int4));
	writepipe((char *) &ysize, sizeof(int4));
	if (xsize * ysize > 0) {
		writepipe(pixcol, xsize * ysize);
	}
	return(0);
}

/*
 *	tv_snake
 *
 *	Function:	- plot a multicoloured snake-like region
 *			- xmtv version
 *
 *	Accepts:	- starting point ordinate
 *			- length of snake
 *			- array of pixel colours
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_snake(x, len, pixcol)

int4			x,
			len;
char			*pixcol;

{
	int		req;

	req = TVQSNAKE;
	writepipe((char *) &req, sizeof(int));
	writepipe((char *) &x, sizeof(int4));
	writepipe((char *) &len, sizeof(int4));
	if (len > 0) {
		writepipe(pixcol, len);
	}
	return(0);
}

/*
 *	tv_text
 *
 *	Function:	- plot a text string
 *			- xmtv version
 *
 *	Accepts:	- horizontal text ordinate of first letter
 *			- vertical text ordinate of first letter
 *			- string colour
 *			- string
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_text(x1, y1, colour, mesg)

int4			x1,
			y1,
			colour;
char			*mesg;

{
	int		req, len;

	req = TVQTEXT;
	writepipe((char *) &req, sizeof(int));
	writepipe((char *) &x1, sizeof(int4));
	writepipe((char *) &y1, sizeof(int4));
	writepipe((char *) &colour, sizeof(int4));
	len = strlen(mesg);
	writepipe((char *) &len, sizeof(int));
	if (len > 0) {
		writepipe(mesg, len);
	}
	return(0);
}

/*
 *	tv_state
 *
 *	Function:	- get mtvd state information
 *
 *	Accepts:	- ptr to state structure
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_state(status)

struct tvstate		*status;

{
	int		req;

	req = TVQSTATE;
	writepipe((char *) &req, sizeof(int));

	readback((char *) &(status->tvs_xlen), sizeof(int4));
	readback((char *) &(status->tvs_ylen), sizeof(int4));
	readback((char *) &(status->tvs_ncolour), sizeof(int4));
	readback((char *) &(status->tvs_xtext), sizeof(int4));
	readback((char *) &(status->tvs_ytext), sizeof(int4));

	return(0);
}

/*
 *	xmtv
 *
 *	Function:	- mainline of mtv X Window process
 *			- initializes and then enters event loop
 *
 *	Accepts:	- argc, argv from command line
 */
static void
xmtv(argc, argv)

int			argc;
char			*argv[];

{
	Arg		args[8];	/* resource arguments */
	int		i;		/* favourite index */
	XGCValues	gcval;		/* GC value fields */

	String		trans =		/* event translations */
		"<Expose>:	redraw()";

	static XtActionsRec		/* action table */
			window_actions[] = {
		{"redraw", redraw},
	};
/*
 * Initialization block.
 */
	me = argv[0];

	wtoplevel = XtInitialize(
		me,			/* not really used */
		"XMtvd",		/* application class */
		options,		/* resource mgr options */
		XtNumber(options),	/* # of RM options */
		&argc,			/* # of args */
		argv			/* command line */
	);

	if (argc > 1) {
		errno = EUSAGE;
		terror(argv[0]);
		exit(errno);
	}

	XtGetApplicationResources(wtoplevel,
		&app_res,
		resources,
		XtNumber(resources),
		NULL,
		0);

	if ((app_res.ap_xsize <= 0) || (app_res.ap_ysize <= 0)) {
		errno = EINVAL;
		terror(me);
		exit(errno);
	}

	winxsize = app_res.ap_xsize;
	winysize = app_res.ap_ysize;
/*
 * Create widgets.
 */
	i = 0;
	XtSetArg(args[i], XtNallowHoriz, TRUE);   i++;
	XtSetArg(args[i], XtNallowVert, TRUE);   i++;
	XtSetArg(args[i], XtNuseBottom, TRUE);   i++;
	XtSetArg(args[i], XtNuseRight, TRUE);   i++;
	XtSetArg(args[i], XtNwidth,
		(winxsize > DEFXSIZE) ? DEFXSIZE : winxsize);   i++;
	XtSetArg(args[i], XtNheight,
		(winysize > DEFYSIZE) ? DEFYSIZE : winysize);   i++;

	wviewport = XtCreateManagedWidget(
		"viewport",		/* instance name */
		viewportWidgetClass,	/* widget class */
		wtoplevel,		/* parent */
		args,			/* resources */
		i			/* # of resources */
	);

	i = 0;
	XtSetArg(args[i], XtNtranslations,
		XtParseTranslationTable(trans));   i++;
	XtSetArg(args[i], XtNwidth, winxsize);   i++;
	XtSetArg(args[i], XtNheight, winysize);   i++;

	wdraw = XtCreateManagedWidget(
		"mtv",		
		widgetClass,
		wviewport,
		args,	
		i
	);
/*
 * Create pixmap window and graphic pen (context).
 */
	win = XCreatePixmap(XtDisplay(wdraw),
		RootWindowOfScreen(XtScreen(wdraw)),
		winxsize, winysize,
		DefaultDepthOfScreen(XtScreen(wdraw)));

	gcval.foreground = BlackPixelOfScreen(XtScreen(wviewport));
	pen = XCreateGC(XtDisplay(wviewport),
		RootWindowOfScreen(XtScreen(wviewport)),
		GCForeground, &gcval);
/*
 * Set up the colour table and clear the screen to black.
 */
	setup_colour();
	XFillRectangle(XtDisplay(wdraw), win, pen, 0, 0, winxsize, winysize);
/*
 * Hook up the exit handler, window actions and input pipe.
 */
	XSetIOErrorHandler(Bail);
	XtAddActions(window_actions, XtNumber(window_actions));
	XtAddInput(t2xfd[0], XtInputReadMask, trequest, NULL);
/*
 * Realize widgets.
 */
	XtRealizeWidget(wtoplevel);
/*
 * X event loop
 */
	XtMainLoop();
}

/*
 *	setup_colour
 *
 *	Function:	- allocate and set up the colour table
 *			- different screen depths are handled
 *			  in a reasonably portable manner
 *			- maximum of 256 colours allowed
 */
static void
setup_colour()

{
	int		i;		/* favorite counter */
/*
 * Check the bounds on allowed window depth.
 */
	windepth = DefaultDepthOfScreen(XtScreen(wdraw));
	if (windepth > MAXWINDEPTH) {
		windepth = MAXWINDEPTH;
	}
/*
 * Allocate the colour lookup table.
 */
	ncolours = (1 << windepth);
	colours = (unsigned long *)
		malloc((unsigned) ncolours * sizeof(unsigned long));
	if (colours == (unsigned long *) 0) {
		fprintf(stderr, "%s: cannot malloc colour table\n", me);
		exit(errno);
	}

	if (windepth == 1) {
/*
 * Monochrome mode.
 */
		colours[0] = name2colour("black");
		colours[1] = name2colour("white");
	}

	else if (windepth == 2) {
/*
 * Define the four colours: black, blue, green and red.
 */
		colours[0] = name2colour("black");
		colours[1] = name2colour("blue");
		colours[2] = name2colour("green");
		colours[3] = name2colour("red");
	}

	else {
/*
 * Use list of colours to set up the colour table.
 */
		for (i = 0; i < ncolours; i++) {
			colours[i] = name2colour(colour_list[i]);
		}
		colours[ncolours - 1] = name2colour("white");
	}
}

/*
 *	expose_rect
 *
 *	Function:	- expose specified screen region (rectangle)
 *			- wrapper for the redraw function
 *
 *	Accepts:	- upper left corner coordinates
 *			- rectangle width
 *			- rectangle height
 */
static void
expose_rect(x, y, width, height)

int			x,
			y,
			width,
			height;

{
	XExposeEvent	xev;

	xev.x = x;
	xev.y = y;
	xev.width = width;
	xev.height = height;

	redraw(wdraw, &xev);
}

/*
 *	redraw
 *
 *	Function:	- redraws exposed portion of widget window
 *
 *	Accepts:	- widget handle
 *			- expose event
 */
static void
redraw(w, event)

Widget			w;
XExposeEvent		*event;

{
	int		x, y;
	unsigned int	width, height;

	if (event) {
		x = event->x;
		y = event->y;
		width = event->width;
		height = event->height;
	} else {
		x = 0;
		y = 0;
		width = winxsize;
		height = winysize;
	}

	XCopyArea(XtDisplay(w), win, XtWindow(w), pen,
		x, y, width, height, x, y);
}

/*
 *	trequest
 *
 *	Function:	- handle requests from the Trollius side
 *			- reads graphics requests from pipe from mtvd
 *			  and invoke the requested service
 */
static void
trequest()

{
	int		reqcode;	/* mtvd request code */
/*
 * Get the request code and invoke the requested service.
 */
	readpipe((char *) &reqcode, sizeof(int));

	switch(reqcode) {
		case TVQCLR:
			do_clr();
			break;
		case TVQRECT:
			do_rect();
			break;
		case TVQSNAKE:
			do_snake();
			break;
		case TVQTEXT:
			do_text();
			break;
		case TVQSTATE:
			do_state();
			break;
		default:
			break;
	}
}

/*
 *	do_clr
 *
 *	Function:	- clear the screen to black.
 */
static void
do_clr()

{
	setpen(0);
	XFillRectangle(XtDisplay(wdraw), win, pen, 0, 0, winxsize, winysize);
	expose_rect(0, 0, winxsize, winysize);
}

/*
 *	do_rect
 *
 *	Function:	- plot a multicoloured rectangle
 *			- handle boundary clipping
 */
static void
do_rect()

{
	int4		x1, y1;		/* upper right corner coordinates */
	int4		xlen, ylen;	/* rectangle width and height */
	int4		xmax, ymax;	/* upper rectangle bounds */
	int4		xmin, ymin;	/* lower rectangle bounds */
	int4		x, y;		/* favorite counters */
	char		*p;		/* favorite pointer */
/*
 * Get the rectangle information.
 */
	readpipe((char *) &x1, sizeof(int4));
	readpipe((char *) &y1, sizeof(int4));
	readpipe((char *) &xlen, sizeof(int4));
	readpipe((char *) &ylen, sizeof(int4));
	if (xlen * ylen > 0) {
		readpipe(inbuf, xlen * ylen);
	} else {
		return;
	}
/*
 * Plot the rectangle.
 */
	xmax = x1 + xlen - 1;
	ymax = y1 + ylen - 1;

	if ((x1 >= 0) && (y1 >= 0) && (xmax < winxsize) && (ymax < winysize)) {
/*
 * If the full rectangle is within the screen, plot it with no pixel checks.
 */
		p = inbuf;
		for (y = y1; y <= ymax; y++) {
			for (x = x1; x <= xmax; x++) {
				setpen((int) *(p++));
				XDrawPoint(XtDisplay(wdraw), win, pen, x, y);
			}
		}
		expose_rect(x1, y1, xlen, ylen);
	}
	else if ((x1 >= winxsize) || (y1 >= winysize) ||
		(xmax < 0) || (ymax < 0)) {
/*
 * Do nothing if the full rectangle is outside the screen.
 */
	}
	else {
/*
 * The rectangle has a portion outside the screen, check each pixel.
 */
		for (y = y1; y <= ymax; y++) {
			for (x = x1; x <= xmax; x++) {
				if ((x >= 0) && (y >= 0) &&
					(x < winxsize) && (y < winysize)) {

					p = inbuf + ((y - y1) * xlen) + x - x1;
					setpen((int) *(p++));
					XDrawPoint(XtDisplay(wdraw), win,
						pen, x, y);
				}
			}
		}
		xmin = (x1 < 0) ? 0 : x1;
		xmax = (x1 + xlen > winxsize) ? winxsize : x1 + xlen;
		ymin = (y1 < 0) ? 0 : y1;
		ymax = (y1 + ylen > winysize) ? winysize : y1 + ylen;
		expose_rect(xmin, ymin, xmax - xmin, ymax - ymin);
	}
}

/*
 *	do_snake
 *
 *	Function:	- plot a multicoloured snake-like region
 *			- handle boundary clipping
 */
static void
do_snake()

{
	int		i;		/* favorite counter */
	int4		x, y;		/* point coordinates */
	int		ymin;		/* minimum vertical ordinate */
	int4		len;		/* snake length */
	char		*p;		/* favorite pointer */
/*
 * Get the snake information.
 */
	readpipe((char *) &x, sizeof(int4));
	readpipe((char *) &len, sizeof(int4));
	if (len > 0) {
		readpipe(inbuf, len);
	}
	p = inbuf;
/*
 * Do nothing if the entire snake is outside the screen.
 */
	if (x >= (winxsize * winysize)) {
		return;
	}
/*
 * Clip the head of the snake if it is outside the screen.
 */
	if (x < 0) {
		p -= x;
		len += x;
		x = 0;
	}
/*
 * Clip the tail of the snake if it is outside the screen.
 */
	if ((x + len) > (winxsize * winysize)) {
		len = (winxsize * winysize) - x;
	}
/*
 * Plot the snake.
 */
	ymin = y = x / winxsize;
	x = x % winxsize;
	for (i = 0; i < len; i++) {
		setpen((int) *(p++));
		XDrawPoint(XtDisplay(wdraw), win, pen, x, y);
		if (++x >= winxsize) {
			x = 0;
			y++;
		}
	}

	expose_rect(0, ymin, winxsize, y - ymin + 1);
}

/*
 *	do_text
 *
 *	Function:	- plot a text string
 *			- handles boundary clipping
 */
static void
do_text()

{
	int4		x, y;		/* first letter coordinates */
	int4		colour;		/* string colour */
	int		len;		/* string length */
	int		xmin;		/* smallest horizontal ordinate */
	char		*p;		/* favorite pointer */

	readpipe((char *) &x, sizeof(int4));
	readpipe((char *) &y, sizeof(int4));
	readpipe((char *) &colour, sizeof(int4));
	readpipe((char *) &len, sizeof(int));
	if (len > 0) {
		readpipe(inbuf, len);
	}
	p = inbuf;
/*
 * If the string is outside the screen, do nothing.
 */
	if ((x >= winxsize/MTVXCHAR) || (y >= winysize/MTVYCHAR) ||
		(x + len <= 0) || (y < 0)) {

		return;
	}
/*
 * Clip the head of the string if needed.
 */
	if (x < 0) {
		p -= x;
		len += x;
		x = 0;
	}
/*
 * Clip the tail of the string if needed.
 */
	if (x + len > winxsize/MTVXCHAR) {
		len = winxsize/MTVXCHAR - x;
		p[len] = '\0';
	}
	xmin = x;
/*
 * Plot the string.
 */
	for ( ; *p; p++, x++) {
		plotchar(x, y, colour, *p);
	}

	expose_rect(xmin * MTVXCHAR, y * MTVYCHAR, len * MTVXCHAR, MTVYCHAR);
}

/*
 *	do_state
 *
 *	Function:	- return state information to the Trollius side
 *			- send the data in the back pipe x2tfd
 */
static void
do_state()

{
	int4		reply;

	reply = winxsize;
	writeback((char *) &reply, sizeof(int4));
	reply = winysize;
	writeback((char *) &reply, sizeof(int4));
	reply = ncolours;
	writeback((char *) &reply, sizeof(int4));
	reply = winxsize / MTVXCHAR;
	writeback((char *) &reply, sizeof(int4));
	reply = winysize / MTVYCHAR;
	writeback((char *) &reply, sizeof(int4));
}

/*
 *	Bail
 *
 *	Function:	- kills mtvd parent and exits
 */
static void
Bail()

{
	if (kinit(0) < 0) {
		terror(me);
		exit(errno);
	}

	if (kdoom(parentid, SIGUDIE) < 0) {
		terror(me);
		kexit(errno);
	}

	(void) close(t2xfd[0]);
	(void) close(x2tfd[1]);

	kexit(errno);
}

/*
 *	name2colour
 *
 *	Function:	- allocate given colour name
 *			- get colour pixel value
 *			- assign white in case of error
 *
 *	Accepts:	- colour name
 *
 *	Returns:	- pixel value
 */
static unsigned long
name2colour(colour)

char			*colour;

{
	XColor		col_def,		/* colour definition */
			dev_null;		/* unused functionality */

	if (! XAllocNamedColor(XtDisplay(wdraw),
		DefaultColormapOfScreen(XtScreen(wdraw)),
		colour, &col_def, &dev_null)) {

		return(name2colour("white"));
	}
	else {
		return(col_def.pixel);
	}
}

/*
 *	setpen
 *
 *	Function:	- set the pen to the given colour
 *
 *	Accepts:	- colour table index
 */
static void
setpen(colour)

int			colour;

{
	XSetForeground(XtDisplay(wdraw), pen, colours[colour]);
}

/*
 *	plotchar
 *
 *	Function:	- plot a character on the screen
 *
 *	Accepts:	- horizontal character position
 *			- vertical character position
 *			- colour
 *			- character code
 */
static void
plotchar(xc, yc, col, char_code)

int4			xc,
			yc,
			col;
char			char_code;

{
	int		x, y;		/* char. pixel coordinates */
	int		i;		/* favorite counter */
	unsigned char	row;		/* char. row pixel map */
	unsigned char	*p;		/* ptr character map */
/*
 * Do nothing if the character is not valid.
 */
	i = (int) char_code;
	if ((i < 0) || (i > MTVNCHAR)) {
		return;
	}
/*
 * Loop over the rows of the character to be plotted.
 */
	setpen(col);
	p = charset[i].cm_rows;
	y = yc * MTVYCHAR;
	for (i = 0; i < MTVYCHAR; i++, p++, y++) {
		x = xc * MTVXCHAR;
		for (row = *p; row != 0; row <<= 1, x++) {
			if (row & 0x80) {
				XDrawPoint(XtDisplay(wdraw), win, pen, x, y);
			}
		}
	}
}

/*
 *	readpipe
 *
 *	Function:	- read a number of bytes from the t2xfd pipe
 *			- place the bytes in the given buffer
 *			- exit if error or pipe is closed
 *
 *	Accepts:	- ptr to buffer
 *			- number of bytes
 */
static void
readpipe(buf, nbytes)

char			*buf;
int			nbytes;

{
	int		n;

	n = read(t2xfd[0], buf, nbytes);

	if (n < 0) {
		perror(me);
		exit(errno);
	}
	else if (n < nbytes) {
		exit(0);
	}
}

/*
 *	writepipe
 *
 *	Function:	- write a number of bytes to the t2xfd pipe
 *			- exit if error or pipe is closed
 *
 *	Accepts:	- ptr to buffer
 *			- number of bytes
 */
static void
writepipe(buf, nbytes)

char			*buf;
int			nbytes;

{
	int		n;

	n = write(t2xfd[1], buf, nbytes);

	if (n < 0) {
		terror(me);
		kexit(errno);
	}
}

/*
 *	readback
 *
 *	Function:	- read a number of bytes from the x2tfd pipe
 *			- place the bytes in the given buffer
 *			- exit if error or pipe is closed
 *
 *	Accepts:	- ptr to buffer
 *			- number of bytes
 */
static void
readback(buf, nbytes)

char			*buf;
int			nbytes;

{
	int		n;

	n = read(x2tfd[0], buf, nbytes);

	if (n < 0) {
		perror(me);
		kexit(errno);
	}
	else if (n < nbytes) {
		kexit(0);
	}
}

/*
 *	writeback
 *
 *	Function:	- write a number of bytes to the x2tfd pipe
 *			- exit if error or pipe is closed
 *
 *	Accepts:	- ptr to buffer
 *			- number of bytes
 */
static void
writeback(buf, nbytes)

char			*buf;
int			nbytes;

{
	int		n;

	n = write(x2tfd[1], buf, nbytes);

	if (n < 0) {
		perror(me);
		exit(errno);
	}
}
