/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	ttrace.c,v $
 * Revision 2.2  92/04/24  16:28:39  trillium
 * Ohio Release
 * 
 *	Function:	- control and unload traced
 *	Accepts:	- see the "usage" variable
 *	Examples:	- ttrace n10
 *			  unloads all trace data on node 10
 *			- ttrace n10 -k
 *			  unloads but does not forget trace data on node 10
 *			- ttrace n10 -s
 *			  forgets all trace data on node 10
 *			- ttrace n10 -g
 *			  tells traced to prompt its clients with a signal
 */

#include <fcntl.h>
#include <portable.h>
#include <stdio.h>
#include <terror.h>
#include <trreq.h>
#include <typical.h>

/*
 * external variables
 */
extern int 	errno;		/* global error variable */

/*
 * external functions
 */
extern void	validopts();	/* set the valid options */
extern void	terror();
extern char	*mnemonic();
extern int	do_args();	/* parse the valid options */
extern int	getnode();	/* get next node entry */
extern int	opt_taken();	/* Was a particular option selected? */
extern int	parse_nodes();	/* parse the node identifiers */
extern int	rtrsignal();	/* signal clients */
extern int	rtrsweep();	/* forget only */
extern int4	rtrfforget();	/* unload and forget */
extern int4	rtrfget();	/* unload only */

/*
 * local functions
 */
static void	help();

/*
 * local variables
 */
static char *_copyright =
	"Copyright 1991 The Ohio State University $Revision: 2.2 $";
/*
 * syntax for ttrace command
 */
static char *usage = "ttrace [-ghksv] [<filename>] <nodes>";

main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int4		n_index;		/* index of node entry */
	int4		nodeid;			/* value of node entry */
	int4		n_flags;		/* extra node entry info */
	int		fd;			/* output file descriptor */
/*
 * Parse the command line.
 */
	validopts("ghksv");
	exclusive("ghks");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		} else {
			terror("ttrace (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Open file if not -s or -g.
 */
	if (!opt_taken('s') && !opt_taken('g')) {

		if (argc != 2) {
			fprintf(stderr, "usage: %s\n", usage);
			exit(EUSAGE);
		}

		fd = open(argv[1], O_WRONLY | O_CREAT | O_APPEND, 0644);

		if (fd < 0) {
			terror("ttrace (open)");
			exit(errno);
		}
	}
/*
 * Attach to kernel.
 */
	if (kinit(0) < 0) {
		terror("ttrace (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	(void) getnode(&n_index, &nodeid, &n_flags);

	if (n_index >= 0) {

		do {
			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(errno);
			}

			if (opt_taken('s')) {

				if (rtrsweep(nodeid) < 0) {
					terror("ttrace (rtrsweep)");
					kexit(errno);
				}
			}

			else if (opt_taken('g')) {

				if (rtrsig(nodeid) < 0) {
					terror("ttrace (rtrsig)");
					kexit(errno);
				}
			}

			else if (opt_taken('k')) {
				if (opt_taken('v')) printf(
				"ttrace: unload (keep) node %s ...\n",
				mnemonic(nodeid));

				if (rtrfget(nodeid, fd) < 0) {
					terror("ttrace (rtrfget)");
					kexit(errno);
				}
			}

			else {
				if (opt_taken('v')) printf(
				"ttrace: unload node %s ...\n",
				mnemonic(nodeid));

				if (rtrfforget(nodeid, fd) < 0) {
					terror("ttrace (rtrfforget)");
					kexit(errno);
				}
			}

			(void) getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	} else {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	kexit(0);
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the ttrace command
 */
static void
help()
{
	printf("\nSynopsis:	ttrace [options] <nodes> [<filename>]\n");
	printf("\nDescription:	Unload or control remote traced.\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\t\t-g	signal traced clients\n");
	printf("\t\t-k	unload, but maintain traces\n");
	printf("\t\t-s	sweep clean traces\n");
	printf("\t\t-v	operate verbosely\n");
	mnusage();
	printf("\t\th\t(local host node)\n");
	printf("\nExample:	ttrace n3-9,11\n");
	printf("\t\t\t\"Unload trace data from nodes 3 thru 9 & 11.\"\n\n");
}
