/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	brendlib.c,v $
 * Revision 2.2  92/04/24  16:32:45  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:14:10  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:37:05  gdburns
 * Homo sapien
 * 
 *
 *	Function:	- Implementation of BRENDA via Trollius 2.0
 *			- the library functions
 */

#include "brenda.h"

#include <castreq.h>
#include <t_types.h>
#include <terror.h>

#ifdef DEBUG
#include <stdio.h>
#endif

#ifdef DEBUG
PRIVATE int
error (pt, hd)
	int pt;
	struct nmsg *hd;
{
	printf ("brendalib: runtime error %d at point %d\n", errno, pt);
	if (hd == 0)  return (B_ERROR);
	printf ("last message parameters:\n");
	printf ("event=0x%x, type=0x%x, flags=0x%x, buf=0x%x, length=%d\n",
	   hd->nh_event, hd->nh_type, hd->nh_flags, hd->nh_msg, hd->nh_length);
	return (B_ERROR);
}
#else
PRIVATE int
error (pt, hd)
	int pt;
	struct nmsg *hd;
{
	return (B_ERROR);
}
#endif

/*
 *  The function to calculate storage node.
 *  First time this is called it gets the list
 *  of (ITB) nodes from the router deamon.
 */
PUBLIC long
hashnode (tag)
	long tag;
{
	static int standalone = 0;
	static int done_getnet = 0;
	static int nnodes;			/* number of storage nodes */
	static long *nodetable;
	static long smalltable[NODETABLESIZE];
	static long bcastid = NOTNODEID;
	int r;
	long id;
	unsigned int amount;

	if (standalone)			/* stand-alone Trollius */
		return (LOCAL);		/* it's ALL local */

	if (! done_getnet) {
		nnodes = getngrid();		/* use only GRID nodes */
		if (nnodes < 1) {
			standalone = 1;		/* host only */
			return (LOCAL);
		}
		if (nnodes > NODETABLESIZE) {
			amount = nnodes * sizeof(long);
			nodetable = (long *) rmalloc (LOCAL, amount, 0, 0);
			if (nodetable == NULL)
				return (NOTNODEID);
		} else {
			nodetable = smalltable;
		}
		if (getgnodes (nodetable, nnodes) != 0)
			return (NOTNODEID);
		done_getnet = 1;
	}

	if (tag == ANYTAG) {			/* bcast ID wanted */
		if (bcastid != NOTNODEID)
			return (bcastid);	/* not first call */
		if (getnodeid() == getorigin()) {
			bcastid = HOST2GRID;
		} else {
			/* pick ID for group (hopefully unique): */
			id = BRENDA + getnodeid();
			/* register the bcast with Trollius: */
			r = rcast (LOCAL,id,CBN_ERR,nodetable,nnodes);
			if (r != 0)
				return (NOTNODEID);
			bcastid = id;
		}
		return (bcastid);
	}

	/* normal non-broadcast server call: hash on tag */
	return (nodetable [RANDTAG(tag) % nnodes]);
}


/*
 *	The common Brenda function, for the user calls
 *		(PUT, GRAB, LOOK, etc).
 *
 *	Always sends out a request, unless parameters invalid.
 *	For a small, unverified, PUT, that's all.
 *	Otherwise an ack and/or an object is then received.
 *	In the latter case "buf" holds new data upon return.
 *	This function returns B_OK or B_ERROR.
 */
PUBLIC int
brenda (request, tag, buffer, size, flags, snode)
	int	request, flags;
	long	tag, size, snode;
	char	*buffer;
{
	long	data, msize, node, event, type;
	int	s, isput, small, global;
	char	*msg;
	struct	nmsg nhead;
	struct	brenda_req *req;

	if (request > B_BUMP) {
		errno = EBADREQUEST;
		return (error(10,NULLP));
	}

	if (request != B_BUMP && buffer != NULL && size < sizeof(long)) {
		errno = EMSGLEN;
		return (error(15,NULLP));
	}

	if (tag == ANYTAG)
		flags |= B_SETTLED;

	isput = (request == B_PUT);

	if (isput) {
		if (buffer == NULL) {
			data = size;
			buffer = (char *) &data;
			size = sizeof(long);
		}
		small = (size <= MAXNMSGLEN && (flags & B_VERIFY) == 0);
		if (! small)
			request = _LPUT_;
	}

	global = (flags & B_GLOBAL);

	if (global) {
		flags |= B_SETTLED;
		if (isput) {
			snode = hashnode(ANYTAG);	/* broadcast */
			if (small)
				request = _GPUT_;
			else
				request = _LGPUT_;
		} else if (request == B_GRAB) {
			snode = hashnode(ANYTAG);	/* broadcast */
			request = _GGRAB_;
		} else if (request == B_LOOK || request == B_TRYLOOK) {
			snode = LOCAL;
		} else {
			errno = EBADREQUEST;
			return (error(20,NULLP));
		}
		if (snode == NOTNODEID)
			return (error(22,NULLP));
	}

	if (flags & B_LOCAL) {
		flags |= B_SETTLED;
		snode = LOCAL;
	}

	req = (struct brenda_req *) nhead.nh_data;

	req->node    = getnodeid();
	req->event   = getpid();
	req->request = request;
	req->tag     = tag;
	req->size    = size;

	if (isput && small) {
		if (size == sizeof(long)) {
			req->data = *((long *) buffer);
			msg = NULLP;
			msize = 0;
		} else {
			msg = buffer;
			msize = size;
		}
	} else {
		if (isput && size == sizeof(long))	/* VERIFY-ed int put */
			req->data = *((long *) buffer);
		msg = NULLP;
		msize = 0;
	}

	if ((flags & B_SETTLED) == 0) {
		snode = hashnode(tag);
		if (snode == NOTNODEID)
			return (error(25,NULLP));
	}

	if (flags & B_TORCH)
		type = (_FROZEN_ | _REQ_);	/* cut through the ice */
	else
		type = _REQ_;

	nhead.nh_node   = snode;
	nhead.nh_event  = BRENDA;
	nhead.nh_type   = type;
	nhead.nh_flags  = DINTDATA;		/* do ltot() on nh_data */
	nhead.nh_msg    = msg;
	nhead.nh_length = msize;
/*0*/	if (nsend (&nhead) != 0)
		return (error(30,&nhead));

	if (isput) {					/*** it's a put ***/

		if (small)				/* && unverified */
			return (B_OK);			/* assume OK */

		/* get ack message */

		if (global) {
			/* -- no ACK message from deamon */
			/* determine node and event to send data to: */
			node = snode;			/* broadcast again */
			event = (getnodeid() ^ getpid());
		} else {
			nhead.nh_event  = getpid();
			nhead.nh_type   = _REPLY_;
			nhead.nh_flags  = DINTDATA;
			nhead.nh_msg    = NULLP;
			nhead.nh_length = 0;
/*2*/			if (nrecv (&nhead) != 0)
				return (error(40,&nhead));
			if (req->request != _ACK_) {
				errno = EIMPOSSIBLE;
				return (error(50,&nhead));
			}
			s = req->data;
			if (s != 0) {		/* not-ACK */
				errno = s;
				return (error(60,&nhead));
			}
			node  = req->node;
			event = req->event;
		}

		/* send the data if separate */

		if (size == sizeof(long))
			return (B_OK);		/* data already sent */

		if (! global && size > req->size)
			size = req->size;	/* send only what is wanted */

		nhead.nh_node   = node;
		nhead.nh_event  = event;
		nhead.nh_flags  = DRAWDATA;
		nhead.nh_type   = _DATA_;
		nhead.nh_msg    = buffer;
		nhead.nh_length = size;
/*3*/		if (nsend (&nhead) != 0)
			return (error(70,&nhead));

		return (B_OK);		/* assume OK since we got ack above */
	}

	/*** it's not a put ***/

	if (request == B_BUMP)		/* size held the increment */
		size = sizeof(long);	/* we now receive an int */

	if (size == sizeof(long)) {
		msg = NULLP;
		msize = 0;
	} else {
		msg = buffer;
		msize = (msg? size : 0);
	}

	nhead.nh_event  = getpid();
	nhead.nh_type   = _REPLY_;
	nhead.nh_flags  = DINTDATA;
	nhead.nh_msg    = msg;
	nhead.nh_length = msize;
/*1*/	if (nrecv (&nhead) != 0)	/* this may block for a long time */
		return (error(80,&nhead));

	request = req->request;

	if (request == _FOUND_) {	/* a small object received */

		if (size == sizeof(long) && buffer != NULL)
			*((long *) buffer) = req->data;

		return (B_OK);
	}

	if (request == _LFOUND_) {		/* a long object coming */

		if (req->size < size)		/* not _that_ long */
			size = req->size;

		/* receive the data */
		nhead.nh_event  = getpid();
		nhead.nh_type   = _DATA_;
		nhead.nh_flags  = DRAWDATA;
		nhead.nh_msg    = buffer;
		nhead.nh_length = size;
/*4*/		if (nrecv (&nhead) != 0)
			return (error(90,&nhead));

		return (B_OK);
	}

	/** other requests are errors or EWOULDBLOCK **/

	s = (int) req->data;
	if (s)
		errno = s;
	else
		errno = EIMPOSSIBLE;
	return (error(100,NULLP));
}


/*
 *	The freeze/thaw functions.
 *
 *	These are accepted even when Brenda is frozen.
 *	Consecutive identical calls have no effect,
 *	and no requests get stuck in buffers either.
 *
 *	The argument must be _FREEZE_ or _THAW_.
 *	If not, returns B_ERROR.  Returns B_OK upon success.
 *	But note that there is no reply from the Brenda
 *	deamons.  That is because we broadcast to all of them.
 */
PRIVATE int
b_freeze (action)
	int	action;
{
	long	node;
	struct	nmsg nhead;
	struct	brenda_req *req;

	if (action != _FREEZE_ && action != _THAW_)
		return (B_ERROR);

	req = (struct brenda_req *) nhead.nh_data;

	req->request = action;

	node = hashnode(ANYTAG);		/* broadcast */
	if (node == NOTNODEID)
		return (error(105,NULLP));

	nhead.nh_node   = node;
	nhead.nh_event  = BRENDA;
	nhead.nh_type   = (_FROZEN_ | _REQ_);
	nhead.nh_flags  = DINTDATA;
	nhead.nh_msg    = NULLP;
	nhead.nh_length = 0;
/*0*/	if (nsend (&nhead) != 0)
		return (error(107,&nhead));

	return (B_OK);				/* no reply message */
}

PUBLIC int
freeze()
{
	return (b_freeze (_FREEZE_));
}

PUBLIC int
thaw()
{
	return (b_freeze (_THAW_));
}


/*
 *	Brenda function for non-user (debugging) requests.
 *	These are accepted even when Brenda is frozen.
 *
 *	This function returns B_OK or B_ERROR.
 *	It also modifies the tag and size fields in the header.
 *	The size field is set to the number of found objects (>=0).
 */
PUBLIC int
_brenda (argblk)
	struct brenda_pb *argblk;
{
	int	s;
	long	reqtype, tag, size, flags, msize, snode;
	long	*ip;
	char	*msg;
	struct	nmsg nhead;
	struct	brenda_req *req;

	reqtype = argblk->request;
	tag     = argblk->tag;
	size    = argblk->size;
	flags   = argblk->flags;
	snode   = argblk->snode;

	if (reqtype < _OBJ_SUM) {
		errno = EBADREQUEST;
		return (error(110,NULLP));
	}

	if (flags & B_GLOBAL) {
		errno = EBADREQUEST;
		return (error(120,NULLP));
	}

	if (tag == ANYTAG)
		flags |= B_SETTLED;

	if (flags & B_LOCAL) {
		flags |= B_SETTLED;
		snode = getnodeid();
	}

	if ((flags & B_SETTLED) == 0) {
		snode = hashnode(tag);
		if (snode == NOTNODEID)
			return (error(125,NULLP));
	}

	req = (struct brenda_req *) nhead.nh_data;

	req->node    = getnodeid();
	req->event   = getpid();
	req->request = reqtype;
	req->tag     = tag;
	req->size    = size;

	nhead.nh_node   = snode;
	nhead.nh_event  = BRENDA;
	nhead.nh_type   = (_FROZEN_ | _REQ_);
	nhead.nh_flags  = DINTDATA;
	nhead.nh_msg    = NULLP;
	nhead.nh_length = 0;
/*0*/	if (nsend (&nhead) != 0)
		return (error(130,&nhead));

	if (reqtype==_OBJ_LST || reqtype==_REQ_LST) {
		msg   = argblk->buffer;
		msize = size;
	} else {
		msg   = NULLP;
		msize = 0;
	}

	nhead.nh_event  = getpid();
	nhead.nh_type   = _REPLY_;
	nhead.nh_flags  = DINTDATA;
	nhead.nh_msg    = msg;
	nhead.nh_length = msize;
/*1*/	if (nrecv (&nhead) != 0)
		return (error(140,&nhead));

	reqtype = req->request;
	if (reqtype == _ERROR_) {
		s = (int) req->data;
		if (s)
			errno = s;
		else
			errno = EIMPOSSIBLE;
		return (error(150,NULLP));
	}

	argblk->tag   = req->tag;
	argblk->snode = req->node;
	argblk->size  = req->size;
	argblk->data  = req->data;

	if (reqtype==_OBJ_SUM || reqtype==_REQ_SUM || reqtype==_CLEAR_)
		return (B_OK);

	/* _OBJ_LST or _REQ_LST: */

	if (req->data == 0)			/* report: none exist */
		return (B_OK);

	/* need to do byte-ordering on the reported list: */
	if (reqtype == _OBJ_LST)
		msize = sizeof (struct obj_lst);
	else
		msize = sizeof (struct req_lst);
	msize *= req->data;
	if (msize > size)  msize = size;
	msize /= sizeof(long);
	ip = (long *) argblk->buffer;
	for (s=0; s<msize; s++, ip++)
		*ip = ttol (*ip);
	return (B_OK);
}
