/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	bcast.c,v $
 * Revision 2.2  92/04/24  16:38:14  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:17:49  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:51  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: broadcast().
 *
 *	Broadcast() does NOT use Trollius multicasting.
 *	Instead, a simple scheme, specific to grids, is implemented here.
 *	This is to allow welding the grid into virtual circuits for
 *	the duration of the I/O-free portion of the application.
 *
 *	The broadcast algorithm:  the data is spread from the source node
 *	along each axis, on at a time, in a fixed order.  After each axis
 *	is done with there are more parallel threads working on the next.
 *	Within each axis the direction along which there are more nodes
 *	is sent to first.  That makes a difference in the total time to
 *	complete the broadcast since we cannot send on more than one link
 *	at a time.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- broadcast() does not return an error if the incoming data
 *		is larger than the specified nbytes.  The data passed
 *		to nodes downstream, though, is limited to nbytes!
 *	- broadcast() DOES pick up the message directed to the processor
 *		even if given a NULL buffer pointer, although the data is
 *		lost, of course.  Unlike the original CrOS, though, it does
 *		not return the number of bytes that the CP sent.  Instead it
 *		returns the number of bytes received.
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>
#include <malloc.h>


PRIVATE int
bsend (buffer, mask, nbytes)
	char *buffer;
	unsigned int mask;
	int nbytes;
{
	struct nmsg nhead;

	if (sendnbr (mask, &nhead) != 0)
		return (-1);

	if (nhead.nh_node == NOTNODEID)		/* at edge of no-wrap grid */
			return (0);

	nhead.nh_type     = CBCAST;
	nhead.nh_msg      = buffer;
	nhead.nh_length   = nbytes;
	if (csend (&nhead) != OK)
		return (-1);

	return (nbytes);
}

PRIVATE int
brecv (buffer, mask, nbytes)
	char *buffer;
	unsigned int mask;
	int nbytes;
{
	struct nmsg nhead;

	if (recvnbr (mask, &nhead) != 0)
		return (-1);

	if (nhead.nh_event == NOTEVENT)		/* at edge of grid */
			return (-1);

	nhead.nh_type   = CBCAST;
	nhead.nh_msg    = buffer;
	nhead.nh_length = nbytes;
	if (crecv (&nhead) != OK)
		return (-1);

	return (nhead.nh_length);
}


PUBLIC int
/*
broadcast (char *buffer, int src, unsigned int mask, int nbytes)
*/
broadcast (buffer, src, mask, nbytes)
	char *buffer;
	int src;
	unsigned int mask;
	int nbytes;
{
	int i, ndims, here, issrc, sign, size, malbuf;
	unsigned int submask, mask0, mask1, mask2;
	struct cubenv env;
	int *mycoord;
	int dimsize[MAXDIMS];
	int mycoorda[MAXDIMS];
	int srccoord[MAXDIMS];

	if (mask == 0 || (mask & ~fullmask()) != 0) {
		errno = EBADMASK;
		return (-1);
	}

	/* get dimensions of grid: */

	if (cparam (&env) != 0)
		return (-1);
	if (griddims (dimsize) != 0)
		return (-1);

	ndims = env.doc;
	here  = env.procnum;
	issrc = (here == src);		/* source node for broadcast */

	/* special case: NULL buffer: */

	malbuf = FALSE;
	if (buffer == NULL) {
		if (issrc && nbytes != 0) {
			errno = EBADBUF;
			return (-1);
		}
		buffer = malloc ((unsigned int) nbytes);
		if (buffer == NULL)
			return (-1);
		malbuf = TRUE;
	}

	/* get grid coordinates of local and source nodes: */

	if (gridcoord (src, srccoord) != 0)
		return (-1);
	if (issrc) {
		mycoord = srccoord;
	} else {
		if (gridcoord (here, mycoorda) != 0)
			return (-1);
		mycoord = mycoorda;
	}

	/* do the broadcast: */

	if (issrc) {

		/* set up variables for the loop below: */
		i = ndims-1;
		submask = (3 << (2*i));

	} else {

		/* Find the direction to receive from: */

		submask = 3;				/* first two bits */
		for (i=0; i<ndims; i++, submask<<=2) {
			if ((submask & mask) == 0)	/* this axis not used */
				continue;
			sign = mycoord[i] - srccoord[i];
			if (sign != 0) {
				if (sign > 0)
					mask0 = rightbit(submask);
				else
					mask0 = leftbits(submask);
				mask1 = (submask ^ mask0);
				if ((mask1 & mask) == 0) {
					/*
					 *  src & mask parameters imply
					 *  that this node is excluded!
					 */
					errno = EBADMASK;
					return (-1);
				}
				break;
			}
			/*
			 *  continue looping over dimensions in which this
			 *  node has same position as source node, stop at
			 *  the first (used) dimension where they differ.
			 */
		}

		/*
		 *  - now mask0 is the channel to receive from, and mask1
		 *  is the channel, on same axis, to send on.  Do it:
		 */

		nbytes = brecv (buffer, mask0, nbytes);
		if (nbytes < 0)
			return (-1);
		if (sign > 0 && mycoord[i] < dimsize[i]-1
		||  sign < 0 && mycoord[i] > 0) {
			if (bsend (buffer, mask1, nbytes) < 0)
				return (-1);
		}
		/* else we are at end of axis, do not send */

		--i;
		submask >>= 2;
	}

	/*
	 *  Whether source or not, we need to send on
	 *  all the (remaining) channels in order:
	 */

	for (; i>=0; i--, submask>>=2) {
		if ((submask & mask) == 0)		/* this axis not used */
			continue;
		size = dimsize[i] - 1;
		if (mycoord[i] == 0)			/* at bottom of axis */
			mask0 = 0;			/* can't go further  */
		else
			mask0 = rightbit(submask);
		if (mycoord[i] == size)			/* at top of axis */
			mask1 = 0;			/* can't go further  */
		else
			mask1 = leftbits(submask);
		if (2*srccoord[i] < size) {		/* closer to bottom,  */
			mask2 = mask0;			/* so swap the order: */
			mask0 = mask1;			/* - send up first    */
			mask1 = mask2;
		}
		if ((mask0 & mask) != 0) {
			if (bsend (buffer, mask0, nbytes) < 0)
				return (-1);
		}
		if ((mask1 & mask) != 0) {
			if (bsend (buffer, mask1, nbytes) < 0)
				return (-1);
		}
	}

	if (malbuf)
		free (buffer);

	return (nbytes);
}
