/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	combine.c,v $
 * Revision 2.2  92/04/24  16:38:19  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:17:59  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:55  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: combine().
 *
 *	Combine() does NOT use Trollius multireeling.
 *	Instead, a simple scheme, specific to grids, is implemented here.
 *	This is to allow welding the grid into virtual circuits for
 *	the duration of the I/O-free portion of the application.
 *
 *	The algorithm used: the data from the nodes is combined in all the
 *	nodes along an axis, in parallel for all positions along other axes.
 *	This is first done for the highest-index axis, and then repeated
 *	along the next highest-indexed axis, and so on.  At each step the
 *	data moved is the combined data from the preceding step.
 *	The combining of the data within an axis is accomplished by the
 *	nodes at both ends sending their data towards the middle.  The
 *	receiving nodes combine it with their data and then send it further
 *	towards the middle.  The process is continued until the two streams
 *	of data meet and then continue all the way to the other end of the
 *	axis.  (After they meet the final result is what travels the rest
 *	of the way.)
 *
 *	Combine() requires that the buf parameter be valid (not NULL),
 *	and it is assumed that the supplied buf has space for at least
 *	(size * nitems) bytes.  A NULL buf is an error (unless nitems = 0).
 *	A temporary buffer of the same size is also needed, malloc() is used.
 *
 *	Note that if nitems is 0 then the combining func() is never called,
 *	and the messages passed are of 0 length.  This sort of combine() is
 *	handy as a synchronization barrier.  In the case of 0 nitems, the
 *	buf parameter may be NULL.
 *
 *	The coding of the algorithm here relies heavily on the fact that
 *	the nodes are numbered in row-major order.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>
#include <malloc.h>


/*
 *  Special send/recv routines for combine.
 */

PRIVATE int
bsend (buffer, mask, nbytes)
	char *buffer;
	unsigned int mask;
	int nbytes;
{
	struct nmsg nhead;

	if (sendnbr (mask, &nhead) != 0)
		return (-1);

	if (nhead.nh_node == NOTNODEID)		/* at edge of no-wrap grid */
			return (-1);

	nhead.nh_type     = CCOMBINE;
	nhead.nh_msg      = buffer;
	nhead.nh_length   = nbytes;
	if (csend (&nhead) != OK)
		return (-1);

	return (0);
}

PRIVATE int
brecv (buffer, mask, nbytes)
	char *buffer;
	unsigned int mask;
	int nbytes;
{
	struct nmsg nhead;

	if (recvnbr (mask, &nhead) != 0)
		return (-1);

	if (nhead.nh_event == NOTEVENT)		/* at edge of grid */
			return (-1);

	nhead.nh_type   = CCOMBINE;
	nhead.nh_msg    = buffer;
	nhead.nh_length = nbytes;
	if (crecv (&nhead) != OK)
		return (-1);

	return (0);
}


/*
 *  Combine two buffers through the user's function.
 *  The result is in buf0.  Return errors from func().
 */
PRIVATE int
do_combine (buf0, buf1, func, size, nitems)
	char *buf0, *buf1;
	int (*func)();
	int size, nitems;
{
	int i, s;

	s = 0;
	for (i=0; i<nitems; i++) {
		s |= (*func)(buf0, buf1, size);
		buf0 += size;
		buf1 += size;
	}
	return (s);
}


/*
 *  Combine everybody's data everywhere along an axis.
 *  Returns OK (0) or ERROR (-1);
 */
PRIVATE int
do_axis (axis, mycoord, dimsize, buf, tbuf, func, size, nitems)
	int axis, mycoord, dimsize, size, nitems;
	char *buf, *tbuf;
	int (*func)();
{
	int k, s, nbytes;
	unsigned int upmask, dnmask, mask0, mask1;

	if (dimsize == 1)
		return (0);

	nbytes = size * nitems;

	dnmask = (1 << (2*axis));
	upmask = (dnmask << 1);
	if (mycoord == 0)
		dnmask = 0;
	if (mycoord == dimsize-1)
		upmask = 0;

	k = 2*mycoord - dimsize + 1;	/* deviation from middle of axis */

	if (k < 0) {			/* closer to 0-end of axis */
		mask0 = dnmask;
		mask1 = upmask;
	} else {
		mask0 = upmask;
		mask1 = dnmask;
	}

	s = 0;

	if (mask0) {			/* not at either end */

		/*
		 *  Combine our data with data from below
		 */
		s |= brecv (tbuf, mask0, nbytes);
		s |= do_combine (tbuf, buf, func, size, nitems);

		/*
		 *  Send combined data up.  Make sure to avoid deadlock
		 *  in the middle with both sides sending at the same time.
		 *  If dimsize is even, the data from both ends meets in
		 *  the TWO nodes dimsize/2 and dimsize/2-1, with k values
		 *  1 and -1 respectively.  If dimsize is odd, then the k<0
		 *  condition above makes the middle node behave as if it is
		 *  in the upper half, and the data meets in nodes with
		 *  k values 0 and -2.  Combine all data in those nodes,
		 *  and send that final combined result back towards ends.
		 */

		if (k==-1 || k==-2) {		/* the data streams meet here */

			/* send combined (below & ours) up */
			s |= bsend (tbuf, mask1, nbytes);
			/* receive combined (above) */
			s |= brecv (buf, mask1, nbytes);
			/* combine (below & ours) & above --> final result */
			s |= do_combine (buf, tbuf, func, size, nitems);

		} else if (k==1 || k==0) {	/* data streams meet here too */

			/* receive combined (above) */
			s |= brecv (buf, mask1, nbytes);
			/* send combined (below & ours) up */
			s |= bsend (tbuf, mask1, nbytes);
			/* combine (below & ours) & above --> final result */
			s |= do_combine (buf, tbuf, func, size, nitems);

		} else {

			/* send combined (below & ours) up */
			s |= bsend (tbuf, mask1, nbytes);
			/* receive the final result all cooked */
			s |= brecv (buf, mask1, nbytes);
		}

		/* send final result down */
		s |= bsend (buf, mask0, nbytes);

	} else {			/* edge node: the buff starts here */

		if (dimsize < 4 && k < 2) {

			/* avoid deadlock of both sending */
			if (k < 0) {
				s |= bsend (buf, mask1, nbytes);
				s |= brecv (tbuf, mask1, nbytes);
			} else {
				s |= brecv (tbuf, mask1, nbytes);
				s |= bsend (buf, mask1, nbytes);
			}
			s |= do_combine (buf, tbuf, func, size, nitems);

		} else {		/* no local combining needed */

			s |= bsend (buf, mask1, nbytes);
			s |= brecv (buf, mask1, nbytes);
		}

	}

	return (s);
}


PUBLIC int
/*
combine (char *buf, int (*func)(), int size, int nitems)
*/
combine (buf, func, size, nitems)
	char *buf;
	int (*func)();
	int size, nitems;
{
	int i, ndims, n, here, s, total, nbytes;
	char *tbuf;
	struct cubenv env;
	int dimsize[MAXDIMS];
	int mycoord[MAXDIMS];

	if (buf == NULL && nitems != 0) {
		errno = EBADBUF;
		return (-1);
	}

	/* get dimensions of grid, local coords, etc: */

	if (cparam (&env) != 0)
		return (-1);
	if (griddims (dimsize) != 0)
		return (-1);
	ndims = env.doc;
	n     = env.nproc;
	here  = env.procnum;
	if (gridcoord (here, mycoord) != 0)
		return (-1);

	/* get work space: */

	nbytes = size * nitems;
	if (nbytes != 0) {
		tbuf = malloc ((unsigned int) nbytes);
		if (tbuf == NULL)
			return (-1);
	} else {
		buf = NULL;
		tbuf = NULL;
	}

	/* do the thing along each axis: */

	for (i=ndims-1; i>=0; i--) {
	   s = do_axis (i,mycoord[i],dimsize[i],buf,tbuf,func,size,nitems);
	   if (s != 0)
		return (-1);
	}

	if (tbuf != NULL)
		free (tbuf);

	return (nitems);
}
