/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	concat.c,v $
 * Revision 2.2  92/04/24  16:38:21  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:01  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:56  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: concat().
 *
 *	Concat() does NOT use Trollius multireeling.
 *	Instead, a simple scheme, specific to grids, is implemented here.
 *	This is to allow welding the grid into virtual circuits for
 *	the duration of the I/O-free portion of the application.
 *
 *	The algorithm used: the data from the nodes is spread to all other
 *	nodes along an axis, in parallel for all positions along other axes.
 *	This is first done for the highest-index axis, and then repeated
 *	along the next highest-indexed axis, and so on.  At each step the
 *	data moved includes all the data collected in the preceding step.
 *	The spreading of the data within an axis is accomplished by the
 *	nodes at both ends sending their data towards the middle.  The
 *	receiving nodes add their data (prepending or appending it as
 *	appropriate) and then send it further towards the middle.  The
 *	process is continued until the two growing streams of data cross
 *	each other and continue all the way to the other end of the axis.
 *
 *	The coding of the algorithm here relies heavily on the fact that
 *	the nodes are numbered in row-major order.
 *
 *	Note: in the common case where nbytes equals binsz and is the same
 *	in all nodes, there is no problem with leaving szarray NULL, and it
 *	is even more efficient, in this implementation, to do so.  But if
 *	szarray is NULL in any node it must be NULL in all the nodes!
 *
 *	Portability note:  we use bcopy() here.  We should really be using
 *	the ANSI standard memcpy().  But for now we need to deal with
 *	non-ANSI libraries.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- concat() does not return an error if the incoming data
 *		is larger than the specified binsz.
 *	- concat() requires that the inbuf parameter be valid (not NULL),
 *		and it is assumed that the supplied inbuf has space for
 *		at least (n * binsz) bytes, where n is the number of nodes
 *		in the grid.  A NULL inbuf is an error.
 *	- concat() is rather inefficient if nbytes is much smaller than binsz,
 *		since binsz is the amount actually passed in messages.
 *		The only way to avoid that would be to send each node's data
 *		in a separate message.  We opted for large combined messages.
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>

extern bcopy();

/*
 *  Special send/recv routines for concat.
 *  Besides the data messages, auxillary messages are sent
 *  carrying the sizes of the data blocks from each node.
 */

PRIVATE int
xsend (szarray, buffer, size, binsz, offset, num, mask)
	int *szarray;
	char *buffer;
	int size, binsz, offset, num;
	unsigned int mask;
{
	struct nmsg nhead;

	if (mask == 0)
		return (0);

	if (sendnbr (mask, &nhead) != 0)
		return (-1);

	offset *= size;

	if (szarray != NULL) {
		nhead.nh_type   = CCONCATS;
		nhead.nh_msg    = (char *) &szarray[offset];
		nhead.nh_length = num * size * sizeof(int);
		if (csend (&nhead) != OK)
			return (-1);
	}
	nhead.nh_type   = CCONCAT;
	nhead.nh_msg    = buffer + (offset * binsz);
	nhead.nh_length = num * size * binsz;
	if (csend (&nhead) != OK)
		return (-1);

	return (0);
}

PRIVATE int
xrecv (szarray, buffer, size, binsz, offset, num, mask)
	int *szarray;
	char *buffer;
	int size, binsz, offset, num;
	unsigned int mask;
{
	struct nmsg nhead;

	if (mask == 0)
		return (0);

	if (recvnbr (mask, &nhead) != 0)
		return (-1);

	offset *= size;

	if (szarray != NULL) {
		nhead.nh_type   = CCONCATS;
		nhead.nh_msg    = (char *) &szarray[offset];
		nhead.nh_length = num * size * sizeof(int);
		if (crecv (&nhead) != OK)
			return (-1);
	}
	nhead.nh_type   = CCONCAT;
	nhead.nh_msg    = buffer + (offset * binsz);
	nhead.nh_length = num * size * binsz;
	if (crecv (&nhead) != OK)
		return (-1);

	return (0);
}

/*
 *  Spread everybody's data everywhere along an axis.
 *  Records the number of new bytes received in szarray.
 *  Returns OK (0) or ERROR (-1);
 */
PRIVATE int
do_axis (axis, mycoord, dimsize, buf, size, binsz, szarray)
	int axis, mycoord, dimsize, size, binsz;
	char *buf;
	int *szarray;
{
	int k, s, off0, off1, off2, off3, num0, num1, num2, num3;
	unsigned int upmask, dnmask, mask0, mask1;

	dnmask = (1 << (2*axis));
	upmask = (dnmask << 1);
	if (mycoord == 0)
		dnmask = 0;
	if (mycoord == dimsize-1)
		upmask = 0;

	k = 2*mycoord - dimsize + 1;	/* deviation from middle of axis */

	if (k < 0) {			/* closer to 0-end of axis */

		mask0 = dnmask;
		mask1 = upmask;
		off0 = 0;
		off1 = 0;
		off2 = mycoord + 1;
		off3 = mycoord;
		num0 = mycoord;
		num1 = mycoord + 1;
		num2 = dimsize - mycoord - 1;
		num3 = dimsize - mycoord;

	} else {

		mask0 = upmask;
		mask1 = dnmask;
		off0 = mycoord + 1;
		off1 = mycoord;
		off2 = 0;
		off3 = 0;
		num0 = dimsize - mycoord - 1;
		num1 = dimsize - mycoord;
		num2 = mycoord;
		num3 = mycoord + 1;

	}

	/*
	 *  Make sure to avoid deadlock in the middle with
	 *  both sides sending at the same time...
	 */

	s = xrecv (szarray, buf, size, binsz, off0, num0, mask0);

	if (k==0 || k==1) {	/* middle or just above middle */

		s |= xrecv (szarray, buf, size, binsz, off2, num2, mask1);
		s |= xsend (szarray, buf, size, binsz, off1, num1, mask1);

	} else {

		s |= xsend (szarray, buf, size, binsz, off1, num1, mask1);
		s |= xrecv (szarray, buf, size, binsz, off2, num2, mask1);

	}

	s |= xsend (szarray, buf, size, binsz, off3, num3, mask0);

	return (s);
}


PUBLIC int
/*
concat (char *outbuf, int nbytes, char *inbuf, int binsz, int *szarray)
*/
concat (outbuf, nbytes, inbuf, binsz, szarray)
	char *outbuf, *inbuf;
	int nbytes, binsz;
	int *szarray;
{
	int i, ndims, n, here, size, r, total;
	struct cubenv env;
	int dimsize[MAXDIMS];
	int mycoord[MAXDIMS];

	if (outbuf == NULL && nbytes > 0) {
		errno = EBADBUF;
		return (-1);
	}

	if (inbuf == NULL) {
		errno = EBADBUF;
		return (-1);
	}

	if (nbytes > binsz) {
		errno = EDATASIZE;
		return (-1);
	}

	/* get dimensions of grid, local coords, etc: */

	if (cparam (&env) != 0)
		return (-1);
	if (griddims (dimsize) != 0)
		return (-1);
	ndims = env.doc;
	n     = env.nproc;
	here  = env.procnum;
	if (gridcoord (here, mycoord) != 0)
		return (-1);

	/* do the concat: */

	/* point to our place in arrays: */

	inbuf += here * binsz;
	if (szarray != NULL)
		szarray += here;

	/* copy local data from output buffer into input buffer: */

	if (outbuf != inbuf)
		bcopy (outbuf, inbuf, nbytes);
	if (szarray != NULL)
		*szarray = nbytes;

	/* do the thing along each axis: */

	size = 1;
	for (i=ndims-1; i>=0; i--) {
	   /* slide down to start of relevant subarray: */
	   inbuf -= mycoord[i] * binsz * size;
	   if (szarray != NULL)
	     szarray -= mycoord[i] * size;
	   /* communicate: */
	   r = do_axis (i, mycoord[i], dimsize[i], inbuf, size, binsz, szarray);
	   if (r < 0)
		return (-1);
	   /* adjust block size for next step: */
	   size *= dimsize[i];
	}

	/*
	 *  We do NOT collapse the data, so it may not be consecutive
	 *  inside inbuf in the case where not all blocks are a full
	 *  binsz bytes long.  (Same as done in "the book", page 250.)
	 *  We do need to calculate the total number of bytes:
	 */

	if (szarray != NULL) {
		total = 0;
		for (i=0; i<n; i++)
			total += szarray[i];
	} else {
		total = n * binsz;
	}

	return (total);
}
