/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	cshift.c,v $
 * Revision 2.2  92/04/24  16:38:25  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:05  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:57  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: cshift()
 *
 *	The implementation of cshift() depends on how we identify
 *	channels and grid positions.  What we must avoid is parties
 *	to an exchange both doing "recv" first, leading to deadlock.
 *	If both do a "send" first then Trollius buffers save the day,
 *	but that is not very efficient.  Here we try to maximize the
 *	cases where buffering is avoided.  To do that we have the
 *	even nodes along the shift axis send first, with the odd
 *	nodes receiving, then the reverse.  This is based on the
 *	inmask, which may not have more than one bit set.  Care
 *	must be taken when outmask has multiple bits set.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- returns once outgoing data is sent out and incoming data
 *		arrives, but outgoing data may not have been
 *		received yet at its destination.
 *	- no error if received less data than was sent.
 */


#include <net.h>
#include <cros.h>
#include <malloc.h>

extern bcopy();

PUBLIC int
/*
cshift (char *inbuf,  unsigned int inmask,  int inbytes,
		char *outbuf, unsigned int outmask, int outbytes)
*/
cshift (inbuf, inmask, inbytes, outbuf, outmask, outbytes)
	char *inbuf;
	unsigned int inmask;
	int inbytes;
	char *outbuf;
	unsigned int outmask;
	int outbytes;
{
	int bytes, oddnode;
	char *buffer;

	/* if no I/O at all, just copy data from outbuf to inbuf: */
	if (inmask == 0 && outmask == 0) {
		bytes = outbytes;
		if (inbytes < outbytes)
			bytes = inbytes;
		if (inbuf != outbuf)
			bcopy (outbuf, inbuf, bytes);
		return (bytes);
	}

	/* special case: output but no input */
	if (inmask == 0)
		return (cwrite (buffer, outmask, outbytes));

	/* another special case: input but no output */
	if (outmask == 0)
		return (cread (inbuf, inmask, 0, inbytes));

	/* handle overlapping buffers: */
	if (inbuf == outbuf) {
		/* need to copy one of the buffers to prevent overwrite: */
		buffer = malloc ((unsigned) outbytes);
		if (buffer == NULL)
			return (-1);
		bcopy (outbuf, buffer, outbytes);
	} else {
		/* distinct buffers -- check for overlap: */
		buffer = outbuf;
		if (inbuf < outbuf && inbuf + inbytes > outbuf) {
			errno = EOVERLAP;
			return (-1);
		}
		if (outbuf < inbuf && outbuf + outbytes > inbuf) {
			errno = EOVERLAP;
			return (-1);
		}
	}

	/* avoid deadlock: transfer on even hops first: */
	oddnode = isoddnode (inmask);
	if (oddnode < 0)
		return (-1);
	if (oddnode) {
		bytes = cread (inbuf, inmask, 0, inbytes);
		if (bytes < 0)
			return (-1);
		if (cwrite (buffer, outmask, outbytes) < 0)
			return (-1);
	} else {
		if (cwrite (buffer, outmask, outbytes) < 0)
			return (-1);
		bytes = cread (inbuf, inmask, 0, inbytes);
	}

	if (inbuf == outbuf)
		free (buffer);

	return (bytes);
}
