/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	cweld.c,v $
 * Revision 2.2  92/04/24  16:38:27  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:08  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:59  gdburns
 * Homo sapien
 * 
 *
 *	Functions to weld grid into virtual circuits
 */

#include <net.h>
#include <events.h>
#include <cros.h>

#ifdef itb

/*
 *  Helper functions for do_weld():
 */

PRIVATE int
do_send (mask, flags)
	int mask, flags;
{
	struct nmsg nhead;

	if (sendnbr (mask, &nhead) != OK)
		return (ERROR);

	if (nhead.nh_node == NOTNODEID) {
		/* at edge of no-wrap grid */
		return (OK);
	}

	nhead.nh_type   = CWELD;
	nhead.nh_flags  = flags;
	nhead.nh_msg    = NULL;
	nhead.nh_length = 0;
	return (dsend (&nhead));
}

PRIVATE int
do_recv (mask, flags)
	int mask, flags;
{
	struct nmsg nhead;

	if (recvnbr (mask, &nhead) != OK)
		return (ERROR);

	if (nhead.nh_event == NOTEVENT) {
		/* at edge of no-wrap grid */
		return (OK);
	}

	nhead.nh_type   = CWELD;
	nhead.nh_flags  = flags;
	nhead.nh_msg    = NULL;
	nhead.nh_length = 0;
	return (drecv (&nhead));
}

/*
 *  Check for more logical dimensions than physical:
 */
PRIVATE int
check_weld (doc)
	int doc;
{
	int mask, i, dir;
	struct nmsg up, dn;
	long dlevs[2*MAXDIMS];
	static int mynode = -1;

	if (mynode >= 0)
		return (mynode);

	mynode = (int) getnodeid();

	for (dir=0; dir<doc; dir++) {

		mask = gridchan (mynode, dir,  1);
		if (sendnbr (mask, &up) != OK)
			return (-1);

		mask = gridchan (mynode, dir, -1);
		if (sendnbr (mask, &dn) != OK)
			return (-1);

		dlevs[2*dir]   = up.nh_dl_event;
		dlevs[2*dir+1] = dn.nh_dl_event;

		for (i=0; i<2*dir; i++) {
			if (dlevs[i] != NOTEVENT
			&& (dlevs[i] == up.nh_dl_event
			   || dlevs[i] == dn.nh_dl_event)) {
				errno = EFAKEGRID;
				return (-1);
			}
		}
	}

	return (mynode);
}

/*
 *  Send and recv on all channels.
 *  Used by cweld() and cunweld() below.
 */
PRIVATE int
do_weld (flags)
	int flags;
{
	int i, dir, mynode, odd, upmask, dnmask, tmp;
	struct cubenv env;

	if (cparam (&env) != OK)
		return (ERROR);

	mynode = check_weld (env.doc);
	if (mynode < 0)
		return (ERROR);

	/*
	 *  On each dimension, exchange messages with both neighbors:
	 */

	for (dir=0; dir<env.doc; dir++) {

		upmask = gridchan (mynode, dir,  1);
		dnmask = gridchan (mynode, dir, -1);
		odd = isoddnode (upmask);

		for (i=0; i<2; i++) {			/* up then down */

			if (odd) {			/* recv first */
				if (do_recv (dnmask, flags) != OK)
					return (ERROR);
				if (do_send (upmask, flags) != OK)
					return (ERROR);
			} else {			/* send first */
				if (do_send (upmask, flags) != OK)
					return (ERROR);
				if (do_recv (dnmask, flags) != OK)
					return (ERROR);
			}

			/* swap for second round: */
			tmp = upmask;
			upmask = dnmask;
			dnmask = tmp;
		}
	}

	cwelded = flags;

	return (OK);
}


/*
 *  Set up the grid as virtual circuits.
 *  MUST be called in all the grid nodes synchronously!
 */
PUBLIC int
cweld()
{
	/* synchronize all nodes before starting VCs: */
	if (combine (NULLPTR, (int (*)()) 0, 0, 0) != 0)
		return (ERROR);
	return (do_weld(HOLD));
}

/*
 *  Break up the virtual circuits.
 *  MUST be called in all the grid nodes synchronously!
 */
PUBLIC int
cunweld()
{
	return (do_weld(0));
}

#else	/* not itb, i.e., for standalone */

/*
 *  There isn't really any point in setting up the virtual circuits
 *  while debugging CrOS code on the host with stand-alone Trollius,
 *  so we just fake it here:
 */

PUBLIC int
cweld()
{
	if (combine (NULLPTR, (int (*)()) 0, 0, 0) != 0)
		return (ERROR);
	cwelded = DINTDATA;		/* nonzero but benign */
	return (OK);
}

PUBLIC int
cunweld()
{
	cwelded = 0;
	return (OK);
}

#endif  /* itb */
