/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	cwrite.c,v $
 * Revision 2.2  92/04/24  16:38:28  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:10  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:37:00  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: cread() and cwrite().
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- cwrite() returns once the data is sent out, although
 *		the receiver may not have received it yet.
 *	- cwrite() returns 0 if the 'mask' parameter is 0.
 *		It is not clear to me what the original CrOS returns.
 *	- cread() returns 0 if the 'inmask' parameter is 0.
 *		It is not clear to me what the original CrOS returns.
 *	- cread() does NOT forward the tail of the message if it
 *		did not have a large enough buffer to receive it.
 *		This is because Trollius discards that portion.
 *	- cread() does not return an error if the incoming data
 *		is larger than the specified buffer.
 *	- cread() cannot operate unless given an actual buffer.
 *		If given NULLPTR it returns an error.
 *		This is because the needed memory amount is unknown.
 *	- no check is made here for data transmission errors.
 *	- rdstat() is not supported, since Trollius does not provide
 *		a way to test for a message w/o picking it up.
 */


#include <net.h>
#include <events.h>
#include <cros.h>

PUBLIC int
/*
cwrite (char *buf, unsigned int mask, int nbytes)
*/
cwrite (buf, mask, nbytes)
	char *buf;
	unsigned int mask;
	int nbytes;
{
	unsigned int submask;
	struct nmsg nhead;
	int nsent;

	if (mask == 0)
		return (0);

	if (buf == NULL) {
		errno = EBADBUF;
		return (-1);
	}

	nsent = 0;

	while (mask != 0) {

		submask = rightbit(mask);
		if (sendnbr (submask, &nhead) != 0)
			return (-1);

		if (nhead.nh_node != NOTNODEID) {

			/* not at edge of no-wrap grid */

			nhead.nh_type   = CRDWR;
			nhead.nh_msg    = buf;
			nhead.nh_length = nbytes;

			if (csend (&nhead) != OK)
				return (-1);

			nsent = nbytes;
		}

		mask = leftbits(mask);
	}

	return (nsent);
}


PUBLIC int
/*
cread (char *buf, unsigned int inmask, unsigned int outmask, int nbytes)
*/
cread (buf, inmask, outmask, nbytes)
	char *buf;
	unsigned int inmask, outmask;
	int nbytes;
{
	struct nmsg nhead;

	if (inmask == 0)
		return (0);

	if (buf == NULL) {
		errno = EBADBUF;
		return (-1);
	}

	if (recvnbr (inmask, &nhead) != 0)
		return (-1);

	if (nhead.nh_event == NOTEVENT)		/* at edge of no-wrap grid */
		return (0);

	nhead.nh_type   = CRDWR;
	nhead.nh_msg    = buf;
	nhead.nh_length = nbytes;

	if (crecv (&nhead) != OK)
		return (-1);

	nbytes = nhead.nh_length;

	if (outmask != 0) {
		if (cwrite (buf, outmask, nbytes) < 0)
			return (-1);
	}

	return (nbytes);
}
