/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	grid.c,v $
 * Revision 2.2  92/04/24  16:38:34  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:19:43  gdburns
 * Ohio release
 * 
 * Revision 2.0.1.1  91/05/01  16:13:19  raghu
 * removed local declaration of variable hostnode in get_neighbors()
 * 
 * Revision 2.0  90/09/26  12:37:02  gdburns
 * Homo sapien
 * 
 *
 *	Grid setup and utility functions.
 *
 *	This implementation of the CrOS library uses the Trollius
 *	network level, datalink sublevel, services.
 *
 *	We assume that the physical machine has been booted to
 *	support the desired grid, hopefully with direct links
 *	everywhere, and that the nodes have been numbered 0...n-1
 *	so as to fit the grid order, in row-major format.  E.g.,
 *	a 2-rows-by-3-cols 2-d mesh has node IDs:
 *		(0)	(1)	(2)	<--- second index
 *	(0)	 0	 1	 2
 *	(1)	 3	 4	 5
 *	 ^---------------------------------- first index
 *	so that, e.g., node 4 is at grid position (1,1).
 *	Additionally, it is assumed that node 0 is directly
 *	connected to the host.
 *
 *	The first call must be gridinit(), in the standard CrOS
 *	syntax, but there is no way for this code to tell whether
 *	the above assumptions are correct.  Runtime errors will
 *	result if the node IDs are not 0..n-1, but no errors
 *	will result if message forwarding is implicitly invoked.
 *
 *	Trollius messages use nh_events that identify the direction.
 *	That allows setting up all grid links as virtual circuits.
 *	Note that the choice of events is nonportable: it takes
 *	into account the fact that DLO events (mod 16) are 0,4,8,12.
 *	Therefore, the nh_events are set to ((M<<16) + 4*K + 2 + S),
 *	where M is the destination (recipient) node ID, K is the index
 *	of the grid axis travelled along, and S=1 when the message is
 *	going up the axis, -1 when going down.  Also, if the message
 *	is wrapping around the end of the grid, then 16 is added to the
 *	event, to distinguish it from messages travelling the other way
 *	in the case where the wrapping is emulated via message forwarding.
 *
 *	This file is kinda long but the functions in it share variables.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- cparam() is not void (as is seems to be in the book).
 *		Instead it returns 0 (OK) or -1 (ERROR).
 *	- the host processor ID is whatever its Trollius node ID
 *		is, not necessarily 1024.
 *	- channel IDs (mask bit positions) are not the same as they
 *		would be on a hypercube.  Instead, looking up along
 *		grid axis K is channel 2*K+1, down is 2*K.  Note that
 *		sender and receiver do NOT use the same channel!
 *	- wrap-around at the edges of the grid is supported by
 *		default, but may involve message-forwarding and
 *		thus be inefficient.  If an element in the array
 *		of grid sizes passed to gridinit() is negative,
 *		it is taken to mean that no wraparound is to be
 *		done on that axis, so relevant channel masks
 *		will be set to 0.
 *	- there is no "errcnt" variable since we cannot put
 *		anything useful in it.
 */


/*********************
 **    Preamble:    **
 *********************/

#include <net.h>
#include <events.h>
#include <malloc.h>
#include <castreq.h>

#include <cros.h>

/*
 *  Macro to return the number of nodes in the "grid".
 *  We only care about the ITB and WASTE bits, and they
 *  must be set and clear, respectively:
 */
#define getngrid() getntype(NT_ITB & ~NT_WASTE, NT_ITB | NT_WASTE)


/*
 *  Types and structures:
 */

struct cnbor {
	long	nodeid;				/* Trollius node ID,	*/
	long	dlevent;			/* Trollius link to,	*/
	long	sevent;				/* Trollius nh_event to,*/
	long	revent;				/* nh_event from,	*/
	int	cmask;				/* & CrOS channel for	*/
						/* a nearest neighbor.	*/
	int	oddcoord;			/* for cshift()		*/
};


/*
 *  Public global variables:
 */

PUBLIC int	cwelded = 0;	/* when virtual circuits are on this is
				   set to HOLD (or DINTDATA if standalone) */

/*
 *  Private static variables shared between functions in this file:
 */

PRIVATE int	gridflag = FALSE;		/* TRUE if gridinit() called */
PRIVATE int	griddim;			/* dimensionality of grid */
PRIVATE int	gridsize[MAXDIMS];		/* nodes along each axis */
PRIVATE int	gridmap[MAXDIMS];		/* nodes in each subgrid */
PRIVATE int	dowrap[MAXDIMS];		/* whether to wrap on ea dim */
PRIVATE int	nproc;				/* total number of processors */
PRIVATE int	mynodeid;			/* our Trollius node ID, cast */
PRIVATE long	mypid;				/* process (not processor) ID */
PRIVATE int	mycoord[MAXDIMS];		/* where are we on the grid */
PRIVATE long	hostnode;			/* the host node ID */
PRIVATE int	onhost;				/* are we the host node? */
PRIVATE int	standalone;			/* simulated network? */
PRIVATE unsigned int	allmask;		/* all valid bits */
PRIVATE struct cnbor	neighbor[MAXDIMS][2];	/* neighbors of a cube node */
PRIVATE struct cnbor	forhost;		/* node 0 to/from host */
PRIVATE struct net_info	net_inf;		/* Trollius, from getnet() */
PRIVATE struct cubenv	cube_env;		/* CrOS, for cparam() */



/****************************
 **  Auxillary functions:  **
 ****************************/

/*
 *  Register a Trollius multicast node ID for broadcast
 *  from the host to all nodes in the grid.
 */
PRIVATE long
/*
get_bcastid (int ngrid)
*/
get_bcastid (ngrid)
	int ngrid;
{
	int i, nitb;
	long castid;
	long *nodetable;

	if (standalone)
		return (NOTNODEID);
	nitb = getnitb();
	if (ngrid == nitb)			/* all ITB nodes included */
		return (HOST2ITB);
#ifdef HOST2GRID
	if (ngrid == getngrid())
		return (HOST2GRID);
#endif
	nodetable = (long *) malloc (ngrid * sizeof(long));
	if (nodetable == NULL)
		return (NOTNODEID);
	for (i=0; i<ngrid; i++)
		nodetable[i] = i;
	castid = nitb + 3;			/* hopefully unique */
 	if (rcast (ORIGIN, castid, CBN_ERR, nodetable, ngrid) != OK)
		return (NOTNODEID);
	free (nodetable);
	return (castid);
}


/*
 *  Identify the neighbors of the node.
 *  Prepare a table of info needed later.
 */
PRIVATE int
/*
get_neighbors (void)
*/
get_neighbors()
{
	int i, node, sign, wrapping;
	unsigned int submask;
	long event;
	struct cnbor *nbrp;
	int coord[MAXDIMS];

	/* if on the host, only fill in the forhost struct: */
	if (onhost) {
		if (standalone) {
			event = CFROMHOST;
		} else {
			event = getroute ((long) 0, NOTEVENT);
			if (event == NOTEVENT)
				return (ERROR);
		}
		nbrp = &forhost;
		nbrp->nodeid  = 0;			/* always */
		nbrp->dlevent = event;
		nbrp->sevent  = CFROMHOST;
		nbrp->revent  = CTOHOST;
		nbrp->cmask   = 1 << (2*MAXDIMS);
		return (OK);
	}

	/* else, in the box: */
	submask = 3;
	allmask = 0;
	for (i=0; i<griddim; i++) {			  /* for each axis... */
		for (sign = -1; sign < 2; sign += 2) {	  /* down then up */
			(void) gridcoord (mynodeid, coord);
			coord[i] += sign;
			if (coord[i] < 0) {
				if (dowrap[i]) {
					wrapping = 16;
					coord[i] = gridsize[i] - 1;
					node = gridproc (coord);
				} else {
					node = NOTNODEID;
				}
			} else if (coord[i] >= gridsize[i]) {
				if (dowrap[i]) {
					wrapping = 16;
					coord[i] = 0;
					node = gridproc (coord);
				} else {
					node = NOTNODEID;
				}
			} else {
				wrapping = 0;
				node = gridproc (coord);
			}
			nbrp = &neighbor[i][sign>0? 1 : 0];
			if (node == NOTNODEID) {
				nbrp->nodeid = NOTNODEID;
				nbrp->dlevent = NOTEVENT;
				nbrp->sevent = NOTEVENT;
				nbrp->revent = NOTEVENT;
			} else {
				event = getroute ((long) node, NOTEVENT);
				if (event == NOTEVENT && ! standalone)
					return (ERROR);
				nbrp->nodeid = (long) node;
				nbrp->dlevent = event;
				event = 4*i + 2 + wrapping;
				nbrp->sevent = event + sign + (node << 16);
				nbrp->revent = event - sign + (mynodeid << 16);
			}
			nbrp->cmask = (1 << (2*i + (sign>0? 1 : 0)));
			nbrp->oddcoord = (mycoord[i] & 1);
		}
		allmask |= submask;	/* accumulate the fullmask value */
		submask <<= 2;
	}

	/* if connected to host, fill in the forhost struct: */
	if (mynodeid == 0) {
		if (standalone) {
			event = CTOHOST;
		} else {
			event = getroute (hostnode, NOTEVENT);
			if (event == NOTEVENT)
				return (ERROR);
		}
		nbrp = &forhost;
		nbrp->nodeid  = hostnode;
		nbrp->dlevent = event;
		nbrp->sevent  = CTOHOST;
		nbrp->revent  = CFROMHOST;
		nbrp->cmask   = 1 << (2*MAXDIMS);
	}

	return (OK);
}

/*
 *  Return a pointer to the correct cnbor struct
 *  given a channel mask.  The mask must have
 *  exactly one bit set.  Returns NULL on error.
 */
PRIVATE struct cnbor *
/*
getnbor (unsigned int mask)
*/
getnbor (mask)
	unsigned int mask;
{
	int i;
	struct cnbor *p;

	if (! cwelded) {
		/* check for errors */
		if (! gridflag) {
			errno = ENOGRID;
			return (NULL);
		}
		if (mask==0 || ismultibit(mask)) {
			errno = EBADMASK;
			return (NULL);
		}
		if (onhost) {
			if (mask == forhost.cmask) {
				return (&forhost);
			} else {
				errno = EBADCHAN;
				return (NULL);
			}
		}
	}

	for (i=0; i<griddim; i++) {
		p = neighbor[i];
		if (mask == p[0].cmask)
			return (&p[0]);
		if (mask == p[1].cmask)
			return (&p[1]);
	}
	if (mynodeid == 0 && mask == forhost.cmask)
			return (&forhost);
	errno = EBADCHAN;
	return (NULL);
}


/*************************************************
 **  Public functions: the CrOS library calls:  **
 *************************************************/

PUBLIC int
/*
gridcoord (int proc, int *coord)
*/
gridcoord (proc, coord)
	int proc;
	int *coord;
{
	int i;

	if (proc < 0 || proc >= nproc)
		return (ERROR);

	if (proc == mynodeid) {
		for (i=0; i<griddim; i++)
			coord[i] = mycoord[i];
		return (OK);
	}

	/* if 'proc' is not local, it is a bit harder... */
	for (i=0; i<griddim; i++) {
		coord[i] = proc / gridmap[i];
		proc     = proc % gridmap[i];
	}
	return (OK);
}


PUBLIC int
/*
gridproc (int *coord)
*/
gridproc (coord)
	int *coord;
{
	int i, j, node;

	i = griddim-1;
	node = coord[i];
	if (node >= gridsize[i] || node < 0)
		return (ERROR);
	for (--i; i>=0; i--) {
		j = coord[i];
		if (j >= gridsize[i] || j < 0)
			return (ERROR);
		node += j * gridmap[i];
	}
	return (node);
}


PUBLIC int
/*
gridchan (int proc, int dir, int sign)
*/
gridchan (proc, dir, sign)
	int proc;			/* this is actually ignored */
	int dir;
	int sign;
{
	int mask;

	if (dir >= griddim || dir < 0)
		return (ERROR);
	if (sign != 1 && sign != -1)
		return (ERROR);
	mask = (1 << (dir << 1));	/* bit position = 2*dir */
	if (sign > 0)
		mask <<= 1;
	return (mask);
}


PUBLIC int
/*
gridinit (int dim, int *num)
*/
gridinit (dim, num)
	int dim;
	int *num;
{
	int i;
	long here, castid;

	/* determine grid size and shape: */
	if (dim > MAXDIMS)
		return (ERROR);
	griddim = dim;
	for (i=0; i<dim; i++) {
		dowrap[i] = TRUE;
		if (num[i] < 0) {
			dowrap[i] = FALSE;
			num[i] = -num[i];
		}
		gridsize[i] = num[i];
	}
	gridmap[dim-1] = 1;
	for (i=dim-1; i>0; i--)
		gridmap[i-1] = gridmap[i] * gridsize[i];
	nproc = 1;
	for (i=0; i<dim; i++)			/* calc total nodes */
		nproc *= num[i];
	if (nproc == 0)	{			/* num[] has zeros */
		errno = ENUMPROC;
		return (ERROR);
	}

	/* get and check basic Trollius network info: */
	if (getnet (&net_inf) != OK)
		return (ERROR);
	here     = net_inf.ni_nodeid;
	hostnode = net_inf.ni_origin;
	if (here == hostnode) {
		onhost = TRUE;
		standalone = (getroute (0, NOTEVENT) == NOTEVENT);
		/* establish host-->grid multicast: */
		castid = get_bcastid (nproc);
		if (! standalone && castid == NOTNODEID)
			return (ERROR);
	} else {
		onhost = FALSE;
		/* force gridcoord() to fill in mycoord[]: */
		mynodeid = (int) ~here;
		if (gridcoord ((int) here, mycoord) == ERROR)
			return (ERROR);
		standalone = (getroute (hostnode, NOTEVENT) == NOTEVENT);
	}
	mynodeid = (int) here;
	mypid = getpid();
	if (standalone) {
		if (getnitb() < nproc) {	/* not enough fake ITB nodes */
			errno = ENUMPROC;
			return (ERROR);
		}
	} else {
		if (getngrid() < nproc) {	/* not enough nonwasted nodes */
			errno = ENUMPROC;
			return (ERROR);
		}
	}

	/* identify neighbors: */
	if (get_neighbors() == ERROR)
		return (ERROR);

	/* fill in cube environment: */
	cube_env.doc = dim;			/* (not a real hypercube) */
	cube_env.procnum = mynodeid;
	cube_env.nproc = nproc;
	if (mynodeid == 0)			/* connected to host */
		cube_env.cpmask = forhost.cmask;
	else
		cube_env.cpmask = 0;		/* no direct connection */
	if (onhost) {
		cube_env.cubemask = forhost.cmask;
		cube_env.cubeid   = castid;
	} else {
		cube_env.cubemask = 0;		/* not applicable */
		cube_env.cubeid   = NOTNODEID;
	}
	gridflag = TRUE;
	return (OK);
}


PUBLIC int
/*
cparam (struct cubenv *env)
*/
cparam (env)
	struct cubenv *env;
{
	if (! gridflag) {
		errno = ENOGRID;
		return (ERROR);
	}
	env->doc      = cube_env.doc;
	env->procnum  = cube_env.procnum;
	env->nproc    = cube_env.nproc;
	env->cpmask   = cube_env.cpmask;
	env->cubemask = cube_env.cubemask;
	env->cubeid   = cube_env.cubeid;
	return (OK);
}


/**
 **	Public functions added to the standard CrOS ones, mostly
 **	for use by the implementation of other CrOS functions
 **	that are in other files.
 **/

/*
 *  Return the mask for a complete grid broadcast:
 */
PUBLIC unsigned int
fullmask()
{
	return (allmask);
}

/*
 *  Return the size of the grid along each dimension.
 *  This is needed because it is not a hypercube, so this
 *  information is not implicit in what cparam() returns.
 */

PUBLIC int
/*
griddims (int *num)
*/
griddims (num)
	int *num;
{
	int i;

	if (! gridflag) {
		errno = ENOGRID;
		return (ERROR);
	}
	for (i=0; i<griddim; i++)
		num[i] = gridsize[i];
	return (OK);
}

/*
 *  Return the parameters needed for a cwrite to a neighbor:
 */
PUBLIC int
/*
sendnbr (unsigned int mask, struct nmsg *nheader)
*/
sendnbr (mask, nheader)
	unsigned int mask;
	struct nmsg *nheader;
{
	struct cnbor *p;

	if ((p = getnbor (mask)) == NULL)
		return (ERROR);
	if (standalone) {
		nheader->nh_dl_event = p->sevent;
		nheader->nh_flags    = NOBUF;
	} else {
		nheader->nh_dl_event = p->dlevent;
		nheader->nh_flags    = cwelded;
	}
	nheader->nh_node  = p->nodeid;
	nheader->nh_event = p->sevent;
	return (OK);
}

/*
 *  Return the parameters needed for a cread from a neighbor:
 */
PUBLIC int
/*
recvnbr (unsigned int mask, struct nmsg *nheader)
*/
recvnbr (mask, nheader)
	unsigned int mask;
	struct nmsg *nheader;
{
	struct cnbor *p;

	if ((p = getnbor (mask)) == NULL)
		return (ERROR);
	nheader->nh_event = p->revent;
	nheader->nh_flags = cwelded;
	return (OK);
}

/*
 *  Tell whether we are an odd node on the axis identified by a mask.
 *  Returns 1 (true), 0 (false), or -1 (error).  Called by cshift().
 */
PUBLIC int
/*
isoddnode (unsigned int mask)
*/
isoddnode (mask)
	unsigned int mask;
{
	struct cnbor *p;

	p = getnbor (mask);
	if (p == NULL)
		return (-1);
	if (p == &forhost) {
		errno = EBADCHAN;
		return (-1);
	}
	if (p->oddcoord)
		return (1);
	return (0);
}
