/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	vwrite.c,v $
 * Revision 2.2  92/04/24  16:38:39  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:35  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:37:05  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: vread(), vwrite(), and vshift().
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- vwrite() returns once the data is sent out, although
 *		the receiver may not have received it yet.
 *	- vwrite() returns 0 if the 'mask' parameter is 0.
 *		It is not clear to me what the original CrOS returns.
 *	- vread() returns 0 if the 'inmask' parameter is 0.
 *		It is not clear to me what the original CrOS returns.
 *	- vread() does NOT forward the tail of the message if it
 *		did not have a large enough buffer to receive it.
 *		This is because Trollius discards that portion.
 *	- vread() does not return an error if the incoming data
 *		is larger than the specified buffer.
 *	- vread() cannot operate unless given an actual buffer.
 *		If given NULLPTR it returns an error.
 *		This is because the needed memory amount is unknown.
 *	- both vread() and vwrite() use malloc() to allocate a big
 *		block to temporarily hold the whole vector.  If this
 *		malloc() fails the vread() or vwrite() call fails.
 *	- vshift() requires two malloc()s, for input and output
 *		gathered forms of the vectors.
 *	- functions return once outgoing data is sent out and incoming
 *		data arrives, but outgoing data may not have been
 *		received yet at its destination.
 *	- no error if received less data than was sent.
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>
#include <malloc.h>

#ifdef t800
#include <c.h>			/* for Move2D */
#endif

#ifdef Move2D
/* use Logical Systems' wrapper for t800 2d move: */
#define gather(b,v,s,w,n)  Move2D(v,b,s,n,w,s)
#else
extern bcopy();
PRIVATE void
/*
gather (char *blk, char *vec, int size, int offset, int nitems)
*/
gather (blk, vec, size, offset, nitems)
	char *blk;
	char *vec;
	int size;
	int offset;
	int nitems;
{
	int i;
	register char *p, *q;

	if (size == sizeof(int))
		for (i=0, p=blk, q=vec; i<nitems; i++, p+=size, q+=offset)
			*((int *) p) = *((int *) q);
	else if (size == sizeof(float))
		for (i=0, p=blk, q=vec; i<nitems; i++, p+=size, q+=offset)
			*((float *) p) = *((float *) q);
	else if (size == sizeof(double))
		for (i=0, p=blk, q=vec; i<nitems; i++, p+=size, q+=offset)
			*((double *) p) = *((double *) q);
	else
		for (i=0, p=blk, q=vec; i<nitems; i++, p+=size, q+=offset)
			bcopy (q, p, size);
}
#endif

#ifdef Move2D
/* use Logical Systems' wrapper for t800 2d move: */
#define scatter(b,v,s,w,n)  Move2D(b,v,s,n,s,w)
#else
PRIVATE void
/*
scatter (char *blk, char *vec, int size, int offset, int nitems)
*/
scatter (blk, vec, size, offset, nitems)
	char *blk;
	char *vec;
	int size;
	int offset;
	int nitems;
{
	int i;
	register char *p, *q;

	if (size == sizeof(int))
		for (i=0, p=blk, q=vec; i<nitems; i++, p+=size, q+=offset)
			*((int *) q) = *((int *) p);
	else if (size == sizeof(float))
		for (i=0, p=blk, q=vec; i<nitems; i++, p+=size, q+=offset)
			*((float *) q) = *((float *) p);
	else if (size == sizeof(double))
		for (i=0, p=blk, q=vec; i<nitems; i++, p+=size, q+=offset)
			*((double *) q) = *((double *) p);
	else
		for (i=0, p=blk, q=vec; i<nitems; i++, p+=size, q+=offset)
			bcopy (p, q, size);
}
#endif


PUBLIC int
/*
vwrite (char *buf, unsigned int mask, int size, offset, nitems)
*/
vwrite (buf, mask, size, offset, nitems)
	char *buf;
	unsigned int mask;
	int size, offset, nitems;
{
	int bytes;
	char *msgbuf;

	if (mask == 0)
		return (0);

	if (buf == NULL) {
		errno = EBADBUF;
		return (-1);
	}

	bytes = size * nitems;
	msgbuf = malloc ((unsigned) bytes);
	if (msgbuf == NULL)
		return (-1);

	gather (msgbuf, buf, size, offset, nitems);

	bytes = cwrite (msgbuf, mask, bytes);
	if (bytes < 0)
		return (-1);

	free (msgbuf);

	return (bytes / size);
}


PUBLIC int
/*
vread (char *buf, unsigned int inmask, outmask, int size, offset, nitems)
*/
vread (buf, inmask, outmask, size, offset, nitems)
	char *buf;
	unsigned int inmask, outmask;
	int size, offset, nitems;
{
	int bytes;
	char *msgbuf;

	if (inmask == 0)
		return (0);

	if (buf == NULL) {
		errno = EBADBUF;
		return (-1);
	}

	bytes = size * nitems;
	msgbuf = malloc ((unsigned) bytes);
	if (msgbuf == NULL)
		return (-1);

	bytes = cread (msgbuf, inmask, outmask, bytes);
	if (bytes < 0)
		return (-1);

	/* (cread() does the output part of vread() for us too!) */

	scatter (msgbuf, buf, size, offset, nitems);

	free (msgbuf);

	return (bytes / size);
}


PUBLIC int
/*
vshift (char *inbuf,  unsigned int inmask,  int insize,  inoffset,  initems,
	char *outbuf, unsigned int outmask, int outsize, outoffset, outitems)
*/
vshift (inbuf,  inmask,  insize,  inoffset,  initems,
        outbuf, outmask, outsize, outoffset, outitems)
	char *inbuf, *outbuf;
	unsigned int inmask, outmask;
	int insize, outsize;
	int inoffset, outoffset;
	int initems, outitems;
{
	int inbytes, outbytes;
	char *inblk, *outblk;

	/* if no I/O, still need to copy data from outbuf to inbuf: */
	/* - we punish the foolish user and do it indirectly via cshift(). */

	if (inbuf == outbuf) {
		/* man page says it's OK only if the rest matches: */
		if (insize != outsize || inoffset != outoffset) {
			errno = EOVERLAP;
			return (-1);
		}
		/* no need to copy one of the buffers to prevent overwrite,
			since we do gather/scatter anyway... */
	} else {
		/* distinct buffers -- should check for overlap
			but it seems too hard for now... */
	}

	inbytes = insize * initems;
	inblk = malloc ((unsigned) inbytes);
	if (inblk == NULL)
		return (-1);

	outbytes = outsize * outitems;
	outblk = malloc ((unsigned) outbytes);
	if (outblk == NULL)
		return (-1);

	gather (outblk, outbuf, outsize, outoffset, outitems);

	inbytes = cshift (inblk, inmask, inbytes, outblk, outmask, outbytes);
	if (inbytes < 0)
		return (-1);

	initems = inbytes / insize;

	scatter (inblk, inbuf, insize, inoffset, initems);

	free (inblk);
	free (outblk);

	return (initems);
}
