/*
 *	Office of Research Computing				Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	ipcdr.c,v $
 * Revision 2.2  92/04/24  16:45:25  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/01/24  11:22:41  raja
 * Modified it to work locally without accessing the router.
 * 
 * Revision 2.1.1.1  91/10/03  01:35:21  raja
 * Stricter checking of message ordering: handle duplicate messages.
 * Some cosmetic changes including the use of the DBUG macro in <debug.h>.
 * 
 * Revision 2.1  91/05/02  14:40:04  gdburns
 * Ohio release
 * 
 * Revision 2.0.1.1  91/01/31  15:22:55  raja
 * Added the option to check the order of the received messages.
 * 
 * Revision 2.0  90/09/26  12:36:48  gdburns
 * Homo sapien
 *  
 *	Function:	- IPC diagnosis tool
 *			- send or receive messages of given length
 *			- report timings for message transfers
 *			- optional use of virtual circuits and buffers
 *
 *	Accepts:	- See the "usage" variable.
 *
 *	Examples:	- ipcdr -vt -m 10 -l 200 -s n0
 *			- ipcdr -vct -m 5 -l 10
 */


#include <stdio.h>
#include <malloc.h>
#include <terror.h>				/* error codes */
#include <events.h>				/* messages event */
#include <net.h>				/* network constants */
#include <typical.h>				/* common constants */
#include <debug.h>

/*
 * external variables
 */
extern int		errno;			/* global error variable */

/*
 * external functions
 */
extern void		exclusive();		/* set exclusive options */
extern void		mnusage();		/* node usage help menu */
extern void		terror();
extern void		validopts();		/* set the valid options */
extern int		do_args();		/* parse the valid options */
extern int		intparam();		/* interger option parameter */
extern int		kinit();
extern int		opt_chosen();		/* pick from exclusive opts */
extern int		opt_taken();		/* was option selected ? */
extern int		parse_nodes();		/* parse the node list */
extern double		ttime();		/* get current time value */

/*
 * local functions
 */
static void		fill_msg();		/* fill the message body */
static void		help();			/* print help menu */
static int		check_msg();		/* check message contents */
static int		do_receiver();		/* run as a receiver */
static int		do_sender();		/* run as a sender */
static int		msg_recv();		/* receive a message */
static int		msg_send();		/* send a message */
static int		set_defaults();		/* set the option defaults */

/*
 * local variables
 */
static char *_copyright = "Copyright 1990 The Ohio State University $Revision: 2.2 $";
static char *usage = "ipcdr [-bcdhotvCDFINRV][-e event] [-l length] [-m messages] [-s <nodes>]";

static char		*progname;		/* command name */
static int		type_size;		/* data type size */
static int		event;			/* event number */
static int		fl_buffer;		/* buffering flag */
static int		fl_check;		/* check message */
static int		fl_debug;		/* debugging flag */
static int		fl_msgtype;		/* message data type */
static int		fl_msgorder;		/* check message ordering */
static int		fl_pretend;		/* pretend flag */
static int		fl_rtrip;		/* round-trip flag */
static int		fl_sender;		/* sender flag */
static int		fl_timing;		/* timing flag */
static int		fl_verbose;		/* verbose flag */
static int		fl_virtual;		/* virtual circuit */
static int		length;			/* message length */
static int		num_messages;		/* number of loops */
static int		total_errors;		/* transmission errors */
static struct nmsg	nhead;			/* message header */

/*
 * The program starts here.
 */
int
main(argc, argv)

int			argc;
char			*argv[];

{
	progname = argv[0];
/*
 * Parse the command line.
 */
	validopts("bcdhostvNRV");
	followed("elm");
	exclusive("CDFI");

	if (do_args(&argc, argv) == ERROR) {
		fprintf(stderr, "usage: %s\n", usage);
		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Attach to the kernel.
 */
	if (kinit(0) == ERROR) {
		terror(progname);
		exit(errno);
	}
/*
 * Continue parsing the command line for nodes.
 */
	if (opt_taken('s')) {
		if ((parse_nodes(&argc, argv) == ERROR) ||
			(errno = (argc == 1) ? 0 : EUSAGE)) {
			fprintf(stderr, "usage: %s\n", usage);
			kexit(errno);
		}
	}
/*
 * Set the default options.
 */
	if (set_defaults()) {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(ERROR);
	}
/*
 * Fill the message header.
 */
	nhead.nh_event	= event;
	nhead.nh_type	= 0;
	nhead.nh_length	= length * type_size;
	nhead.nh_flags	= fl_msgtype |
			(fl_buffer ? 0 : NOBUF) | 
			(fl_virtual ? KHOLD : 0);
	if (nhead.nh_length > 0) {
		if ((nhead.nh_msg = malloc(nhead.nh_length)) == NULL) {
			terror(progname);
			kexit(errno);
		}
	}
	else {
		nhead.nh_msg = NULL;
	}
/*
 * Send of receive depending on the required role.
 */
	if (fl_sender) {
		kexit(do_sender());
	}
	else {
		kexit(do_receiver());
	}
}

/*
 *	do_sender
 *
 *	Function:	- send messages to the given nodes
 *			- report the timing if needed
 *
 *	Returns:	- 0 or ERROR
 */
static int
do_sender()
{
	int		i, loops;
	int		n_index, nodeid, n_flags, my_nodeid;
	double		timing;	

/*
 * Initialize.
 */
	total_errors = loops = 0;
	fill_msg(&nhead);
/*
 * Get the first nodeid and check for usage.
 */
	getnode(&n_index, &nodeid, &n_flags);
	if (n_index == ERROR) {
		fprintf(stderr, "usage: %s\n", usage);
		if (length > 0) free(nhead.nh_msg);
		return(ERROR);
	}

	timing = ttime();
/*
 * Loop over the number of messages to process.
 */
	for (i=0; i < num_messages; i++) {
/*
 * Write in each message header the message sequence number.
 */
		nhead.nh_data[1] = i;
/*
 * Loop over the given destination nodes.  Update loops count.
 */
		do {
			loops++;

			if ((nhead.nh_node = nodeid) != LOCAL) {
				*nhead.nh_data = getnodeid();
			} else {
				*nhead.nh_data = LOCAL;
			}

			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				if (length > 0) free(nhead.nh_msg);
				return(ERROR);
			}
/*
 * Disconnect a possible virtual circuit if it is the last message.
 */
			if (i == num_messages-1) {
				nhead.nh_flags &= ~KHOLD;
			}
/*
 * Send the message.  For a round trip, receive the message.
 */
			if (msg_send(&nhead)) return(ERROR);
			if (fl_rtrip) {
				if (msg_recv(&nhead)) return(ERROR);
				DBUG("round trip: msg %d, len %d",
					i, nhead.nh_length);
			}

			getnode(&n_index, &nodeid, &n_flags);
		} while(n_index);

	} /* end for */
/*
 * Get the elapsed time and report timing.
 */
	timing = ttime() - timing;
	if (fl_rtrip) loops *= 2;
	if (fl_timing) {
		printf("total time: %f\taverage message time: %f\n", 
			timing, timing/loops);
		if (nhead.nh_length > 0)
			printf("\t\t\taverage byte time: %f\n", 
				timing/loops/nhead.nh_length);
	}
/*
 * Report the number of transmission errors.
 */
	if (length > 0) free(nhead.nh_msg);
	printf("\n");
	if (fl_verbose && fl_check && fl_rtrip)
		printf("total transmission errors: %d\n", total_errors);
	if (total_errors > 0) {
		tprint("ipcdr: transmission errors\n");
		return(ERROR);
	}
	else
		return(0);
}

/*
 *	do_receiver
 *
 *	Function:	- receive messages and echo back if needed
 *			- report timings if needed
 *
 *	Returns:	- 0 or ERROR.
 */
static int
do_receiver()
{
	int		i, loops;
	int		seq_errors, next_seq;
	double		timing;

/*
 * Initialize.
 */
	total_errors = seq_errors = loops = 0;
	timing = ttime();
/*
 * Loop over the number of messages to receive.
 */
	for (i=0, next_seq = 0; i < num_messages; i++, next_seq++) {
/*
 * Disconnect a possible virtual circuit if it is the last message.
 */
		if (i == num_messages-1) {
			nhead.nh_flags &= ~KHOLD;
		}
/*
 * Receive the message and check errors.  Update loops count.
 * Check the sequencing of the messages if required and report the number
 * of messages out of order.  For a round trip, send the message back.
 */
		loops++;
		if (msg_recv(&nhead)) return(ERROR);
		DBUG("received: message %d, seq. # %d, length %d",
			i, nhead.nh_data[1], nhead.nh_length);
		if (fl_msgorder && (nhead.nh_data[1] != next_seq)) {
			seq_errors += abs(nhead.nh_data[1] - next_seq);
			next_seq = nhead.nh_data[1];
			DBUG("    <- break in message order\n");
		}
		else {
			DBUG("\n");
		}
		if (fl_rtrip) {
			nhead.nh_node = *nhead.nh_data;
			if (msg_send(&nhead)) return(ERROR);
		}

	} /* end for */
/*
 * Get the elapsed time and report timing.
 */
	timing = ttime() - timing;
	if (fl_rtrip) loops *= 2;
	if (fl_timing) {
		printf("total time: %f\taverage message time: %f\n", 
			timing, timing/loops);
		if (nhead.nh_length > 0)
			printf("\t\t\taverage byte time: %f\n", 
				timing/loops/nhead.nh_length);
	}
/*
 * Report the number of transmission and message ordering errors.
 */
	if (length > 0) {
		free(nhead.nh_msg);
	}
	printf("\n");
	if (fl_verbose && fl_check) {
		printf("total transmission errors: %d\n", total_errors);
	}
	if (fl_verbose && fl_msgorder) {
		printf("total msg ordering errors: %d\n", seq_errors);
	}
	if ((total_errors > 0) || (seq_errors > 0)) {
		if (total_errors > 0) {
			tprint("ipcdr: transmission errors\n");
		}
		if (seq_errors > 0) {
			tprint("ipcdr: message ordering errors\n");
		}
		return(ERROR);
	}
	else {
		return(0);
	}
}

/*
 *	fill_msg
 *
 *	Function:	- fill the body of the message
 *			- handles messages of char, int, float and double
 *
 *	Accepts:	- ptr to a message header
 */
static void
fill_msg(header)

struct nmsg		*header;

{
	int		i, elements;
	int		*p_imsg;
	char		*p_msg;
	float		*p_fmsg;
	double		*p_dmsg;

	if (header->nh_flags & DDBLMSG) {
/*
 * Fill the message with doubles.
 */
		p_dmsg = (double *) header->nh_msg;
		elements = header->nh_length / sizeof(double);
		for (i=0; i < elements; i++, p_dmsg++) {
			*p_dmsg = (double) 12345.6789;
		}
	}
	else if (header->nh_flags & DFLTMSG) {
/*
 * Fill the message with floats.
 */
		p_fmsg = (float *) header->nh_msg;
		elements = header->nh_length / sizeof(float);
		for (i=0; i < elements; i++, p_fmsg++) {
			*p_fmsg = (float) 1234.5678;
		}
	}
	else if (header->nh_flags & DINTMSG) {
/*
 * Fill the message with integers.
 */
		p_imsg = (int *) header->nh_msg;
		elements = header->nh_length / sizeof(int);
		for (i=0; i < elements; i++, p_imsg++) {
			*p_imsg = 123456789;
		}
	}
	else {
/*
 * Fill the message with characters.
 */
		p_msg = header->nh_msg;
		for (i=0; i < header->nh_length; i++, p_msg++) {
			*p_msg = (char) 90;
		}
	}
}

/*
 *	check_msg
 *
 *	Function:	- check the contents of the message
 *			- handles messages of char, int, float and double
 *
 *	Accepts:	- ptr to a message header
 *
 *	Returns:	- 0 or ERROR
 */
static int
check_msg(header)

struct nmsg		*header;

{
	int 		i, elements;
	int		*p_imsg;
	char 		*p_msg;
	float		*p_fmsg;
	double		*p_dmsg;

	if (header->nh_flags & DDBLMSG) {
/*
 * Check a message of doubles.
 */
		p_dmsg = (double *) header->nh_msg;
		elements = header->nh_length / sizeof(double);
		for (i=0; i < elements; i++, p_dmsg++) {
			if (*p_dmsg != (double) 12345.6789) return(ERROR);
		}
	}
	else if (header->nh_flags & DFLTMSG) {
/*
 * Check a message of floats.
 */
		p_fmsg = (float *) header->nh_msg;
		elements = header->nh_length / sizeof(float);
		for (i=0; i < elements; i++, p_fmsg++) {
			if (*p_fmsg != (float) 1234.5678) return(ERROR);
		}
	}
	else if (header->nh_flags & DINTMSG) {
/*
 * Check a message of integers.
 */
		p_imsg = (int *) header->nh_msg;
		elements = header->nh_length / sizeof(int);
		for (i=0; i < elements; i++, p_imsg++) {
			if (*p_imsg != 123456789) return(ERROR);
		}
	}
	else {
/*
 * Check a message of characters.
 */
		p_msg = header->nh_msg;
		for (i=0; i < header->nh_length; i++, p_msg++) {
			if (*p_msg != (char) 90) return(ERROR);
		}
	}

	return(0);
}

/*
 *	help
 *
 *	Function:	- print the help menu for the tool
 */
static void
help()
{
	printf("\nSynopsis:\tipcdr [options]\n");
	printf("\nDescription:\tTest and time Inter-Process Communication.\n");
	printf("\nOptions:\t-h                this help message\n");
	printf("\t\t-b                disable buffering\n");
	printf("\t\t-c                check received message\n");
	printf("\t\t-d                turn on debugging\n");
	printf("\t\t-o                check order of messages\n");
	printf("\t\t-t                time message exchange\n");
	printf("\t\t-v                verbose mode\n");
	printf("\t\t-C                message of characters\n");
	printf("\t\t-D                message of doubles\n");
	printf("\t\t-F                message of floats\n");
	printf("\t\t-I                message of integers\n");
	printf("\t\t-N                pretend mode\n");
	printf("\t\t-R                message round trip\n");
	printf("\t\t-V                virtual circuit mode\n");
	printf("\t\t-e event          synchronize on event\n");
	printf("\t\t-l length         message length\n");
	printf("\t\t-m messages       send this many messages\n");
	printf("\t\t-s <nodes>        send to these nodes\n");
	mnusage();
}

/*
 *	set_defaults
 *
 *	Function:	- set the default values of key variables
 *			- checks most of the command line arguments
 *
 *	Returns:	- 0 or ERROR
 */
static int
set_defaults()
{
/*
 * Set the flags.
 */
	if (opt_taken('b')) {
		fl_buffer = FALSE;
	}
	else {
		fl_buffer = TRUE;
	}

	if (opt_taken('c')) {
		fl_check = TRUE;
	}
	else {
		fl_check = FALSE;
	}

	if (opt_taken('d')) {
		fl_debug = TRUE;
	}
	else {
		fl_debug = FALSE;
	}

	if (opt_taken('o')) {
		fl_msgorder = TRUE;
	}
	else {
		fl_msgorder = FALSE;
	}

	if (opt_taken('t')) {
		fl_timing = TRUE;
	}
	else {
		fl_timing = FALSE;
	}

	if (opt_taken('v')) {
		fl_verbose = TRUE;
	}
	else {
		fl_verbose = FALSE;
	}

	if (opt_taken('N')) {
		fl_pretend = TRUE;
	}
	else {
		fl_pretend = FALSE;
	}

	if (opt_taken('R')) {
		fl_rtrip = TRUE;
	}
	else {
		fl_rtrip = FALSE;
	}

	if (opt_taken('V')) {
		fl_virtual = TRUE;
	}
	else {
		fl_virtual = FALSE;
	}

	if (opt_taken('s')) {
		fl_sender = TRUE;
	}
	else {
		fl_sender = FALSE;
	}
/*
 * Set among the mutual exclusive options.
 */
	switch(opt_chosen('C')) {
		case 'C':
			fl_msgtype = DRAWMSG;
			type_size = sizeof(char);
			break;
		case 'D':
			fl_msgtype = DDBLMSG;
			type_size = sizeof(double);
			break;
		case 'F':
			fl_msgtype = DFLTMSG;
			type_size = sizeof(float);
			break;
		case 'I':
			fl_msgtype = DINTMSG;
			type_size = sizeof(int);
			break;
		case ERROR:
			fl_msgtype = 0;
			type_size = sizeof(char);
			break;
		default:
			return(ERROR);
	}
/*
 * Set the parameters.
 */
	if (opt_taken('e')) {
		if (intparam('e', &event)) {
			return(ERROR);
		}
	}
	else {
		event = EVIPCDIAG;
	}

	if (opt_taken('m')) {
		if (intparam('m', &num_messages)) {
			return(ERROR);
		}
	}
	else {
		num_messages = 1;
	}

	if (opt_taken('l')) {
		if (intparam('l', &length)) {
			return(ERROR);
		}
	}
	else {
		length = 0;
	}

	return(0);
}

/*
 *	msg_send
 *
 *	Function:	- send the given message if not pretending
 *			- print debugging information if needed
 *
 *	Accepts:	- ptr to the message header
 *
 *	Returns		- 0 or ERROR
 */
static int
msg_send(header)

struct nmsg		*header;

{
	DBUG("send: node %d, length %d\n", header->nh_node, header->nh_length);
	if (!fl_pretend) {
		if (nsend(header)) {
			if (header->nh_length > 0) free(header->nh_msg);
			terror(progname);
			return(ERROR);
		}
	}
	return(0);
}

/*
 *	msg_recv
 *
 *	Function:	- receive a message
 *			- check the message contents if needed
 *
 *	Accepts:	- ptr to a message header
 *
 *	Returns:	- 0 or ERROR
 */
static int
msg_recv(header)

struct nmsg		*header;

{
	if (!fl_pretend) {
		if (nrecv(header)) {
			if (header->nh_length > 0) free(header->nh_msg);
			terror(progname);
			return(ERROR);
		}
		if (fl_check) {
			if (check_msg(header)) total_errors++;
		}
	}
	return(0);
}
