/*
 *      OSU Research Computing					Trillium
 *      Copyright 1991 The Ohio State University
 *      GDB
 *
 *      $Log:	trproc.c,v $
 * Revision 2.2  92/04/24  16:50:34  trillium
 * Ohio Release
 * 
 *	Function:	- functions that manipulate the traced process
 *			  data structure
 */

#include <all_list.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <trreq.h>
#include <typical.h>
#include <shared.h>

/*
 * external variables
 */
extern int4		errno;
extern char		buf[];			/* traced I/O buffer */

/*
 * external functions
 */
extern void		send_reply();

/*
 * local structures
 */
struct trproc {
	int4		trp_pid;
	int4		trp_length;
	int4		trp_sig;
};

#define TRPNULL ((struct trproc *) 0)

/*
 * local variables
 */
static LIST		*trp_list;
static int4		det_length;		/* length of trace data
						   for detached processes */
/*
 * local functions
 */
static int4		trpcompare();		/* compare two list entries */

/*
 *	trpinit
 *
 *	Function:	- initializes trace process list
 */
void
trpinit()

{
	trp_list = al_init(sizeof(struct trproc), trpcompare);

	if (trp_list == (LIST *) 0) {
		panic(errno);
	}

	det_length = 0;
}

/*
 *	trqattach
 *
 *	Function:	- attaches client to trace daemon
 *			- no reply message
 *
 *	Accepts:	- ptr to traced request
 */
void
trqattach(request)

struct trreq		*request;

{
	struct trproc	new;

	new.trp_pid = request->trq_pid;
	new.trp_length = 0;
	new.trp_sig = 0;

	if (al_find(trp_list, &new) != 0) {
		return;
	}

	if (al_insert(trp_list, &new) == 0) {
		panic(errno);
	}
}

/*
 *	trqdetach
 *
 *	Function:	- detaches client from trace daemon
 *
 *	Accepts:	- ptr to traced request
 */
void
trqdetach(request)

struct trreq		*request;

{
	struct trproc	old;
	struct trproc	*match;

	old.trp_pid = request->trq_pid;
	match = (struct trproc *) al_find(trp_list, &old);

	if (match == TRPNULL) {
		return;
	}

	det_length += match->trp_length;

	if (al_delete(trp_list, match) < 0) {
		panic(errno);
	}
}

/*
 *	trqsig
 *
 *	Function:	- signals all current clients
 *			- a big hint to generate store requests
 */
void
trqsig()

{
	struct trproc	*curr;

	curr = (struct trproc *) al_top(trp_list);

	while (curr != TRPNULL) {

		if (kdoom(curr->trp_pid, SIGTRACE) < 0) {
		}

		curr->trp_sig = 1;
		curr = (struct trproc *) al_next(trp_list, curr);
	}
}

/*
 *	trpinc
 *
 *	Function:	- increments total trace length
 *
 *	Accepts:	- client pid
 *			- incremental trace length
 *
 *	Returns:	- ERROR if client not found, else 0
 */
int4
trpinc(pid, length)

int4			pid;
int4			length;

{
	struct trproc	match;
	struct trproc	*found;

	match.trp_pid = pid;

	if ((found = (struct trproc *) al_find(trp_list, &match)) == TRPNULL) {
		return(ERROR);
	}

	found->trp_length += length;
	found->trp_sig = 0;

	return(0);
}

/*
 *	trpzero
 *
 *	Function:	- resets all traces lengths to zero
 */
void
trpzero()

{
	struct trproc	*curr;

	curr = (struct trproc *) al_top(trp_list);

	while (curr != TRPNULL) {
		curr->trp_length = 0;
		curr = (struct trproc *) al_next(trp_list, curr);
	}

	det_length = 0;
}

/*
 *	trqstate
 *
 *	Function:	- traced status
 *			- sends status messages to client
 *			- assumes MAXNMSGLEN > sizeof(struct trstate)
 *
 *	Accepts:	- ptr to traced request
 */
void
trqstate(request)

struct trreq		*request;

{
	struct trreply	reply;
	struct trproc	*curr;
	struct trstate	*pstate;

	curr = (struct trproc *) al_top(trp_list);
	pstate = (struct trstate *) buf;
	reply.trr_nstate = 0;
	reply.trr_errno = 0;
	reply.trr_length = det_length;

	while (curr != TRPNULL) {
		pstate->trs_pid = curr->trp_pid;
		pstate->trs_length = curr->trp_length;
		pstate->trs_sig = curr->trp_sig;
		reply.trr_nstate += 1;

		if (((reply.trr_nstate + 1) * sizeof(struct trstate))
			> MAXNMSGLEN) {

			reply.trr_errno = 0;
			send_reply(request, &reply, reply.trr_nstate *
				sizeof(struct trstate));
			reply.trr_nstate = 0;
			pstate = (struct trstate *) buf;
		} else {
			pstate += 1;
		}

		curr = (struct trproc *) al_next(trp_list, curr);
	}

	reply.trr_errno = EEOF;
	send_reply(request, &reply, reply.trr_nstate * sizeof(struct trstate));
}

/*
 *	trpcompare
 *
 *	Function:	- compares two trp list entries by pid
 *
 *	Accepts:	- ptr to list entry A
 *			- ptr to list entry B
 *
 *	Returns:	- -1 if A < B
 *			- 0 if A == B
 *			- 1 if A > B
 */
static int4
trpcompare(a, b)

struct trproc		*a;
struct trproc		*b;

{
	if (a->trp_pid < b->trp_pid) {
		return(-1);
	} else if (a->trp_pid == b->trp_pid) {
		return(0);
	} else {
		return(1);
	}
}
