/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: fft64.c,v 2.2 92/04/21 11:01:49 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/************************************************************************/
/*									*/
/*	Routine:	fft						*/
/*									*/
/*	Perform 64 bit forward FFT operation on complex array "data".	*/
/*									*/
/*	The forward fft function uses a radix 2 method.  It evaluates:	*/
/*									*/
/*			 N - 1						*/
/*			 \----						*/
/*			  \	  -j(2(PI)/N)kn				*/
/*		X(k)   =   | x(n)e					*/
/*			  /						*/
/*			 /____						*/
/*			 n = 0						*/
/*									*/
/*	This compares with the inverse fft function (function "ifft"),	*/
/*	which evaluates:						*/
/*									*/
/*			 N - 1						*/
/*			 \----						*/
/*			  \	  j(2(PI)/N)kn				*/
/*		X(k)   =   | x(n)e					*/
/*			  /						*/
/*			 /____						*/
/*			 n = 0						*/
/*									*/
/*	Parameters:	"data"		Input data array (COMPLEX).	*/
/*			"logbase2"	The log base 2 of the # of data	*/
/*					elements which are to be done.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/*	Algorithm:	"C" implementation of standard Cooley-Tukey	*/
/*			decimation-in-time FFT with what has now come	*/
/*			to be called the Danielson-Lanczos divide-and-	*/
/*			conquer strategy.				*/
/*									*/
/*	WARNING:	In the literature there are many variations in	*/
/*			definition between which transform is called	*/
/*			"forward" and which "inverse" (not to mention	*/
/*			how the composite 1/N scaling factor should be	*/
/*			applied)!  No scaling is provided in the	*/
/*			forward and reverse transforms in the Logical	*/
/*			Systems package and scaling MUST be done as	*/
/*			desired externally (see transform definitions	*/
/*			above).						*/
/*									*/
/*	Written by Kirk Bailey of Logical Systems.  Copyright 1988 by	*/
/*	Logical Systems.						*/
/*									*/
/************************************************************************/

#include	<stdio.h>
#include	<math.h>

void
fft(COMPLEX data[],int logbase2)
	{
	register	int	i,j,k,m,n,o;
	COMPLEX			*dp,*dk;
	double			r_real,r_imag,s_real,s_imag,t_real,t_imag;
/*
 *	First scramble the input data so that when the fft is done the data
 *	will be in the correct positions.
 */
	m = 1 << logbase2;	/* # of data points to scramble */
	n = m >> 1;
	m--;			/* # of data points to scramble - 1 */
	for(i = j = 0; i < m; j += k, i++)
		{
		if(i < j)
			{
			t_real = data[j].real;
			t_imag = data[j].imag;
			data[j].real = data[i].real;
			data[j].imag = data[i].imag;
			data[i].real = t_real;
			data[i].imag = t_imag;
			}
		for(k = n; k <= j; j -= k, k >>= 1) ;
		}
/*
 *	Now for the meat of the fft.
 */
	for(n = 1, i = 0; i < logbase2; i++)
		{
		o = n;			/* From 1 to 2**(logbase2-1) */
		n <<= 1;		/* From 2 to 2**logbase2 */
		r_real = s_real = cos(t_real = (3.1415926535897932) / o);
		r_imag = s_imag = -sin(t_real); /* Change sign for inverse */
/*
 *	This first loop eliminates the complex multiplies when we know
 *	the value of "r" is unity.
 */
		for(dk = &data[0]; dk < &data[m]; dk += n)
			{
			dp = dk + o;
			t_real = dp->real;
			t_imag = dp->imag;
			dp->real = dk->real - t_real;
			dp->imag = dk->imag - t_imag;
			dk->real += t_real;
			dk->imag += t_imag;
			}
/*
 *	This stuff handles the general case where "r" is not unity.
 */
		for(j = 1; j < o; j++)
			{
			for(dk = &data[j]; dk < &data[m]; dk += n)
				{
				dp = dk + o;
				t_real = (dp->real * r_real) -
						(dp->imag * r_imag);
				t_imag = (dp->real * r_imag) +
						(dp->imag * r_real);
				dp->real = dk->real - t_real;
				dp->imag = dk->imag - t_imag;
				dk->real += t_real;
				dk->imag += t_imag;
				}
			t_real = (r_real * s_real) - (r_imag * s_imag);
			r_imag = (r_real * s_imag) + (r_imag * s_real);
			r_real = t_real;
			}
		}
	}
