/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: ifft32.c,v 2.2 92/04/21 11:01:51 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/************************************************************************/
/*									*/
/*	Routine:	ifftf						*/
/*									*/
/*	Perform 32 bit inverse FFT operation on complex array "data".	*/
/*									*/
/*	The inverse fft function uses a radix 2 method.  It evaluates:	*/
/*									*/
/*			 N - 1						*/
/*			 \----						*/
/*			  \	  j(2(PI)/N)kn				*/
/*		X(k)   =   | x(n)e					*/
/*			  /						*/
/*			 /____						*/
/*			 n = 0						*/
/*									*/
/*	This compares with the forward fft function (function "fftf"),	*/
/*	which evaluates:						*/
/*									*/
/*			 N - 1						*/
/*			 \----						*/
/*			  \	  -j(2(PI)/N)kn				*/
/*		X(k)   =   | x(n)e					*/
/*			  /						*/
/*			 /____						*/
/*			 n = 0						*/
/*									*/
/*	Parameters:	"data"		Input data array (COMPLEXF).	*/
/*			"logbase2"	The log base 2 of the # of data	*/
/*					elements which are to be done.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/*	Algorithm:	"C" implementation of standard Cooley-Tukey	*/
/*			decimation-in-time FFT with what has now come	*/
/*			to be called the Danielson-Lanczos divide-and-	*/
/*			conquer strategy.  Hand-optimized assembly	*/
/*			language version provided for T800.		*/
/*									*/
/*	WARNING:	In the literature there are many variations in	*/
/*			definition between which transform is called	*/
/*			"forward" and which "inverse" (not to mention	*/
/*			how the composite 1/N scaling factor should be	*/
/*			applied)!  No scaling is provided in the	*/
/*			forward and reverse transforms in the Logical	*/
/*			Systems package and scaling MUST be done as	*/
/*			desired externally (see transform definitions	*/
/*			above).						*/
/*									*/
/*	Written by Kirk Bailey of Logical Systems.			*/
/*									*/
/*	Copyright 1988, 1989 by Logical Systems.			*/
/*									*/
/*  EXCEPT "COSF/SINF" CALLS, CONVERTED TO PORTABLE CALLING CONVENTION	*/
/*									*/
/************************************************************************/

#include	<stdio.h>
#include	<math.h>
#include	<inline.h>

void
ifftf(COMPLEXF data[],int logbase2)
	{
#ifdef	DOUBLE32
	#pragma	asm
	.pub	ifft
ifft:
	#pragma	endasm
#endif	DOUBLE32

#ifdef	T800
	#pragma	asm
	.ext	cosf
	.ext	sinf
	_asm_ajw(-14)
;m = 1 << logbase2;	/* # of data points to scramble */
	.LDC	+1
	LDL	[logbase2]
	SHL
	STL	4	;m
;n = m >> 1;
	LDL	4	;m
	.LDC	+1
	SHR
	.LDC	8
	PROD
	STL	5	;n
;m--;			/* # of data points to scramble - 1 */
	LDL	4	;m
	ADC	-1
	.LDC	8
	PROD
	STL	4	;m
;
;	Scramble input data so when FFT is finished the data will be correctly
;	ordered.
;
;i = j = 0;
	.LDC	0
	STL	2	;j
	.LDC	0
	STL	1	;i
?fft4
;i < m
	LDL	4	;m
	LDL	1	;i
	GT
	CJ	@?fft3
;if(i < j)
	LDL	2	;j
	LDL	1	;i
	GT
	CJ	@?fft5
;t_real = data[j].real;
;data[j].real = data[i].real;
;data[i].real = t_real;
	LDL	2	;j
	LDL	[data]
	ADD
	STL	7	;&data[j]
	LDL	1	;i
	LDL	[data]
	ADD		;&data[i]
	DUP
	STL	8	;&data[i]
	LDNL	0	;data[i].real
	LDL	7	;&data[j]
	LDNL	0	;data[j].real
	LDL	8	;&data[i]
	STNL	0	;data[i].real = data[j].real
	LDL	7	;&data[j]
	STNL	0	;data[j].real = (old) data[i].real
;t_imag = data[j].imag;
;data[j].imag = data[i].imag;
;data[i].imag = t_imag;
	LDL	8	;&data[i]
	LDNL	1	;data[i].imag
	LDL	7	;&data[j]
	LDNL	1	;data[j].imag
	LDL	8	;&data[i]
	STNL	1	;data[i].imag = data[j].imag
	LDL	7	;&data[j]
	STNL	1	;data[j].imag = (old) data[i].imag
?fft5
;k = n;
	LDL	5	;n
	STL	3	;k
?fft8
;k <= j;
	LDL	3	;k
	LDL	2	;j
	GT
	EQC	0
	CJ	@?fft7
?fft6
;j -= k, k >>= 1;
	LDL	2	;j
	LDL	3	;k
	SUB
	STL	2	;j
	LDL	3	;k
	.LDC	+1
	SHR
	STL	3	;k
	J	@?fft8
?fft7
?fft2
;j += k, i++
	LDL	2	;j
	LDL	3	;k
	ADD
	STL	2	;j
	LDL	1	;i
	ADC	+8
	STL	1	;i
	J	@?fft4
?fft3
;
;	Compute "&data[m]" and put in place of "m".
;
	LDL	4	;m
	LDL	[data]
	ADD
	STL	4	;m->&data[m]
;
;	Beginning of FFT main loop.
;
	.LDC	8
	STL	6	;o=8
	.LDC	1
	STL	13	;o'=1
	.LDC	16
	STL	5	;n=16
	.LDC	0
	STL	1	;i=0
?fft11
;i < logbase2;
	LDL	[logbase2]
	LDL	1	;i
	GT
	CJ	@?fft10
;r_real = s_real = cosf(t_real = (3.1415926535897932) / o);
	.LDC	?fft12
	FPLDNLSN
	LDLP	13	;o'
	FPI32TOR32
	FPDIV
	LDLP	0	;t_real
	FPSTNLSN
	LDL	0
#ifdef	STATIC_LINK
	LDL	[?slb]
#endif	/* STATIC_LINK */
	CALL	@cosf
	FPDUP
	LDLP	9	;r_real
	LDLP	11	;s_real
	FPSTNLSN
	FPSTNLSN
;r_imag = s_imag = sinf(t_real); /* Change sign for forward */
	LDL	0	;t_real
#ifdef	STATIC_LINK
	LDL	[?slb]
#endif	/* STATIC_LINK */
	CALL	@sinf
	FPDUP
	LDLP	12	;s_imag
	FPSTNLSN
	LDLP	10	;r_imag
	FPSTNLSN
;
;	Loop to handle case where "r" is unity.
;
;dk = &data[0];
	LDL	[data]
	STL	8	;dk
?fft15
;dk < &data[m];
	LDL	4	;&data[m]
	LDL	8	;dk
	GT
	CJ	@?fft14
;dp = dk + o;
	LDL	6	;o
	LDL	8	;dk
	ADD
	DUP
	STL	7	;dp
;t_real = dp->real;
;dp->real = dk->real - t_real;
;dk->real += t_real;
	FPLDNLSN	;dp->real
	FPDUP		;In both FAreg & FBreg
	LDL	8	;dk
	FPLDNLSN	;dk->real
	FPREV
	FPSUB
	LDL	8	;dk
	DUP
	LDL	7	;dp
	FPSTNLSN	;dp->real = dk->real - dp->real
	FPLDNLADDSN	;dk->real + (old) dp->real
	LDL	8	;dk
	ADC	4	;&(dk->imag)
	STL	8	;dk
	LDL	7	;dp
	ADC	4	;&(dp->imag)
	DUP
	STL	7	;dp
	REV
	FPSTNLSN	;dk->real = previous
;t_imag = dp->imag;
;dp->imag = dk->imag - t_imag;
;dk->imag += t_imag;
;dk += (n - 4)
	FPLDNLSN	;dp->imag
	FPDUP		;In both FAreg & FBreg
	LDL	8	;dk
	FPLDNLSN	;dk->imag
	FPREV
	FPSUB
	LDL	8	;dk
	DUP
	LDL	7	;dp
	FPSTNLSN	;dp->imag = dk->imag - dp->imag
	FPLDNLADDSN	;dk->imag + (old) dp->imag
	LDL	5	;n
	ADC	-4	;Compensate for earlier addition of 4
	LDL	8	;dk
	ADD
	STL	8	;dk
	FPSTNLSN	;dk->imag = previous
	J	@?fft15
?fft14
;
;	Loop(s) to handle general case of "r" not unity.
;
;j = 8;
	.LDC	+8
	STL	2	;j
?fft18
;j < o
	LDL	6	;o
	LDL	2	;j
	GT
	CJ	@?fft17
;dk = &data[j];
	LDL	2	;j
	LDL	[data]
	ADD
	STL	8	;dk
?fft21
;dk < &data[m];
	LDL	4	;&data[m]
	LDL	8	;dk
	GT
	CJ	@?fft20
;dp = dk + o;
	LDL	6	;o
	LDL	8	;dk
	ADD
	DUP
	STL	7	;dp
;t_real = (dp->real * r_real) - (dp->imag * r_imag);
;t_imag = (dp->real * r_imag) + (dp->imag * r_real);
;dp->imag = dk->imag - t_imag;
	FPLDNLSN
	LDLP	9	;r_real
	FPLDNLMULSN
	LDLP	10	;r_imag
	LDL	7	;dp
	ADC	+4
	FPLDNLSN
	FPLDNLMULSN
	LDLP	10	;r_imag
	LDL	7	;dp
	LDLP	0	;t_real
	FPSUB
	FPSTNLSN
	FPLDNLSN
	FPLDNLMULSN
	LDLP	9	;r_real
	LDL	7	;dp
	ADC	+4
	FPLDNLSN
	FPLDNLMULSN
	LDL	8	;dk
	ADC	+4
	DUP		;&(dk->imag) in Areg & Breg
	DUP		;&(dk->imag) in Areg & Breg & Creg
	FPADD		;t_imag in FAreg
	FPDUP		;and FBreg
	FPLDNLSN	;dk->imag in FAreg, t_imag in FBreg & FCreg
	FPREV
	FPSUB
	LDL	7	;dp
	ADC	+4
	FPSTNLSN	;Store into dp->imag
;dk->imag += t_imag;
;dp->real = dk->real - t_real;
;dk->real += t_real;
;dk += n
	FPLDNLADDSN	;t_imag in FAreg: dk->imag+t_imag
	LDLP	0	;&t_real
	REV		;&(dk->imag) in Areg,&t_real in Breg
	LDL	8	;&(dk->real)
	REV		;&(dk->imag) in Areg,&(dk->real) in Breg,&t_real in C
	FPSTNLSN	;Store into dk->imag
	FPLDNLSN
	FPDUP
	FPLDNLSN
	FPSUB
	LDL	8	;dk
	LDLP	0	;t_real
	LDL	7	;dp
	FPSTNLSN
	FPLDNLADDSN
	DUP		;&dk in Areg & Breg
	LDL	5	;n
	ADD
	STL	8	;dk
	FPSTNLSN	
	J	@?fft21
?fft20
;t_real = (r_real * s_real) - (r_imag * s_imag);
;r_imag = (r_real * s_imag) + (r_imag * s_real);
;r_real = t_real;
	LDLP	9	;r_real
	FPLDNLSN
	LDLP	11	;s_real
	FPLDNLMULSN
	LDLP	12	;s_imag
	LDLP	10	;r_imag
	FPLDNLSN
	FPLDNLMULSN
	LDLP	12	;s_imag
	LDLP	9	;r_real
	LDLP	0	;t_real
	FPSUB
	FPSTNLSN
	FPLDNLSN
	FPLDNLMULSN
	LDLP	10	;r_imag
	LDLP	11	;s_real
	LDLP	10	;r_imag
	FPLDNLSN
	FPLDNLMULSN
	LDL	0	;t_real
	STL	9	;r_real
	FPADD
	FPSTNLSN
?fft16
;j+=8
	LDL	2	;j
	ADC	+8
	STL	2	;j
	J	@?fft18
?fft17
?fft9
;i++
	LDL	1	;i
	ADC	+1
	STL	1	;i
	LDL	5	;n
	DUP
	STL	6	;o=n (ie. o<<=1)
	DUP
	ADD
	STL	5	;n<<=1
	LDL	13	;o'
	DUP
	ADD
	STL	13	;o'<<=1
	J	@?fft11
?fft10
	_asm_ajw(14)
	RET
	.ALIGN
?fft12
	.REAL32	3.141592653589793
	#pragma	endasm
#else	/* T800 */
	register	int	i,j,k,m,n,o;
	COMPLEXF		*dp,*dk;
	double			r_real,r_imag,s_real,s_imag,t_real,t_imag;
/*
 *	First scramble the input data so that when the fft is done the data
 *	will be in the correct positions.
 */
	m = 1 << logbase2;	/* # of data points to scramble */
	n = m >> 1;
	m--;			/* # of data points to scramble - 1 */
	for(i = j = 0; i < m; j += k, i++)
		{
		if(i < j)
			{
			t_real = data[j].real;
			t_imag = data[j].imag;
			data[j].real = data[i].real;
			data[j].imag = data[i].imag;
			data[i].real = t_real;
			data[i].imag = t_imag;
			}
		for(k = n; k <= j; j -= k, k >>= 1) ;
		}
/*
 *	Now for the meat of the fft.
 */
	for(n = 1, i = 0; i < logbase2; i++)
		{
		o = n;			/* From 1 to 2**(logbase2-1) */
		n <<= 1;		/* From 2 to 2**logbase2 */
		r_real = s_real = cosf(t_real = (3.1415926535897932) / o);
		r_imag = s_imag = sinf(t_real); /* Change sign for forward */
/*
 *	This first loop eliminates the complex multiplies when we know
 *	the value of "r" is unity.
 */
		for(dk = &data[0]; dk < &data[m]; dk += n)
			{
			dp = dk + o;
			t_real = dp->real;
			t_imag = dp->imag;
			dp->real = dk->real - t_real;
			dp->imag = dk->imag - t_imag;
			dk->real += t_real;
			dk->imag += t_imag;
			}
/*
 *	This stuff handles the general case where "r" is not unity.
 */
		for(j = 1; j < o; j++)
			{
			for(dk = &data[j]; dk < &data[m]; dk += n)
				{
				dp = dk + o;
				t_real = (dp->real * r_real) -
						(dp->imag * r_imag);
				t_imag = (dp->real * r_imag) +
						(dp->imag * r_real);
				dp->real = dk->real - t_real;
				dp->imag = dk->imag - t_imag;
				dk->real += t_real;
				dk->imag += t_imag;
				}
			t_real = (r_real * s_real) - (r_imag * s_imag);
			r_imag = (r_real * s_imag) + (r_imag * s_real);
			r_real = t_real;
			}
		}
#endif	/* T800 */
	}
