/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	bootd.c,v $
 * Revision 2.2  92/04/30  10:19:42  trillium
 * Ohio Release
 * 
 *	Function:	- boots neighbour ITB nodes
 *			- maintains "objects" that are needed for booting
 *			  - strap, the initial bootstrap program
 *			  - stage, a table of programs that constitute
 *				the neighbour's system
 *			  - spawn, a table of programs that constitute
 *				the local system
 *			  - route, the complete route table of the whole machine
 *			  - frozen, a frozen version of stage, including
 *				moses commands
 *			- spawn table must only contain re-entrant programs,
 *			  if it is to used for booting
 *			- size of spawn table must reconcile with maconfig.h
 *			- prefix code is bt
 */

#include "bootd.h"
#include <bootreq.h>
#include <events.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <shared.h>

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern int		btaddfrozen();
extern int		btaddstage();
extern int		btaddstrap();
extern int		btaddroute();
extern int		btboot();		/* boot neighbour */
extern int		btflushfrozen();
extern int		btflushstage();
extern int		btflushstrap();
extern int		btflushroute();
extern int		btmap();		/* map stage table */
extern int		btrout();		/* trout neighbour */
extern int		btstagestage();
extern int		btstagestrap();
extern int		btstageroute();
extern void		btinitfrozen();
extern void		btinitstage();
extern void		btinitstrap();
extern void		btinitroute();
extern void		btstate();

/*
 * local functions
 */
static int		btqadd();
static int		btqboot();
static int		btqflush();
static int		btqspread();
static int		btqstage();
static int		btqstate();

/*
 * local variables
 */
static int		(*do_req[BTQNUM])() = {	/* service functions */
				btqboot,	/* boot neighbour ITB node */
				btqadd,		/* add programs, data */
				btqflush,	/* delete all programs */
				btqstage,	/* stage/load programs */
				btqspread,	/* recursive booting */
				btqstate	/* bootd status */
			};

static struct nmsg	nhead;			/* kenya descriptor */

/*
 * global variables
 */
struct bootreq		*request;		/* request from client */
struct bootreply	*reply;			/* reply to client */

main()

{
	int4		btreq;			/* bootd request code */

	if (kinit(0)) {
		panic(errno);
	}

	if (rpattach(LOCAL)) {
		panic(errno);
	}
/*
 * Initialize object packages.
 */
	btinitfrozen();
	btinitstage();
	btinitstrap();
	btinitroute();

	reply = (struct bootreply *) nhead.nh_data;
	request = (struct bootreq *) nhead.nh_data;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;
/*
 * main request processing loop
 */
	for (;;) {
/*
 * Receive a request.
 */
		nhead.nh_event = EVBOOTD;

		if (nrecv(&nhead)) {
			panic(errno);
		}

		btreq = request->btq_request;
		nhead.nh_node = request->btq_src_node;
		nhead.nh_event = request->btq_src_event;
			
		if ((btreq >= 0) && (btreq < BTQNUM)) {
			reply->btr_reply = ((*do_req[btreq])() == ERROR) ?
				(int4) errno : 0;
		} else {
			reply->btr_reply = EBADREQUEST;
		}
/*
 * Send a reply.
 */
		if (btreq != BTQSTATE) {

			if (nsend(&nhead)) {
				panic(errno);
			}
		}
	}
}

/*
 *	btqadd
 *
 *	Function:	- notes the loading of a program or route file
 *	Returns:	- 0 or ERROR
 */

static int
btqadd()

{
	switch(request->btq_object) {

		case BTO_FROZEN:
		return(btaddfrozen());
		break;

		case BTO_STAGE:
		return(btaddstage());
		break;

		case BTO_STRAP:
		return(btaddstrap());
		break;

		case BTO_ROUTE:
		return(btaddroute());
		break;

		default:
		errno = EINVAL;
		return(ERROR);
		break;
	}

	return(0);
}

/*
 *	btqflush
 *
 *	Function:	- frees space controlled by bootd programs
 *			  and/or files
 *	Returns:	- 0 or ERROR
 */

static int
btqflush()

{
	if (request->btq_object &
			~(BTO_FROZEN | BTO_STAGE | BTO_STRAP | BTO_ROUTE)) {
		errno = EINVAL;
		return(ERROR);
	}

	if (request->btq_object & BTO_FROZEN) {
		if (btflushfrozen() < 0) return(ERROR);
	}

	if (request->btq_object & BTO_STAGE) {
		if (btflushstage() < 0) return(ERROR);
	}

	if (request->btq_object & BTO_STRAP) {
		if (btflushstrap() < 0) return(ERROR);
	}

	if (request->btq_object & BTO_ROUTE) {
		if (btflushroute() < 0) return(ERROR);
	}

	return(0);
}

/*
 *	btqstage
 *
 *	Function:	- loads objects onto neighbouring nodes
 *	Returns:	- 0 or ERROR
 */

static int
btqstage()

{
	if (request->btq_object &
			~(BTO_FROZEN | BTO_STAGE | BTO_STRAP | BTO_ROUTE)) {
		errno = EINVAL;
		return(ERROR);
	}

	if (request->btq_object & BTO_FROZEN) {
		if (btstagefrozen(request->btq_node) < 0) return(ERROR);
	}

	if (request->btq_object & BTO_STAGE) {
		if (btstagestage(request->btq_node) < 0) return(ERROR);
	}

	if (request->btq_object & BTO_STRAP) {
		if (btstagestrap(request->btq_node) < 0) return(ERROR);
	}

	if (request->btq_object & BTO_ROUTE) {
		if (btstageroute(request->btq_node) < 0) return(ERROR);
	}

	return(0);
}

/*
 *	btqboot
 *
 *	Function:	- boots neighbour ITB node
 *	Returns:	- 0 or ERROR
 */

static int
btqboot()

{
	if ((request->btq_ilink >= NLINKS) || (request->btq_ilink < 0)) {
		errno = EINVAL;
		return(ERROR);
	}
/*
 * Map the system image, if using the stage table.
 */
	if (request->btq_object & BTO_STAGE) {

		if (btmap() < 0) {
			return(ERROR);
		}
	}
/*
 * Boot the neighbour.
 */
	if (btboot(request->btq_ilink) < 0) {
		return(ERROR);
	}

	return(0);
}

/*
 *	btqspread
 *
 *	Function:	- boots all links in the boot spanning tree
 *			  and then continues recursively down the tree
 *			- always uses BTO_SPAWN downstream
 *	Returns:	- 0 or ERROR
 */

static int
btqspread()

{
	int4		i;		/* favourite index */
	int4		nboot;		/* # of boot neighbours */
	int4		pboot[NLINKS];	/* boot neighbour array */
	int4		ilink;		/* link number */

	getsmart();

	if ((nboot = getntype(NT_BOOT, NT_BOOT)) < 0) {
		return(ERROR);
	}

	if (nboot > 0) {

		if (getnodes(pboot, nboot, NT_BOOT, NT_BOOT) < 0) {
			return(ERROR);
		}

		if (request->btq_object & BTO_STAGE) {

			if (btmap() < 0) {
				return(ERROR);
			}
		}
	}

	for (i = 0; i < nboot; ++i) {
		ilink = getroute(pboot[i]);
		ilink = dl_link(ilink);

		if (btboot(ilink) < 0) {
			return(ERROR);
		}

		if (btrout(pboot[i]) < 0) {
			return(ERROR);
		}

		if (btstagestrap(pboot[i]) < 0) {
			return(ERROR);
		}

		if (btstageroute(pboot[i]) < 0) {
			return(ERROR);
		}
	}

	if (nboot > 0) {

		if (rbtspread(pboot, nboot, BTO_SPAWN | BTO_STRAP | BTO_ROUTE)
				< 0) {
			return(ERROR);
		}
	}
/*
 * Flush all objects.
 */
	if (btflushstrap() < 0) return(ERROR);
	if (btflushroute() < 0) return(ERROR);

	if (request->btq_object & BTO_STAGE) {
		if (btflushstage() < 0) return(ERROR);
	} else if (request->btq_object & BTO_FROZEN) {
		if (btflushfrozen() < 0) return(ERROR);
	}

	return(0);
}

/*
 *	btqstate
 *
 *	Function:	- returns status on bootd
 */

static int
btqstate()

{
	btstate();
}
