/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	btprog.c,v $
 * Revision 2.2  92/04/30  10:19:43  trillium
 * Ohio Release
 * 
 *	Function:	- functions that manipulate a boot program table
 *			- the meat of bootd
 *			- still transputer specific in some spots but
 *			  easy to change
 *			- attempts to be load module format independent
 *			  but there is still some work to be done on this
 */

#include "bootd.h"
#include <bootreq.h>
#include <kio.h>
#include <machine/maconfig.h>
#include <net.h>
#include <pcb.h>
#include <shared.h>
#include <t_types.h>
#include <terror.h>
#include <typical.h>

/*
 * global functions
 */
int			btaddstage();	/* add staged program */
int			btboot();	/* boot neighbour ITB node */
int			btflushstage();	/* flush stage table */
int			btmap();	/* map stage table */
int			btstagestage();	/* stage the stage table */
void			btinitstage();	/* initialize stage table */
void			btstate();	/* send btprog status */

/*
 * external functions
 */
extern int		btaddstage();	/* add program to bootd */
extern int		btexport();	/* boot bootstrap */
extern int		btreset();	/* reset neighbour node */
extern int		btthaw();	/* frozen boot neighbour node */
extern void		btclock();	/* synchronize clocks */
extern void		btmosescall();	/* moses call command */
extern void		btmosesload();	/* moses load command */
extern void		btmosesrunp();	/* moses runp command */
extern void		btmoseswrite();	/* moses write command */
extern void		*toutdataptr();	/* ptr to data segment */
extern void		toutpatch();	/* relocate data segment */
extern void		*toutrelptr();	/* ptr to relocation & symbol tables */
extern void		*touttextptr();	/* ptr to text segment */
extern int4		toutbsssize();	/* size of bss segment */
extern int4		toutdatasize();	/* size of data segment */
extern int4		toutentry();	/* entry point offset */
extern int4		touthtxtsize();	/* size of header + text segment */
extern int4		toutloadsize();	/* total load module size */
extern char		*strncpy();
extern char		*rindex();

/*
 * local functions
 */
static void		btcboot();	/* boot one program */

/*
 * external variables
 */
extern int		errno;
extern struct bootreq	*request;	/* bootd request */

/*
 * local variables
 */
static struct btprog	tstage[BTPMAX];	/* staged boot program table */
static struct btprog	tload[BTPMAX];	/* mapped boot program table */
static int4		nstage;		/* # filled entries in tstage/tload */

/*
 *	btinitstage
 *
 *	Function:	- initializes staged program table
 */

void
btinitstage()

{
	nstage = 0;
}

/*
 *	btaddstage
 *
 *	Function:	- adds new entry in staged program table
 *			- argv must be loaded onto the local node with flatd
 *	Returns:	- 0 or ERROR
 */

int
btaddstage()

{
	int4		i;		/* favourite index */
	struct btprog	*p;		/* favourite ptr */
	char		*arg;		/* argument ptr */

	if (nstage >= BTPMAX) {
		errno = ENOBTDESCRIPTORS;
		return(ERROR);
	}

	p = &tstage[nstage];
	p->btp_argv = (char **) rforget(LOCAL, request->btq_tag);

	if (p->btp_argv == (char **) 0) {
		return(ERROR);
	}
/*
 * Set header/text load module address.
 */
	if (request->btq_flags & BTSHARE) {
/*
 * Look for previous staged program that used the same tag.
 */
		for (i = 0; (i < nstage) &&
			(tstage[i].btp_tag != request->btq_tag); ++i);

		if (i < nstage) {
			p->btp_load = tstage[i].btp_load;
		} else {
			errno = ESHARE;
			return(ERROR);
		}

	} else {
		p->btp_load = (void *) rforget(LOCAL, request->btq_tag);

		if (p->btp_load == (void *) 0) {
			return(ERROR);
		}
	}

	p->btp_tag = request->btq_tag;
	p->btp_data = toutdataptr(p->btp_load);
	p->btp_stksize = request->btq_stksize;
	p->btp_flags = request->btq_flags;
	p->btp_priority = request->btq_priority;
/*
 * Determine argc from the argv structure.
 * Determine the size of the entire argv structure.
 */
	p->btp_avsize = sizeof(int4);

	for (i = 0; p->btp_argv[i]; ++i);
	p->btp_argc = i;

	p->btp_avsize = (p->btp_argc + 1) * sizeof(char *);
	arg = (char *) (p->btp_argv + p->btp_argc + 1);

	for (i = 0; i < p->btp_argc; ++i) {
		p->btp_avsize += strlen(arg) + 1;
		arg += strlen(arg) + 1;
	}

	nstage++;
	return(0);
}

/*
 *	btflushstage
 *
 *	Function:	- frees everything in the staged program table
 *	Returns:	- 0 or ERROR
 */
 
int
btflushstage()

{
	int4		i;		/* favourite index */

	for (i = 0; i < nstage; ++i) {

		if (! (tstage[i].btp_flags & BTSHARE)) {

			if (rfree(LOCAL, (char *) tstage[i].btp_load) < 0) {
				return(ERROR);
			}
		}

		if (rfree(LOCAL, (char *) tstage[i].btp_argv) < 0) {
			return(ERROR);
		}
	}

	nstage = 0;
	return(0);
}

/*
 *	btmap
 *
 *	Function:	- maps images into memory locations and patches
 *			- calculates boot address for all segments
 *			  based on the current state of the staged
 *			  program table
 *			- all sizes and DEF* are assumed to be word
 *			  aligned except argv strings
 *
 *			- memory layout
 *
 *			  -------------------------	higher address
 *			  | argv strings          | 
 *			  -------------------------
 *			  | argv pointers         |
 *			  -------------------------
 *			  | stack		  |
 *			  -------------------------
 *			  | static data/bss       |
 *			  -------------------------
 *			  | header/text           |
 *			  -------------------------	lower address
 *
 *			- header/text may not be present if shared
 *			  from previous program (BTSHARE)
 *
 *	Returns:	- 0 or ERROR
 */

int
btmap()

{
	int4			i;		/* favourite index variable */
	int4			j;		/* second favourite */
	int4			length;		/* of an argument string */
	int4			onchipmark;	/* high water mark onchip */
	int4			watermark;	/* current high water mark */
	char			*arg;		/* ptr to an argument string */
	struct btprog		*local;		/* ptr in stage table */
	struct btprog		*remote;	/* ptr in remote map table */

	watermark = DEFMEMSTART;
	onchipmark = DEFONCHIPSTART;

	for (i = 0; i < nstage; ++i) {
		local = tstage + i;
		remote = tload + i;
/*
 * Set header/text load address.
 */
		if (local->btp_flags & BTSHARE) {

			for (j = 0;
				(j < i) &&
				(local->btp_load != tstage[j].btp_load);
				 j++);

			if (j < i) {
				remote->btp_load = tload[j].btp_load;
			} else {
				errno = ESHARE;
				return(ERROR);
			}
		} else {
			remote->btp_load = (void *) watermark;
			watermark += touthtxtsize(local->btp_load);
		}
/*
 * Set data load address.
 */
		remote->btp_data = (void *) watermark;
		watermark += toutdatasize(local->btp_load) +
			toutbsssize(local->btp_load);
/*
 * The special memory region, ONCHIP, is transputer specific.
 * We need to generalize the placement of segments into memory regions.
 */
		if (local->btp_flags & BTONCHIPSTK) {
			remote->btp_stack = (void *) onchipmark;
			onchipmark += local->btp_stksize;
		} else {
			remote->btp_stack = (void *) watermark;
			watermark += local->btp_stksize; 
		}

		remote->btp_stksize = local->btp_stksize;
/*
 * Set argv load address.
 */
		remote->btp_argc = local->btp_argc;
		remote->btp_avsize = local->btp_avsize;

		remote->btp_argv = (char **) watermark;
		watermark += (local->btp_argc + 1) * sizeof(char *);
		arg = (char *) (local->btp_argv + local->btp_argc + 1);

		for (j = 0; j < local->btp_argc; ++j) {
			local->btp_argv[j] = (char *) watermark;
			length = strlen(arg) + 1;
			watermark += length;
			arg += length;
		}
/*
 * Round high water up to a word address.
 */
		watermark = (watermark + (sizeof(int4) - 1))
			& ~(sizeof(int4) - 1);
/*
 * Patch load module.
 */
		toutpatch(local->btp_load, toutrelptr(local->btp_load),
			local->btp_data, touttextptr(remote->btp_load),
			remote->btp_data);

		remote->btp_flags = local->btp_flags;
		remote->btp_priority = local->btp_priority;
		remote->btp_tag = local->btp_tag;
	}

	if ((watermark > MAIN_START) || (onchipmark > ONCHIP_START)) {
		errno = EHIGHWATER;
		return(ERROR);
	}

	return(0);
}

/*
 *	btboot
 *
 *	Function:	- boots programs over a link with moses running
 *			  on the other side
 *			- uses either spawn table or stage/load tables,
 *			  as selected by client
 *			- calls btcboot() on each entry in a program table
 *			- see btmap() for a memory layout map
 *
 *	Accepts:	- link number to neighbour node
 *	Returns:	- 0 or ERROR
 */

int
btboot(ilink)

int4			ilink;

{
	struct btprog	*local;		/* local program addresses */
	struct btprog	*remote;	/* remote program addresses */
	int4		nprog;		/* # programs in table */
	int4		n;		/* favourite counter */

	if ((ilink >= NLINKS) || (ilink < 0)) {
		errno = EINVAL;
		return(ERROR);
	}
/*
 * Set local and remote boot program table.
 */
	if (request->btq_object & BTO_SPAWN) {
		local = (struct btprog *) BOOTSPAWN;
		remote = (struct btprog *) BOOTSPAWN;
		nprog = *((int4 *) BOOTNSPAWN);
	} else if (request->btq_object & BTO_STAGE) {
		local = tstage;
		remote = tload;
		nprog = nstage;
	}
/*
 * Export bootstrap.
 */
	if (request->btq_object & BTO_STRAP) {

		if (btreset(ilink) < 0) {
			return(ERROR);
		}

		if (btexport(ilink) < 0) {
			return(ERROR);
		}
/*
 * Synchronize the clocks.
 */
		btclock(ilink);
	}
/*
 * Frozen boot.  This must include the spawn table if spreading later.
 */
	if (request->btq_object & BTO_FROZEN) {

		if (btthaw(ilink) < 0) {
			return(ERROR);
		}
	}
/*
 * Load spawn table and boot all processes in list.
 */
	else if (request->btq_object & (BTO_STAGE | BTO_SPAWN)) {
		btmoseswrite(ilink, (int4 *) BOOTNSPAWN, nprog);
		btmosesload(ilink, (void *) remote, (void *) BOOTSPAWN,
			(int4) (nprog * sizeof(struct btprog)));

		for (n = 0; n < (nprog - 1); ++n) {
			btcboot(ilink, local + n, remote + n, btmosesrunp);
		}

		if (n < nprog) {
			btcboot(ilink, local + n, remote + n, btmosescall);
		}
	}

	return(0);
}

/*
 *	btcboot
 *
 *	Function:	- boots a load module through a link
 *			- local is a program table entry with local addresses
 *			  of a load module to be booted
 *			- remote is a program table entry with remote addresses
 *			  of a load module to be booted
 *			- assumes load module has been patched
 *			- load module must be re-entrant if spawn table
 *			  is used as local
 *
 *	Accepts:	- link index
 *			- local program descriptor ptr
 *			- remote program descriptor ptr
 *			- ptr to btmosesrunp() or btmosescall()
 */

static void
btcboot(ilink, local, remote, create)

int4				ilink;		/* link number */
struct btprog			*local;		/* local program addresses */
struct btprog			*remote;	/* remote program addresses */
void				(*create)();	/* 1 of 2 moses create funcs */

{
	int4			ip;		/* instruction pointer */
	int4			stacktop;	/* top of stack address */
	struct pcb_t		newpcb;		/* newborn PCB */

/*
 * Load header and text (the load segment) if necessary.
 */
	if (! (local->btp_flags & BTSHARE)) {
		btmosesload(ilink, local->btp_load, remote->btp_load,
			touthtxtsize(local->btp_load));
	}
/*
 * Load data.
 */
	btmosesload(ilink, local->btp_data, remote->btp_data,
		toutdatasize(local->btp_load));
/*
 * Load argv.
 */
	btmosesload(ilink, (void *) local->btp_argv, (void *) remote->btp_argv,
		local->btp_avsize);
/*
 * Initialize process control block.
 */
	stacktop = (int4) remote->btp_stack + remote->btp_stksize;
/*
 * The PCB structure has not been updated to modern times, so
 * we do some stupid casting here.
 */
	newpcb.pc_pid = (int) ((stacktop + 1) & INT4_MAX);
	newpcb.pc_header = (struct exec *) remote->btp_load;
	newpcb.pc_nodeid = NOTNODEID;
	newpcb.pc_argc = (int) remote->btp_argc;
	newpcb.pc_argv = remote->btp_argv;
	newpcb.pc_rtf = RTF_SYSGEN;
	newpcb.pc_stack = (char *) remote->btp_stack;
	newpcb.pc_stksize = (int) remote->btp_stksize;
	newpcb.pc_stdin = 0;
	newpcb.pc_stdout = 1;
	newpcb.pc_stderr = 2;

	stacktop -= sizeof(newpcb);
	btmosesload(ilink, (void *) &newpcb, (void *) stacktop,
		(int4) sizeof(newpcb));
/*
 * This section of code is strongly related to the transputer.
 * I have not yet figured out how to make it portable without great pain.
 */
	btmoseswrite(ilink, ((int4 *) stacktop) - 1, stacktop);
	btmoseswrite(ilink, ((int4 *) stacktop) - 2, (int4) remote->btp_data);
/*
 * Start load module running.  Use a call to transfer control away
 * from moses if this is the last program.  Otherwise create a new process.
 */
	ip = (int4) remote->btp_load;
	ip += toutentry(local->btp_load);

	if (local->btp_priority) {
		stacktop = (stacktop - (3 * sizeof(int4))) + 1;
	} else {
		stacktop = stacktop - (3 * sizeof(int4));
	}

	(*create)(ilink, (void *) ip, (void *) stacktop);
/*
 * end of transputer related code
 */
}

/*
 *	btstagestage
 *
 *	Function:	- loads all programs in stage table onto remote node
 *			  and registers each with remote bootd
 *			- uses rflat() to load program and argv structure
 *	Accepts:	- destination nodeid
 *	Returns:	- 0 or ERROR
 */

int
btstagestage(nodeid)

int4			nodeid;

{
	struct btprog		*pstage;	/* ptr to staged program */
	int4			n;		/* favourite counter */
	int4			loadsize;	/* total load module size */

	for (n = 0; n < nstage; ++n) {

		pstage = tstage + n;
/*
 * Load the argv table.
 */
		if (rflat(nodeid, (char *) pstage->btp_argv,
			pstage->btp_avsize, pstage->btp_avsize,
			pstage->btp_tag) < 0) {

			return(ERROR);
		}
/*
 * Do we want to load the load module?
 */
		if (!(pstage->btp_flags & BTSHARE)) {
			loadsize = toutloadsize(pstage->btp_load);

			if (rflat(nodeid, (char *) pstage->btp_load,
				loadsize, loadsize, pstage->btp_tag) < 0) {

				return(ERROR);
			}
		}
/*
 * Add the loaded program to the remote bootd.
 */
		if (rbtaddstage(nodeid, pstage->btp_tag,
			pstage->btp_flags, pstage->btp_stksize,
			pstage->btp_priority) < 0) {

			return(ERROR);
		}
	}

	return(0);
}

/*
 *	btstate
 *
 *	Functions:	- sends state information on the program tables
 */

void
btstate()

{
	int4			n;		/* favourite counter */
	int4			nprog;		/* # programs in table */
	struct btprog		*pprog;		/* ptr to program desc. */
	struct bootreply	*reply;
	struct nmsg		nhead;
	struct bootstate	state;
	char			*arg0;		/* argv[0] */
	char			*tail;		/* tail of argv[0] */

	if (request->btq_object & BTO_SPAWN) {
		pprog = (struct btprog *) BOOTSPAWN;
		nprog = *((int4 *) BOOTNSPAWN);
	} else {
		pprog = tstage;
		nprog = nstage;
	}

	if (nprog > request->btq_maxprog) nprog = request->btq_maxprog;

	reply = (struct bootreply *) nhead.nh_data;
	reply->btr_reply = 0;
	reply->btr_nprog = nprog;

	nhead.nh_node = request->btq_src_node;
	nhead.nh_event = request->btq_src_event;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = sizeof(struct bootstate);
	nhead.nh_msg = (char *) &state;

	if (nprog == 0) {

		if (nsend(&nhead) < 0) {
			panic(errno);
		}
	} else {

		for (n = 0; n < nprog; ++n) {
			state.bts_load = (void *) ltot((int4) pprog->btp_load);
			state.bts_data = (void *) ltot((int4) pprog->btp_data);
			state.bts_stksize = ltot(pprog->btp_stksize);
			state.bts_flags = ltot(pprog->btp_flags);
			state.bts_priority = ltot(pprog->btp_priority);
			
			if (pprog->btp_argc > 0) {
				arg0 = (char *)
					(pprog->btp_argv + pprog->btp_argc + 1);
				tail = rindex(arg0, STRDIR);

				if (! tail) tail = arg0;
				else ++tail;

				(void) strncpy(state.bts_name,
					tail, BTMAXNAME);
				state.bts_name[BTMAXNAME - 1] = '\0';
			} else {
				state.bts_name[0] = '\0';
			}

			if (nsend(&nhead) < 0) {
				panic(errno);
			}

			pprog++;
		}
	}
}
