/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	pstate.c,v $
 * Revision 2.2  92/04/30  10:31:35  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  03:02:26  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/19  23:55:36  gdburns
 * Be more careful with NOBUF.
 * 
 * Revision 2.0.1.2  90/11/29  19:46:47  gdburns
 * Shared.h is dangerous and should be listed last.
 * 
 * Revision 2.0.1.1  90/07/13  19:27:05  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:15  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:02:28  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0.1.1  88/02/29  17:14:27  gdburns
 * State would not work with processes not attached to kenyad.
 * 
 * Revision 1.0  87/12/19  15:43:29  trillium
 * Initial revision
 * 
 *
 *	Function:	- implement the PSTATE request
 *			- may send more than one reply if several process
 *			  states are requested
 *			- check for a match on each used process descriptor
 *			- builds state and sends reply for each descriptor
 *			  that matches true
 *			- if no process is matched and a pid is supplied,
 *			  attempts to acquire state from the kernel
 */

#include <typical.h>
#include <terror.h>
#include "kenya.h"
#include <kreq.h>
#include <preq.h>
#include <shared.h>

/*
 * external variables
 */
extern struct pprocess	ptable[];		/* kenya process table */
extern struct preq	request;		/* kenya request */
extern int		errno;

/*
 * external functions
 */
extern char		*strncpy();

/*
 * local variables
 */
static struct pstate	reply_state;		/* kenya state reply */


void
do_pstate()

{
	int		i,			/* favourite index variable */
			matched,		/* result of match function */
			found_match = FALSE;	/* did anything match? */


	reply_state.ps_reply = 0;

	for (i = 0; (i < PMAX) && !reply_state.ps_reply; ++i) {

	    matched = match(ptable + i, request.pq_sel_method,
				request.pq_sel_value);

	    if (matched) {
		found_match = TRUE;

		if (matched == ERROR) {
		    reply_state.ps_reply = 0;
		}

		else if (build_pstate(i)) {

		    if (errno != ENOTPROCESS) {
			reply_state.ps_reply = errno;
		    }
		}
/*
 * Send a nice clean state reply back to the client.
 */
		else if (send_reply((char *) &reply_state, 
					sizeof(reply_state), NOBUF)) {
		    panic(errno);
		}
	    }
	}

	if (!found_match) {
/*
 * No matching kenya process was found.  Try to build kernel state only
 * if a pid is supplied (because the pid is the only key the kernel will take).
 */
	    if (request.pq_sel_method & SELECT_PID) {

		if (kstate(request.pq_sel_value, &reply_state.ps_kernel)) {
			reply_state.ps_reply = errno;
		}

		else {
/*
 * Fill in the non-existant kenya info.
 */
		    reply_state.ps_kenya.p_pid = request.pq_sel_value;
		    reply_state.ps_kenya.p_loadpt = (struct exec *) 0;
		    reply_state.ps_kenya.p_stack = (char *) 0;
		    reply_state.ps_kenya.p_static = (char *) 0;
		    reply_state.ps_kenya.p_flags = 0;
		    reply_state.ps_index = 0;

		    (void) strncpy(reply_state.ps_name, "<unknown>", PSMAXNAME);
/*
 * Send a state reply back to the client.
 */
		    if (send_reply((char *) &reply_state,
				sizeof(reply_state), NOBUF)) {
			panic(errno);
		    }
		}
	    }
	}
/*
 * Send EEOF if no previous error has been detected.
 */
	if (! reply_state.ps_reply) {
	    reply_state.ps_reply = EEOF;
	}
/*
 * This is the last reply and the code will always indicate an error.
 */
	if (send_reply((char *) &reply_state, sizeof(reply_state), NOBUF)) {
	    panic(errno);
	}
}


/*
 *	build_pstate
 *
 *	Function:	- build the complete state of a process
 *			- state information is acquired partly from kenya's
 *			  process table and partly via the kstate kernel request
 *			- fills in argv[0] as the program name if present
 *
 *	Accepts:	- index of a process descriptor
 *
 *	Returns:	- 0 or ERROR
 */

static int
build_pstate(istate)

int			istate;			/* index of state process */

{
	struct pprocess	*pstate;		/* pointer to state process */


	pstate = ptable + istate;
/*
 * Fill in the kenya info.
 */
	bcopy((char *) pstate, (char *) &reply_state.ps_kenya,
		sizeof(struct pprocess));
/*
 * Fill in process index.
 */
	reply_state.ps_index = istate + 1;
/*
 * Fill in argv[0], usually the program name, if it exists.
 */
	if (pstate->p_argc != 0) {
	    (void) strncpy(reply_state.ps_name, *pstate->p_argv, PSMAXNAME);
	}

	else {
	    (void) strncpy(reply_state.ps_name, "<unknown>", PSMAXNAME);
	}
/*
 * Fill in the kernel info.
 */
	if (kstate(pstate->p_pid, &reply_state.ps_kernel)) {

	    if (errno == ENOTPROCESS) {
/*
 * We assume the process has not attached to the kernel.
 */
		reply_state.ps_kernel.ks_state = SFREE;
		reply_state.ps_kernel.ks_event = 0;
		reply_state.ps_kernel.ks_type = 0;
		reply_state.ps_kernel.ks_priority = 0;
		reply_state.ps_kernel.ks_signal = 0;
	    }

	    else {
		return(ERROR);
	    }
	}

	return(0);
}
