/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	madesc.c,v $
 * Revision 2.2  92/04/30  10:39:43  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/11/10  23:35:53  gdburns
 * Minor overhaul fixes several problems.
 * 
 * Revision 2.1.1.1  91/10/09  22:30:03  gdburns
 * Allow descriptor blocks to be reported by mstate.
 * 
 * Revision 2.1  91/03/20  03:03:12  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/09/13  11:19:36  gdburns
 * Removed machine dependencies into separate routines and header files.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:40  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:02  trillium
 * Homo erectus
 *
 *	Function:	- handle the peculiar problems of the descriptor list
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <kernel.h>
#include <machine/maconfig.h>
#include <mareq.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* error number */
extern struct memblock	*empty;		/* empty descriptor list */
extern struct memblock	*tbldesc;	/* descriptor table list */
extern struct memblock	*memalloc[];	/* allocated descriptor list */
extern struct memblock	*memfree[];	/* free descriptor list */
extern int		nempty;		/* # empty descriptors */

/*
 * external functions
 */
extern struct memblock	*delete();
extern void		insert();

/*
 * global functions
 */
void			initdesc();
struct memblock		*getdesc();

/*
 *	getdesc
 *
 *	Function:	- get a descriptor from the empty list
 *			- usually a trivial operation
 *			- If we run out of descriptors, we have to get more
 *			  by allocating some of the same memory that we are
 *			  managing for clients.
 *
 *	Returns:	- ptr to a memory block descriptor
 */

struct memblock
*getdesc()

{
	struct memblock	*auxdesc;	/* auxiliary descriptor table */
	struct memblock	*pdesc;		/* ptr to descriptor */
	int		tblsize;	/* #bytes in descriptor table */
	int		i;		/* favourite index */
	int		key[2];		/* search key for BYSIZE */

	if (! empty) {
/*
 * We are fresh out of descriptors!
 * Look for a block of new descriptors.
 */
		tblsize = MAMAX * sizeof(struct memblock);
		auxdesc = MANULL;
		key[0] = tblsize;
		key[1] = sizeof(int);

		for (i = 0; (i < NMAIN) && ! auxdesc; ++i) {
			auxdesc = delete(&memfree[i], (int) key, BYSIZE);
		}

		if (auxdesc) {
			initdesc((struct memblock *) auxdesc->ma_start, i - 1);
			auxdesc->ma_start += tblsize;
			auxdesc->ma_size -= tblsize;
	
			if (auxdesc->ma_size > 0) {
				insert(&memfree[i - 1], auxdesc, NOCOMPACT);
			} else {
				auxdesc->ma_start = (char *) 0;
				insert(&empty, auxdesc, NOCOMPACT);
				++nempty;
			}
		}

		else {
/*
 * We don't have enough space to allocate more descriptors.
 */
			errno = ENOMEM;
			return(MANULL);
		}
	}

	pdesc = delete(&empty, (int) empty, BYPTR);
	return(pdesc);
}

/*
 *	initdesc
 *
 *	Function:	- initialize empty memory block descriptor table
 *
 *	Accepts:	- ptr to descriptor table
 *			- region index from whence table came
 *			  (array index, not actual region number,
 *			   since this is completely internal)
 */

void
initdesc(tbl, iregion)

struct memblock		tbl[];
int			iregion;

{
	int		i;			/* favourite index variable */
	struct memblock	*new;			/* desc. table entry */

	for (i = 0; i < (MAMAX - 1); ++i) {
		tbl[i].ma_next = &tbl[i + 1];
		tbl[i].ma_start = (char *) 0;
	}

	tbl[MAMAX - 1].ma_next = MANULL;
	tbl[MAMAX - 1].ma_start = (char *) 0;
/*
 * Refresh the empty list.
 */
	empty = tbl;
	nempty += MAMAX;
/*
 * The purpose of the tbldesc is to keep a list of the descriptor *tables*
 * so that we can send a few large contiguous messages in reply to an
 * MSTATE request.
 */
	new = delete(&empty, (int) empty, BYPTR);
	new->ma_start = (char *) tbl;
	new->ma_pid = getpid();
	new->ma_iregion = iregion;
	insert(&tbldesc, new, NOCOMPACT);
	--nempty;
}

/*
 *	freedesc
 *
 *	Function:	- compacts descriptors such that whole blocks
 *			  of descriptors, as allocated in getdesc(),
 *			  can be freed
 */

void
freedesc()

{
	struct memblock	*tbl;			/* descriptor table */
	struct memblock	*old;			/* existing used desc. */
	struct memblock	*new;			/* new unused desc. */
	struct memblock	**toplist;		/* free or alloc list */
	struct memblock	**ptop;			/* ptr to top of list  */
	int		i;			/* favourite index */
	int		j;			/* second favourite index */
	int		iregion;		/* region index */

	tbl = delete(&tbldesc, 0, BYLAST);
	iregion = tbl->ma_iregion;
	tbl = (struct memblock *) tbl->ma_start;

	for (i = 1; i < MAMAX; ++i) {
/*
 * If descriptor is used, delete it from the correct list, copy it to
 * an empty descriptor and re-insert it in the list.
 */
		if (tbl[i].ma_start) {
			new = delete(&empty, (int) empty, BYPTR);

			if (! new) {
				panic(EIMPOSSIBLE);
			}

			new->ma_start = tbl[i].ma_start;
			new->ma_size = tbl[i].ma_size;
			new->ma_pid = tbl[i].ma_pid;
			new->ma_node = tbl[i].ma_node;

			toplist = (tbl[i].ma_pid == NOTPROCESS) ?
				memfree : memalloc;
			old = MANULL;

			for (j = 0; (j < NREGIONS) && !old; ++j) {
				ptop = &toplist[j];
				old = delete(ptop, (int) &tbl[i], BYPTR);
			}

			if (! old) {
				panic(EIMPOSSIBLE);
			}

			insert(ptop, new, NOCOMPACT);
		}
/*
 * If descriptor is empty, delete it from the empty list and forget it.
 */
		else {
			(void) delete(&empty, (int) &tbl[i], BYPTR);
		}

		--nempty;
	}
/*
 * Obtain an empty descriptor and insert the now totally unused former
 * descriptor block into a free list.
 */
	if (! empty) {
		panic(EIMPOSSIBLE);
	}

	new = delete(&empty, (int) empty, BYPTR);
	--nempty;
	new->ma_start = (char *) tbl;
	new->ma_pid = NOTPROCESS;
	new->ma_size = MAMAX * sizeof(struct memblock);
	insert(&memfree[iregion], new, COMPACT);
}
