/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	malloc.c,v $
 * Revision 2.2  92/04/30  11:49:15  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/03/16  10:30:47  gdburns
 * Add i860 hooks.
 * Replace bcopy() with memcpy().
 * 
 * Revision 2.1.1.1  91/12/16  21:58:34  gdburns
 * Use RTF_MALLOC as the initialization switch.
 * 
 * Revision 2.1  91/03/20  13:11:46  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:32:58  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:58:00  trillium
 * Homo erectus
 *
 *	Function:	- Moshe's malloc
 *			- allows for restart when spawned to another node
 *
 *			- our own malloc() scheme, quick for small blocks.
 *			- calls rmalloc() only when necessary.
 *			- adapted from K&R, p174.
 * 
 *			- modified 890523 to _do_ implicit rfree()s, at the
 *			  expense of higher memory overhead.
 *			- also moved some things to a header file, malloc.h.
 *
 *			- further modified 890629 to detect changes in the PID
 *			  and start afresh (malloc) or return an error (free).
 *			  This is necessary for using malloc/free safely in
 *			  threads started by kpar_Fe().
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *			- 2.0 Author: MB
 */

#include <kio.h>
#include <net.h>
#include <terror.h>
#include <shared.h>

#include <malloc.h>

#ifdef i860
#define REGION		3
#else
#define REGION		0
#endif

/******** type definitions: ********/

union alsize {
	unsigned int	size;
	ALIGNER		aligner;
};

union alptr {
	struct mheader	*ptr;		/* if free: next free block */
					/* if malloced: rmalloced block */
	ALIGNER		aligner;
};

struct mheader {
	union alptr	p;
	union alsize	s;
};

typedef struct mheader MHEADER;


/******** macro definitions: ********/

/*
 *  This is an arbitrary value that we put in the upper 8 bits
 *  of size fields to serve as a magic number.  That gives us
 *  a good chance of detecting corruption of memory early.
 */
#define FREE	0x1D000000

#undef NOTPID
#define NOTPID	0xFFFF1E00

#undef NULL
#define NULL 0

#define mnext(q)  ((q)->p.ptr)
#define mcore(q)  ((q)->p.ptr)
#define msize(p) (((p)->s.size)&0x00FFFFFF)
#define magic(p) (((p)->s.size)&0xFF000000)
#define tomagic(p) (((p)<<16)&0xFF000000)


/******** global variables: ********/

static int		pidmagic;
static MHEADER		base;
static MHEADER		*mallocp;


/******** external variables and functions: ********/

extern int		errno;
extern struct kio_t	_kio;


/******** functions: ********/


/*
 *  This function is called from freeall().
 *  The idea is to keep mallocp static...
 */
void
_reset_malloc()
{
	mallocp = NULL;
}


/*
 *  This is the implementation of the standard library function.
 */
int
free (ap)
	char *ap;
{
	register MHEADER *p, *q, *r;
	MHEADER *core;
	unsigned int count;
/*
 *  Check the rtf.  If reset then restart.
 */
	if (! (_kio.ki_rtf & RTF_MALLOC)) {
		_kio.ki_rtf |= RTF_MALLOC;
		pidmagic = tomagic(getpid());
		mallocp = NULL;
	}

	/*
	 *  Check arguments for sanity.
	 */
	if (ap == NULL || mallocp == NULL) {
		errno = ENOTMALLOCED;
		return (-1);
	}
	ap -= sizeof (MHEADER);
	p = (MHEADER *) ap;
	if (magic(p) != pidmagic) {
		errno = ENOTMALLOCED;
		return (-1);
	}

	/*
	 *  Find the block being free()d in the malloc list
	 *  and free it.
	 */
	for (q=mallocp, r=mnext(q);  p<=q || p>=r;  q=r, r=mnext(q)) {
		if (q >= r && (p > q || p < r))
			break;
	}
	mallocp = q;
	p->s.size = (FREE | msize(p));		/* not malloced any more */

	/*
	 *  The pointer to the "core" mheader in a free block is stored in
	 *  the 2nd "unit", while in a malloced block it is stored in the
	 *  first, so need to copy it.  Decrement the count of malloc()s
	 *  on the rmalloc()ed core block upon free().
	 */
	core = mcore(p);
	if (magic(core) != pidmagic) {
		errno = EGARBAGE;
		return (-1);
	}
	mcore(p+1) = core;
	count = msize(core) - 1;
	core->s.size = (pidmagic | count);

	/*
	 *  Coalesce free()d block with blocks above and below
	 *  if they are adjacent.
	 */
	if (p+msize(p) == r) {
		p->s.size += msize(r);
		mnext(p) = mnext(r);
		r = p;
	} else
		mnext(p) = r;
	if (q+msize(q) == p) {
		q->s.size += msize(p);
		mnext(q) = mnext(p);
		p = q;
	} else
		mnext(q) = p;

	if (count > 0)
		return (0);

	/*
	 *  If the count is down to 0, rfree() the core block, returning
	 *  it to the system.  But first must unlink the core block
	 *  from the malloc list.
	 */
	if (p != q)			/*  p did not coalesce with q */
		mnext(q) = mnext(p);
	else {
		/* search for q's predecessor */
		while (mnext(p) != q)
			p = mnext(p);
		mnext(p) = mnext(q);
		mallocp = p;
	}
	(void) rfree (LOCAL, (char *) core);
	return (0);
}


/*
 *  This function gets a new big block from the system.
 */
static MHEADER *
morecore (nu)
	unsigned int nu;	/* minimum number of units needed */
{
	char *cp;
	MHEADER *up, *core;
	int rnu;

	/*
	 *  Add one unit for the "core" header.
	 *  Round the amount up to a multiple of ALLOCBLK.
	 *  Make it at least ALLOCMIN.
	 */
	rnu = ALLOCBLK * (((nu + 1) + (ALLOCBLK - 1)) / ALLOCBLK);
	if (rnu < ALLOCMIN)
		rnu = ALLOCMIN;

	/*
	 *  Get the block:
	 */
	cp = (char *) rmalloc (LOCAL, rnu * sizeof(MHEADER), REGION, 0);
	if (cp == NULL)
		return (NULL);

	/*
	 *  Reserve the "core" header.
	 *  Initialize the core and the pointer to the core.
	 *  Insert the new block into the malloc lists via free().
	 */
	core = (MHEADER *) cp;
	core->s.size = (pidmagic | 9999);	/* fake it to avoid rfree */
	up = core + 1;
	up->s.size = (pidmagic | (rnu-1));	/* simulate malloc()ed block */
	mcore(up) = core;			
	if (free (cp + 2 * sizeof (MHEADER)))
		return (NULL);
	core->s.size = (pidmagic | 0);		/* now the real count */

	return (mallocp);
}


/*
 *  This is the implementation of the standard library function.
 */
char *
malloc (nbytes)
	unsigned int nbytes;
{
	register MHEADER *p, *q;
	register unsigned int nunits;
	MHEADER *core;
	unsigned int count;
/*
 *  Check the rtf.  If reset then restart.
 */
	if (! (_kio.ki_rtf & RTF_MALLOC)) {
		_kio.ki_rtf |= RTF_MALLOC;
		pidmagic = tomagic(getpid());
		mallocp = NULL;
	}

	/*
	 *  Check arguments for sanity.
	 */
	if (nbytes == 0) {
		errno = ENOMEM;
		return (NULL);
	}

	/*
	 *  Round up to units needed.
	 *  Add one unit for overhead.
	 */
	nunits = 1 + (nbytes+(sizeof(MHEADER)-1)) / sizeof(MHEADER);

	/*
	 *  Find the place in the list to search from.
	 *  If first call (detect by NULL mallocp), initialize things.
	 */
	q = mallocp;
	if (q == NULL) {
		base.p.ptr = mallocp = q = &base;
		base.s.size = (FREE | 0);
	}

	/*
	 *  Find a big enough hole in free list.
	 *  If none available, ask system for more core.
	 *  When found, copy "core" pointer over "next" pointer,
	 *  and increment count of malloc()s on core block.
	 */
	for (p=mnext(q);  /* forever */ ;  q=p, p=mnext(p)) {
		if (magic(p) != FREE) {
			errno = EGARBAGE;	/* list is corrupted */
			return (NULL);
		}
		if (msize(p) >= nunits) {
			mallocp = q;
			core = mcore(p+1);
			if (magic(core) != pidmagic) {
				errno = EGARBAGE;
				return (NULL);
			}
			if (msize(p) <= nunits+2) {
				mnext(q) = mnext(p);
				p->s.size = (pidmagic | msize(p));
			} else {
				p->s.size -= nunits;
				p += msize(p);
				p->s.size = (pidmagic | nunits);
			}
			mcore(p) = core;
			count = msize(core) + 1;
			core->s.size = (pidmagic | count);
			return ((char *) &p[1]);
		}
		if (p == mallocp) {
			p = morecore (nunits);
			if (p == NULL)
				return (NULL);
		}
	}
}


/*
 *  This is a standard library function.
 */
char *
calloc (nelem, elsize)
	unsigned int nelem, elsize;
{
	register ALIGNER *p, *top;
	register unsigned int nbytes, nwords;
	char *cp;

	nbytes = nelem * elsize;
	cp = malloc (nbytes);
	if (cp == NULL)
		return (NULL);
	/* need to fill with 0's: do it a word at a time! */
	nwords = (nbytes + (sizeof(ALIGNER)-1)) / sizeof(ALIGNER);
	p = (ALIGNER *) cp;
	top = p + nwords;
	while (p < top)
		*p++ = (ALIGNER) 0;
	return (cp);
}


/*
 *  This implementation of the standard library function realloc()
 *  MUST be given a pointer that was malloc()ed and NOT free()d.
 */
char *
realloc (ap, nbytes)
	char *ap;
	unsigned int nbytes;
{
	register MHEADER *p, *q, *r;
	register unsigned int nunits, size;
	MHEADER *core;
	char *cp;
	unsigned int obytes;
/*
 *  Check the rtf.  If reset then restart.
 */
	if (! (_kio.ki_rtf & RTF_MALLOC)) {
		_kio.ki_rtf |= RTF_MALLOC;
		pidmagic = tomagic(getpid());
		mallocp = NULL;
		errno = ENOTMALLOCED;
		return (NULL);
	}

	/*
	 *  Round the amount up.
	 *  Check for errors.
	 */
	if (nbytes == 0) {
		errno = ENOMEM;
		return (NULL);
	}
	nunits = 1 + (nbytes+(sizeof(MHEADER)-1)) / sizeof(MHEADER);
	if (ap == NULL) {
		errno = ENOTMALLOCED;
		return (NULL);
	}
	cp = ap - sizeof (MHEADER);
	p = (MHEADER *) cp;
	if (magic(p) != pidmagic) {
		errno = ENOTMALLOCED;
		return (NULL);
	}

	/*
	 *  Find p's place in the free list
	 */
	for (q=mallocp, r=mnext(q);  p<=q || p>=r;  q=r, r=mnext(q)) {
		if (magic(q) != FREE) {
			errno = EGARBAGE;	/* list is corrupted */
			return (NULL);
		}
		if (q >= r && (p > q || p < r))
			break;
	}
	mallocp = q;

	/*
	 *  Resize the block:
	 */
	obytes = msize(p) * sizeof (MHEADER) - sizeof (ALIGNER);
	/* if there is any space just above, grab it: */
	if (p+msize(p) == r) {
		p->s.size += msize(r);
		mnext(q) = mnext(r);
	}
	/* satisfy the request in-place, if possible: */
	size = msize(p);
	if (size >= nunits) {
		if (size <= nunits+2)	/* we added about as much as needed */
			return (ap);
		/* we added too much, release the excess: */
		p->s.size = (pidmagic | nunits);
		core = mcore(p);
		if (magic(core) != pidmagic) {
			errno = EGARBAGE;
			return (NULL);
		}
		p += nunits;
		p->s.size = (FREE | (size - nunits));
		mnext(p) = mnext(q);
		mnext(q) = p;
		mcore(p+1) = core;
		return (ap);
	}
	/* else we need to move the block: */
	cp = malloc (nbytes);
	if (cp == NULL) {
		errno = ENOMEM;
		return (NULL);		/* original intact! */
	}
	memcpy (cp, ap, (int) obytes);	/* obytes < nbytes guaranteed */
	if (free (ap) != 0)
		return (NULL);		/* just in case */
	return (cp);
}

