/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	VSD
 *
 *	$Log:	dl_moron.c,v $
 * Revision 2.2  92/04/30  12:18:08  trillium
 * Ohio Release
 * 
 *	Function:	- creates datalink input and output process pairs 
 *			- datalink configuration obtained from dl_mom
 *			- does one centralized open for each datalink device
 *			- functions as a pipe between datalink processes
 *			  and the physical link.
 *			- This module is introduced primarily to take care of
 *			  device drivers which do not support concurrent
 *			  access by more than one process.
 */

#include <stdio.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <typical.h>
#include <net.h>
#include <debug.h>
#include <t_types.h>
#include <signal.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
extern void		dli_dev();	/* datalink input */
extern void		dlo_dev();	/* datalink output */

/*
 * local functions
 */
static void		suicide();	/* die on SIGCHLD */
static void		rdwr_dev();	/* transfer a frame */
static int		mread();	/* pipe resistant read */

/*
 * local variables
 */
static int		fd;		/* device descriptor */


void
dl_moron (devname, event)

char			*devname;
int			event;

{
	int		pid;		/* child's or my pid */
	int		ret;		/* return value from wait() */
	int		pdi[2];		/* pipe btn dl_moron and dli */
	int		pdo[2];		/* pipe btn dl_moron and dlo */
/*
 * Local variables for the select() call.
 */
	int		width;		/* #of bits to be checked in each
					   bit mask that represent a file
					   descriptor */
	fd_set		readfds;	/* descriptor for reading */

/*
 * When child dies, exit.
 */
	signal(SIGCHLD, suicide);

/*
 * Create two pipes -
 * pdi : between datalink device (devname) and datalink input process (dli),
 * pdo : between datalink device and datalink output process (dlo).
 * dl_moron will read data from 'devname' and write the same data on 'pdi'
 * which will then be read by dli process.
 * dlo will write data onto 'pdo' which is next read by dl_moron which 
 * writes the same onto devname to be sent out.
 */
	if ((ret = pipe(pdi)) < 0) {
		terror("dl_moron(pipe_i)");
		exit(errno);
	}

	if ((ret = pipe(pdo)) < 0) {
		terror("dl_moron(pipe_o)");
		exit(errno);
	}

	db(("dl_moron: created two pipes\n"));
/*
 * Open the device.
 */
	if ((fd = open(devname, O_RDWR)) < 0) {
		terror(devname);
		exit(errno);
	}
	db(("dl_moron: opened link device\n"));
/*
 * Fork the datalink input process and give it the 'read' component 
 * of 'pdi' pipe.
 */
	if ((pid = fork()) < 0) {
		terror("dl_moron(fork-dli)");
		exit(errno);
	}

	else if (pid == 0) {
		db(("dl_moron: dli started\n"));
		dli_dev(pdi[0]);
		exit(0);
	}
/*
 * Fork the datalink output process and give it the 'write' component 
 * of 'pdo' pipe and the event number.
 */
	if ((pid = fork()) < 0) {
		terror("dl_moron(fork-dlo)");
		exit(errno);
	}

	else if (pid == 0) {
		db(("dl_moron: dlo started\n"));
		dlo_dev(pdo[1], event);
		exit(0);
	}
/*
 * Initialization for the select call.
 */
	width = getdtablesize();
	FD_ZERO (&readfds);
/*
 * Loop forever.
 */
	for (;;) {
/*
 * Include the two read file descriptors in readfds.
 */
		FD_SET (fd, &readfds);
		FD_SET (pdo[0], &readfds);
/*
 * Poll (select between) the two reads, from link device and from dlo process.
 */
		db(("dl_moron: select reads\n"));
		if ((ret = select (width, &readfds, (fd_set *) 0,
			(fd_set *) 0, (struct timeval *) 0)) < 0) {

			terror("dl_moron(select)");
			exit(errno);
		}
/*
 * Check which file descriptor is ready for reading if at all any.
 */
		if ((ret > 0) && (FD_ISSET (fd, &readfds) != 0)) {
/*
 * Link device is ready for reading, so read data from link. 
 * Write the same data onto pdi[1].
 */
			db(("dl_moron: link device ready\n"));
			rdwr_dev(fd, pdi[1]);

		} /* if fd ready for reading */

		if ((ret > 0) && (FD_ISSET (pdo[0], &readfds) != 0)) {
/*
 * dlo process has output data, so read it from the associated pipe.
 * Write the same data onto link device, fd. The data is to be sent out 
 * to another node.
 */
			db(("dl_moron: dlo ready\n"));
			rdwr_dev(pdo[0], fd);
		}
	}
}

/*
 *	rdwr_dev
 *
 *	Function:	- reads and write from/to specified devices
 *	Accepts:	- device descriptors 
 */

static void
rdwr_dev(rd_desc, wr_desc)

int			rd_desc; 	/* read device descriptor */
int			wr_desc;	/* write device descriptor */

{
/*
 * Local variables for reading/writing frames of data.
 */
        char            ack;            /* acknowledgement byte */
        struct netframe frame;          /* network frame (header and data) */
	int		msglen;		/* length of nh_msg in frame */
	int		msgflags;	/* flags of message in frame */
	int		nbytes;		/* return value from write */

/*
 * Read network frame header from rd_desc.
 */
	if (mread(rd_desc, (char *) &frame.nf_header, sizeof(struct nmsg))) {
		exit(errno);
	}
/*
 * Convert message flags and length to local byte ordering.
 */ 
	msglen = frame.nf_header.nh_length;
	msgflags = frame.nf_header.nh_flags;
	msglen = ttol(msglen);
	msgflags = ttol(msgflags);
/*
 * Read network frame data from device.
 */
	if (msglen > 0) {

		if (mread(rd_desc, frame.nf_data, msglen)) {
                            exit(errno);
		}

		frame.nf_header.nh_msg = frame.nf_data;
	}    
/*
 * Read acknowledgement byte from device if not continuing a VC.
 * This is needed for compatibility with dl_t4.
 */
/*
 *	if (! (msgflags & KHOLD)) {
 * 
 *		if (mread(rd_desc, &ack, sizeof(ack))) {
 *			exit(errno);
 *		}
 *	}
 */
/*
 * Write the same data onto wr_desc.
 */
	nbytes = write(wr_desc, (char *) &frame, sizeof(struct nmsg) + msglen);
 
	if (nbytes == ERROR) {
		terror("dl_moron(write)");
		exit(errno);
	} else if (nbytes < msglen) {
		fprintf(stderr, "dl_moron: write(fd, b, n) < n\n");
		exit(errno);
	}
/*
 * Write acknowledgement byte to the device if we are not continuing
 * a virtual circuit.   
 * This is needed for compatibility with dl_t4.
 */
/*
 *	if (! (msgflags & KHOLD)) {
 *
 *		if (write(wr_desc, &ack, sizeof(ack)) == ERROR) {
 *			exit(errno);
 *		}
 *	}
 */
}

/*
 *      mread
 *
 *      Function:       - execute multiple reads up to a given length
 *      Accepts:        - file descriptor
 *                      - length
 *                      - buffer pointer
 *      Returns:        - 0 or ERROR
 */

static int
mread(fdes, q, length)

int                     fdes;
int                     length;
char                    *q;

{
        int             ret;            /* # of bytes actually read */

        while ((ret = read(fdes, q, length)) < length) {

                if (ret == ERROR) {
                        terror("dl_moron(read)");
                        return(ERROR);
                }
 
                else if (ret == 0) {
                        fprintf(stderr, "dli_dev: premature EOF\n");
                        return(ERROR);
                }
 
                else {
                        length -= ret;
                        q += ret;
                }
        }

        return(0);
}

/*
 *	suicide
 *
 *	Function:	- clean up and exit
 *			- a signal from one child assumes the other is dead too
 */

static void
suicide()

{
	db(("dl_moron: recd SIGCHLD\n"));

	if (close(fd) == ERROR) {
		terror("dl_moron(close)");
	}

	exit(0);
}
