/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	netkill.c,v $
 * Revision 2.2  92/04/30  12:21:47  trillium
 * Ohio Release
 * 
 *	Function:	- resource deallocator for Trollius/Voltop
 *			- officially, a topology layer tool
 */

#include <fcntl.h>
#include <stdio.h>
#include <net.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <terror.h>
#include <typical.h>
#include <unistd.h>
#include <nail.h>

/*
 * external variables
 */
extern int			errno;
extern FILE			*yyin;		/* lex input file */

/*
 * local static variables
 */
static char			*f_bnail;	/* boot schema filename */
static char			**p_bnail;	/* schema pathname prefix */
static int			fl_verbose;	/* verbosity? */

static char *usage = "usage: netkill [-h] [-c <bschema>]\n";

/*
 * global variables
 */
char				*me;

/* 
 * external functions
 */
extern void			followed();
extern void			validopts();
extern char			*getparam();
extern char			*findpath();
extern int			addarg();
extern int			do_args();
extern int			opt_taken();
extern int			parse_nail();

/*
 * local functions
 */
static void		help();
static int		set_defaults();

main(argc, argv)  

int			argc;
char			*argv[];

{
	struct bschema	nail;		/* parsed boot schema */
	struct t_node	*tnode;		/* node information */
	char		*pf_bnail;	/* full pathname for boot schema */
	char		*dentry;	/* host link name */
	int		pid;		/* child pid */
	int		busy;		/* entry busy flag */
	union wait	status;		/* file status buffer */
	HASH		*tnhd;		/* node hash table */

	me = argv[0];
/*
 * Parse the command line.
 */
	validopts("hv");
	followed("c");

	if ((do_args(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
		fprintf(stderr, usage);
		return(errno);
	}

	if (opt_taken('h')) {
		help();
		return(0);
	}
/*
 * Set the mode of operation.
 */
	if (set_defaults()) {
		return(errno);
	}
/* 
 * Locate the boot schema file.
 */
	if ((pf_bnail = findpath(f_bnail, p_bnail, R_OK)) == NULL) {
		fprintf(stderr, "netkill: cannot locate bnail file \"%s\"\n",
			f_bnail);
		return(1);
	}
/*
 * Open the boot schema file.
 */
	if ((yyin = fopen(pf_bnail, "r")) == NULL) {
		perror("netkill");
		return(errno);
	}
/*
 * Parse the boot schema.
 */
	if (parse_nail(&nail)) {
		fprintf(stderr, "netkill: cannot parse the NaIL syntax\n");
		return(1);
	}
/*
 * Close the boot schema file.
 */
	if (fclose(yyin)) {
		perror("netkill");
		return(errno);
	}
/*
 * Find the name of the first and only link from the host node.
 */
	tnhd = nail.bs_tnhd;
	tnode = (struct t_node *) ah_top(tnhd);

	while (tnode) {

		if (tnode->tn_specs->ns_type == 0) {
			nail.bs_hostnode = tnode;
		}

		tnode = (struct t_node *) ah_next(tnhd, (void *) tnode);
	}

	dentry = (nail.bs_hostnode)->tn_specs->ns_lkout->lk_link->tl_srcname;
/*
 * Execute tkill and wait for the link to become available.
 */
	VERBOSE("netkill: tkill ...\n");

	if ((pid = fork()) < 0) {
		perror("netkill");
		return(errno);
	}
/*
 * child
 */
	else if (pid == 0) {
		execlp("tkill", "tkill", (char *) 0);
		return(1);
	}
/*
 * parent
 */
	else {
		while ((wait(&status) != 0) && (! WIFEXITED(status)));

		if (status.w_retcode != 0) {
			(void) fprintf(stderr,
				"netkill: tkill exit status %d\n",
				(int) status.w_retcode);
			return((int) status.w_retcode);
		}
	}
/*
 * Wait for the link to become available by continuously
 * executing netwhat until it returns a 0 status.
 * There are simpler ways to do this, but the Archipel
 * driver is giving me problems.
 */
	VERBOSE("netkill: wait for Trollius termination ...\n");
	busy = TRUE;

	while (busy) {

		if ((pid = fork()) < 0) {
			perror("netkill");
			return(errno);
		}
/*
 * child
 */
		else if (pid == 0) {
			execlp("netwhat", "netwhat", dentry, (char *) 0);
			return(1);
		}
/*
 * parent
 */
		else {
			while ((wait(&status) != 0) && (! WIFEXITED(status)));

			if (status.w_retcode == 0) {
				busy = FALSE;
			} else if (status.w_retcode != EBUSY) {
				(void) fprintf(stderr,
					"netkill: netwhat exit status %d\n",
					(int) status.w_retcode);
				return((int) status.w_retcode);
			}
		}
	}
/*
 * Execute the Voltop netlogout command on the link.
 */
	VERBOSE("netkill: netlogout, release resources ...\n");

	if ((pid = fork()) < 0) {
		perror("netkill");
		return(errno);
	}
/*
 * child
 */
	else if (pid == 0) {
		execlp("netlogout", "netlogout", "-d", dentry, (char *) 0);
		return(1);
	}
/*
 * parent
 */
	else {
		while ((wait(&status) != 0) && (! WIFEXITED(status)));

		if (status.w_retcode != 0) {
			(void) fprintf(stderr,
				"netkill: netlogout exit status %d\n",
				(int) status.w_retcode);
			return((int) status.w_retcode);
		}
	}

	return(0);
}

/*
 *	help
 *
 *	Function:	- Print usage information.
 */
static void 
help()
{
	printf("\nSynopsis:\tnetkill [options]\n");
	printf("\nDescription:\tTerminate Trollius and release resources.\n");
	printf("\nOptions:\t-h\t\tthis message\n");
	printf("\t\t-c <bschema>\tspecify boot schema\n");
	printf("\t\t-v\t\tverbose mode on\n");
}

/*
 *	set_defaults
 *
 *	Function:	- set default paths, filenames, tools, flags
 *	Returns:	- 0 or ERROR
 */

static int 
set_defaults()

{
	int		n_bnail;		/* # args in p_bnail */
/*
 * pathnames
 */
	n_bnail = 0;
	p_bnail = (char **) 0;
	addarg(&n_bnail, &p_bnail, "");
	addarg(&n_bnail, &p_bnail, "$TROLLIUSHOME/boot");
	addarg(&n_bnail, &p_bnail, DEFPBNAIL);
/*
 * filenames
 */
	if (opt_taken('c')) {
		f_bnail = getparam('c');
	} else {
		f_bnail = DEFFBNAIL;
	}
/*
 * flags
 */
	fl_verbose = opt_taken('v');
	return(0);
}
