/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RKM/RBD/GDB
 *
 *	$Log:	solder.c,v $
 * Revision 2.2  92/04/30  12:57:53  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  10:15:13  gdburns
 * overhaul of machine independent code
 * 
 * Revision 2.1.1.1  91/04/01  18:24:44  raja
 * Upgraded to handle multiple device names.
 * 
 * Revision 2.1  91/03/11  14:42:20  gdburns
 * Ohio Release
 * 
 *	Function:	- system specific topology reconfiguration tool
 *			- topology is specified using NaIL
 *			- INMOS B014 version
 */

#include <stdio.h>
#include <portable.h>
#include <unistd.h>
#include <string.h>
#include <terror.h>
#include <typical.h>
#include <debug.h>
#include <nail.h>
#include <net.h>
#include <machine/machine.h>
#include <solder.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <fcntl.h>

/*
 * global variables
 */
char			*me;		/* tool name */

/*
 * external variables
 */
extern int		errno;
extern FILE		*yyin;		/* lex input file */

/*
 * static variables
 */
static struct bschema	nail;		/* network description */
static int		fl_debug;	/* debug option flag */
static int		fl_pretend;	/* pretend option flag */
static int		fl_verbose;	/* verbose option flag */
static char		*f_bnail;	/* boot schema filename */
static char		*fullbnail;	/* full boot schema filename */
static char		f_solder[24];	/* soldering command file */
static char		*t_solder;	/* link connect tool */
static char		**p_bnail;	/* schema pathname prefix */
static FILE		*cmd_file;	/* solder command file */

static struct im_hwnodeinfo
			*src_node;	/* ptr to hw node specs */
static struct im_hwnodeinfo
			*dest_node;	/* ptr to hw node specs */
static struct im_board	*pboard;	/* ptr to root of board array */

static struct im_board	*pb1, *pb2;	/* ptr to boards */
static struct slot_connections
			*ps1, *ps2;	/* ptr to slots array */

static char *usage = "usage: %s [-dhvNV] [-c <bnail>]\n";

/* 
 * external functions
 */
extern void		followed();
extern void		validopts();
extern char		*findpath();
extern char		*getenv();
extern char		*getparam();
extern char		*mktemp();
extern char		*sprintf();
extern char		*strcpy();
extern char		*strrchr();
extern int		addarg();
extern int		cnfexec();
extern int		do_args();
extern int		opt_taken();
extern int		parse_nail();
extern int		unlink();
extern int4		stoi();

/*
 * static functions
 */
static void		do_connect();
static void		help_solder();
static void		prt_defaults();
static void		set_defaults();
static int		check_connect();
static int		do_error();
static int		do_solder();
static struct im_hwnodeinfo
			*check_node();

/*
 * main program
 */
int
main(argc, argv)  

int			argc;
char			*argv[];

{
/*
 * Assign the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("dhvNV");
	followed("c");

	if ((do_args(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		fprintf(stderr, usage, me);
		return(errno);
	}

	if (opt_taken('h')) {
		help_solder();
		return(0);
	}
/*
 * Set the mode of operation.
 */
	set_defaults();

	if (fl_debug) {
		prt_defaults();
	}
/* 
 * Locate the boot schema file.
 */
	if ((fullbnail = findpath(f_bnail, p_bnail, R_OK)) == NULL) {
		fprintf(stderr, "%s: cannot locate bnail file \"%s\"\n",
			me, f_bnail);
		return(1);
	}
/*
 * Open the boot schema file.
 */
	DBUG("%s: opening the boot schema file: %s ...\n", me, f_bnail);
	if ((yyin = fopen(fullbnail, "r")) == NULL) {
		perror(me);
		return(errno);
	}
/*
 * Parse the boot schema.
 */
	if (parse_nail(&nail) < 0) {
		fprintf(stderr, "%s: cannot parse the NaIL syntax\n", me);
		return(errno);
	}
/*
 * Close the boot schema.
 */
	if (fclose(yyin) == EOF) {
		perror(me);
		return(errno);
	}
/*
 * Print the boot schema out if needed.
 */
	if ((opt_taken('V')) || (fl_debug)) {
		pretty_bnail(&nail);

		if (opt_taken('V')) {
			return(0);
		}
	}
/*  
 * Solder the network.
 */
	DBUG("%s: soldering ...\n", me);
	if (do_solder() < 0) {
		fprintf(stderr, "%s: unable to solder\n", me);
		return(errno);
	}

	DBUG("%s: soldered OK\n", me);
	return(0);
}

/*
 *	help_solder
 *
 *	Function:	- prints the help menu
 */

static void 
help_solder()
{
	printf("\nSynopsis:\t%s [options]\n", me);
	printf("\nDescription:\tConfigure the topology of a multicomputer.\n");
	printf("\nOptions:\t-h\t\tPrint this message.\n");
	printf("\t\t-d\t\tPrint debug information.\n");
	printf("\t\t-v\t\tBe verbose.\n");
	printf("\t\t-N\t\tPretend to take action.\n");
	printf("\t\t-V\t\tPretty print the bnail file.\n");
	printf("\t\t-c <bnail>\tUse <bnail> as boot schema.\n");
}

/*
 *	set_defaults
 *
 *	Function:	- sets the values of key variables controlling
 *			  the operation of the tool
 */

static void 
set_defaults()

{
	int		n_bnail;

	n_bnail = 0;
	p_bnail = (char **) 0;
	addarg(&n_bnail, &p_bnail, "");
	addarg(&n_bnail, &p_bnail, "$TROLLIUSHOME/boot");
	addarg(&n_bnail, &p_bnail, DEFPBNAIL);
/*
 * Set the tool name.
 */
	t_solder = DEFTLCONN;
/*
 * Set the file names.
 */
	if (opt_taken('c')) {
		f_bnail = getparam('c');
	} else if (f_bnail = getenv("TROLLIUSBNAIL")) {
	} else {
		f_bnail = DEFFBNAIL;
	}
/*
 * Set the flags.
 */
	fl_verbose = opt_taken('v');
	fl_pretend = opt_taken('N');

	if (opt_taken('d')) {
		fl_debug = TRUE;
		fl_verbose = TRUE;
	} else {
		fl_debug = FALSE;
	}
}

/*
 *	prt_defaults
 *
 *	Function:	- prints the values of key variables
 *			- mostly used for debugging
 */

static void 
prt_defaults()

{
	printf("\nDefault Settings\n\n");
	printf("boot schema: %s\n", f_bnail);
	printf("debug flag: %d\n", fl_debug);
	printf("verbose flag: %d\n", fl_verbose);
	printf("pretend flag: %d\n", fl_pretend);
	printf("local solder command: %s\n", t_solder);
}

/*
 *	do_solder
 *
 *	Function:	- reconfigure the system links
 *			- system depend code: b014 board version
 *			- generates check command lines for C014
 *
 *	Returns:	- 0 or ERROR
 */
static int
do_solder()
{
	int		i, j, k1, k2;		/* loop indices */
	int		cmd_des, null_des;	/* file descriptors */
	int		pid, n_cmd;
	char		*devname;		/* device name */
	char		**v_cmd;
	char		tmp_string[80];
	union wait	status;			/* status after forking */
	struct t_link	*tlink;			/* ptr to tlink element */
	struct link	*devlink;		/* ptr to device link */

/*
 * Initialize.
 */
	pboard = (struct im_board *) 
			malloc( sizeof(struct im_board) * NUMBEROFBOARDS );

	if ( pboard == (struct im_board *) 0) {
		return(ERROR);
	}
	
	pboard->bd_slots = *slots_b014;
	pboard->bd_edges = edges_b014;
	for (i = 0; i < NUMBEROFSWITCHES ; i++) {
		for (j=0; j < NUMBEROFSWLINKS; j++) {
			    pboard->bd_switch[i][j]  = '-';
		}
	}

/*
 * Issue the link reconfiguration commands.   Some links should not
 * be reconfigured (system specific) as determined by the
 * check_connect() function.
 */
	tlink = nail.bs_tlink;
	for (i = 0; i < nail.bs_ntlink; i++, tlink = tlink->tl_next) {
		if (check_connect(tlink)) {
			do_connect();
		}

		if (src_node != NULLHW) {		/* clean up please */
			free(src_node);	
		}
		if (dest_node != NULLHW) {
			free(dest_node);	
		}
	}

/* 
 * Open a check configuration file. A lot of ifs and if-nots exist if multiple
 * boards exist. Only one file is opened at the moment.
 */
/* 
 * Create and open the temporary command file. 
 */
	(void) strcpy(f_solder, "/tmp/solderXXXXXX");
	(void) mktemp(f_solder);

	if ((cmd_file = fopen(f_solder, "w")) == NULL) {
		perror(me);
		return(ERROR);
	}
/*
 *  Only one file is written to at the moment.
 */
	if (find_host(&nail)) {
		fprintf(stderr, "%s: cannot find the network host\n", me);
		return(ERROR);
	}
	DBUG("found hostnode %d\n", nail.bs_hostnode->tn_nodeid);
	devlink = ((nail.bs_hostnode)->tn_specs)->ns_lkout;
	if (devlink) {
		devname = (devlink->lk_link)->tl_srcname;
		if (! devname) {
			devname = DEFDEVNAME;
		}
	}
	else {
		devname = DEFDEVNAME;
	}

	for(i=0; i < NUMBEROFBOARDS; i++) {	

	    fprintf(cmd_file,"Using %s check 2.53\n", devname);
   	    fprintf(cmd_file,
		"   # Part rate Mb Bt [  Link0  Link1  Link2  Link3 ]\n");
	    fprintf(cmd_file,
		"   0 T800c-20 0.37 0 [   HOST    1:1    ...    ... ]\n");
	    fprintf(cmd_file,
		"   1 T2   -17 0.92 1 [    2:C    0:1    ...    3:C ]\n");

/* 
 * The following code is "check specific". 
 */
	    for(j=0; j < NUMBEROFSWITCHES ; j++) {
		    fprintf(cmd_file,"   %d C004b   [",j+2);
		for (k1=0; k1 < 4 ;k1++) { 
		    fprintf(cmd_file," ");
		    for (k2=0; k2 < ( NUMBEROFSWLINKS / 4 ); k2++) {
			fprintf(cmd_file,"%c",
				pboard->bd_switch[j][k1 * 8 + k2]);
		    }
		}	
		fprintf(cmd_file," ]\n");
	    }
	}

	free(pboard);					/* free ptr */

	if(fclose(cmd_file)) { 				/* close file */
		perror(me);
		return(ERROR);
	}

/* 
 * Execute the system soldering command.
 */
	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_solder);		/* "program" */
	addarg(&n_cmd, &v_cmd, DEFSOPT);		/* "program -opt" */
	addarg(&n_cmd, &v_cmd, "-L");			/* specify device */
	addarg(&n_cmd, &v_cmd, devname);

	if (fl_debug) {
		printfv(n_cmd, v_cmd);
	}

	VERBOSE("soldering ...\n");
	if (! fl_pretend) {

		if ((pid = fork()) < 0) {		/* error */
		   sprintf(tmp_string, "cannot fork %s", t_solder);
		   exit(do_error(tmp_string));
		}
		else if (pid == 0) {			/* child */

			if ((cmd_des = open(f_solder,O_RDONLY)) == ERROR) {
				exit(do_error("open failed"));
			}

			if(dup2(cmd_des,0) == ERROR){	/* redirect stdin */
				close(cmd_des);
				exit(do_error("dup2 failed"));
			}

			close(cmd_des);

			if ((null_des = open("/dev/null")) == ERROR) {
				exit(do_error("open failed"));
			}
							/* silence reqd. ? */

			if( !fl_verbose && !fl_debug ) {

							/* redirect stdout */
				if(dup2(null_des,1) == ERROR ){
					close(null_des);
					exit(do_error("dup2 failed"));
				}
			}

			VERBOSE("dispatching %s ...\n",t_solder);
			execvp(t_solder, v_cmd);	/* no ret code chk */
		}
		else {					/* parent */
			while ((wait(&status) != 0) &&
				(! WIFEXITED(status)));
			DBUG("%s: status: %d\n", t_solder,
				status.w_retcode);
		}
		if (status.w_retcode != 0) {
			exit((int) status.w_retcode);
		}

		if (unlink(f_solder)) {
			perror(me);
			return(ERROR);

		}


	}
 
	return(0);
}

/*
 *	check_connect
 *
 *	Function:	- check if link connection is allowed
 *			- system specific
 *			- b014 version
 *
 *	Accepts:	- ptr to link description
 *
 *	Returns:	- TRUE or FALSE
 */
static int
check_connect(link)

struct t_link		*link;
{
	struct nspecs		*src_spec, *dest_spec ;

	src_spec = (link->tl_srcnode)->tn_specs;
	dest_spec = (link->tl_destnode)->tn_specs;

/*
 * Check the node types being linked.  If any one is an otb node
 * the connection should be inhibited.
 */
	if (tisotb(src_spec->ns_type) || tisotb(dest_spec->ns_type)) {
		return(FALSE);
	}

/*
 * Check the syntax of the system specific node names.
 */
	if ( (src_node = check_node(src_spec->ns_sysnode)) == NULLHW){
		fprintf(stderr, "%s: illegal system node syntax: %s\n", 
			me, src_spec->ns_sysnode);
		return(FALSE);
	}

	if ( (dest_node = check_node(dest_spec->ns_sysnode)) == NULLHW){
		fprintf(stderr, "%s: illegal system node syntax: %s\n", 
			me, dest_spec->ns_sysnode);
		return(FALSE);
	}

/*
 * Check if both nodes have the same host.  For now links are connected
 * only when they share the same host.  This restriction should be
 * removed later.
 */
	if (src_spec->ns_host && dest_spec->ns_host) {
		if (strcmp(src_spec->ns_host, dest_spec->ns_host)) {
			return(FALSE);
		}
	}
	else if (src_spec->ns_host || dest_spec->ns_host) {
		return(FALSE);
	}


/* 
 * Locate the board where the src_node resides. Point to the place where
 * a description of src_node exists.
 */
	pb1 = pboard + src_node->hw_board ;
	ps1 = pb1->bd_slots + src_node->hw_slot * NUMBEROFSLOTLINKS 
		+ link->tl_srclink ;

/* 
 * Locate the board where the dest_node resides. Point to the place where
 * a description of dest_node exists.
 */
	pb2 = pboard + dest_node->hw_board ;
	ps2 = pb2->bd_slots + dest_node->hw_slot * NUMBEROFSLOTLINKS 
		+ link->tl_destlink ;

/*
 * Some connections cannot be made. If links are not connected to a link
 * of a switch, then connections cannot be made.
 */
	if ( ps1->sl_switch == CBAD ) {
		fprintf(stderr,"%s: link %d of node %s unsuable.\n",
			me, link->tl_srclink, src_spec->ns_sysnode);	
		return(FALSE);
	}
	if ( ps2->sl_switch == CBAD ) {
		fprintf(stderr,"%s: link %d of node %s unsuable.\n",
			me, link->tl_destlink, dest_spec->ns_sysnode);	
		return(FALSE);
	}

/*
 *	No hardwired links. But some connections cannot be made. For instance,
 *	on a B014 only links 1,2 and 0,3 can be connected. This is because 
 *	links 1, 2 are connected through a switch while links 0 and 3 are 
 *	connected through another. 
 */
	if ( ps1->sl_switch != ps2->sl_switch ) {
		fprintf(stderr,"%s: nodes %s and %s cannot be connected.\n",
			me, src_spec->ns_sysnode, dest_spec->ns_sysnode);	
		return(FALSE);
	}
/*
 * Check if the link is a duplicate.  A link is duplicate if the
 * source link number is greater than the destination link number or if
 * they are equal but the source nodeid is greater or equal to the
 * destination nodeid.  As nodeids are unique, this forms a complete
 * ordering among (node,link) pairs and the connection is allowed
 * iff (srcnode, srclink) < (destnode, destlink).
 * Raja's idea --- a masterpiece.
 */
	if (link->tl_srclink > link->tl_destlink) {
		return(FALSE);
	}
	else if (link->tl_srclink < link->tl_destlink) {
		return(TRUE);
	}
	else if ((link->tl_srcnode)->tn_nodeid <
		(link->tl_destnode)->tn_nodeid) {
		return(TRUE);
	}
	else {
		return(FALSE);
	}
}

/*
 *	do_connect
 *
 *	Function:	- Create the C004 lines.
 *			- system specific
 *			- b014 version
 */
static void
do_connect()
{

	int		src_switch, dest_switch ;
	int		src_link, dest_link ;

/*
 * Finally do the "connecting" by doing
 * 	switch[src_link] =  dest_link;
 *	switch[dest_link] = src_link;
 */
	src_switch = ps1->sl_switch;
	dest_switch = ps2->sl_switch;

	src_link =  ps1->sl_sw_link;
	dest_link = ps2->sl_sw_link;

	if ( src_node->hw_board == dest_node->hw_board ) {

		pb1->bd_switch[src_switch][src_link] = 
			(dest_link < 10 ) ? ( dest_link + '0' ): 
				( dest_link - 10 + 'A' ) ;

		pb2->bd_switch[dest_switch][dest_link] = 
			(src_link < 10 ) ? ( src_link + '0' ): 
				( src_link - 10 + 'A' ) ;
	}

/*
 *	else { 
 *		do nothing for now.... Waiting for Brest....
 *	}
 */

}

/*
 *	check_node
 *
 *	Function:	- check system specific node name syntax
 *			- B014 version.
 *			- Allowed syntax 
 *				"board.module" -- for TRAMS
 *
 *	Accepts:	- node name
 *
 *	Returns:	- pointer to struct hwnodeinfo
 *			- NULL pointer
 */
static struct im_hwnodeinfo
*check_node(name)

char			*name;

{
	int				i;
	char				*p_name;
	char				temp[5];	/* buffer for board # */
	struct	im_hwnodeinfo		*node;		/* ptr to node */
/*
 *	Shall we malloc ?
 */

	node=(struct im_hwnodeinfo *) malloc(sizeof(struct im_hwnodeinfo)) ;

/*
 * Check if first character is a digit.
 */
	temp[0] = *name;
	if ( isdigit ( (int) *(name++) ) ) {

	    for (i=1 ; isdigit((int) *name); i++, name++) {
		temp[i] = *name;
	    }

	    temp[i] = '\0';				/* null terminated */
	    node->hw_board = stoi(temp);

	    if ( node->hw_board >= NUMBEROFBOARDS ) {	/* correct board # */
		return(NULLHW);
	    }

	    if ( *(name++) == '.' ) {			/* is it a '.' */

		if ( isdigit( (int) *(name) ) ) {

		    p_name = name;
		    name++;
		    for (i=1; isdigit((int) *name); i++, name++) { }

		    if (*name == '\0' ) {		/* no junk please */

		    	node->hw_slot = stoi(p_name);
		    	if (  node->hw_slot >= NUMBEROFSLOTS ){
				return(NULLHW);
		    	}

		        return(node);			/* board.num */
		    }
		    else {
		  	    return(NULLHW);		/*board.JUNK */
		    }
		} 
	    }
	    else {
		return(NULLHW);				/* boardJUNK */
	    }
	}
	else {
	    return(NULLHW);				/* JUNK */
	}
	return(NULLHW);
}
/*
 *	do_error
 *
 *	Function:	- report error messages from do_solder()
 *			- close and unlink the soldering file
 *
 *	Accepts:	- the error message
 *
 *	Returns:	- ERROR
 */
static int
do_error(message)

char			*message;

{
	fprintf(stderr, "%s (do_solder)", me);
	if (message) {
		fprintf(stderr, ": %s\n", message);
	}
	else {
		fprintf(stderr, "\n");
	}
	unlink(f_solder);
	return(ERROR);
}
