/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University
 *
 *	VSD/GDB
 *
 *	$Log:	dl_idiot.c,v $
 * Revision 2.2  92/04/30  14:01:28  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/04/24  13:09:28  gdburns
 * vmtm version
 * 
 *
 *	Function:	- creates datalink input and output process pairs 
 *			- datalink configuration obtained from dl_mom
 *			- does one centralized open for each datalink device
 *			- functions as a pipe between datalink processes
 *			  and the physical link.
 *			- uses the Inmos "link.c" interface
 *			- This is the polling solution for substandard or
 *			  non-existant device drivers.
 */

#include <stdio.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <typical.h>
#include <net.h>
#include <debug.h>
#include <t_types.h>
#include <signal.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
extern void		dli_dev();	/* datalink input */
extern void		dlo_dev();	/* datalink output */
extern int		OpenLink();	/* open the device */
extern int		CloseLink();	/* close the device */
extern int		WriteLink();	/* write to the device */
extern int		ReadLink();	/* read from the device */
extern char		*lerror();	/* report device error */

/*
 * local functions
 */
static void		suicide();	/* die on SIGCHLD */
static void		rdwr_dev();	/* transfer a frame */
static int		mread();	/* pipe resistant read */

/*
 * local variables
 */
static int		fd;		/* device descriptor */


void
dl_idiot (devname, event)

char			*devname;
int			event;

{
	int		pid;		/* child's or my pid */
	int		bdev;		/* activity on device */
	int		bpipe;		/* activity on dlo pipe */
	int		pdi[2];		/* pipe btn dl_idiot and dli */
	int		pdo[2];		/* pipe btn dl_idiot and dlo */
/*
 * Local variables for the select() call.
 */
	int		width;		/* #of bits to be checked in each
					   bit mask that represent a file
					   descriptor */
	fd_set		readfds;	/* descriptor for reading */
	struct timeval	timeout;	/* for polling dlo pipe */
	char		*p;		/* favourite ptr */

/*
 * When child dies, exit.
 */
	signal(SIGCHLD, suicide);
/*
 * Create two pipes -
 * pdi : between datalink device (devname) and datalink input process (dli),
 * pdo : between datalink device and datalink output process (dlo).
 * dl_idiot will read data from 'devname' and write the same data on 'pdi'
 * which will then be read by dli process.
 * dlo will write data onto 'pdo' which is next read by dl_idiot which 
 * writes the same onto devname to be sent out.
 */
	if (pipe(pdi) == ERROR) {
		terror("dl_idiot(pipe_i)");
		exit(errno);
	}

	if (pipe(pdo) == ERROR) {
		terror("dl_idiot(pipe_o)");
		exit(errno);
	}

	db(("dl_idiot: created two pipes\n"));
/*
 * Open the device.
 * Replace '/' with ' ' in Parsytec link name.
 */
	for (p = devname; *p; ++p) if (*p == '/') *p = ' ';

	if ((fd = OpenLink(devname, O_RDWR)) < 0) {
		fprintf(stderr, "dl_idiot: cannot open link %s\n", devname);
		fprintf(stderr, "dl_idiot: %s\n", lerror(fd));
		exit(1);
	}
	db(("dl_idiot: opened link device\n"));
/*
 * Fork the datalink input process and give it the 'read' component 
 * of 'pdi' pipe.
 */
	if ((pid = fork()) < 0) {
		terror("dl_idiot(fork-dli)");
		exit(errno);
	}

	else if (pid == 0) {
		db(("dl_idiot: dli started\n"));
		dli_dev(pdi[0]);
		exit(0);
	}
/*
 * Fork the datalink output process and give it the 'write' component 
 * of 'pdo' pipe and the event number.
 */
	if ((pid = fork()) < 0) {
		terror("dl_idiot(fork-dlo)");
		exit(errno);
	}

	else if (pid == 0) {
		db(("dl_idiot: dlo started\n"));
		dlo_dev(pdo[1], event);
		exit(0);
	}
/*
 * Initialization for the select call.
 * TUNE THIS FOR BETTER PERFORMANCE!
 */
	width = getdtablesize();
	FD_ZERO (&readfds);
	timeout.tv_sec = 1;
	timeout.tv_usec = 0;
/*
 * Loop forever.
 */
	for (;;) {
/*
 * Poll the two reads, from link device and from dlo process.
 */
		if ((bdev = TestRead(fd)) < 0) {
			fprintf(stderr, "dl_idiot(TestRead): failed\n");
			exit(1);
		}

		else if (bdev) {
/*
 * Link device is ready for reading, so read data from link. 
 * Write the same data onto pdi[1].
 */
			db(("dl_idiot: link device ready\n"));
			rdwr_dev(fd, pdi[1]);
			timeout.tv_sec = 0;
		}

		else {
			timeout.tv_sec = 1;
		}

		FD_SET (pdo[0], &readfds);
			
		if ((bpipe = select (width, &readfds, (fd_set *) 0,
			(fd_set *) 0, &timeout)) == ERROR) {

			terror("dl_idiot(select)");
			exit(errno);
		}

		if (bpipe) {
/*
 * dlo process has output data, so read it from the associated pipe.
 * Write the same data onto link device, fd. The data is to be sent out 
 * to another node.
 */
			db(("dl_idiot: dlo ready\n"));
			rdwr_dev(pdo[0], fd);
		}
	}
}

/*
 *	rdwr_dev
 *
 *	Function:	- reads and write from/to specified devices
 *
 *	Accepts:	- device descriptors 
 */

static void
rdwr_dev(rd_desc, wr_desc)

int	rd_desc; 	/* read device descriptor */
int	wr_desc;	/* write device descriptor */

{
/*
 * Local variables for reading/writing frames of data.
 */
        struct netframe frame;          /* network frame (header and data) */
	int		msglen;		/* length of nh_msg in frame */
	int		msgflags;	/* flags of message in frame */
	int		nbytes;		/* return value from write */

/*
 * Read network frame header from rd_desc.
 */
	if (mread(rd_desc, (char *) &frame.nf_header, sizeof(struct nmsg))) {
		exit(errno);
	}
/*
 * Convert message flags and length to local byte ordering.
 */ 
	msglen = frame.nf_header.nh_length;
	msgflags = frame.nf_header.nh_flags;
	msglen = ttol(msglen);
	msgflags = ttol(msgflags);
/*
 * Read network frame data from device.
 */
	if (msglen > 0) {
		if (mread(rd_desc, frame.nf_data, msglen)) {
                            exit(errno);
		}
		frame.nf_header.nh_msg = frame.nf_data;
	}    
/*
 * Write the same data onto wr_desc.
 */
	nbytes = (wr_desc == fd) ?
	WriteLink(wr_desc, (char *) &frame, sizeof(struct nmsg) + msglen, 0) :
	write(wr_desc, (char *) &frame, sizeof(struct nmsg) + msglen);
 
	if (nbytes < 0) {
		fprintf(stderr, "dl_idiot: cannot write device\n");
		exit(1);
	}
 
	else if (nbytes < msglen) {
		fprintf(stderr, "dl_idiot: write(fd, b, n) < n\n");
		exit(1);
	}
}

/*
 *      mread
 *
 *      Function:       - execute multiple reads up to a given length
 *
 *      Accepts:        - file descriptor
 *                      - length
 *                      - buffer pointer
 *
 *      Returns:        - 0 or ERROR
 */

static int
mread(fdes, q, length)

int                     fdes;
int                     length;
char                    *q;

{
        int             ret;            /* # of bytes actually read */

        while ((ret = (fdes == fd) ? ReadLink(fdes, q, length, 0) :
		read(fdes, q, length)) < length) {

                if (ret < 0) {
                        fprintf(stderr, "dl_idiot: cannot read device\n");
                        return(ERROR);
                }
 
                else if (ret == 0) {
                        fprintf(stderr, "dl_idiot: premature EOF\n");
                        return(ERROR);
                }
 
                else {
                        length -= ret;
                        q += ret;
                }
        }

        return(0);
}

/*
 *	suicide
 *
 *	Function:	- clean up and exit
 *			- a signal from one child assumes the other is dead too
 */

static void
suicide()

{
	db(("dl_idiot: recd SIGCHLD\n"));

	if (CloseLink(fd) < 0) {
		fprintf(stderr, "dl_idiot: cannot close device\n");
		lerror("dl_idiot");
		exit(1);
	}

	exit(0);
}
