.TH BSEND l "April 8, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
bsend, brecv \- basic Trollius message passing
.SH SYNOPSIS
.nf
.ft B
#include <net.h>

int	bsend (struct nmsg *header);
int	brecv (struct nmsg *header);
.fi
.SH DESCRIPTION
.I Bsend()
and
.I brecv()
provide very simple message passing at the Trollius network level.
They allow the user to make point-to-point data transfers between two nodes.
These functions are easier to use than their nsend(l) and nrecv(l) counterparts
but, in the same token, they are less powerful.
They are meant to be used in pairs, each message sent by
.I bsend()
being received by
.I brecv().
.SS THE MESSAGE STRUCTURE
The functions accept a pointer to a message structure which is the same
one used by nsend(l) and nrecv(l) and is defined in <net.h>.
The difference is that only a small subset of the structure members is used by
.I bsend()
and
.I brecv().
Thus, for the purposes of using the basic Trollius message passing functions,
the structure looks as follows:
.PP
.nf
.RS
.DT
struct nmsg {
	int	nh_node,
		nh_length,
		nh_flags,
	char	*nh_msg;
};
.RE
.fi
.PP
The
.I nh_node
field is used by
.I bsend()
to identify the destination node running the intended receiving process, local
or remote.
The
.I nh_node
field is used by
.I brecv()
to identify the source node running the process that is sending the message.
A receiving process thus can directly specify its synchronizing sending process.
Two special node identifiers are defined in <net.h>.
LOCAL refers to the local node and
ORIGIN refers to the host node from which the Trollius multicomputer was booted.
The
.I nh_node
field remains unchanged after calling
.I bsend()
or
.I brecv().
Since the sender and receiver processes are only identified by the nodes on
which they run (source node and destination node), this limits the applicability
of these functions to the case of a single user process per node.
For more elaborate message passing, the use of nsend(l) and nrecv(l) is
recommended.
.PP
The
.I nh_length
field holds the length of the message as a number of data elements.
For example, if the message to be sent is an array of 20 floating point
numbers,
.I nh_length
should be set to 20.
If the sender and the receiver specify different lengths,
the lesser amount will be transferred.
The
.I nh_length
field remains unchanged after calling
.I bsend(),
but is set to the minimum of the sender's and receiver's lengths after calling
.I brecv().
.PP
The
.I nh_flags
field is used to specify the type of data in the message and remains unchanged
after calling either
.I bsend()
or
.I brecv().
The functions transmit the message between any two Trollius nodes.
On nodes of different types, data may have different representations.
For example, integers may be stored with the most significant byte
first in memory (big-endian) or with the most significant byte last in
memory (little-endian).
.I Bsend()
and
.I brecv()
automatically modify the byte ordering to suit the internal data
representation of each node.
The data types supported, defined in <net.h>, are:
.TP 14
DINTMSG
The message is an array of 32-bit integers.
.TP
DFLTMSG
The message is an array of single precision real numbers.
.TP
DDBLMSG
The message is an array of double precision real numbers.
.TP
DRAWMSG
The message representation will not be changed.
.PP
DRAWMSG is typically used when sending an array of bytes (e.g. an ASCII
text string) or when the message array contains a mixture of data types.
In this later case, the user will have to specify the size of the array
in bytes and perform any necessary changes in the data representation by
using the several byte-order translation functions available in Trollius
(see ttol(l)).
.PP
The
.I nh_msg
field holds the address of the first byte of data to be sent.
The data must be stored contiguously in memory.
This field remains unchanged after calling
.I bsend()
or
.I brecv().
.SH EXAMPLE USAGE
The following is an example of a typical usage of
.I bsend()
and
.I brecv()
to transmit an array of single precision floating point numbers
between two nodes (n3 and n7).
.PP
The sending process on node 3 sends the contents of the data array of length
5 to the receiving process on node 7 by executing:
.nf
.ft B

float		data[5];
struct nmsg	header;

header.nh_node = 7;
header.nh_msg = (char *) data;
header.nh_length = 5;
header.nh_flags = DFLTMSG;
bsend(&header);
.fi
.PP
The receiving process on node 7, not knowing how many floating point numbers
are going to be sent, sets a maximum limit of 20.
.I Brecv()
modifies the value of
.I nh_length
in the header to indicate how many numbers were actually received.
On the receiver side,
.I nh_node
has to hold the source nodeid thus indicating which node to communicate with.
The code executed by the receiver is:
.nf
.ft B

float		indata[20];
struct nmsg	header;

header.nh_node = 3;
header.nh_msg = (char *) indata;
header.nh_length = 20;
header.nh_flags = DFLTMSG;
brecv(&header);
.fi
.SH BLOCKING
A process calling
.I brecv()
blocks until the message sent by the process calling
.I bsend()
arrives.
A process calling
.I bsend()
blocks only until its message is picked up by the Trollius message routing
system.
The only thing that is guaranteed by a successful return from
.I bsend()
is that the message has entirely left the calling process, thus the sender
is free to reuse the space of the message array.
In the example above, the user can modify the values in the
.I data
array after
.I bsend()
returns.
.SH "RETURN VALUE
Upon successful completion,
.I bsend()
and
.I brecv()
return 0.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.I Brecv()
also changes the length of the message to match
the number of data elements received (i.e. the lesser amount between
the length of the array sent and the length of the array expected).
.SH ERRORS
.TP 23
[EINVAL]
The given data type is invalid.
.TP
[ENOEVDESCRIPTORS]
The Trollius kernel ran out of message descriptor space.
.TP
[EBADNODE]
The specified destination node does not exist.
.SH "SEE ALSO"
nsend(l), ttol(l)
