/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	bfstate.c,v $
 * Revision 2.2  92/04/30  14:06:47  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  11:28:16  gdburns
 * Monitor user space threshold.
 * 
 * Revision 2.1  91/03/20  02:49:39  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:26:26  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/03/06  17:07:55  gdburns
 * Added -l option.
 * 
 * Revision 2.0  90/02/20  11:53:53  trillium
 * Homo erectus
 *
 *	Function:	- command to print Trillium buffer status info
 *			- nodes are selected in a system dependent fashion
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *
 *	Accepts:	- see the "usage" variable
 *
 *	Examples:	- bfstate n10
 *			- describe all messages buffered on node 10
 */

#include <bufreq.h>
#undef NULL
#include <net.h>
#include <stdio.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * external variables
 */
extern int 		errno;		/* global error variable */

/*
 * external functions
 */
extern void		validopts();	/* set the valid options */
extern void		terror();
extern void		intparam();
extern void		followed();
extern char		*mnemonic();
extern char		*malloc();
extern int		do_args();	/* parse the valid options */
extern int		getnode();	/* get next node entry */
extern int		opt_taken();	/* Was a particular option selected? */
extern int		parse_nodes();	/* parse the node identifiers */
extern int		rbfstate();	/* remote bufferd status */

/*
 * local functions
 */
static void		header();
static void		help();
static void		do_bfstate();

/*
 * local variables
 */
static char		*_copyright =
	"Copyright 1992 The Ohio State University $Revision: 2.2 $";

static struct bfstate	*bftable;		/* buffer state descriptor */
static int		bfmax;			/* max entries */

/*
 * syntax for bfstate command
 */
static char		*usage =
	"bfstate [-bhl] [-B <maxmsgs>] <nodes>";


main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int		n_index;		/* index of node entry */
	int		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
/*
 * Parse the command line.
 */
	validopts("bhlB");
	followed("B");

	if ((do_args(&argc, argv) < 0) ||
		(parse_nodes(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		} else {
			terror("bfstate (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Allocate buffer state table.
 */
	if (opt_taken('B')) {
		 intparam('B', &bfmax);
	} else {
		bfmax = BSMAX;
	}

	bftable = (struct bfstate *) 
		malloc((unsigned) (bfmax * sizeof(struct bfstate)));

	if (! bftable) {
		terror("bfstate (malloc)");
	}
/*
 * Attach to kernel.
 */
	if (kinit(0) < 0) {
		terror("bfstate (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index >= 0) {
/*
 * Print status header.
 */
		header();

		do {
			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(errno);
			}

			do_bfstate(nodeid);
			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}

	else {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	kexit(0);
}

/*
 *	do_bfstate
 *
 *	Function:	- find and print the buffer state of a remote node
 *			- table of buffer state descriptors has one spacer
 *			  descriptor which separates the pending descriptors
 *			  from the waiting descriptors
 *
 *	Accepts:	- node identifier
 */

static void
do_bfstate(nodeid)

int			nodeid;

{
	int		size;		/* # of return blocks */
	int		limit;		/* # of blocks to print */
	int		b_spacer;	/* found the spacer? */
	int		b_blkd;		/* -b option? */
	int		i;		/* favourite index */
	struct bfparms	parms;		/* buffer parameters */

	if ((size = rbfstate(nodeid, bftable, bfmax, &parms)) < 0) {
		terror("bfstate");
		kexit(errno);
	}

	b_spacer = FALSE;
	b_blkd = opt_taken('b');
	limit = (size < bfmax) ? size : bfmax;

	for (i = 0; i < limit; ++i) {
/*
 * Look for the spacer descriptor.
 */
		if (bftable[i].bfs_node == NOTNODEID) {
			b_spacer = TRUE;
		}

		else if (!b_spacer || !b_blkd) {
			printf("%-9.9s", mnemonic(nodeid));
			printf(" %-9.9s", mnemonic(bftable[i].bfs_node));
			printf(" %-11d", bftable[i].bfs_event);
			printf(" %-8X", bftable[i].bfs_type);
			printf(" %-7d", bftable[i].bfs_length);

			if (b_spacer) {
				printf(" waiting\n");
			} else {
				printf(" blocked\n");
			}
		}
	}

	if ((size > bfmax) && (!b_blkd || !b_spacer)) {
		printf("...and more buffered messages...\n");
	}

	if (opt_taken('l')) {
	    printf("\n%-9.9s", mnemonic(nodeid));
	    printf(" max workers = %d, used workers = %d, full workers = %d\n",
		parms.bfp_maxworkers, parms.bfp_nworkers, parms.bfp_nfull);
	    printf("%-9.9s", mnemonic(nodeid));
	    printf(" max space = %d, max user space = %d, used space = %d\n\n",
		parms.bfp_maxspace, parms.bfp_maxusr, parms.bfp_nspace);
	}
}

/*
 *	  header
 *
 *	  Function:	- prints the bfstate table header
 */

static void
header()
{
	printf("NODE      DEST      EVENT       TYPE     LENGTH  STATE\n");
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the bfstate command
 */

static void
help()
{
	printf("\nSynopsis:	bfstate [options] <nodes>\n");
	printf("\nDescription:	Print information on message buffers.\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\t\t-b	information on blocked messages only\n");
	printf("\t\t-l	also print buffer levels and maximums\n");
	printf("\t\t-B <#>	set maximum # of messages to print\n");
	mnusage();
	printf("\t\th\t(local host node)\n");
	printf("\nExample:	bfstate h n3-9,11\n");
	printf("\t\t\t\"Get buffer status information on the local host\n");
	printf("\t\t\t node and nodes 3 thru 9 and 11.\"\n");
}
