/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB/DLF
 *
 *	$Log:	doom.c,v $
 * Revision 2.2  92/04/30  14:12:13  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  11:40:43  raja
 * new multicast procedure
 * 
 * Revision 2.1  91/03/20  02:52:18  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:26:07  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:38  trillium
 * Homo erectus
 * 
 *
 *	Function:	- command to deliver signals to Trillium processes
 *			- processes can be selected by index, load address,
 *			  identifier or blocked event
 *			- absence of process specifier selects all processes
 *			- nodes are selected in a system dependent fashion
 *			- absence of signal number defaults to SIGUDIE
 *
 *	Accepts:	- see the "usage" variable
 *
 *	Examples:	- doom n10 -4 %1 %2 p20300
 *			- send a SIGUDIE to process indices
 *			  1 & 2 and process ID 20300 on node 10
 */


#include <stdio.h>
#include <terror.h>		/* error codes */
#include <typical.h>		/* commonly used constants */
#include <preq.h>		/* kenya interface constants */
#include <ksignal.h>		/* signal constants */
#include <net.h>		/* network constants */


/*
 * external variables
 */
extern int 	errno;		/* global error variable */

/*
 * external functions
 */
extern void	validopts();	/* set the valid options */
extern void	terror();
extern int	do_args();	/* parse the valid options */
extern int	getnode();	/* get next node entry */
extern int	getnodeid();
extern int	getproc();	/* get next process entry */
extern int	kdoom();	/* local process doom */
extern int	kinit();
extern int	opt_taken();	/* Was a particular option selected? */
extern int	parse_nodes();	/* parse the node identifiers */
extern int	parse_procs();	/* parse the process identifiers */
extern int	predef();
extern int	rdoom();	/* remote process doom */

/*
 * local functions
 */
static void	help();
static void	doom_lnode();
static void	doom_rnode();

/*
 * syntax for state command
 */
static char *usage = "doom [-h] [-<signal>] <nodes> [<processes>]";


/*
 *******MAIN
 */

main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int		n_index;		/* index of node entry */
	int		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
	int		signo;			/* signal number */
/*
 * Parse the command line.
 */
	validopts("h0");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(parse_procs(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		}

		else {
			terror("doom (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Set default signal.
 */
	if (opt_taken('0')) {
		if (intparam('0', &signo) == ERROR) {
			fprintf(stderr, "usage: %s\n", usage);
			exit(EUSAGE);
		}
	}

	else {
		signo = SIGUDIE;
	}
/*
 * Attach to kernel.
 */
	if (kinit(0) == ERROR) {
		terror("doom (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index != ERROR) {

		do {
			if ((nodeid == LOCAL) || (nodeid == getnodeid())) {
				doom_lnode(signo);
			}

			else {
/*
 * Register the HOST2ITB castid (predef) if required.
 * HOST2OTB and HOST2ALL are not allowed here.
 */
				if ((nodeid == HOST2OTB) ||
						(nodeid == HOST2ALL)) {
					errno = EBADNODE;
					terror("doom");
					kexit(errno);
				}
				else if (predef(nodeid)) {
					terror("doom (predef)");
					kexit(errno);
				}

				doom_rnode(nodeid, signo);
			}

			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}

	else {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	kexit(0);
}


/*
 *	doom_lnode
 *
 *	Function:	- deliver a signal to a local otb process
 *			- makes doom request directly to local kernel
 *
 *	Accepts:	- signal number
 */

static void
doom_lnode(signo)

int			signo;

{
	int		fnotpid = FALSE;	/* non pid selectors */
	int		p_index;		/* index of process entry */
	int		p_value;		/* value of process entry */
	int		p_flags;		/* process selection method */


	getproc(&p_index, &p_value, &p_flags);

	if (p_index == ERROR) {
	    fnotpid = TRUE;
	}

	else {
	    do {
		if (p_flags == SELECT_PID) {

		    if (kdoom(p_value, signo) == ERROR) {
			terror("doom (kdoom)");
			kexit(errno);
		    }
		}

		else {
		    fnotpid = TRUE;
		}

		getproc(&p_index, &p_value, &p_flags);
	    } while(p_index);
	}
/*
 * Were any unsupported host node selection methods used?
 */
	if (fnotpid) {
	    fprintf(stderr, "\nHost node processes must be selected by PID.\n");
	}
}


/*
 *	state_rnode
 *
 *	Function:	- find and print the state of a remote itb process
 *
 *	Accepts:	- node identifier
 *			- signal number
 */

static void
doom_rnode(nodeid, signo)

int			nodeid;
int			signo;

{
	int		p_index;		/* index of process entry */
	int		p_value;		/* value of process entry */
	int		p_flags;		/* process selection method */

	getproc(&p_index, &p_value, &p_flags);

	if (p_index == ERROR) {
		p_flags = SELECT_APPL;
	}

	do {
		if (rdoom (nodeid, p_flags, p_value, signo)) {
			terror("doom (rdoom)");
			kexit(errno);
		}

		getproc(&p_index, &p_value, &p_flags);
	} while(p_index > 0);
}


/*
 *	  help
 *
 *	  Function:	- prints helpful information on the doom command
 */

static void
help()
{
	printf("\nSynopsis:	doom [options] [-<sig>] <nodes> [processes]\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\nDescription:	Deliver a signal to Trillium processes.\n");
	printf("\nProcesses:	p<list> (identifiers)\n");
	printf("\t\t%%<list> (indices)\n");
	printf("\t\t@<list> (load addresses)\n");
	printf("\t\t#<list> (blocked events)\n");
	mnusage();
	printf("\t\th	(local host nodes)\n");
	printf("\t\tC	(all compute nodes)\n");
	printf("\nSignals:	4 (UDIE), 5 (ARREST), 6 (RELEASE),\n");
	printf("\t\t7 (user defined A), 8 (user defined B)\n");
	printf("\nExample:	doom -5 n0-7,11 #6 %%2 p0x2A340\n");
	printf("\t\t\t\"Deliver signal 5 to PID 0x2A340,\n");
	printf("\t\t\tindex 1, and all processes blocked on event 6\n");
	printf("\t\t\tfor nodes 3 thru 9 and 11.\"\n");
}
