/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	VSD/GDB
 *
 *	$Log:	launch.c,v $
 * Revision 2.2  92/04/30  14:12:55  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/06  10:36:38  gdburns
 * Let execv(3) locate programs.
 * 
 * Revision 2.1.1.1  91/12/18  15:03:49  gdburns
 * Redecorate and bring in line with 2.2 state of the art.
 * 
 * Revision 2.1  91/03/20  02:52:22  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/08/14  11:30:07  dixit-v
 * Fixed bug - if no nodes specified in cmd line, return error.
 * 
 * Revision 2.0.1.2  90/07/13  19:26:08  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/21  12:59:29  raghu
 * Replaced exit() with kexit() and kexit(0) in the main program module.
 * 
 * Revision 2.0  90/02/20  11:53:39  trillium
 * Homo erectus
 *
 * Revision 2.0.1.2  90/06/21  12:29:21  raghu
 * Included a final kexit(0), for cleaning up IPC message queues.
 * 
 * Revision 2.0.1.1  90/06/20  16:16:29  raghu
 * Replaced exit with kexit after kinit. 
 * 
 *	Function:	- Launch (execute) processes for all the nodes in the
 *			  topology onto the host node.
 *			  It is used to emulate a multicomputer system on a 
 *			  single host node. It is used in conjunction with 
 *			  the Stand-alone router (sa_rtr).
 *			  With the launch command, one can specify the nodes
 *			  on which the processes are to be eventually executed
 *			  in the multicomputer environment. Launch however runs
 *			  all processes for all nodes only on the host node. 
 *			  The sa_rtr emulates all remote node message passing 
 *			  on the host node itself. This enables the Trollius
 *			  user to run and debug his parallel code using a single
 *			  node, and finally run them on the multicomputer system
 *			  without having to make any changes in his code.
 *
 *			- In case multiple nodes and/or multiple files are
 *			  specified, all files with the same (single/multiple)
 *			  argument(s) are executed on the host node for all 
 *			  the nodes. To execute different files on different 
 *			  nodes, or for different files to have different 
 *			  arguments, or for any other combination, multiple 
 *			  launch commands must be used.
 *
 *	Usage:		launch [-dhvw] <nodeids> <files> [-- args]
 *
 *	Example:	launch n2-4,10-13 test1 test2 arg1 arg2
 */

#include <debug.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <typical.h>
#include <terror.h>
#include <unistd.h>

/*
 * external functions
 */
extern void 		validopts();
extern int		terror();
extern int 		opt_taken();
extern int 		do_args();
extern int		addarg();
extern int 		parse_nodes();
extern int 		getnode();
extern int		rrhold();
extern int		rrattach();
extern int		rrdetach();

/*
 * local functions
 */
static void		help_launch();

/*
 * external variables
 */
extern int		errno;
extern int		argtailc;	/* # of args to created processes */
extern char		**argtailv;	/* arg vector to created processes */

/*
 * local variables
 */
static char *usage = "launch [-dhvw] <nodeids> <files> [-- args]";

static int 		fl_debug;	/* -d option */
static int 		fl_verbose;	/* -v option */

main(argc, argv)

int			argc;
char			*argv[];

{
	int 		num = 0;	/* total # of created processes */
	int 		i;		/* favourite index */
	int		j;		/* second favourite index */
	int		n_pid;		/* child pid */
	int 		n_index;	/* index of cmd line nodeid */
	int		n_nodeid;	/* cmd line node ID */
	int		n_flags;	/* for getnode() */
	int	 	w_status;	/* exit status of child process */

	int		n_cmd;		/* # of cmd args for child */
	char	      **v_cmd;		/* cmd arg vector for child */

/*
 * Parse command line arguments.
 */
	validopts("hdvw");
/*
 * Example: launch -vw n2-6,10-13 n50 test1 test2 -- arg1 arg2
 * do_args() removes options and 'args' from argv. It returns proper
 * values for opt_taken() and also fills in argtailc, and argtailv 
 * with 'args'.
 * After do_args(), argv = launch n2-6,10-13 n50 test1 test2
 * and argtailv = arg1 arg2
 *
 * parse_nodes() extracts any valid nodeid arguments beginning with 'n',
 * if there is a file name beginning with 'n' then it is left untouched.
 * It also takes care of nodeids specified as multiple arguments.
 * It then returns proper values for getnode().
 * After parse_nodes(), argv = launch test1 test2
 */   
	if ((do_args(&argc, argv) < 0) ||
		(parse_nodes(&argc, argv) < 0)) {

		if (errno == EUSAGE)
			fprintf(stderr, "Usage: %s\n", usage);
		else
			terror("launch (arguments)");

		exit(errno);
	}

	if (opt_taken('h')) {
		help_launch();
		exit(0);
	}

	fl_debug = opt_taken('d');
	fl_verbose = opt_taken('v');
/*
 * argv = launch <files>
 */
	++argv;
	--argc;
/*
 * argv = <files>
 */
	if (argc == 0) {
		fprintf(stderr, "Usage: %s\n", usage);
		exit(EUSAGE);
	}
/*
 * Execute n copies of every file onto host node, where n = number of nodes.
 */
	getnode(&n_index, &n_nodeid, &n_flags);

	if (n_index < 0) {
		fprintf(stderr, "Usage  : %s\n", usage);
		exit(EUSAGE);
	}
/*
 * Attach to kernel.
 */
	if (kinit (0) < 0) {
		terror("launch (kinit)");
		exit(errno);
	}

	for (i = 0; i < argc; ++i) {
/*
 * Form command line.
 */
		n_cmd = 0;
		v_cmd = (char **) 0;
		addarg (&n_cmd, &v_cmd, argv[i]);

		for (j = 0; j < argtailc; j++) {
			addarg (&n_cmd, &v_cmd, argtailv[j]);
		}

		if (fl_debug) {
			printf("launch: executing ");

			for (j = 0; j < n_cmd; j++) {
				printf("%s ", v_cmd[j]);
			}

			printf("...\n");
		}
/*
 * Loop thru' each node.
 */
		do {
			num++;
/*
 * Hold the router.
 */
			if (rrhold() < 0) {
				terror("launch (rrhold)");
				kexit(errno);
			}
/*
 * Fork a child process.
 */
			if ((n_pid = fork()) < 0) {
				perror("launch (fork)");
				kexit(errno);
      			}

			if (n_pid != 0) { 
/* 
 * parent process - Send RQATTACH (n_pid, n_nodeid) message to sa_rtr.
 */
				DBUG("launch: rrattach(%d, %d) ...\n",
					n_pid, n_nodeid);

				if (rrattach(n_pid, n_nodeid) < 0) {
					terror("launch (rrattach)");
					kexit(errno);
				}

				VERBOSE("[0] %d %s running on node %d\n",
					n_pid, argv[i], n_nodeid);
			}

			else {		/* child */
				execvp(argv[i], v_cmd);
				perror(argv[i]);
				exit(0);
      			}

			getnode(&n_index, &n_nodeid, &n_flags);

		} while (n_index > 0);
	}
/* 
 * Wait for children to terminate if -w is not present.
 */
	if (! opt_taken('w')) {
	  	DBUG("launch: wait on %d children ...\n", num);

		for (i = 0; i < num; i++) {
			n_pid = wait(&w_status);

			DBUG("launch: rrdettach(%d) ...\n", n_pid);

			if (rrdetach(n_pid) < 0) {
				terror("launch (rrdetach)");
				kexit(errno);
			}
		}
	}

	kexit(0);
	return(0);
}

/*
 *	help_launch
 *
 *	Function:	- Print options for launch.
 */

static void 
help_launch()

{
	printf("\nSynopsis:\tlaunch [options] <nodes> <progs> [-- <args>]\n");
	printf("\nDescription:\tRun programs under Stand-alone Trollius.\n");
	printf("\nOptions: \t-h\tPrint this message.\n");
	printf("\t\t-d\tPrint debug information.\n");
	printf("\t\t-v\tBe verbose.\n");
	printf("\t\t-w\tDo not wait for processes to terminate.\n");
	mnusage();
	printf("\nExample: \tlaunch n0-3 prog1 -- arg1 arg2\n");
	printf("\t\t\t\"Execute 'prog1' on emulated nodes 0 thru 3\n");
	printf("\t\t\twith arguments 'arg1' and 'arg2'.\"\n");
}
