/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	mstate.c,v $
 * Revision 2.2  92/04/30  14:13:52  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/11/10  23:36:21  gdburns
 * Minor overhaul fixes several problems.
 * 
 * Revision 2.1  91/03/20  02:52:29  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:26:17  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:45  trillium
 * Homo erectus
 * 
 *	Function:	- command to print Trillium memory status information
 *			- processes are selected by ID
 *			- nodes are selected in a system dependent fashion
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *
 *	Accepts:	- see the "usage" variable
 *
 *	Examples:	- mstate n10 p0x20300
 *			- print all memory blocks allocated by process
 *			  ID 0x20300 on node 10
 */

#include <stdio.h>
#include <kernel.h>
#include <terror.h>		/* error codes */
#include <typical.h>		/* commonly used constants */
#include <mareq.h>		/* mallocd interface constants */
#include <preq.h>

/*
 * external variables
 */
extern int 	errno;		/* global error variable */

/*
 * external functions
 */
extern void	terror();
extern void	validopts();	/* set the valid options */
extern int	do_args();	/* parse the valid options */
extern int	getnode();	/* get next node entry */
extern int	getproc();	/* get next process entry */
extern int	opt_taken();	/* Was a particular option selected? */
extern int	parse_nodes();	/* parse the node identifiers */
extern int	parse_procs();	/* parse the process identifiers */
extern int	rmstate();	/* remote mallocd status */
extern char	*malloc();
extern char	*mnemonic();	/* ascii representation of nodeid */

/*
 * local functions
 */
static void	header();
static void	help();
static void	do_mstate();
static void	mstate_out();

/*
 * local variables
 */
static char *_copyright = "Copyright 1991 The Ohio State University $Revision: 2.2 $";

static struct memblock	*mastbl;		/* memory block descriptors */

/*
 * syntax for mstate command
 */
static char *usage = "mstate [-hf] [-M <#blocks>] <nodes> [<processes>]";

main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int		n_index;		/* index of node entry */
	int		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
	int		nblks;			/* # block desc. to alloc */
/*
 * Parse the command line.
 */
	validopts("hf");
	followed("M");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(parse_procs(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		}

		else {
			terror("mstate (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Allocate status descriptor table.
 */
	if (opt_taken('M')) {
		intparam('M', &nblks);
	} else {
		nblks = MASMAX;
	}

	mastbl = (struct memblock *) malloc((unsigned) nblks *
		sizeof(struct memblock));

	if (! mastbl) {
		terror("mstate (malloc)");
		exit(errno);
	}
/*
 * Attach to kernel.
 */
	if (kinit(0) == ERROR) {
		terror("mstate (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index != ERROR) {
/*
 * Print status header.
 */
		header();

		do {
			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(errno);
			}

			do_mstate(nodeid, nblks);
			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}

	else {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	kexit(0);
}

/*
 *	do_mstate
 *
 *	Function:	- find and print the memory state of a remote node
 *			  for all specified processes
 *			- no specified processes will cause all node allocated
 *			  memory to be displayed
 *
 *	Accepts:	- node identifier
 *			- maximum # of memory blocks to print
 */

static void
do_mstate(nodeid, nblks)

int			nodeid;
int			nblks;

{
	int		p_index;		/* index of process entry */
	int		p_value;		/* value of process entry */
	int		p_flags;		/* process selection method */
	int		size;			/* # of return blocks */

	if ((size = rmstate(nodeid, mastbl, nblks)) == ERROR) {
		terror("mstate");
		kexit(errno);
	}

	getproc(&p_index, &p_value, &p_flags);

	if (p_index == ERROR) {
		p_value = NOTPROCESS;
		p_flags = SELECT_PID;
		mstate_out(nodeid, p_value, (size > nblks) ? nblks : size);
	}

	else {
		if (opt_taken('f')) {
			fprintf(stderr, "usage: %s\n", usage);
			exit(EUSAGE);
		}

		do {
			if (p_flags != SELECT_PID) {
				fprintf(stderr,
				    "mstate: must select processes by PID\n");
				kexit(EUSAGE);
			}

			mstate_out(nodeid, p_value,
				(size > nblks) ? nblks : size);
			getproc(&p_index, &p_value, &p_flags);
		} while(p_index > 0);
	}

	if (size > nblks) {
		printf("%-9.9s ...more...\n", mnemonic(nodeid));
	}
}

/*
 *
 *	  mstate_out
 *
 *	  Function:	- print out memory state information
 *			- either print allocated or unallocated blocks
 *
 *	  Accepts:	- node ID
 *			- process ID
 *			- total number of block descriptors
 */

static void
mstate_out(nodeid, pid, size)

int			nodeid;
int			pid;
int			size;

{
	int		i;			/* favourite index variable */
/*
 * Check every block we received from mallocd.
 */
	for (i = 0; i < size; ++i) {
/*
 * Make sure descriptor is in use.
 */
	    if (mastbl[i].ma_start) {

		if (opt_taken('f')) {

		    if (mastbl[i].ma_pid == NOTPROCESS) {
/*
 * Print free block.
 */
			printf("%-9.9s", mnemonic(nodeid));
			printf("%-9.9s", " free");
			printf(" %-8X %-8X\n", (int) mastbl[i].ma_start,
				mastbl[i].ma_size);
		    }
		}

		else if ((mastbl[i].ma_pid != NOTPROCESS) &&
			((pid == NOTPROCESS) || (pid == mastbl[i].ma_pid))) {
/*
 * Print allocated block.
 */
		    printf("%-9.9s", mnemonic(nodeid));
		    printf(" %-8d", mastbl[i].ma_pid);
		    printf(" %-8X %-8X\n", (int) mastbl[i].ma_start,
				mastbl[i].ma_size);
		}
	    }
	}
}


/*
 *	  header
 *
 *	  Function:	- print the mstate command table header
 */

static void
header()
{
	printf("NODE      PID      ADDRESS  SIZE\n");
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the mstate command
 */

static void
help()
{
	printf("\nSynopsis:	mstate [options] <nodes> [processes]\n");
	printf("\nDescription:	Print information on memory allocation.\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\t\t-f	print information on free memory\n");
	printf("\t\t-M #	print # memory blocks\n");
	printf("\nProcesses:	p<list> (identifiers)\n");
	mnusage();
	printf("\nRestriction:\tUse mstate only with compute nodes.\n");
	printf("\nExample:	mstate n3-9,11 p303\n");
	printf("\t\t\t\"Get memory status information for PID 303\n");
	printf("\t\t\t on nodes 3 thru 9 and 11.\"\n");
}
