/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	spawn.c,v $
 * Revision 2.2  92/04/30  14:14:03  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  02:52:33  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:26:31  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:56  trillium
 * Homo erectus
 * 
 *
 *	Function:	- command to spawn existing kenya processes
 *			- processes can be selected by index, load address,
 *			  identifier or blocked event
 *			- absence of process specifier selects all processes
 *			- nodes are selected in a system dependent fashion
 *			- a single node indicates a local spawn
 *			- exactly two nodes indicates a remote spawn
 *			- more than two nodes is an error
 *
 *	Accepts:	- see the "usage" variable
 *
 *	Examples:	- spawn n10 n11 %1 %2 p20300
 *			- spawn process indices 1 & 2 and process ID 20300
 *			  from node 10 to node 11
 */


#include <stdio.h>
#include <terror.h>		/* error codes */
#include <typical.h>		/* commonly used constants */
#include <preq.h>		/* kenya interface constants */
#include <net.h>		/* network constants */
#include <kio.h>
#include <freq.h>

#include <sys/file.h>


/*
 * external variables
 */
extern int 		errno;		/* global error variable */
extern int		argtailc;	/* # of args after "--" */
extern char		**argtailv;	/* args after "--" */
extern struct kio_t	_kio;

/*
 * external functions
 */
extern void		validopts();	/* set the valid options */
extern void		terror();
extern int		do_args();	/* parse the valid options */
extern int		getnode();	/* get next node entry */
extern int		getnodeid();
extern int		getproc();	/* get next process entry */
extern int		kinit();
extern int		opt_taken();	/* Was a particular option selected? */
extern int		parse_nodes();	/* parse the node identifiers */
extern int		parse_procs();	/* parse the process identifiers */
extern int		rspawn();	/* remote process doom */
extern char		**pack_argv();	/* pack runtime arguments */
extern char		*ttyname();

/*
 * local functions
 */
static void		help();

/*
 * local variables
 */
static char *_copyright = "Copyright 1989 The Ohio State University $Revision: 2.2 $";

/*
 * syntax for state command
 */
static char *usage = "spawn [-h] <src_node> [<dest_node>] [<processes>] [-- <args>]";


/*
 *******MAIN
 */

main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int		n_index;		/* index of node entry */
	int		n_flags;		/* extra node entry info */
	int		p_value;		/* process selector value */
	int		src_node;		/* source node */
	int		dest_node;		/* destination node */
	int		argtailn;		/* size of packed args */
	int		argtag;			/* flat tag for args */
	int		px;			/* new process index */
	int		pid;			/* new process ID */

/*
 * Parse the command line.
 */
	validopts("hv");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(parse_procs(&argc, argv) == ERROR)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		}

		else {
			terror("spawn (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Attach to kernel.
 */
	if (kinit(0)) {
		terror("spawn (kinit)");
		exit(errno);
	}
/*
 * Get the source node.
 */
	getnode(&n_index, &src_node, &n_flags);

	if ((n_index == ERROR) || n_flags) {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	else {
/*
 * Get the source node.
 */
		getnode(&n_index, &dest_node, &n_flags);

		if (n_flags) {
			fprintf(stderr, "usage: %s\n", usage);
			kexit(errno);
		}

		else if (! n_index) {
			dest_node = LOCAL;
		}
	}
/*
 * Prepare runtime arguments for loading onto the destination node.
 * Get the runtime arguments loaded onto the destination node.
 */
	if (argtailc) {
		*(argtailv - 1) = "<fill argv[0] in here>";
		argtailc += 1;
		argtailv = pack_argv(argtailc, argtailv - 1, &argtailn);
		argtag = getpid();
	}
/*
 * Open stdin, stdout, and stderr for the Trollius remote filesystem.
 */
/*
	_kio.ki_stdin = topen("/dev/tty", O_RDONLY | O_REUSE);

	if (_kio.ki_stdin == ERROR) _kio.ki_stdin = 0;

	_kio.ki_stdout = topen("/dev/tty", O_WRONLY | O_REUSE);

	if (_kio.ki_stdout == ERROR) _kio.ki_stdout = 1;

	_kio.ki_stderr = topen("/dev/tty", O_WRONLY | O_REUSE);

	if (_kio.ki_stderr == ERROR) _kio.ki_stderr = 2;
*/

	getproc(&n_index, &p_value, &n_flags);

	do {
		if ((n_flags != SELECT_PID) && (n_flags != SELECT_INDEX)) {
		fprintf(stderr, "spawn: must select process by ID or index\n");
			kexit(1);
		}

		if (argtailc) {
			if (rflat(src_node, (char *) argtailv, argtailn,
					argtailn, argtag)) {

				terror("spawn (rflat)");
				kexit(errno);
			}
		}

		if ((px = rspawn(src_node, n_flags, p_value, dest_node,
				argtailc, argtag, &pid)) == ERROR) {

			terror("spawn (rspawn)");
			kexit(errno);
		}

		else {
			if (opt_taken('v')) {

				if (dest_node == LOCAL) {
					dest_node = src_node;
				}

				printf("[%d] %d running on node %x\n",
					px, pid, dest_node);
			}
		}

		getproc(&n_index, &p_value, &n_flags);
	} while(n_index > 0);
}


/*
 *	  help
 *
 *	  Function:	- prints helpful information on the spawn command
 */

static void
help()
{
	printf("\nSynopsis:\tspawn [options] <src> [<dest>] [processes] [-- <args>]\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\t\t-v	turn on verbose mode\n");
	printf("\nDescription:	Spawn a Trollius process.\n");
	printf("\nProcesses:	p<list> (identifiers)\n");
	printf("\t\t%%<list> (indices)\n");
	printf("\t\t@<list> (load addresses)\n");
	printf("\t\t#<list> (blocked events)\n");
        printf("\nNode:		n<constant>\n");
        printf("\t\t\t(any single compute node)\n");
        printf("\t\t\twhere <constant> = decimal or hexadecimal (0x)\n");
	printf("\nRestriction:\tUse spawn only with compute nodes.\n");
	printf("\nExample:	spawn n7 n8 p567\n");
	printf("\t\t\t\"Spawn process ID 567 on node 7 to node 8.\n");
}
