/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	state.c,v $
 * Revision 2.2  92/04/30  14:14:16  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/02  10:04:39  gdburns
 * Give adequate space for temporary formatting buffer.
 * 
 * Revision 2.1.1.1  91/12/18  16:20:57  gdburns
 * Juggle sprintf value declarations.
 * 
 * Revision 2.1  91/03/20  02:52:34  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:26:12  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/05/16  13:25:48  gdburns
 * Print very negative event number in hexadecimal notation.
 * 
 * Revision 2.0  90/02/20  11:53:40  trillium
 * Homo erectus
 *
 *	Function:	- command to print Trollius process status information
 *			- processes can be selected by index, load address,
 *			  identifier or blocked event
 *			- absence of process specifier selects all processes
 *			- nodes are selected in a system dependent fashion
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *
 *	Accepts:	- see the "usage" variable
 *
 *	Examples:	- state n10 %1 %2 p20300
 *			- get status on process indices
 *			  1 & 2 and process ID 20300 on node 10
 */

#include <stdio.h>
#include <events.h>		/* for EVSTOP and EVPAUSE */
#include <kernel.h>		/* for NSTATES */
#include <terror.h>		/* error codes */
#include <typical.h>		/* commonly used constants */
#include <preq.h>		/* kenya interface constants */
#include <net.h>		/* network constants */
#include <ksignal.h>		/* signal constants */

/*
 * external variables
 */
extern int 	errno;		/* global error variable */

/*
 * external functions
 */
extern void	validopts();	/* set the valid options */
extern void	terror();
extern int	do_args();	/* parse the valid options */
extern int	getnode();	/* get next node entry */
extern int	getnodeid();
extern int	getproc();	/* get next process entry */
extern int	kinit();
extern int	kstate();	/* local process status */
extern int	opt_taken();	/* Was a particular option selected? */
extern int	parse_nodes();	/* parse the node identifiers */
extern int	parse_procs();	/* parse the process identifiers */
extern int	rstate();	/* remote process status */
extern char	*rindex();	/* used to get filename	w/o path */

/*
 * local functions
 */
static void	header();
static void	help();
static void	state_lnode();
static void	state_out();
static void	state_rnode();

/*
 * local variables
 */
static char *_copyright = "Copyright 1989 The Ohio State University $Revision: 2.2 $";

static char	*kstates[NSTATES] = {
		"RM",
		"RS",
		"RE",
		"RA",
		"R ",
		"BS",
		"BR",
		"A ",
		"F "
	};

static struct pstate	ptable[PMAX];		/* kenya process table */

/*
 * syntax for state command
 */
static char *usage = "state [-ghlt] <nodes> [<processes>]";


/*
 *******MAIN
 */

main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int		n_index;		/* index of node entry */
	int		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
/*
 * Parse the command line.
 */
	validopts("ghlt");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(parse_procs(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		}

		else {
			terror("state (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Attach to kernel.
 */
	if (kinit(0) == ERROR) {
		terror("state (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index == ERROR) {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	else {
/*
 * Print status header
 */
		header();

		do {
			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(errno);
			}

			if ((nodeid == LOCAL) || (nodeid == getnodeid())) {
				state_lnode();
			}

			else {
				state_rnode(nodeid);
			}

			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}

	kexit(0);
}


/*
 *	state_lnode
 *
 *	Function:	- find and print the state of a local otb process
 *			- obtains state information from the kernel
 *			- kenya state is assigned null values
 */

static void
state_lnode()

{
	struct pstate	ps_local;		/* local state */
	int		fnotpid = FALSE;	/* non pid selectors */
	int		p_index;		/* index of process entry */
	int		p_value;		/* value of process entry */
	int		p_flags;		/* process selection method */


	getproc(&p_index, &p_value, &p_flags);

	if (p_index == ERROR) {
	    fnotpid = TRUE;
	}

	else if (opt_taken('g') || opt_taken('t')) {
	    fnotpid = TRUE;
	}

	else {
	    do {
		if (p_flags == SELECT_PID) {

		    if (kstate(p_value, &ps_local.ps_kernel) == ERROR) {
			terror("state (kstate)");
			kexit(errno);
		    }

		    else {
			ps_local.ps_index = 0;
			ps_local.ps_kenya.p_pid = p_value;
			ps_local.ps_kenya.p_stksize = 0;
			ps_local.ps_kenya.p_stack = 0;
			ps_local.ps_kenya.p_static = 0;
			ps_local.ps_kenya.p_loadpt = 0;
			ps_local.ps_kenya.p_flags = 0;
			ps_local.ps_name[0] = '\0';
			state_out(&ps_local, LOCAL);
		    }
		}

		else {
		    fnotpid = TRUE;
		}

		getproc(&p_index, &p_value, &p_flags);
	    } while(p_index);
	}
/*
 * Were any unsupported host node selection methods used?
 */
	if (fnotpid) {
	    fprintf(stderr, "\nHost node processes must be selected by PID.\n");
	}
}


/*
 *	state_rnode
 *
 *	Function:	- find and print the state of a remote itb process
 *
 *	Accepts:	- node identifier
 */

static void
state_rnode(nodeid)

int			nodeid;

{
	int		p_index;		/* index of process entry */
	int		p_value;		/* value of process entry */
	int		p_flags;		/* process selection method */
	int		nprocs;			/* # processes returned */
	int		i;			/* favourite index variable */

	getproc(&p_index, &p_value, &p_flags);

	if (p_index == ERROR) {

		if (opt_taken('g')) {
			p_flags = SELECT_ALL;
		}

		else if (opt_taken('t')) {
			p_flags = SELECT_SYS;
		}

		else {
			p_flags = SELECT_APPL;
		}
	}

	do {
		nprocs = rstate(nodeid, p_flags, p_value, ptable, PMAX);

		if (nprocs == ERROR) {
			terror("state (rstate)");
			kexit(errno);
		}

		else {
			for (i = 0; i < nprocs; ++i) {
				state_out(ptable + i, nodeid);
			}
		}

		getproc(&p_index, &p_value, &p_flags);
	} while(p_index > 0);
}


/*
 *
 *	  state_out
 *
 *	  Function:	- print out state information
 *
 *	  Accepts:	- completed state structure
 *			- node identifier
 */

static void
state_out(state, nodeid)

struct pstate		*state;
int			nodeid;

{
	char	buffer[32];	/* used for formatting purposes */
	char	*shortname;	/* pointer to filename */
	char	*p;		/* favourite pointer */
	int	i;		/* favourite index variable */
	

	if ((nodeid == LOCAL) || (nodeid == getnodeid())) {
		printf("h        ");
	}

	else {
		printf("%-9.9s", mnemonic(nodeid));
	}
		
	(void) sprintf(buffer, "[%d]", state->ps_index);
	printf(" %-6.6s", buffer);

	if (opt_taken('l')) {
		printf(" %-8X %-8X %-8X", state->ps_kenya.p_stack,
			state->ps_kenya.p_static, state->ps_kenya.p_loadpt);

		if (state->ps_kernel.ks_signal & ~(-1 << TNSIG)) {

			p = buffer;

			for (i = 1; (i <= TNSIG); ++i) {

				if (state->ps_kernel.ks_signal & 1) {
					(void) sprintf(p, "%1.1d,", i);
					p += 2;
				}

				state->ps_kernel.ks_signal >>= 1;
			}

			*(p - 1) = '\0';
			printf(" %-5.5s", buffer);
		}

		else {
			printf(" <nil>");
		}
	}

	else {
		printf(" %-8d", state->ps_kenya.p_pid);
		printf(" %-6d", state->ps_kernel.ks_priority);

		if (state->ps_kernel.ks_event == NOTEVENT) {
			(void) sprintf(buffer, "%s",
				kstates[state->ps_kernel.ks_state - 1]);
		}

		else if ((state->ps_kernel.ks_event == EVSTOP) &&
			(state->ps_kernel.ks_state == SRECV_BLKD)) {
			(void) sprintf(buffer, "<stopped>");
		}

		else if ((state->ps_kernel.ks_event == EVPAUSE) &&
			(state->ps_kernel.ks_state == SRECV_BLKD)) {
			(void) sprintf(buffer, "<paused>");
		}

		else if ((state->ps_kernel.ks_event & 0xFF000000) == MININT) {
			(void) sprintf(buffer, "%s (X%X, %X)",
				kstates[state->ps_kernel.ks_state - 1],
				state->ps_kernel.ks_event,
				state->ps_kernel.ks_type);
		}

		else {
			(void) sprintf(buffer, "%s (%d, %X)",
				kstates[state->ps_kernel.ks_state - 1],
				state->ps_kernel.ks_event,
				state->ps_kernel.ks_type);
		}

		printf(" %-26.26s", buffer);
	}
/*
 * Print the program name without any pathname prefix.
 */
	if ((shortname = rindex(state->ps_name, '/')) == (char *) 0) {
		shortname = state->ps_name;
	} else {
		shortname++;
	}

	printf(" %.19s\n", shortname);
}

/*
 *	  header
 *
 *	  Function:	- print the state command table header
 */

static void
header()
{
	if (opt_taken('l')) {
	  printf("NODE      INDEX  STACK    STATIC   LOADPT   SIGS  PROGRAM\n");
	}

	else {
		printf("NODE      INDEX  PID      KPRI   KSTATE                     PROGRAM\n");
	}
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the state command
 */

static void
help()
{
	printf("\nSynopsis:	state [options] <nodes> [processes]\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\t\t-l	give load address information\n");
	printf("\t\t-g	select all processes\n");
	printf("\t\t-t	select only system processes\n");
	printf("\nDescription:	Print information on Trollius processes.\n");
	printf("\nProcesses:	p<list> (identifiers)\n");
	printf("\t\t%%<list> (indices)\n");
	printf("\t\t@<list> (load addresses)\n");
	printf("\t\t#<list> (blocked events)\n");
	mnusage();
	printf("\t\th\t(local host node)\n");
	printf("\nExample:	state n3-9,11 %%1 %%2 p303\n");
	printf("\t\t\t\"Get status information for PID 303 and\n");
	printf("\t\t\tindices 1 & 2 for nodes 3 thru 9 and 11.\"\n");
}
