/*
 *	OSU Research Computing, architects and engineers	Trillium
 *
 *	Copyright 1989 The Ohio State University
 *
 *	GDB
 *
 *	$Log:	fcast.c,v $
 * Revision 2.2  92/04/30  14:18:22  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/02/24  20:59:28  gdburns
 * Ohio Release
 * 
 *
 *	Function:	- handles casts or groups of nodes that
 *			  synchronously access a file in either multi or
 *			  singl mode
 *			- initially, only HOST2ITB cast is provided
 *			- full blown implementation to come later, if needed
 *			- casts are kept in a singly linked list
 *			- searches are simply linear, since this not
 *			  a speed critical module
 */

#include "filed.h"
#include <net.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern char		*malloc();

/*
 * global functions
 */
struct fcast		*fcadd();	/* add a new cast ID */
struct fcast		*fcdef();	/* prefined cast IDs */
struct fcast		*fcfind();	/* find a cast entry */
int			fcrem();	/* remove an old cast ID */

/*
 * local variables
 */
struct fcast		*pfctop = FCNULL;
					/* top of fcast list */
/*
 * local functions
 */
static int		intcompare();	/* for sorting */

/*
 *	fcadd
 *
 *	Function:	- allocate, initialze and insert a new
 *			  file cast entry into the list
 *			- insertion is always at head of list
 *
 *	Accepts:	- new file cast ID
 *			- array of new file cast members
 *			- number of new file cast members
 *
 *	Returns:	- ptr to new file cast entry or NULL ptr
 */

struct fcast *
fcadd(id, members, num)

int4			id;
int4			*members;
int4			num;

{
	struct fcast	*pfc;		/* new fcast entry */

	pfc = (struct fcast *) malloc((unsigned int) sizeof(struct fcast));

	if (pfc == FCNULL) {
		return(FCNULL);
	}

	pfc->fc_id = id;
	pfc->fc_num = num;
	pfc->fc_members = (int4 *) malloc((unsigned int) (num * sizeof(int4)));

	if (pfc->fc_members == (int4 *) 0) {
		return(FCNULL);
	}

	bcopy((char *) members, (char *) pfc->fc_members,
		(int) (num * sizeof(int4)));
	pfc->fc_next = pfctop;
	pfctop = pfc;
	return(pfc);
}

/*
 *	fcrem
 *
 *	Function:	- remove a file cast entry from the list
 *			- uses a simple linear search to find the old entry
 *
 *	Accepts:	- old file cast identifier
 *
 *	Returns:	- 0 or ERROR
 */

int
fcrem(id)

int4			id;

{
	struct fcast	*curr;		/* current cast entry */
	struct fcast	*prev;		/* previous cast entry in list */


	curr = prev = pfctop;

	while ((curr != FCNULL) && (curr->fc_id != id)) {
		prev = curr;
		curr = curr->fc_next;
	}

	if (curr == FCNULL) {
		errno = EDELETE;
		return(ERROR);
	}

	if (curr == pfctop) {
		pfctop = curr->fc_next;
	}

	else {
		prev->fc_next = curr->fc_next;
	}

	free((char *) curr->fc_members);
	free((char *) curr);
	return(0);
}

/*
 *	fcdef
 *
 *	Function:	- check for pre-defined file casts
 *			- only handles HOST2ITB for now, maybe forever
 *			- consults router for members of pre-defined cast
 *			- if router won't cooperate, assume
 *			  0 thru n-1 node numbering
 *
 *	Accepts:	- possibly predefined file cast identifier
 *
 *	Returns:	- ptr to new file cast entry or NULL ptr
 */

struct fcast *
fcdef(id)

int4			id;

{
	int4		*members;	/* predef cast members */
	int4		num;		/* number of members in predef cast */
	int4		i;		/* favourite counter */

	if (id == HOST2ITB) {
		num = (int4) getnitb();
		members = (int4 *) malloc((unsigned) (num * sizeof(int4)));

		if (! members) {
			return(FCNULL);
		}

		if (getnodes(members, num, NT_ITB, NT_ALL) == ERROR) {

			if (errno == ESERVICE) {

				for (i = 0; i < num; ++i) {
					members[i] = i;
				}
			} else {
				return(FCNULL);
			}
		}

		qsort((char *) members, (int) num, sizeof(int4), intcompare);
		return(fcadd(HOST2ITB, members, num));
	}

	else {
		return(FCNULL);
	}
}

/*
 *	fcfind
 *
 *	Function:	- find cast entry using cast ID as search key
 *
 *	Accepts:	- cast ID
 *
 *	Returns:	- ptr to found cast entry or FCNULL
 */

struct fcast *
fcfind(id)

int4			id;

{
	struct fcast	*p;		/* favourite ptr */

	p = pfctop;

	while ((p != FCNULL) && (p->fc_id != id)) {
		p = p->fc_next;
	}

	if (p != FCNULL) {
		return(p);
	} else {
		return(FCNULL);
	}
}

static int intcompare(i, j) int4 *i, *j; { return(*i - *j); }
