/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fdesc.c,v $
 * Revision 2.2  92/04/30  14:18:27  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/06  10:55:19  gdburns
 * Add flags capability to sendreply().
 * Change organization of descriptor list.
 * 
 * Revision 2.1.1.1  91/10/27  20:54:11  gdburns
 * Change lseek to Sys V.
 * 
 * Revision 2.1  91/02/24  20:59:35  gdburns
 * Ohio Release
 * 
 *	Function:	- helper routines for managing Trollius file descriptors
 *			- Active descriptors are always at the top of the
 *			  list sorted in LRU order.
 *			- Inactive descriptors are always at the bottom of
 *			  the list in no particular order.
 */

#include <errno.h>
#include "filed.h"
#include <freq.h>
#include <stdio.h>
#include <typical.h>
#include <sys/types.h>
#include <unistd.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
char			*malloc();

/*
 * global variables
 */
struct fdesc		*pfdtop = FDNULL;
					/* top of descriptor list */
struct fdesc		*pfdbot = FDNULL;
					/* bottom of descriptor list */
/*
 * local variables
 */
static int4		tfd = 0;	/* now serving descriptor number ... */

/*
 * local functions
 */
static void		fdappend();
static void		fddelete();
static void		fdprepend();

/*
 *	fdnew
 *
 *	Function:	- allocate a new file descriptor
 *	Returns:	- ptr to new file descriptor
 */

struct fdesc *
fdnew()

{
	struct fdesc	*new;		/* ptr new file descriptor */

	new = (struct fdesc *) malloc((unsigned int) sizeof(struct fdesc));

	if (new == FDNULL) {
		fprintf(stderr, "filed: can't malloc new descriptor\n");
		perror("filed");
		kexit(errno);
	}
/*
 * Since the descriptor starts out inactive, append it to the end of pfdtop.
 */
	new->f_tfd = tfd++;
	fdappend(new);

	return(new);
}

/*
 *	fdcut
 *
 *	Function:	- removes old file descriptor
 *	Accepts:	- ptr to old file descriptor
 */

void
fdcut(old)

struct fdesc		*old;

{
	fddelete(old);

	if (old->f_pcubix != CXNULL) {
		free((char *) old->f_pcubix);
	}

	free(old->f_name);
	free((char *) old);
}

/*
 *	fdfind
 *
 *	Function:	- finds descriptor in list using handle (tfd)
 *			  as the search key
 *	Accepts:	- file handle
 *	Returns:	- ptr to found file descriptor
 */

struct fdesc *
fdfind(tfd)

int4			tfd;

{
	struct fdesc	*curr;			/* search ptr */

	curr = pfdtop;

	while ((curr != FDNULL) && (curr->f_tfd != tfd)) {
		curr = curr->f_next;
	}

	return(curr);
}

/*
 *	fdfindname
 *
 *	Function:	- finds descriptor in list using file name
 *			  and open flags as the search keys
 *	Accepts:	- file name
 *			- matching flags
 *	Returns:	- ptr to found file descriptor
 */

struct fdesc *
fdfindname(name, tflags)

char			*name;
int4			tflags;

{
	struct fdesc	*curr;			/* search ptr */
	int4		lock;			/* O_LOCK bit */

	curr = pfdtop;
/*
 * FACTIVE is not considered.
 * O_LOCK is considered only if the caller asks for it.
 */
	tflags &= ~FACTIVE;
	lock = (tflags & O_LOCK) | ~O_LOCK;

	while ((curr != FDNULL) && ((strcmp(name, curr->f_name) != 0) ||
		(((curr->f_tflags & ~FACTIVE) & lock) != tflags))) {
		curr = curr->f_next;
	}

	return(curr);
}

/*
 *	fdactivate
 *
 *	Function:	- activates a file descriptor
 *			- If the file is active (Unix fd is open),
 *			  there is nothing to do.
 *			  Otherwise the file must be reopened and the
 *			  file pointer set to its former location.
 *			  The number of open Unix fds is limited; if the
 *			  filed limit is reached, a previously active file
 *			  must be de-activated (close Unix fd).
 *			  The de-activated fd is chosen on a LRU basis.
 *			- exits on error
 *
 *	Accepts:	- ptr to file descriptor
 */

void
fdactivate(f)

struct fdesc		*f;

{
	int		nactive;	/* # currently active descs */
	struct fdesc	*p;		/* favourite ptr */

	if (f->f_tflags & FACTIVE) {

		if (f != pfdtop) {
			fddelete(f);
			fdprepend(f);
		}

		return;
	}
/*
 * Count the number of active files.
 */
	nactive = 0;

	for (p = pfdtop; (p != FDNULL) && (p->f_tflags & FACTIVE);
		p = p->f_next) {

		nactive++;
	}
/*
 * Check if we need to de-activate (close) the LRU descriptor to make
 * way for the newly activated file.
 */
	if (nactive >= ACTIVEMAX) {
/*
 * Look backwards thru the list to find the least recently used
 * descriptor that is not locked.
 */
		p = (p == FDNULL) ? pfdbot : p->f_prev;

		for (; p->f_tflags & O_LOCK; p = p->f_prev);

		p->f_tflags &= ~FACTIVE;
		fddelete(p);
		fdappend(p);

		if (myclose(p->f_fd) < 0) {
			fprintf(stderr, "filed: can't close active fdesc\n");
			perror("filed");
			kexit(errno);
		}
	}
/*
 * Open the newly activated file and seek to the current file pointer
 * in order to resume operatons at the right place.
 */
	f->f_fd = myopen(f->f_name, f->f_flags, 0);

	if (f->f_fd < 0) {
		fprintf(stderr, "filed: can't open inactive fdesc\n");
		perror("filed");
		kexit(errno);
	}

	if (mylseek(f->f_fd, f->f_ptr, SEEK_SET) < 0) {
		fprintf(stderr, "filed: can't seek inactive fdesc\n");
		perror("filed");
		kexit(errno);
	}

	f->f_tflags |= FACTIVE;
	fddelete(f);
	fdprepend(f);
}

/*
 *	fdlock
 *
 *	Function:	- locks a descriptor permanently active
 *			- At most, we can tolerate ACTIVEMAX - 1
 *			  locked descriptors.  One slot must be
 *			  left open so the cache concept doesn't seize up.
 *
 *	Accepts:	- ptr to file descriptor
 *
 *	Returns:	- 0 or ERROR
 */

int
fdlock(f)

struct fdesc		*f;

{
	int		nlock;		/* # currently locked descs */
	struct fdesc	*p;		/* favourite ptr */

	if (f->f_tflags & FLOCK) {
		return(0);
	}

	fdactivate(f);
	nlock = 0;

	for (p = pfdtop; p != FDNULL; p = p->f_next) {

		if (p->f_tflags & FLOCK) {
			++nlock;
		}
	}

	if (nlock >= (ACTIVEMAX - 1)) {
		errno = ENFILE;
		return(ERROR);
	}

	f->f_tflags |= FLOCK;
	return(0);
}

/*
 *	fddelete
 *
 *	Function:	- deletes existing descriptor from list
 *	Accepts:	- ptr to existing descriptor
 */

static void
fddelete(old)

struct fdesc		*old;

{
	if (old == pfdtop) {
		pfdtop = old->f_next;
	} else {
		old->f_prev->f_next = old->f_next;
	}

	if (old == pfdbot) {
		pfdbot = old->f_prev;
	} else {
		old->f_next->f_prev = old->f_prev;
	}
}

/*
 *	fdappend
 *
 *	Function:	- appends a descriptor to the end of the list
 *	Accepts:	- new descriptor
 */

static void
fdappend(new)

struct fdesc		*new;

{
	if (pfdbot == FDNULL) {
		pfdtop = new;
		new->f_prev = FDNULL;
	} else {
		pfdbot->f_next = new;
		new->f_prev = pfdbot;
	}

	new->f_next = FDNULL;
	pfdbot = new;
}

/*
 *	fdprepend
 *
 *	Function:	- prepends a descriptor to the front of the list
 *	Accepts:	- new descriptor
 */

static void
fdprepend(new)

struct fdesc		*new;

{
	if (pfdtop == FDNULL) {
		pfdbot = new;
		new->f_next = FDNULL;
	} else {
		pfdtop->f_prev = new;
		new->f_next = pfdtop;
	}

	new->f_prev = FDNULL;
	pfdtop = new;
}
