/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqopen.c,v $
 * Revision 2.2  92/04/30  14:18:42  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.5  92/02/06  10:57:10  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.4  92/02/04  10:42:51  gdburns
 * Fully initialize new file descriptors.
 * 
 * Revision 2.1.1.3  91/12/17  22:48:01  gdburns
 * Remove O_CREAT and O_TRUNC from the saved Unix flags.
 * 
 * Revision 2.1.1.2  91/10/27  20:52:16  gdburns
 * Change O_ASYNC to O_NOCBX.
 * Change lseek to sys V.
 * 
 * Revision 2.1.1.1  91/10/09  15:44:23  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/02/24  20:59:47  gdburns
 * Ohio Release
 * 
 *	Function:	- FQOPEN request
 *			- exits daemon in case of non-user error
 *			- supports Cubix
 *			- activates file after opening
 *			- open flags specific to Trollius:
 *			  O_REUSE - reuse existing descriptor with same
 *			  name; does not mix with Cubix methods
 *			  O_NOCBX - default, non-Cubix method
 *			  O_SINGL - Cubix singl mode, see cubix(l)
 *			  O_MULTI - Cubix multi mode, see cubix(l)
 *			  O_LOCK  - lock file permanently active
 *			  O_1WAY  - no reply sent on NOCBX write
 *			  O_DINT  - data are integers
 *			  O_DFLT  - data are floats
 *			  O_DDBL  - data are floats
 */

#include <events.h>
#include <fcntl.h>
#include "filed.h"
#include <freq.h>
#include <net.h>
#include <stdio.h>
#include <sys/types.h>
#include <terror.h>
#include <typical.h>
#include <unistd.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern char		buf[];		/* multi purpose buffer */

/*
 * external functions
 */
extern struct fdesc	*fdfindname();	/* find a filed descriptor by name */
extern struct fdesc	*fdnew();	/* allocate a new descriptor */
extern void		sendreply();	/* general purpose file reply */
extern void		fdactivate();	/* make sure tfd is active */
extern void		fdcut();	/* clean and free descriptor */
extern struct cxnode	*cxnew();	/* initialize Cubix array */
extern char		*malloc();
extern char		*strcpy();
extern char		*ttyname();
extern int		fdlock();	/* make descriptor permanently active */

/*
 * global functions
 */
void			do_open();
struct fdesc		*fdinit();

/*
 *	do_open
 *
 *	Function:	- handles simple client opens
 *			- handles synchronized Cubix opens
 *			- does not support multiple concurrent opens
 *			  of the same file in Cubix mode (yields ECXREQ)
 *
 *	Accepts:	- ptr to filed request
 */

void
do_open(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target descriptor */
	int		status;		/* Cubix request status */
	int4		i;		/* favourite index */
	int4		fd;		/* initial open descriptor */
	int		err;		/* errno */

/*
 * Cubix opens are synchronized by looking for the same filename.
 */
	if (request->fq_flags & FCUBIX) {
		f = fdfindname(buf, request->fq_flags);
/*
 * Initialize a new Cubix file descriptor.
 */
		if (f == FDNULL) {
			f = fdinit(request->fq_flags);
			f->f_pcubix = cxnew(HOST2ITB, &(f->f_ncubix));
			f->f_req = FQOPEN;
			f->f_nheard = 0;
		}

		f->f_src_node = request->fq_src_node;
		f->f_src_event = request->fq_src_event;
/*
 * Hear the latest Cubix node to synchronize on open.
 */
		status = cxhear(f, request);

		if (status == ERROR) {
			sendreply(request->fq_src_node, request->fq_src_event,
				errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		} else if (status == 0) {
/*
 * Not all Cubix nodes have synchronized yet.  There is nothing to do.
 */
		} else {
/*
 * Open the file first to detect errors.
 */
			fd = myopen(buf, request->fq_flags & ~O_FILED,
				request->fq_mode);

			if (fd != ERROR) {
/*
 * Determine the starting file ptr based on what happened in the test open().
 * This handles O_APPEND.
 */
				f->f_ptr = mylseek(fd, INT4_NIL, SEEK_CUR);
/*
 * Close the file and allow it to be activated naturally by filed.
 */
				(void) myclose(fd);
				fdactivate(f);
/*
 * Lock the file active if requested.
 */
				if ((request->fq_flags & O_LOCK) &&
					(fdlock(f) == ERROR)) {
					err = errno;
					fd = ERROR;
				} else {
					err = 0;
					fd = f->f_tfd;
				}
			}

			else {
				err = errno;
				fdcut(f);
			}
/*
 * Reply to all Cubix nodes.
 */
			for (i = 0; i < f->f_ncubix; ++i) {
				sendreply(f->f_pcubix[i].cx_src_node,
					f->f_pcubix[i].cx_src_event,
					err, fd, INT4_NIL, INT4_NIL);
			}
		}
	}
/*
 * Is this a REUSE situation?
 */
	else if (request->fq_flags & O_REUSE) {
		request->fq_flags &= ~O_REUSE;
		f = fdfindname(buf, request->fq_flags);
/*
 * This is the first open.  Remove O_REUSE and recall do_open().
 */
		if (f == FDNULL) {
			do_open(request);
		}
/*
 * This is a subsequent open.  Bump the reuse counter.  Note the client.
 */
		else {
			f->f_count += 1;
			f->f_src_node = request->fq_src_node;
			f->f_src_event = request->fq_src_event;
			sendreply(request->fq_src_node, request->fq_src_event,
				0, f->f_tfd, INT4_NIL, INT4_NIL);
		}
	}
/*
 * This is a simple open in NOCBX mode.
 */
	else {
/*
 * Open the file first to detect errors.
 */
		fd = myopen(buf, request->fq_flags & ~O_FILED,
			request->fq_mode);

		if (fd != ERROR) {
/*
 * Determine the pointer within the file.
 * Close it, initialize the descriptor and let fdactivate decide active status.
 */
			f = fdinit(request->fq_flags);
			f->f_ptr = mylseek(fd, INT4_NIL, SEEK_CUR);
			(void) myclose(fd);
			f->f_count = 1;
			f->f_src_node = request->fq_src_node;
			f->f_src_event = request->fq_src_event;
			fdactivate(f);
/*
 * Lock the file active if requested.
 */
			if ((request->fq_flags & O_LOCK) &&
				(fdlock(f) == ERROR)) {
				fd = ERROR;
			} else {
				errno = 0;
				fd = f->f_tfd;
			}
		}

		sendreply(request->fq_src_node, request->fq_src_event,
			errno, fd, INT4_NIL, INT4_NIL);
	}
}

/*
 *	fdinit
 *
 *	Function:	- initializes a new file descriptor
 *			- New descriptors cannot be active or locked.
 *			- sets as many fields as possible
 *			- file name comes from I/O buffer
 *	Accepts:	- access flags
 *	Returns:	- ptr to new descriptor
 */

struct fdesc *
fdinit(tflags)

int4			tflags;		/* Trollius and Unix flags */

{
	struct fdesc	*f;		/* ptr to target descriptor */

	f = fdnew();
	f->f_fd = -1;
	f->f_ptr = 0;
	f->f_count = 0;
	f->f_src_node = NOTNODEID;
	f->f_src_event = NOTEVENT;
	f->f_req = -1;
	f->f_nheard = 0;
	f->f_ncubix = 0;
	f->f_pcubix = CXNULL;
	f->f_flow = 0;
	f->f_tflags = tflags & ~(O_LOCK | FACTIVE);
/*
 * Isolate the Unix flags.
 */
	f->f_flags = tflags & ~O_FILED;
/*
 * The effect of O_CREAT and O_TRUNC will be achieved in the test open().
 * We don't want them when activating files.
 */
	f->f_flags &= ~(O_TRUNC | O_CREAT);
/*
 * Copy the filename.
 */
	f->f_name = malloc((unsigned) (strlen(buf) + 1));

	if (f->f_name == NULL) {
		fprintf(stderr, "filed: cannot allocate file name\n");
		perror("filed");
		kexit(errno);
	}
	
	(void) strcpy(f->f_name, buf);
	return(f);
}

/*
 *	fdstdio
 *
 *	Function:	- initializes the standard I/O descriptors,
 *			  stdin (0), stdout (1) and stderr (2)
 *			- stdio are locked and active; therefore it
 *			  is assumed that ACTIVEMAX >= 4
 *			- should never be closed
 */

void
fdstdio()

{
	struct fdesc	*f;		/* ptr to new target descriptor */
/*
 * stdin
 */
	if (isatty(0)) {
		(void) strcpy(buf, ttyname(0));
	} else {
		(void) strcpy(buf, "/dev/tty");
	}

	f = fdinit(O_RDONLY | O_NOCBX);
	f->f_tflags |= FLOCK | FACTIVE;
	f->f_fd = 0;
	f->f_src_node = NOTNODEID;
	f->f_src_event = NOTEVENT;
/*
 * stdout
 */
	if (isatty(1)) {
		(void) strcpy(buf, ttyname(1));
	} else {
		(void) strcpy(buf, "/dev/tty");
	}

	f = fdinit(O_WRONLY | O_NOCBX);
	f->f_tflags |= FLOCK | FACTIVE;
	f->f_fd = 1;
	f->f_src_node = NOTNODEID;
	f->f_src_event = NOTEVENT;
/*
 * stderr
 */
	if (isatty(2)) {
		(void) strcpy(buf, ttyname(2));
	} else {
		(void) strcpy(buf, "/dev/tty");
	}

	f = fdinit(O_RDWR | O_NOCBX);
	f->f_tflags |= FLOCK | FACTIVE;
	f->f_fd = 2;
	f->f_src_node = NOTNODEID;
	f->f_src_event = NOTEVENT;
}
