/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqwrite.c,v $
 * Revision 2.2  92/04/30  14:18:48  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/06  10:57:19  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.1  91/10/09  15:44:42  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/02/24  20:59:54  gdburns
 * Ohio Release
 * 
 *	Function:	- FQWRITE and FQCXWRITE requests
 *			- exits daemon in case of non-user error
 *			- FQCXWRITE is a Cubix call only.
 *			- FQWRITE is the real workhorse.
 *
 *			- This is the most complicated filed operation
 *			  for two reasons.  There is a double synchronization
 *			  in the Cubix method and the written data must be
 *			  packetized on the client's end.  To filed, individual
 *			  packets are separate cxwrite requests.  Thus, the
 *			  operation is non-atomic when the data length exceeds
 *			  the network packet size.  In the second round
 *			  of synchronization on a Cubix file, we must ensure
 *			  that we have seen the last write request before
 *			  we move on.
 *
 *			- The two rounds of synchronization on Cubix write
 *			  are very inefficient, but it prevents the massive
 *			  complication of dealing with a new Cubix operation
 *			  before the last one has completed.
 */

#include <errno.h>
#include <freq.h>
#include "filed.h"
#include <net.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern char		buf[MAXNMSGLEN];/* body of filed request */

/*
 * external functions
 */
struct fdesc		*fdfind();	/* find a filed descriptor */
struct cxnode		*cxfind();	/* find a Cubix descriptor */
void			sendreply();	/* general purpose file reply */
void			fdactivate();	/* make sure tfd is active */
int			cxvalid();	/* validate a Cubix cast client */

/*
 * global functions
 */
void			do_cxwrite();	/* Cubix write hand waving */
void			do_write();	/* meat */

/*
 *	do_write
 *
 *	Function:	- write data to an open file descriptor
 *			- length limited to MAXNMSGLEN
 *			- must be aware of Cubix multi mode
 *
 *	Accepts:	- ptr to filed request
 */

void
do_write(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target descriptor */
	struct cxnode	*p;		/* ptr to Cubix client */
	int4		i;		/* favourite index */
	int4		ret;		/* bytes written by mywrite() */
	int		err;		/* errno */
	int		blast;		/* Is this the last packet? */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			EBADF, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	f->f_src_node = request->fq_src_node;
	f->f_src_event = request->fq_src_event;
/*
 * Ensure that the file is activated.
 */
	fdactivate(f);
	ret = mywrite(f->f_fd, buf, request->fq_length);

	if (ret != ERROR) {
		f->f_flow += ret;
		f->f_ptr += ret;
		err = 0;
		blast = ret < request->fq_length;
	}

	else {
		blast = TRUE;
		err = errno;
	}
/*
 * multi method: Do the carefully synchronized multi dance.
 */
	if (f->f_tflags & FMULTI) {

		p = cxfind(request->fq_src_node, f);

		if (p == CXNULL) {
			sendreply(request->fq_src_node, request->fq_src_event,
				ECXMEMBER, (int4) ERROR, INT4_NIL, INT4_NIL);
			return;
		}

		p->cx_ret = ret;
		p->cx_errno = err;
		p->cx_length -= ret;
		if (p->cx_length == 0) blast = TRUE;
		if (blast) f->f_nheard += 1;
/*
 * Send reply to all Cubix nodes if final member was heard.
 * We are then done with the second round of synchronization.
 */
		if (f->f_nheard == f->f_ncubix) {
			f->f_nheard = 0;
			f->f_req = FQIDLE;
			p = f->f_pcubix;

			for (i = 0; i < f->f_ncubix; ++i) {
				sendreply(p->cx_src_node, p->cx_src_event,
					p->cx_errno, p->cx_ret, INT4_NIL,
					INT4_NIL);
				p++;
			}
		}
/*
 * Tell next Cubix node to go ahead with FQWRITE.
 */
		else if (blast) {
			p++;
			sendreply(p->cx_src_node, p->cx_src_event,
				0, INT4_NIL, INT4_NIL, INT4_NIL);
		}
/*
 * Reply to an intermediate packet.
 */
		else {
			sendreply(p->cx_src_node, p->cx_src_event,
				err, ret, INT4_NIL, INT4_NIL);
		}
	}
/*
 * singl method:  Respond to all members with result of single write.
 * The first member, whose operation was accepted and is waiting at
 * the second round of synchronization, gets the number of bytes written
 * from the last packet.  This is no different from multi method.
 * Everybody else, waiting at the first round of synchronization,
 * gets the error EXSINGL, which tells them not to continue.
 * They are also told how much of their *total* write length was
 * *not* written (normally 0).
 */
	else if (f->f_tflags & FSINGL) {
		p = f->f_pcubix;
		sendreply(p->cx_src_node, p->cx_src_event, err, ret, INT4_NIL,
			INT4_NIL);
		p->cx_length -= ret;

		if (p->cx_length == 0) blast = TRUE;

		if (blast) {
			f->f_req = FQIDLE;

			if (err == 0) {
				err = ECXSINGL;
				ret = p->cx_length;
			}

			for (i = 1; i < f->f_ncubix; ++i) {
				p++;
				sendreply(p->cx_src_node, p->cx_src_event,
					err, ret, INT4_NIL, INT4_NIL);
			}
		}
	}
/*
 * async method: Respond to client.
 */
	else {
		if (! (f->f_tflags & F1WAY)) {
			sendreply(request->fq_src_node, request->fq_src_event,
				err, ret, INT4_NIL, INT4_NIL);
		}
	}
}

/*
 *	do_cxwrite
 *
 *	Function:	- Cubix write operation
 *			- Unless all other cast members have made the same
 *			  request, do_cxwrite() simply notes the event and
 *			  takes no other action.
 *			- If all other members have been "heard" all member's
 *			  requests are replied in order (MULTI) or
 *			  a single write operation is performed and the result
 *			  sent to each (SINGL).
 *
 *	Accepts:	- ptr to filed request
 */

void
do_cxwrite(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target filed desc. */
	int		status;		/* Cubix status from cxhear() */
	
	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			EBADF, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	f->f_src_node = request->fq_src_node;
	f->f_src_event = request->fq_src_event;
	status = cxhear(f, request);

	if (status == ERROR) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
	}

	else if (status == 0) {
/*
 * Not all cast members have been heard from.
 * Nothing to do at this point.
 */
	}
/*
 * Else we have synchronized all cast members and are ready to take
 * FQWRITE requests.
 */
	else {
		f->f_req = FQWRITE;
/*
 * Reply to the first Cubix node in order.
 * This takes care of both singl and multi methods.
 */
		sendreply(f->f_pcubix->cx_src_node, f->f_pcubix->cx_src_event,
			0, INT4_NIL, INT4_NIL, INT4_NIL);
	}
}
