/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RKM/GDB
 *
 *	$Log:	kernelio.msg.c,v $
 * Revision 2.2  92/04/30  14:19:24  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  11:56:51  gdburns
 * enhanced robustness
 * new tkill procedure
 * 
 * Revision 2.1.1.1  91/04/24  13:25:06  gdburns
 * Remove hanging wires from the old filesys.
 * 
 * Revision 2.1  91/03/20  02:58:35  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:23:40  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:51:56  trillium
 * Homo erectus
 *
 *	Function:	- receives kernel requests and sends replys
 *			- OTB message queue version
 */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include <kernel.h>
#include <kreq.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern struct kreq	request;		/* kernel request */
extern struct kreply	reply;			/* kernel reply */

/* 
 * external functions 
 */
extern key_t 		ftok();			/* generate tokens for IPC */
extern char		*getenv();
extern char		*kkillopen();		/* open kill file */
extern char		*memcpy();
extern void		kkillall();		/* clean up clients */

/*
 * global functions
 */
void			init_request();		/* initialize IPC structs */
void 			recv_req();		/* receive requests from proc */
void			send_reply();		/* reply on request to proc */

/*
 * local variables
 */
static int		kqid;			/* kernel's message queue ID */

/*
 * local functions
 */
static void		kcleanup();		/* clean up clients and IPC */

/*
 *	kcleanup
 *
 *	Function:	- removes kernel message queue
 *			- calls kkillall() to kill all clients
 */

static void
kcleanup()

{
	char		f_token[32];		/* pathname to create token */

	if (msgctl(kqid, IPC_RMID, (struct msqid_ds *) 0) < 0) {
		fprintf(stderr, "kernel: can't remove my message queue\n");
		perror("kernel (msgctl)");
	}
/*
 * Remove token creation file.
 */
	sprintf(f_token, "/tmp/%s%dk", getenv("USER"), getuid());
	(void) unlink(f_token);
/*
 * kkillall() does not return.
 */
	kkillall();
}

/*
 *	init_request
 *
 *	Function:	- initialize communication structures for
 *			  sending and receiving kernel requests
 */

void
init_request()

{
	int		fd_token;		/* token file desc. */
	key_t		token;			/* message queue key */	
	char		*f_kill;		/* kill filename */
	char		f_token[32];		/* pathname to create token */
/*
 * Generate the token to create the kernel's message queue.
 */
	sprintf(f_token, "/tmp/%s%dk", getenv("USER"), getuid());

	if ((fd_token = open(f_token, O_CREAT, 0600)) < 0) {
		fprintf(stderr, "kernel: %s\n", f_token);
		perror("kernel (init_request (open))");
		exit(errno);
	}

	if ((token = ftok(f_token, KERNEL_ID)) < 0) {
		fprintf(stderr, "kernel: %c %s\n", KERNEL_ID, f_token);
		perror("kernel (ftok)");
		exit(errno);
	}		

	(void) close(fd_token);
/*
 * Create kernel qid.
 */
	if ((kqid = msgget(token, 0600 | IPC_CREAT | IPC_EXCL)) < 0) {
		fprintf(stderr, "kernel: can't create message queue\n");
		perror("kernel (msgget)");
		exit(errno);
	}
/*
 * Catch common signals so that we can clean up.
 */
	(void) signal(SIGTERM, kcleanup);
	(void) signal(SIGHUP,  kcleanup);
	(void) signal(SIGINT,  kcleanup);
/*
 * Open the kill record file.
 */
	if ((f_kill = kkillopen()) != (char *) 0) {
		fprintf(stderr, "kernel: %s\n", f_kill);
		perror("kernel (kkillopen)");
		exit(errno);
	}
}

/*
 *	recv_req
 *
 *	Function:	- receives the next kernel request
 */

void
recv_req()

{
	int		ret;			/* OTB system return code */

	struct {				/* request message */
		long		mq_type;
		struct kreq	mq_text;
	} msgreq;

/* 
 * Loop through OTB system interrupts, receiving a request. 
 */
	while (((ret = msgrcv(kqid, (struct msgbuf *) &msgreq,
		sizeof(struct kreq), (long) 0, 0)) < 0) && (errno == EINTR));

	if (ret < 0) {
		fprintf(stderr, "kernel: can't receive request\n");
		perror("kernel (msgrcv)");
		exit(errno);
	}

	memcpy((char *) &request, (char *) &msgreq.mq_text,
		sizeof(struct kreq));
}

/*
 *	send_reply
 *
 *	Function:	- sends reply back to the client
 *	Accepts:	- client's kernel communication port
 */

void
send_reply(client_port)

int			client_port;		/* QID of client */

{
	int		ret;

	struct {				/* reply message */
		long		mr_type;
		struct kreply	mr_text;
	} msgreply;

	memcpy((char *) &msgreply.mr_text, (char *) &reply,
		sizeof(struct kreply));
	msgreply.mr_type = KERNEL_REPLY;
/*
 * Loop through OTB system interrupts, sending a reply.
 */
	while (((ret = msgsnd(client_port, (struct msgbuf *) &msgreply,
		sizeof(struct kreply), 0)) < 0) && (errno == EINTR));

/*
 *	if (ret < 0) {
 *		fprintf(stderr, "kernel: can't send reply\n");
 *		perror("kernel (msgsnd)");
 *		exit(errno);
 *	}
 */
}
