/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	VSD/GDB/RBD
 *
 *	$Log:	sa_rtr.c,v $
 * Revision 2.2  92/04/30  14:20:13  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/03/11  09:58:46  gdburns
 * Do not use rmalloc().
 * 
 * Revision 2.1.1.1  91/11/13  17:59:44  gdburns
 * Remove declarations for non-existant functions.
 * 
 * Revision 2.1  91/03/20  02:59:51  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/09/25  23:04:22  raja
 * Added support for the RNODES and RGETSMART requests.
 * Modified send_reply() to accept sending a message.
 * Added count_ntype() as a function used internally.
 * 
 * Revision 2.0.1.2  90/07/13  19:23:35  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/05/19  04:49:24  raja
 * Modified the RGETROUTE request to handle the new getrent functions.
 * Modified the RGETNTYPE request to handle true type masking.
 * 
 * Revision 2.0  90/02/20  11:49:33  trillium
 * Homo erectus
 * 
 *	Function:	- The standalone router is a simplified version of
 *			  the generic router and is used when a single host
 *			  node emulates the multicomputer environment.
 *
 *			- It is used in conjunction with the launch program.
 *			  Using the launch command, one can specify which
 *			  processes are to be executed on which nodes
 *			  eventually in the multicomputer system. The
 *			  sa_rtr records this information and handles the
 *			  various requests that it receives as if it is
 *			  a multiple node system. However, all processes
 *			  for all nodes are run on the single host node.
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <events.h>
#include <terror.h>
#include <typical.h>
#include <rreq.h>
#include <debug.h>
#include <net.h>

#undef NULL
#define NULL		((char *) 0)

#define PNULL		((struct pid_list *) 0)

/*
 * external functions
 */
extern char		*calloc();
extern char		*malloc();
extern int		stoi();

/*
 * local functions
 */
static void		do_getroute();
static void		do_getnet();
static void		do_attach();
static void		do_detach();
static void		do_noservice();
static void		do_hold();
static void		do_nodes();
static void		do_getsmart();
static void		do_getntype();
static void		recv_req();
static void		send_reply();

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * local variables
 */
static char		*_copyright =
	"Copyright 1992 The Ohio State University $Revision: 2.2 $";

static char		*me;		/* argv[0] */
static struct net_info	net;		/* node and network information */
static int		ni_nitb;	/* fake number of itb nodes */
static int		ni_notb;	/* fake number of otb nodes */
static int		ni_ncast;	/* fake number of cast entries */
static struct nmsg	inbound;	/* network message header (requests) */
static struct nmsg	outbound;	/* network message header (replies) */
static struct rreq	*request;	/* router request */
static struct rreply	*reply;		/* router reply */

static void		(*do_req[NREQUESTS])() = {
		do_getroute,
		do_getnet,
		do_noservice,
		do_attach,
		do_detach,
		do_noservice,
		do_noservice,
		do_hold,
		do_nodes,
		do_noservice,
		do_getsmart,
		do_getntype,
};

/*
 * The do_noservice calls are for RSETNET, RLOADROUTES, RSETROUTE,
 * and RREMROUTE reply which are meaningful for the generic router only.
 */

/*
 * data structures for standalone router
 */
static struct pid_list {
	int		p_pid;
	int		p_nodeid;
	struct pid_list	*p_next;
};

static struct pid_list	*p_start;	/* pointer to poid_list */
				

main(argc, argv)

int			argc;
char		      **argv;

{
	me = argv[0];
/*
 * If network information is supplied on the command line, use it.
 */
	ni_nitb = (argc > 1) ? stoi(argv[1]) : 1;
	ni_notb = 1;
	ni_ncast = 0;
	net.ni_nodeid = (argc > 2) ? stoi(argv[2]) : LOCAL;
	net.ni_nodetype = NT_OTB;
	net.ni_origin = net.ni_nodeid;
/*
 * Initialize and attach to the local Trillium kernel.
 */
	request = (struct rreq *) inbound.nh_data;
	reply = (struct rreply *) outbound.nh_data;

	if (kinit(0) < 0) {
		terror(me);
		exit(errno);
	}
/*
 * Continuously receive and process router requests.
 */
	p_start = PNULL;
	inbound.nh_event = EVROUTER;

	for(;;) {
		recv_req();
		inbound.nh_event = EVROUTER;
		db(("%s: recvd request # %d\n", me, request->rq_request));

		if ((request->rq_request < RNORMAL) ||
			request->rq_request > NREQUESTS) {

			reply->rr_reply = EBADREQUEST;
			send_reply(NULL, 0);
		}

		else {
			(*(do_req[request->rq_request - RNORMAL]))();
		}
	}
}

/*
 *	do_getroute
 *
 *	Function:	- RGETROUTE request
 */

static void
do_getroute()

{
	reply->rr_reply = 0;
/*
 * Destination is always the local node, so reply with NOTEVENT since
 * this will make getroute() return the destination event passed to it.
 * The nodetype returned is the same one defined in the net structure.
 */
	reply->rr_event = NOTEVENT;
	reply->rr_nodetype = net.ni_nodetype;
	send_reply(NULL, 0);
}

/*
 *	do_getnet
 *
 *	Function;	- RGETNET request
 *			- replies with all information from the
 *			  "net" structure
 */

static void
do_getnet()

{
        struct pid_list	*right;

	reply->rr_reply = 0;
	reply->rr_nodetype = net.ni_nodetype;
	reply->rr_origin = net.ni_origin;
/*
 * If pid of process requesting net info is found in pid_list, then
 * return the corresponding nodeid, else return default nodeid.
 */
	right = p_start;

	while ((right != PNULL) && (right->p_pid < request->rq_pid)) {
		right = right->p_next;
  	}

	if ((right != PNULL) && (right->p_pid == request->rq_pid)) {
		reply->rr_nodeid = right->p_nodeid;
	} else {
		reply->rr_nodeid = net.ni_nodeid;
	}

	send_reply(NULL, 0);
}

/*
 *	do_noservice
 *
 *	Function:	- not supported by the standalone router
 *			- supported by the generic router
 *			- hence return error
 */

static void
do_noservice()

{
	reply->rr_reply = ESERVICE;
	send_reply(NULL, 0);
}

/*
 *	do_attach
 *
 *	Function:	- RATTACH request
 *			- register (pid,nodeid) pair in linked list
 *	                - this registration msg is recd from launch process
 */

static void
do_attach()

{
        struct pid_list *element;
        struct pid_list *left, *right;

        element = (struct pid_list *)
	                calloc (1, (unsigned) sizeof (struct pid_list));
	element->p_pid = request->rq_pid;
	element->p_nodeid = request->rq_nodeid;

	if (p_start == PNULL) {
		p_start = element;
		element->p_next = PNULL;
	}

	else {
		left = PNULL;
		right = p_start;

		while ((right != PNULL) && (right->p_pid < element->p_pid)) {
			left = right;
			right = right->p_next;
        	}

		if (left == PNULL) {
			element->p_next = p_start;
			p_start = element;
		} else {
			element->p_next = left->p_next;
			left->p_next = element;
		}
	}

	send_reply(NULL, 0);
}

/*
 *	do_detach
 *
 *	Function:	- RDETACH request
 *			- not supported by the generic router
 *			- supported by the stand-alone router
 *			- unregister (pid,nodeid) pair in linked list
 *	                - this registration msg is recd from launch process
 */

static void
do_detach()

{
        struct pid_list *left, *right;
	int             xpid;

	xpid = request->rq_pid;
	right = p_start;

	while ((right != PNULL) && (right->p_pid < xpid)) {
		left = right;
		right = right->p_next;
  	}

	if ((right != PNULL) && (right->p_pid = xpid)) {

		if (right == p_start) p_start = right->p_next;
		else left->p_next = right->p_next;

		free((char *) right);
	}

	send_reply(NULL, 0);
}

/*
 *	do_hold
 *
 *	Function:	- RHOLD request
 *			- forces sa_rtr to wait for the next message from
 *			  the sender of the RHOLD message
 */

static void
do_hold()

{
	inbound.nh_event = request->rq_src_event;
}

/*
 *	do_nodes
 *
 *	Function:	- delivers a list of nodes to the client
 */

static void
do_nodes()

{
	int		*pnodes;		/* node list */
	int		*p;			/* favorite pointer */
	int		i;			/* favorite index */
	int		nnodes;			/* length of node list */
	int		nodetype;		/* requested node type */
	int		typemask;		/* nodetype bit mask */

	reply->rr_reply = 0;
	typemask = request->rq_typemask;
	nodetype = request->rq_nodetype;

	if ((nnodes = count_ntype(nodetype, typemask)) == 0) {
		send_reply(NULL, 0);
		return;
	}

	nodetype &= typemask;
	p = pnodes = (int *) malloc((unsigned) (nnodes * sizeof(int)));

	if (!p) {
		perror(me);
		kexit(errno);
	}

	if (nodetype == (NT_ITB & typemask)) {
/*
 * Request matches the ITB node(s).
 */
		for (i = 0; i < ni_nitb; i++) {
			*p++ = i;
		}
	}

	if (nodetype == (NT_OTB & typemask)) {
/*
 * Request matches the OTB node.
 */
		*p = net.ni_origin;
	}

	send_reply((char *) pnodes, nnodes * sizeof(int));
	free((char *) pnodes);
}

/*
 *	do_getsmart
 *
 *	Function:	- RGETSMART request
 */

static void
do_getsmart()

{
/*
 * Since we received this request, we must be smart and the client is
 * unblocked and happy.
 */
}

/*
 *      do_getntype
 *
 *      Function:       - RGETNTYPE request
 */

static void
do_getntype()

{
        reply->rr_reply = 0;
	reply->rr_ntype = count_ntype(request->rq_nodetype,
		request->rq_typemask);
        send_reply(NULL, 0);
}

/*
 *	recv_req
 *
 *	Function:	- receives a request from a client
 */

static void
recv_req()

{
	inbound.nh_type = 0;
	inbound.nh_flags = 0;
	inbound.nh_length = 0;

	if (nrecv(&inbound) < 0) {
		terror(me);
		kexit(errno);
	}
}

/*
 *	send_reply
 *
 *	Function:	- sends a reply back to the client
 *	Accepts:	- reply message pointer
 *			- reply message length
 */

static void
send_reply(msg, length)

char			*msg;
int			length;

{
	outbound.nh_event = request->rq_src_event;
	outbound.nh_type = 0;
	outbound.nh_length = length;
	outbound.nh_msg = msg;
/*
 * The client is always on the local node.  No sweat here.
 */
	outbound.nh_flags = NOBUF;
	outbound.nh_dl_event = outbound.nh_event;

	if (dsend(&outbound) < 0) {
		terror(me);
		kexit(errno);
	}
}

/*
 *	count_ntype
 *
 *	Function:	- returns the # of nodes that satisfy the given
 *			  nodetype and typemask
 *	Accepts:	- the nodetype bit pattern required
 *			- the mask pattern for the nodetype
 */

static int
count_ntype(nodetype, typemask)

int			nodetype,
			typemask;

{
	int		nnodes;		/* sum of matched nodes */

	nnodes = 0;
	nodetype &= typemask;

	if (nodetype == (NT_ITB & typemask)) {
/*
 * Request matches the ITB node(s).
 */
		nnodes += ni_nitb;
	}

	if (nodetype == (NT_OTB & typemask)) {
/*
 * Request matches the OTB node(s).
 */
		nnodes += ni_notb;
	}

	return(nnodes);
}
