/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	AKP/DLF/GDB
 *
 *	$Log:	tping.c,v $
 * Revision 2.2  92/04/30  14:16:02  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/04/27  11:46:27  raja
 * new multicast interface
 * 
 * Revision 2.1.1.2  91/11/13  17:51:36  gdburns
 * Don't allocate 0 bytes.
 * 
 * Revision 2.1.1.1  91/05/26  20:26:42  raja
 * Fixed typo in the error message when the message size is negative.
 * 
 * Revision 2.1  91/03/20  02:52:39  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/11/29  18:05:14  gdburns
 * Renamed min and max so as not to conflict with defined macros.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:16  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:42  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  00:55:26  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:15:00  trillium
 * Initial revision
 * 
 *
 *	Function : - Talk to the echod process...repetitively.
 *
 *	Accepts  : - argv, argc
 *
 *	Returns  : - summary status information is printed upon interrupt
 *
 *	8/87 Option processing added (DLF)
 *	2/89 new option processing added (GDB)
 */

#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <ereq.h>
#include <terror.h>
#include <net.h>
#include <typical.h>

#define plural(n)		((n != 1) ? "s" : "")

/*
 * external variables
 */
extern int	errno;

/*
 * external functions
 */
extern int	kinit();
extern int	predef();
extern int	recho();
extern char	*malloc();

/*
 * static variables
 */
static char *usage="tping <node> [-vh] [-c <count>] [-l <length>] [-d <delay>]";

static int	interrupt = 0;	/* set when trapping an interrupt.... */
static int	sent;		/* number of messages sent */
static int	arg_count;	/* # of messages to echo */
static int	arg_size;	/* size of each message in bytes */
static int	arg_delay;	/* delay between each echo */
static int	arg_verbose;	/* verbosity flag */
static int	nodeid;		/* destination node ID */
static double	tmin;		/* timing */
static double	tmax;		/* timing */
static double	tot;		/* timing */

/*
 * static functions
 */
static int	checkargs();
static void	help();
static void	trap();


main(argc, argv)
  int   argc;
  char *argv[];
{
    char *p, *q;
    int   dec, r;
    struct timeval start, stop;
    double roundtrip;

/*
 ***GDB***
 *
 * Parse the command line.
 */
        validopts("hv");
	followed("cld");
 
        if ((do_args(&argc, argv) == ERROR) ||
                (parse_nodes(&argc, argv) == ERROR) ||
                (errno = (argc == 1) ? 0 : EUSAGE)) {
 
                if (errno == EUSAGE) {
                        fprintf(stderr, "usage: %s\n", usage);
                }
 
                else {
                        terror("tping (arguments)");
                }
 
                exit(errno);
        }
/*
 * Check for help request.
 */
        if (opt_taken('h')) {
                help();
                exit(0);
        }
/***GDB***/


    /* set defaults and set destintaion node ID */
    if (checkargs() == ERROR) {
	fprintf(stderr, "usage: %s\n", usage);
	exit(errno);
    }
/*
 * Allocate input and output buffers.
 */
	if (arg_size) {
		p = malloc((unsigned) arg_size);
		q = malloc((unsigned) arg_size);

		if ((!p) || (!q)) {
			fprintf(stderr, "tping: can't malloc(%d)\n", arg_size);
			perror("tping (malloc)");
			exit(errno);
		}
	} else {
		p = q = NULL;
	}

    if (arg_count) {
	dec = 1;
    } else {
	arg_count = 1;
	dec   = 0;
    }

    /* here we go... */
    if (kinit(0)) {
	terror("tping (kinit)");
	exit(errno);
    }

    signal(SIGINT, trap);
    signal(SIGTERM, trap);

    tmin =  1.0e+20;		/* i hope that is big enough... */
    tmax = -1.0e-20;		/* i hope that is small enough... */
    tot = 0.0;			/* i hope that is empty enough... */
/*
 * Register any host castids (predefs) required.
 */
    if (predef(nodeid)) {
	terror("tping (predef)");
	kexit(errno);
    }

    for (sent = 0; (arg_count && !interrupt); arg_count -= dec, sent++) {

	/* send it out... */
	gettimeofday(&start, (struct timezone *) 0);

	if ((r = recho(nodeid, p, q, arg_size)) != arg_size) {

	    if (r == ERROR) {
		terror("tping (recho)");
	    }

	    else {
		fprintf(stderr, "tping (recho): length error\n");
	    }

	    kexit(errno);
	}
	gettimeofday(&stop, (struct timezone *) 0);

	/* compute time */
	roundtrip  = (double) (stop.tv_sec - start.tv_sec);
	roundtrip += (double) (stop.tv_usec - start.tv_usec) * 0.000001;

	tot += roundtrip;
	if (roundtrip < tmin) {
	    tmin = roundtrip;
	}
	if (roundtrip > tmax) {
	    tmax = roundtrip;
	}

	if (arg_verbose) {

	    printf("%3d byte%s from ", arg_size, plural(arg_size));

	    if (nodeid == LOCAL) {
		printf("local node: ");
	    }

	    else if (nodeid == HOST2ALL) {
		printf("all nodes: ");
	    }

	    else if (nodeid == HOST2OTB) {
		printf("all host nodes: ");
	    }

	    else if (nodeid == HOST2ITB) {
		printf("all compute nodes: ");
	    }

	    else {
		printf("0x%X: ", nodeid);
	    }

	    printf("%3.3f secs\n", roundtrip);
	}

	if (arg_delay)
	    sleep((unsigned) arg_delay);

    }

    cleanup();
}


static void
trap()
{
    interrupt = 1;
}


static int
cleanup()
{
    int    total_bytes;
    double avg;


    if (sent > 0) {

	/* final stats */
	total_bytes = arg_size * sent;
	avg         = tot / sent;

	printf(
	  "\n%d message%s, %d data byte%s (%4.3fK), %4.3f seconds (%4.3fK/sec)\n",
		sent, plural(sent),
		total_bytes, plural(total_bytes),
		(double) total_bytes / 1024.0,
		tot, ((double) (2 * total_bytes) / 1024.0) / tot
	);
	printf("roundtrip min/avg/max: %3.3f/%3.3f/%3.3f\n",
		tmin, avg, tmax
	);

	kexit(0);
    }
    else {
	fprintf(stderr, "tping: sent 0 messages!!\n");
	kexit(1);
    }
}


static int
checkargs()

{
	int		index;			/* index in nodeid list */
	int		flags;			/* associated nodeid flags */
/*
 * GDB
 *
 * Determine the count of echo messages.
 */
	if (opt_taken('c')) {

		if (intparam('c', &arg_count) == ERROR) {
			return(ERROR);
		}

		else if (arg_count < 0) {
			fprintf(stderr, "tping: count (%d) must be >= 0\n",
				arg_count);
			errno = EUSAGE;
			return(ERROR);
		}
	}

	else {
		arg_count = 0;
	}
/*
 * Determine the delay between messages.
 */
	if (opt_taken('d')) {

		if (intparam('d', &arg_delay) == ERROR) {
			return(ERROR);
		}

		else if (arg_delay < 0) {
			fprintf(stderr, "tping: delay (%d) must be >= 0\n",
				arg_delay);
			errno = EUSAGE;
			return(ERROR);
		}
	}

	else {
		arg_delay = 1;
	}
/*
 * Toggle the verbose flag.
 */
	if (opt_taken('v'))
		arg_verbose = FALSE;
	else
		arg_verbose = TRUE;
/*
 * Determine the echo message length.
 */
	if (opt_taken('l')) {

		if (intparam('l', &arg_size) == ERROR) {
			return(ERROR);
		}

		else if (arg_size < 0) {
			fprintf(stderr, "tping: length (%d) must be >= 0\n",
				arg_size);
			errno = EUSAGE;
			return(ERROR);
		}

		else if (arg_size > MAXNMSGLEN) {
			fprintf(stderr, "tping: using max length (%d)\n",
				MAXNMSGLEN);
			arg_size = MAXNMSGLEN;
		} 
	}

	else {
		arg_size = 1;
	}
/*
 * Determine destination nodeid.
 */
	getnode(&index, &nodeid, &flags);

	if (index == ERROR) {
		errno = EUSAGE;
		return(ERROR);
	}

	return(0);
/***GDB***/
}

 
/*
 *        help
 *
 *        Function:     - prints helpful information on the tping command
 */
 
static void
help()
{
        printf("\nSynopsis:       tping [options] <node>\n");
        printf("\nOptions:        -h	this help message\n");
        printf("\t\t-v      turn off verbose mode\n");
        printf("\t\t-c <#>  echo <#> messages\n");
        printf("\t\t-l <#>  use <#> bytes per message\n");
        printf("\t\t-d <#>  delay <#> seconds between messages\n");
        printf("\nDescription:    Echo a message to a Trillium node.\n");
        printf("\nNode:		n<constant>\n");
        printf("\t\t\t(any single compute node)\n");
        printf("\t\t\twhere <constant> = decimal or hexadecimal (0x)\n");
        printf("\t\th\t(local host node)\n");
        printf("\t\tC\t(all compute nodes)\n");
        printf("\t\tH\t(all host nodes)\n");
        printf("\t\tN\t(all nodes)\n");
        printf("\nExample:        tping n10 -c 100 -l 1024 -d 0\n");
        printf("\t\t\t\"Echo 100 messages of length 1024 to node 10\n");
        printf("\t\t\t without pausing between messages.\"\n");
}
