/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD,GDB
 *
 *	$Log:	trout.c,v $
 * Revision 2.2  92/04/30  14:16:16  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/04/27  11:48:27  gdburns
 * Change messages for new booting system.
 * 
 * Revision 2.1.1.2  91/11/26  16:49:22  raja
 * Use the new routing support functions.
 * 
 * Revision 2.1.1.1  91/05/29  16:53:48  raja
 * Free the dynamically allocated memory in trout_e().
 * Fixed the error message when a malloc() fails.
 * 
 * Revision 2.1  91/03/20  02:52:41  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/10  23:08:29  gdburns
 * Added recursive booting feature.
 * major overhaul
 * 
 * Revision 2.0.1.2  90/11/30  16:38:16  raja
 * Modified DEBUG to DBUG and PRETEND to fl_pretend as part of the 
 * overhaul of the topology layer tools.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:03  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:36  trillium
 * Homo erectus
 *
 *	Function:	- reads the route tables of the specified nodes
 *			  from the route file and sends them to the nodes
 *			- -e functionality added by GDB
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 *
 *      Accepts:        - options and nodeids (using the syntax of parse_nodes).
 */

#include <boot.h>
#include <bootreq.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <fcntl.h>
#include <portable.h>
#include <terror.h>
#include <rreq.h>
#include <net.h>
#include <routeio.h>
#include <debug.h>
#include <t_types.h>
#include <typical.h>

/*
 * global variables
 */
char		*me;			/* tool name */
int		fl_debug;		/* debug option flag */
int		fl_verbose;		/* verbose option flag */

/*
 * external variables
 */
extern int	errno;

/*
 * external functions
 */
extern void	followed();
extern void	prt_route();  
extern void	terror();
extern void	validopts();
extern char	*findpath();
extern char	*getparam();
extern char	*mnemonic();
extern int	rbtadd();
extern int	rsetnet();
extern int	opt_taken();
extern int	do_args();
extern int	addarg();
extern int4	getnode();
extern int	parse_nodes();
extern int	rroute();

/*
 * local functions
 */
static void	set_defaults();
static void	trout_help();
static void	trout_1();
static void	trout_e();
static void	trout_v();

/*
 * local variables
 */
static char	*rtfname;		/* route file name */
static char	*rtfullname;		/* full path route file name */
static char	**path_v;		/* route pathname prefixes */
static int	path_c;			/* # of pathname prefixes */

static char *usage = "usage: %s [-dehtvNV] [-p <routefile>] <nodeids>\n";

main(argc, argv)  

int			argc;
char			*argv[];

{
/*
 * Assign the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * command line option parsing
 */
	validopts("vdetNVh");
	followed("p");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, usage, me);
		} else {
			terror(me);
		}
		exit(errno);
	}
/*
 * Print help and exit.
 */
	if (opt_taken('h')) {
		trout_help();
		exit(0);
	}
/*
 * Set default variables.
 */
	set_defaults();
/* 
 * Locate the route file.
 */
	if ((rtfullname = findpath(rtfname, path_v, R_OK)) == NULL) {
		fprintf(stderr, "%s: cannot locate route file \"%s\"\n",
			me, rtfname);
		exit(1);
	}
	DBUG("%s: found the route file: %s\n", me, rtfullname);

	if (opt_taken('V')) {
		trout_v();
	} else {
		DBUG("%s: attach to kernel ...\n", me);

		if (kinit(0)) {
			terror("trout (kinit)");
			exit(errno);
		}

		if (opt_taken('e')) {
			trout_e();
		} else {
			trout_1();
		}

		kexit(0);
	}

	return(0);
}

/*
 *	trout_v
 *
 *	Function:	- pretty prints one or all route tables
 */	
static void
trout_v()

{
	RTIOFD		*routefd;	/* ptr route file descriptor */
	int4		nodeid;		/* current node ID */
	int		n_index;	/* current node index */
	int		n_flags;	/* current node attribute flags */
/*
 * Open the route file for reading.
 */
	DBUG("%s: opening the route file: %s ...\n", me, rtfullname);

	if ((routefd = rtio_open(rtfullname)) == (RTIOFD *) 0) {
		DBUG("%s: cannot rtio_open file: %s\n", me, rtfullname);
		exit(errno);
	}

	getnode(&n_index, &nodeid, &n_flags);

	if (n_index != ERROR) {
/*
 * Print table of each specified node.
 */
		do {                
			DBUG("%s: node = %d\n", me, nodeid);

			if (rtio_rtread(routefd, nodeid)) {
				rtio_close(routefd);
				exit(errno);
			}

			rtio_rtprint(routefd);

	    		getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}
/*
 * Print table of all nodes.
 */
	else {
		rtio_topnode(routefd);
		while ((nodeid = rtio_thisnode(routefd)) != NOTNODEID) {
			if (rtio_rtread(routefd, nodeid)) {
				rtio_close(routefd);
				exit(errno);
			}

			rtio_rtprint(routefd);

			rtio_nextnode(routefd);
		}
	}
/*
 * Close the route file.
 */
	DBUG("%s: closing the route file: %s ...\n", me, rtfullname);

	rtio_close(routefd);
}

/*
 *	trout_1
 *
 *	Function:	- sends route tables of command line nodes
 *			  to route daemons on command line nodes
 */
static void
trout_1()

{
	RTIOFD		*routefd;	/* ptr route file descriptor */
	struct net_info net;		/* basic network information */
	struct route	*p;		/* favorite pointer */
	int4		nodeid;		/* current node ID */
	int		n_index;	/* current node index */
	int		n_flags;	/* current node attribute flags */

	DBUG("%s: opening the route file: %s ...\n", me, rtfullname);
	if ((routefd = rtio_open(rtfullname)) == (RTIOFD *) 0) {
		DBUG("%s: cannot rtio_open file: %s\n", me, rtfullname);
		exit(errno);
	}

	getnode(&n_index, &nodeid, &n_flags);

	if (n_index == ERROR) {
		fprintf(stderr, usage, me);
		kexit(1);
	}

	net.ni_origin = (routefd->rd_head).rh_rootnode;
	DBUG("%s: origin %d\n", me, net.ni_origin);
/*
 * Trout all command line nodes.
 */
	do {                
		DBUG("%s: node = %d\n", me, nodeid);
/*
 * Read the route table for the current node.
 */
		if (rtio_rtread(routefd, nodeid)) {
			rtio_close(routefd);
			kexit(errno);
		}
/*
 * Get the remaining basic network information
 * and pass it along to the destination router.
 */
		net.ni_nodeid = nodeid;
		p = (struct route *) ahs_find(&(routefd->rd_route), nodeid);
		if (p == (struct route *) 0) {
			fprintf(stderr, "%s: node %d not in route tables\n",
				me, nodeid);
			kexit(EBADNODE);
		}

		net.ni_nodetype = p->r_nodetype;
		DBUG("%s: node %d has type %x\n", me, nodeid, net.ni_nodetype);
/*
 * If the destination node is the origin node, set the destination to
 * LOCAL, since the local origin router has no information or tables.
 */
		nodeid = (nodeid == net.ni_origin) ? LOCAL : nodeid;

		if (! opt_taken('N')) {
/*
 * Ship the basic network information to the destination router.
 */
			VERBOSE("node ID %s ...\n", mnemonic(nodeid));

			if (rsetnet(nodeid, &net)) {
				terror("trout (rsetnet)");
				kexit(errno);
			}
/*
 * Ship the route table to the destination router if not -t.
 */
			if (! opt_taken('t')) {
				VERBOSE("route table %s ...\n",
					mnemonic(nodeid));

				if (rroute(nodeid, routefd->rd_rtable,
					(routefd->rd_head).rh_nentries)) {

					terror("trout (rroute)");
					kexit(errno);
				}
			}
		}

		getnode(&n_index, &nodeid, &n_flags);
	} while (n_index);

	DBUG("%s: closing the route file: %s ...\n", me, rtfullname);

	rtio_close(routefd);
}

/*
 *	trout_e
 *
 *	Function:	- sends entire route file to boot daemon on
 *			  each command line node
 */
static void
trout_e()

{
	int		routefd;	/* open route file desc. */
	struct route    *proute;	/* route table ptr */
	struct rtiohead	*phead;		/* ptr to route file header */
	struct rtiodir	*pdir;		/* ptr to route file directory */
	int4		nodeid;		/* current node ID */
	int4		i;		/* favorite counter */
	int		n_index;	/* current node index */
	int		n_flags;	/* current node attribute flags */
	char		*pfile;		/* file buffer ptr */
	struct stat	status;		/* file status */
/*
 * Get first node on command line.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index == ERROR) {
		fprintf(stderr, usage, me);
		kexit(1);
	}
/*
 * Open route file.
 */
	DBUG("%s: opening the route file: %s ...\n", me, rtfullname);
	if ((routefd = open(rtfullname, O_RDONLY)) == ERROR) {
		terror("trout (open)");
		kexit(errno);
	}
/*
 * Get status on the file.
 */
	if (fstat(routefd, &status)) {
		terror("trout (fstat)");
		kexit(errno);
	}
/*
 * Malloc space for the file, OK with limited node memory.
 */
	if ((pfile = malloc((unsigned) status.st_size)) == NULL) {
		terror("trout (malloc)");
		kexit(errno);
	}
/*
 * Read in the whole file.
 */
	if (read(routefd, pfile, (int4) status.st_size) == ERROR) {
		terror("trout (read)");
		kexit(errno);
	}
/*
 * Close up the file descriptor.
 */
	if (close(routefd) == ERROR) {
		terror("trout (close)");
		kexit(errno);
	}			
/*
 * Convert entire route file to trillium data format.
 */
	phead = (struct rtiohead *) pfile;
	pdir = (struct rtiodir *) (pfile + sizeof(struct rtiohead));

	for (i = 0; i < phead->rh_nentries; ++i) {

		if (pdir[i].rd_nodeid != NOTNODEID) {
			proute = (struct route *) (pfile + pdir[i].rd_rtoffset);

			ltotcpy((int4 *) proute, (int4 *) proute,
				phead->rh_nentries *
				(sizeof(struct route) / sizeof(int4)) );
		}

		pdir[i].rd_nodeid = ltot(pdir[i].rd_nodeid);
		pdir[i].rd_rtoffset = (int4) ltot((int4) pdir[i].rd_rtoffset);
		pdir[i].rd_bootnext = (struct rtiodir *)
			ltot((int4) pdir[i].rd_bootnext);
	}

	phead->rh_magicno = ltot(phead->rh_magicno);
	phead->rh_nentries = ltot(phead->rh_nentries);
	phead->rh_rootnode = ltot(phead->rh_rootnode);

	do {
/*
 * Ship the route file to the destination node.
 */
		VERBOSE("route file %s ...\n", mnemonic(nodeid));
		if (rflat(nodeid, pfile, status.st_size, status.st_size,
				(int4) getpid())) {
			terror("trout (rflat)");
			kexit(errno);
		}
/*
 * Inform bootd on the destination node of the route file tag.
 */
		DBUG("%s: informing bootd on %d ...\n", me, nodeid);
		if (rbtadd(nodeid, BTO_ROUTE, getpid())) {
			terror("trout (rbtadd)");
			kexit(errno);
		}

		getnode(&n_index, &nodeid, &n_flags);
	} while (n_index);

	free(pfile);
}

/*
 *	trout_help
 *
 *	Function:	- prints useful information on the trout command
 */
static void
trout_help()

{
	printf("\nSynopsis:\ttrout [options] [nodeids]\n");
	printf("\nDescription:\tShip route tables to specified nodes.\n");
	printf("\nOptions:\t-h\t\tPrint this message.\n");
	printf("\t\t-d\t\tPrint debug information.\n");
	printf("\t\t-e\t\tLoad entire route file.\n");
	printf("\t\t-t\t\tSkip route shipping step.\n");
	printf("\t\t-v\t\tTurn on verbose mode.\n");
	printf("\t\t-N\t\tPretend to trout.\n");
	printf("\t\t-V\t\tPretty print route tables.\n");
	printf("\t\t-p <file>\tUse route tables in <file>.\n");
	mnusage();
	printf("\nExample:        trout -v n10,0\n");
	printf("\t\t\t\"Ship route tables to nodes 10 and 0.\"\n\n");
}

/*
 *	set_defaults
 *
 *	Functions:	- assesses options and sets default names
 */
static void
set_defaults()

{
/*
 * Set the paths.
 */
	path_c = 0;
	path_v = (char **) 0;
	addarg(&path_c, &path_v, "");				/* absolute */
	addarg(&path_c, &path_v, "$TROLLIUSHOME/boot");		/* env */
	addarg(&path_c, &path_v, DEFPROUTE);			/* default */
/*
 * Set the files.
 */
	if (opt_taken('p')) {
		rtfname = getparam('p');
	} else {
		rtfname = DEFFROUTE;
	}
/*
 * Set the flags.
 */
	fl_debug = opt_taken('d');
	fl_verbose = opt_taken('v');
}
