/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	routeio.c,v $
 * Revision 2.2  92/04/30  14:23:25  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/11/26  14:29:54  raja
 * New design/implementation including support for a second route table.
 * 
 * Revision 2.1  91/03/20  11:30:11  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/20  00:18:14  gdburns
 * Print node type in hexadecimal.
 * 
 * Revision 2.0.1.2  90/11/20  17:15:05  dixit-v
 * Changed DEBUG to DBUG.
 * 
 * Revision 2.0.1.1  90/07/13  19:30:44  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:30  trillium
 * Homo erectus
 * 
 *	Function :	- functions used to create and read route files
 *			- route file has 3 sections:
 *				- header: general topology information
 *				- directory: hash table for locating the
 *				  route table of a given node and a linked
 *				  list to navigate the nodes in boot order
 *				- route tables: one for each node, each being
 *				  a hash table of route entries (struct route)
 *			- the functions to create a route file are:
 *				rtio_init(): create route file descriptor
 *				rtio_header(): write route file header
 *				rtio_dir(): write route file directory
 *				rtio_rtinit(): initialize a route table
 *				rtio_rtwrite(): write a route table
 *				rtio_topnode(): go to start of node list
 *				rtio_thisnode(): node to write its route table
 *				rtio_cleanup(): release route file descriptor
 *			- the functions to read a route file are:
 *				rtio_open(): open an existing route file
 *				rtio_rtread(): read a route table
 *				rtio_rtprint(): pretty print a route table
 *				rtio_topnode(): go to start of node list
 *				rtio_thisnode(): current node in the list
 *				rtio_nextnode(): move to next node in the list
 *				rtio_close(): close route file
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <memory.h>
#include <malloc.h>
#include <portable.h>
#include <typical.h>
#include <routeio.h>
#include <rreq.h>
#include <net.h>
#include <events.h>
#include <nail.h>
#include <debug.h>
#include <all_hash.h>

/*
 * external variables
 */
extern char		*me;		/* tool name */
extern int		fl_debug;	/* debug mode flag */
extern int		fl_verbose;	/* verbose mode flag */
 
/*
 * external functions
 */
extern int4             next_prime();
 

/*
 *	rtio_init
 *
 *	Function: 	- create a route file descriptor
 *			- used only when creating a route file
 *
 *	Accepts:	- name of the route file
 *			- ptr to network description
 *			- extra route table entries
 *
 *	Returns:	- ptr route file descriptor or NULL
 */
RTIOFD *
rtio_init(filename, net, extra)

char			*filename;
struct bschema		*net;
int4			extra;

{
	RTIOFD		*rfd;		/* ptr to route file descriptor */
	struct t_node	*tnode;		/* ptr to node */
	struct route	rtelem;		/* route table entry */
	int4		i;		/* favorite counter */
	uint4		size;
/*
 * Check the trivial cases.
 */
	if (ah_count(net->bs_tnhd) <= INT4_NIL) {
		fprintf(stderr, "%s (rtio_init): network has no nodes\n", me);
		return((RTIOFD *) 0);
	}
/*
 * Allocate space for the route file descriptor and the filename.
 */
	rfd = (RTIOFD *) malloc((uint4) sizeof(RTIOFD) + strlen(filename) + 1);
	if (rfd == (RTIOFD *) 0) {
		fprintf(stderr, "%s (rtio_init): cannot malloc desc.\n", me);
		return((RTIOFD *) 0);
	}

	rfd->rd_filename = ((char *) rfd) + sizeof(RTIOFD);
	(void) strcpy(rfd->rd_filename, filename);
	rfd->rd_dirtable = (struct rtiodir *) 0;
	rfd->rd_dircur = (struct rtiodir *) 0;
	rfd->rd_rtable = (struct route *) 0;
	rfd->rd_net = net;

	rfd->rd_fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0644);
	if (rfd->rd_fd < 0) {
		fprintf(stderr, "%s (rtio_init): cannot open file \"%s\"\n",
			me, filename);
		free((char *) rfd);
		return((RTIOFD *) 0);
	}
/*
 * Set up the internal route table.
 */
	size = next_prime( ((ah_count(net->bs_tnhd) + extra) * 12) / 10 + 1);
	rfd->rd_rtable = (struct route *) malloc(size * sizeof(struct route));
	if (rfd->rd_rtable == (struct route *) 0) {
		fprintf(stderr,
			"%s (rtio_init): cannot malloc route table\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	if (ahs_init(size, (int4) sizeof(struct route), NOTNODEID,
		AHFIXED, (void *) rfd->rd_rtable, (int4 *) 0,
		&(rfd->rd_route)) == (SHASH *) 0) {

		fprintf(stderr,
			"%s (rtio_init): cannot create route table\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	if (rtio_rtinit(rfd)) {
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	return(rfd);
}

/*
 *	rtio_rtinit
 *
 *	Function:	- initialize the internal route table of the
 *			  given route file descriptor
 *			- route table is assumed to have been allocated
 *			- route hash table is assumed to be initialized
 *			- network information is assumed available
 *
 *	Accepts:	- ptr to route file descriptor
 *
 *	Returns:	- 0 or ERROR
 */
int
rtio_rtinit(rfd)

RTIOFD			*rfd;

{
	int4		size;		/* size of route table */
	int4		i;		/* favorite counter */
	struct route	rtelem;		/* route table entry */
	struct route	*p;		/* favorite pointer */
	struct t_node	*tnode;		/* ptr to TNODE information */
/*
 * Clean the route table array.
 */
	size = ahs_size(&(rfd->rd_route));
	for (i = 0, p = rfd->rd_rtable; i < size; i++, p++) {
		p->r_nodeid = NOTNODEID;
		p->r_event = NOTEVENT;
		p->r_event2 = NOTEVENT;
		p->r_nodetype = NOTNODETYPE;
	}
/*
 * Read the TNODE tuples and initialize the route table accordingly.
 */
	tnode = (struct t_node *) ah_top((rfd->rd_net)->bs_tnhd);
	while (tnode) {
		rtelem.r_nodeid = tnode->tn_nodeid;
		rtelem.r_nodetype = (tnode->tn_specs)->ns_type &
			(~ (NT_JONES | NT_BOOT));
		rtelem.r_event = NOTEVENT;
		rtelem.r_event2 = NOTEVENT;

		if (ahs_insert(&(rfd->rd_route), (void *) &rtelem)) {
			fprintf(stderr,
				"%s (rtio_rtinit): cannot hash node %d\n",
				me, tnode->tn_nodeid);
			return(ERROR);
		}
		
		tnode = (struct t_node *) ah_next((rfd->rd_net)->bs_tnhd,
			(void *) tnode);
	}

	return(0);
}

/*
 *	rtio_header
 *
 *	Function:	- fill and write the header of the route file
 *
 *	Accepts:	- ptr to route file descriptor
 *
 *	Returns:	- 0 or ERROR
 */
int
rtio_header(rfd)

RTIOFD			*rfd;

{
/*
 * Fill the header of the route file and write it into the file.
 */
	(rfd->rd_head).rh_magicno = RTIO_MAGIC;
	(rfd->rd_head).rh_nentries = ahs_size(&(rfd->rd_route));
	(rfd->rd_head).rh_rootnode = ((rfd->rd_net)->bs_hostnode)->tn_nodeid;

	DBUG("route file header: magicno: %d\tsize:%d\troot node:%d\n",
		(rfd->rd_head).rh_magicno, (rfd->rd_head).rh_nentries,
		(rfd->rd_head).rh_rootnode);

	if (write(rfd->rd_fd, (char *) &(rfd->rd_head),
		sizeof(struct rtiohead)) < sizeof(struct rtiohead)) {

		fprintf(stderr, "%s (rtio_header): cannot write header\n", me);
		return(ERROR);
	}

	return(0);
}

/*
 *	rtio_dir
 *
 *	Function:	- fill and write the directory of the route file
 *
 *	Accepts:	- ptr to route file descriptor
 *
 *	Returns:	- 0 or ERROR
 */
int 
rtio_dir(rfd)

RTIOFD			*rfd;

{
	struct t_node	*tnode;		/* ptr to node */
	struct rtiodir	*pdir;		/* ptr to directory entry */
	struct route	*proute;	/* ptr to route table entry */
	uint4		rtoffset;	/* table offset in route file */
	uint4		rtincr;		/* offset increment */
	uint4		rtlen;		/* length of route table */
	int4		i;		/* favorite counter */
/*
 * Allocate the route file directory and set its hash table descriptor.
 */
	rtlen = ahs_size(&(rfd->rd_route));
	if ((rfd->rd_dirtable = (struct rtiodir *) malloc(rtlen *
		sizeof(struct rtiodir))) == (struct rtiodir *) 0) {

		fprintf(stderr, "%s (rtio_dir): cannot malloc directory\n", me);
		return(ERROR);
	}

	if (ahs_init(rtlen, (int4) sizeof(struct rtiodir), NOTNODEID,
		AHFIXED, (void *) rfd->rd_dirtable, (int4 *) 0,
		&(rfd->rd_dir)) == (SHASH *) 0) {

		fprintf(stderr, "%s (rtio_dir): cannot create directory\n", me);
		return(ERROR);
	}
/*
 * Fill the directory route table offsets.
 */
	DBUG("route file directory:\nnodeid\trtable offset\n");
	rtoffset = sizeof(struct rtiohead) + (rtlen * sizeof(struct rtiodir));
	rtincr = rtlen * sizeof(struct route);

	pdir = rfd->rd_dirtable;
	proute = rfd->rd_rtable;
	for (i = 0; i < rtlen; i++, pdir++, proute++) {
		pdir->rd_nodeid = proute->r_nodeid;
		if (pdir->rd_nodeid == NOTNODEID) {
			pdir->rd_rtoffset = -1;
			pdir->rd_bootnext = (struct rtiodir *) 0;
		}
		else {
			pdir->rd_rtoffset = rtoffset;
			rtoffset += rtincr;
			DBUG("%d\t%d\n", pdir->rd_nodeid, pdir->rd_rtoffset);
/*
 * Set the pointer to the first node whose route table is expected.
 */
			if (rfd->rd_dircur == (struct rtiodir *) 0) {
				rfd->rd_dircur = pdir;
			}
		}
	}
/*
 * Fill the directory bootnext pointers from the booting order.
 */
	pdir = (struct rtiodir *) ahs_top(&(rfd->rd_dir));
	while (pdir) {
/* 
 * Locate the nodeid in the network description.
 */
		if ((tnode = (struct t_node *) ah_find((rfd->rd_net)->bs_tnhd,
			pdir->rd_nodeid)) == (struct t_node *) 0) {

			fprintf(stderr, "%s (rtio_dir): node %d not in net.\n",
				me, pdir->rd_nodeid);
			return(ERROR);
		}

		tnode = (tnode->tn_specs)->ns_bootnext;
		if (tnode == (struct t_node *) 0) {
/*
 * Last node in the booting order.
 */
			pdir->rd_bootnext = (struct rtiodir *) 0;
		}
		else {
/*
 * Locate the next bootable node in the directory.
 */
			if ((pdir->rd_bootnext = (struct rtiodir *)
				ahs_find(&(rfd->rd_dir), tnode->tn_nodeid))
				== (struct rtiodir *) 0) {

				fprintf(stderr,
					"%s (rtio_dir): node %d not in dir.\n",
					me, tnode->tn_nodeid);
				return(ERROR);
			}
		}

		pdir = (struct rtiodir *)ahs_next(&(rfd->rd_dir), (void *)pdir);
	}
/*
 * Write the directory in the route file.
 */
	i = rtlen * sizeof(struct rtiodir);
	if (write(rfd->rd_fd, (char *) rfd->rd_dirtable, i) < i) {
		fprintf(stderr, "%s (rtio_dir): cannot write directory\n", me);
		return(ERROR);
	}

	return(0);
}

/*
 *	rtio_thisnode
 *
 *	Function:	- get nodeid of current node in the list
 *			- used to write route tables in the required order
 *			- used to read route tables by cycling all nodes
 *			- the nodeid returned changes after a successful call
 *			  to rtio_rtwrite() writing the required route table
 *			  or after a call to rtio_nextnode() when reading the
 *			  route tables of all nodes in the route file
 *
 *	Accepts:	- ptr to route file descriptor
 *
 *	Returns:	- valid nodeid or NOTNODEID
 */
int4
rtio_thisnode(rfd)

RTIOFD			*rfd;

{
	if (rfd->rd_dircur == (struct rtiodir *) 0) {
		return(NOTNODEID);
	}
	else {
		return((rfd->rd_dircur)->rd_nodeid);
	}
}

/*
 *	rtio_topnode
 *
 *	Function:	- initialize the list of nodes to the first
 *			  valid node in the directory
 *
 *	Accepts:	- ptr to route file descriptor
 */
void
rtio_topnode(rfd)

RTIOFD			*rfd;

{
	rfd->rd_dircur = (struct rtiodir *) ahs_top(&(rfd->rd_dir));
}

/*
 *	rtio_nextnode
 *
 *	Function:	- move to the next nodeid in the directory
 *			- used when reading the route tables of all nodes
 *			  in order to cycle through them without requiring
 *			  prior knowledge of the nodeids
 *
 *	Accepts:	- ptr to route file descriptor
 */
void
rtio_nextnode(rfd)

RTIOFD			*rfd;

{
	if (rfd->rd_dircur == (struct rtiodir *) 0) {
/*
 * Do nothing, this is a sink state.
 */
		return;
	}
	else {
/*
 * Move to the next valid directory entry.
 */
		rfd->rd_dircur = (struct rtiodir *) ahs_next(&(rfd->rd_dir),
			(void *) rfd->rd_dircur);
	}
}

/*
 *	rtio_rtwrite
 *
 *	Function:	- write the route table of the given nodeid
 *			- check if the required node order is violated
 *			- update the node order if successful
 *			- the events in the route table are modified
 *
 *	Accepts:	- ptr to route file descriptor
 *			- target nodeid
 *
 *	Returns:	- 0 or ERROR
 */
int
rtio_rtwrite(rfd, nodeid)

RTIOFD			*rfd;
int4			nodeid;

{
	int4		i;		/* favorite counter */
	int4		rtlen;		/* route table length */
	struct route	*p;		/* favorite pointer */
/*
 * Check if a node is expected and if this is the one required.
 * Route tables have to arrive in the required order.
 */
	if ((i = rtio_thisnode(rfd)) == NOTNODEID) {
		fprintf(stderr, "%s (rtio_rtwrite): no more writing\n", me);
		return(ERROR);
	}
	else if (nodeid != i) {
		fprintf(stderr,
			"%s (rtio_rtwrite): wrong order: want n%d not n%d\n",
			me, i, nodeid);
		return(ERROR);
	}
/*
 * Exchange the link numbers stored in the r_event field to 
 * the correct Trollius datalink event number.
 */
	rtlen = ahs_size(&(rfd->rd_route));
	for (i = 0, p = rfd->rd_rtable; i < rtlen; i++, p++) {
		if (p->r_event != NOTEVENT) {
			p->r_event = (p->r_event << 2) + EVDL0;
		}
		if (p->r_event2 != NOTEVENT) {
			p->r_event2 = (p->r_event2 << 2) + EVDL0;
		}
	}
/*
 * Write the route table.
 */
	rfd->rd_rtnode = nodeid;
	i = rtlen * sizeof(struct route);
	if (write(rfd->rd_fd, (char *) rfd->rd_rtable, i) < i) {
		fprintf(stderr, "%s (rtio_rtwrite): cannot write route table\n",
			me);
		return(ERROR);
	}
/*
 * Update the node order for the next call to rtio_thisnode().
 */
	rtio_nextnode(rfd);

	return(0);
}

/*
 *	rtio_rtprint
 *
 *	Function:	- print the internal route table
 *
 *	Accepts:	- prt to route file descriptor
 */
void
rtio_rtprint(rfd)

RTIOFD			*rfd;

{
	struct route	*p;		/* favorite pointer */

	printf("\n       Route Table for Node %d  (NOTEVENT = %8x)\n\n",
		rfd->rd_rtnode, NOTEVENT);
	printf("       Node(dec)   Event(hex)   Event2(hex)   Nodetype(hex)\n");

	p = (struct route *) ahs_top(&(rfd->rd_route));
	while (p) {
		printf(" %12d       %8x     %8x       %8x\n",
			p->r_nodeid, p->r_event, p->r_event2, p->r_nodetype);

		p = (struct route *) ahs_next(&(rfd->rd_route), (void *) p);
	}

	printf("\n");
}

/*
 *	rtio_cleanup
 *
 *	Function:	- free the allocated route file descriptor
 *
 *	Accepts:	- ptr to route file descriptor
 */
void
rtio_cleanup(rfd)

RTIOFD			*rfd;

{
	if (rfd) {
/*
 * Free the directory array.
 */
		if (rfd->rd_dirtable) {
			free((char *) rfd->rd_dirtable);
		}
/*
 * Free the internal route table.
 */
		if (rfd->rd_rtable) {
			free((char *) rfd->rd_rtable);
		}
/*
 * Close the open route file.
 */
		(void) close(rfd->rd_fd);
/*
 * Free the route file descriptor.
 */
		free((char *) rfd);
	}
}

/*
 *	rtio_open
 *
 *	Function:	- open an existing route file for reading
 *
 *	Accepts:	- route file name
 *
 *	Returns:	- ptr to route file descriptor or NULL
 */
RTIOFD *
rtio_open(filename)

char			*filename;

{
	RTIOFD		*rfd;			/* route file descriptor */
	int		sysfd;			/* system file descriptor */
	int4		size;
/*
 * Open the route file.  That also gets us past checking the trivial case.
 */
	if ((sysfd = open(filename, O_RDONLY, 0)) < 0) {
		fprintf(stderr, "%s (rtio_open): cannot open \"%s\"\n",
			me, filename);
		return((RTIOFD *) 0);
	}
/*
 * Allocate and initialize the route file descriptor.
 */
	rfd = (RTIOFD *) malloc((uint4) sizeof(RTIOFD) + strlen(filename) + 1);
	if (rfd == (RTIOFD *) 0) {
		fprintf(stderr, "%s (rtio_open): cannot malloc desc.\n", me);
		(void) close(sysfd);
		return((RTIOFD *) 0);
	}

	rfd->rd_filename = ((char *) rfd) + sizeof(RTIOFD);
	(void) strcpy(rfd->rd_filename, filename);
	rfd->rd_fd = sysfd;
	rfd->rd_dirtable = (struct rtiodir *) 0;
	rfd->rd_dircur = (struct rtiodir *) 0;
	rfd->rd_rtable = (struct route *) 0;
	rfd->rd_net = (struct bschema *) 0;
/*
 * Read the route file header.
 */
	if (read(rfd->rd_fd, (char *) &(rfd->rd_head),
		sizeof(struct rtiohead)) < sizeof(struct rtiohead)) {

		fprintf(stderr, "%s (rtio_open): cannot read header\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}
/*
 * Check the file magic number and extract info from the header.
 */
	if ((rfd->rd_head).rh_magicno != RTIO_MAGIC) {

		if ((rfd->rd_head).rh_magicno == OLDMAGIC) {
			fprintf(stderr,
				"%s: \"%s\" has old route file format\n",
				me, filename);
		}
		else {
			fprintf(stderr, "%s: \"%s\" has incorrect file type\n",
				me, filename);
		}

		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	DBUG("%s: magicno: %d\tnentries: %d\troot node: %d\n", me,
		(rfd->rd_head).rh_magicno, (rfd->rd_head).rh_nentries,
		(rfd->rd_head).rh_rootnode);
/*
 * Allocate the route file directory and setup its hash
 * table descriptor and read it from the file.
 */
	size = (rfd->rd_head).rh_nentries * sizeof(struct rtiodir);
	if ((rfd->rd_dirtable = (struct rtiodir *) malloc((uint4) size))
		== (struct rtiodir *) 0) {

		fprintf(stderr,
			"%s (rtio_open): cannot malloc directory\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	if (ahs_init((rfd->rd_head).rh_nentries, (int4) sizeof(struct rtiodir),
		NOTNODEID, AHFIXED | AHNOINIT, (void *) rfd->rd_dirtable,
		(int4 *) 0, &(rfd->rd_dir)) == (SHASH *) 0) {

		fprintf(stderr, "%s (rtio_open): cannot create dir.\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	if (read(rfd->rd_fd, (char *) rfd->rd_dirtable, size) < size) {
		fprintf(stderr, "%s (rtio_open): cannot read directory\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}
/*
 * Allocate the internal route table and setup its hash table descriptor.
 */
	size = (rfd->rd_head).rh_nentries * sizeof(struct route);
	rfd->rd_rtable = (struct route *) malloc((uint4) size);
	if (rfd->rd_rtable == (struct route *) 0) {
		fprintf(stderr,
			"%s (rtio_open): cannot malloc route table\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	if (ahs_init((rfd->rd_head).rh_nentries, (int4) sizeof(struct route),
		NOTNODEID, AHFIXED | AHNOINIT, (void *) rfd->rd_rtable,
		(int4 *) 0, &(rfd->rd_route)) == (SHASH *) 0) {

		fprintf(stderr,
			"%s (rtio_open): cannot create route table\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}
/*
 * Read in the first route table in the file and store it in the internal
 * route table.  It is going to be used for getting the node types.
 */
	rfd->rd_dircur = (struct rtiodir *) ahs_top(&(rfd->rd_dir));
	if (rfd->rd_dircur == (struct rtiodir *) 0) {
		fprintf(stderr, "%s (rtio_open): no nodes in directory\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}
 
	if (lseek(rfd->rd_fd, (rfd->rd_dircur)->rd_rtoffset, SEEK_SET) < 0) {
		fprintf(stderr,
			"%s (rtio_open): cannot locate route table\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	if (read(rfd->rd_fd, (char *) rfd->rd_rtable, size) < size) {
		fprintf(stderr,
			"%s (rtio_open): cannot read route table\n", me);
		rtio_cleanup(rfd);
		return((RTIOFD *) 0);
	}

	return(rfd);
}

/*
 *	rtio_rtread
 *
 *	Function:	- read the route table of a given nodeid
 *			  into the internal route table space
 *
 *	Accepts:	- ptr to route file descriptor
 *			- target nodeid
 *
 *	Returns:	- 0 or ERROR
 */
int
rtio_rtread(rfd, nodeid)

RTIOFD			*rfd;
int4			nodeid;

{
	struct rtiodir	*p;		/* favorite pointer */
	int4	 	size;
/*
 * Locate and read the route table.
 */
	p = (struct rtiodir *) ahs_find(&(rfd->rd_dir), nodeid);
	if (p == (struct rtiodir *) 0) {
		fprintf(stderr, "%s (rtio_rtread): cannot locate node %d\n",
			me, nodeid);
		return(ERROR);
	}

	if (lseek(rfd->rd_fd, p->rd_rtoffset, SEEK_SET) < 0) {
		fprintf(stderr, "%s (rtio_rtread): cannot find node %d route\n",
			me, nodeid);
		return(ERROR);
	}

	size = (rfd->rd_head).rh_nentries * sizeof(struct route);
	if (read(rfd->rd_fd, (char *) rfd->rd_rtable, size) < size) {
		fprintf(stderr, "%s (rtio_rtread): cannot read node %d route\n",
			me, nodeid);
		return(ERROR);
	}

	rfd->rd_rtnode = nodeid;
	return(0);
}
