/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	bufferd.c,v $
 * Revision 2.2  92/04/30  14:26:46  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  12:14:40  gdburns
 * Add buffer flow control of user messages.
 * 
 * Revision 2.1  91/03/20  11:31:36  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.4  90/07/13  19:31:09  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.3  90/06/11  23:51:31  gdburns
 * Calculation of total used space includes size of buffer descriptors.
 * 
 * Revision 2.0.1.2  90/05/07  00:24:55  raja
 * Added the handling of packetized messages in bfprobe.  If the last
 * packet is not in the buffer returns with no message found.
 * 
 * Revision 2.0.1.1  90/05/06  15:29:42  raja
 * Added the bfprobe service.
 * 
 * Revision 2.0  90/02/20  11:56:47  trillium
 * Homo erectus
 * 
 *	Function:	- buffer daemon
 *			- OTB/ITB portable module
 *			- Bufferd receives all messages to be buffered
 *			  as well as requests to tune buffering characteristics.
 *			- datalink flow control feature disabled by
 *			  maxusr > (maxspace + MAXNMSGLEN)
 *
 *	Usage:		- bufferd [<maxworkers> [<maxspace> [<maxusr>]]]
 */

#include <bufreq.h>
#include <dlreq.h>
#include <events.h>
#include <priority.h>
#include <shared.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern struct buffer	*init();
extern struct buffer	*find();
extern void		insert();
extern void		delete();
extern int		create();
extern char		*malloc();

/*
 * local variables
 */
static int		bjones;			/* neighbours known? */
static int		dlstatus;		/* neighbour dl valve status */
static int		filter;			/* filter bufferd requests */
static int		maxspace;		/* max amount of buffer space */
static int		maxusr;			/* max usr msg space */
static int		maxworkers;		/* max # of buffer processes */
static int		nworkers;		/* # of buffer processes */
static int		nspace;			/* amount of buffer space */
static int		nfull;			/* # of full buffer procs */
static int		njones;			/* # of neighbour nodes */
static int		pjones[MAXNLINK];	/* neighbour node IDs */

static struct buffer	*free_top;		/* top of free list */
static struct buffer	*pend_top;		/* top of pending list */
static struct buffer	*wait_top;		/* top of waiting list */
static struct buffer	*orig_top;		/* original top of free list */

static struct bufreq	*req;			/* bufferd request */
static struct bufreply	*reply;			/* bufferd reply */

static struct nmsg	nhead;			/* central msg descriptor */

/*
 * global variables
 */
char			msg[MAXNMSGLEN];	/* incoming message */

/*
 * local functions
 */
static void		do_buffer();
static void		chkspace();
static void		putmsg();
static void		valve();


main(argc, argv)

int			argc;
char			**argv;

{
	int		ret;			/* return code */
/*
 * Initialize.
 */
	if (kinit(0) < 0) {
		panic(errno);
	}

	if (kattach(PRBUFFERD) < 0) {
		panic(errno);
	}

	if (rpattach(LOCAL) && (errno != ESERVICE)) {
		panic(errno);
	}
/*
 * Consider runtime parameters.
 */
	maxworkers = (argc > 1) ? stoi(argv[1]) : BFDEFWORKERS;
	maxspace = (argc > 2) ? stoi(argv[2]) : BFDEFSPACE;
	maxusr = (argc > 3) ? stoi(argv[3]) : BFDEFUSR;

	filter = BFREQ;
	dlstatus = DLON;
	bjones = FALSE;

	if (maxspace >= MAXNMSGLEN) {
		filter |= (BFSYS | BFUSR | BFDLI);
	}
/*
 * Allocate initial batch of buffer descriptors.
 */
	free_top = init();
	orig_top = free_top;
	pend_top = NULL;
	wait_top = NULL;
	nfull = 0;
	nworkers = 0;
	nspace = 0;

	req = (struct bufreq *) nhead.nh_data;
	reply = (struct bufreply *) nhead.nh_data;
/*
 * Continuously receive and process messages.
 */
	for (;;) {
		nhead.nh_event = EVBUFFERD;
		nhead.nh_type = filter;
		nhead.nh_flags = DFORWARD;
		nhead.nh_length = MAXNMSGLEN;
		nhead.nh_msg = msg;

		if (drecv(&nhead) < 0) {
			panic(errno);
		}

		nhead.nh_flags &= ~KREVPORT;
/*
 * Do we have a bufferd request to process?
 */
		if (nhead.nh_event == EVBUFFERD) {

			switch(req->bfq_req) {

				case BFWORKER:
				ret = do_worker();
				break;

				case BFSTATE:
				ret = do_state();
				break;

				case BFSWEEP:
				ret = do_sweep();
				break;

				case BFSPACE:
				maxspace = req->bfq_maxspace;
				ret = 0;
				break;

				case BFREADY:
				ret = do_ready();
				break;

				case BFPROBE:
				ret = do_probe();
				break;

				case BFUSRSPACE:
				maxusr = req->bfq_maxusr;
				ret = 0;
				break;

				default:
				ret = ERROR;
				errno = EBADREQUEST;
				break;
			}
/*
 * At this time, we are not sending replies to simple buffer control
 * commands.
 */
		} else {
			do_buffer();
		}

		chkspace();
	}
}

/*
 *	do_buffer
 *
 *	Function:	- buffers incoming message
 *			- the bread and butter of bufferd
 */

static void
do_buffer()

{
	struct buffer	*new;		/* incoming buffer descriptor */
	struct buffer	*p;		/* favourite ptr */
/*
 * Obtain a free buffer descriptor.
 */
	if (! free_top) {
		free_top = (struct buffer *)
			malloc((unsigned) sizeof(struct buffer));
		free_top->bf_next = NULL;

		if (! free_top) {
			panic(errno);
		}

		nspace += sizeof(struct buffer);
	}

	new = free_top;
	delete(&free_top, free_top);
/*
 * Initialize a new buffer descriptor.
 */
	bcopy((char *) &nhead, (char *) &new->bf_head, sizeof(struct nmsg));
	new->bf_head.nh_flags |= BYBUF;
/*
 * Search the pending list for a similar message.  We want to keep
 * buffer workers filled with distinct messages.
 */
	p = find(pend_top, new);
/*
 * Are we going to forward to a buffer worker?
 */
	if (!p && (nfull < maxworkers)) {

		insert(&pend_top, new);
/*
 * Do we need to create a new buffer worker?
 */
		if (nfull == nworkers) {

			if (create() < 0) {
				panic(errno);
			}

			++nworkers;
		}
/*
 * Forward message to buffer worker.
 */
		putmsg(&new->bf_head);
/*
 *		new->bf_head.nh_msg = (char *) 0;
 */
		++nfull;
	}

	else {

		if (nhead.nh_length) {
			new->bf_head.nh_msg =
				malloc((unsigned) nhead.nh_length);

			if (! new->bf_head.nh_msg) {
				panic(errno);
			}

			bcopy(nhead.nh_msg, new->bf_head.nh_msg,
				nhead.nh_length);
			nspace += nhead.nh_length;
		}

		insert(&wait_top, new);
	}
}

/*
 *	do_ready
 *
 *	Function:	- A buffer process is reporting that it
 *			  has passed on its message.  What do I do???
 *	Returns:	- 0
 */

static int
do_ready()

{
	struct buffer	*p;		/* favourite pointer */
	struct buffer	key;		/* search descriptor */
/*
 * Locate the buffer descriptor in the pending list.
 */
	key.bf_head.nh_dl_event = req->bfq_dl_event;
	key.bf_head.nh_type = req->bfq_type;
	key.bf_head.nh_event = req->bfq_event;
	p = find(pend_top, &key);

	if (!p) {
		errno = EIMPOSSIBLE;
		panic(errno);
	}
/*
 * Delete the no longer pending buffer descriptor.
 */
	--nfull;
	delete(&pend_top, p);
/*
 * Release descriptor if it is not within the original block.
 */
	if ((p >= orig_top) && (p < (orig_top + BFMAX))) {
		insert(&free_top, p);
	} else {
		(void) free((char *) p);
		nspace -= sizeof(struct buffer);
	}
/*
 * Are we supposed to take this opportunity to pare down the number
 * of buffer worker processes?
 */
	if (nworkers > maxworkers) {
		nhead.nh_event = EVBUFFER;
		putmsg(&nhead);
		--nworkers;
	} else {
/*
 * Look for a buffer descriptor in the waiting list that is distinct
 * from all others in the pending list.
 */
		for (p = wait_top; p && find(pend_top, p); p = p->bf_next) {
		}

		if (p) {
			delete(&wait_top, p);
			insert(&pend_top, p);
			++nfull;

			putmsg(&p->bf_head);

			if (p->bf_head.nh_length) {
				(void) free(p->bf_head.nh_msg);
				p->bf_head.nh_msg = msg;
				nspace -= p->bf_head.nh_length;
			}
		}
	}

	return(0);
}

/*
 *	do_worker
 *
 *	Function:	- Set the limit on the number of buffer workers.
 *			- If new limit falls below current number,
 *			  attempt to extinguish empty buffer workers.
 *			- If new limit falls above current number,
 *			  and all current buffers are full,
 *			  attempt to fill new buffer workers.
 *
 *	Returns:	- 0
 */

static int
do_worker()

{
	struct buffer	*p;		/* favourite ptr */
	struct buffer	*pnext;		/* next desc on list */
	int		i;		/* favourite counter */

	maxworkers = req->bfq_maxworkers;
/*
 * Shall we try to reduce the number of active workers?
 */
	if ((nworkers > maxworkers) && (nfull < nworkers)) {

		for (i = 0; i < (nworkers - nfull); ++i) {
			nhead.nh_event = EVBUFFER;
			putmsg(&nhead);
		}

		nworkers -= i;
	}
/*
 * Shall we try to increase the number of active workers?
 */
	if ((nworkers < maxworkers) && (nfull == nworkers) && wait_top) {
		p = wait_top;

		while (p && (nworkers < maxworkers)) {
			pnext = p->bf_next;
/*
 * Only move distinct event/type combinations into the pending list.
 */
			if (! find(pend_top, p)) {

				if (create() == ERROR) {
					panic(errno);
				}

				++nworkers;
				delete(&wait_top, p);
				insert(&pend_top, p);
				putmsg(&p->bf_head);
				++nfull;

				if (p->bf_head.nh_length) {
					(void) free(p->bf_head.nh_msg);
					p->bf_head.nh_msg = msg;
					nspace -= p->bf_head.nh_length;
				}
			}

			p = pnext;
		}
	}

	return(0);
}

/*
 *	do_sweep
 *
 *	Function:	- empties all full buffer workers
 *	Returns:	- 0
 */

static int
do_sweep()

{
	struct buffer	*p;			/* favourite ptr */
	int		evtmp;			/* temporary for nh_event */

	for (p = pend_top; p; p = p->bf_next) {
/*
 * Try to suck back the message.  If we don't get it, the buffer worker
 * is already empty.
 */
		evtmp = p->bf_head.nh_event;
		p->bf_head.nh_event = p->bf_head.nh_dl_event;

		if (dtry_recv(&p->bf_head)) {

			if (errno != EWOULDBLOCK) {
				panic(errno);
			}
		}

		p->bf_head.nh_event = evtmp;
	}
/*
 * Blow away the wait list.
 */
	while (p = wait_top) {

		if (p->bf_head.nh_length) {
			(void) free(p->bf_head.nh_msg);
			nspace -= p->bf_head.nh_length;
		}

		delete(&wait_top, p);
/*
 * Release descriptor if it is not within the original block.
 */
		if ((p >= orig_top) && (p < (orig_top + BFMAX))) {
			insert(&free_top, p);
		} else {
			(void) free((char *) p);
			nspace -= sizeof(struct buffer);
		}
	}

	return(0);
}

/*
 *	do_state
 *
 *	Function:	- returns state information on all buffer space
 *			- links the pending list to the waiting list
 *			  with a "spacer" descriptor in between
 *			- spacer signals to the client the boundary between
 *			  pending and waiting buffered messages
 *			- removes spacer and restores lists to original state
 *			- if max workers are full, replies are sent
 *			  directly, risking message order integrity of
 *			  the network system, but returning vital debug info
 *	Returns:	- 0 
 */

static int
do_state()

{
	struct buffer	spacer;		/* null descriptor spacer */
	struct buffer	*p;		/* favourite ptr */
	struct buffer	*ptop;		/* remember list top */
	struct bfstate	*pbs;		/* ptr to buffer state */
	int		maxstate;	/* max state descs in one message */
	int		i;		/* favourite index */
	int		bbuffer;	/* buffer repies? */

	bbuffer = nfull < maxworkers;
	pbs = (struct bfstate *) msg;
	maxstate = sizeof(msg) / sizeof(struct bfstate);
	spacer.bf_head.nh_node = NOTNODEID;
	spacer.bf_head.nh_dl_event = NOTEVENT;
	spacer.bf_head.nh_type = 0;
	spacer.bf_head.nh_length = 0;
	spacer.bf_next = NULL;

	if (maxstate <= 0) {
		errno = EIMPOSSIBLE;
		return(ERROR);
	}

	nhead.nh_dl_event = getroute(req->bfq_src_node, req->bfq_src_event);
	nhead.nh_node = req->bfq_src_node;
	nhead.nh_event = req->bfq_src_event;
	nhead.nh_type = 0;
	nhead.nh_flags = bbuffer ? 0 : BYBUF;
	nhead.nh_msg = msg;
	nhead.nh_length = sizeof(struct bfstate) * maxstate;
	reply->bfr_reply = 0;
	reply->bfr_maxworkers = maxworkers;
	reply->bfr_maxspace = maxspace;
	reply->bfr_nworkers = nworkers;
	reply->bfr_nspace = nspace;
	reply->bfr_nfull = nfull;
	reply->bfr_maxusr = maxusr;
/*
 * Send all buffer descriptors (including the spacer) to the client.
 * Loop through the pend list, then include the spacer, then the wait list.
 */
	i = 0;
	ptop = (pend_top == NULL) ? &spacer : pend_top;
	p = ptop;

	while (p) {
		pbs->bfs_node = ltot(p->bf_head.nh_node);
		pbs->bfs_event = ltot(p->bf_head.nh_event);
		pbs->bfs_type = ltot(p->bf_head.nh_type);
		pbs->bfs_length = ltot(p->bf_head.nh_length);
		++pbs;
		++i;
/*
 * Is it time to ship a full state message back to the client?
 */
		if (i >= maxstate) {

			if (bbuffer) {
				do_buffer();
			} else {
				if (nsend(&nhead) < 0) {
					panic(errno);
				}
			}

			pbs = (struct bfstate *) msg;
			i = 0;
		}

		p = p->bf_next;

		if (!p) {
			if (ptop == pend_top) {
				ptop = &spacer;
				p = ptop;
			} else if (ptop == &spacer) {
				ptop = wait_top;
				p = ptop;
			}
		}
	}
/*
 * Send the remaining state message to the client.
 */
	nhead.nh_length = sizeof(struct bfstate) * i;

	if (bbuffer) {
		do_buffer();
	} else {
		if (nsend(&nhead) < 0) {
			panic(errno);
		}
	}
/*
 * Send EEOF message to the client.
 */
	nhead.nh_length = 0;
	reply->bfr_reply = EEOF;

	if (bbuffer) {
		do_buffer();
	} else {
		if (nsend(&nhead) < 0) {
			panic(errno);
		}
	}

	return(0);
}

/*
 *	do_probe
 *
 *	Function:	- checks for the existance of messages with a given
 *			  event/type synchonization point.
 *			- sends a reply message with that message's type,
 *			  length and flags.
 *
 *	Returns:	- 0 
 */

static int
do_probe()

{
	int	event;			/* requested event */
	int	type;			/* requested type */
	struct	buffer	*p;		/* nice pointer */

	nhead.nh_dl_event = getroute(req->bfq_src_node, req->bfq_src_event);
	nhead.nh_node = req->bfq_src_node;
	nhead.nh_event = req->bfq_src_event;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_msg = (char *) 0;
	nhead.nh_length = 0;
	reply->bfr_reply = 0;
/*
 * Search the pending list for the first message that synchronizes
 * with the event/type.  The events must be equal, and if both types 
 * are non-zero, their logical AND must be non-zero.
 */
	event = req->bfq_dl_event;
	type = req->bfq_type;

	for (p = pend_top;
		(p && ((p->bf_head.nh_dl_event != event) ||
		(type && p->bf_head.nh_type && 
		! (type & p->bf_head.nh_type))));
		p = p->bf_next) {
	}

	if (p) {
/*
 * Found a message, return its characteristics.
 */
		reply->bfr_reply = 1;
		reply->bfr_type = type = p->bf_head.nh_type;
		reply->bfr_length = p->bf_head.nh_length;
		if ((reply->bfr_flags = p->bf_head.nh_flags) & NMORE) {
/*
 * The message is packetized.  Search the waiting list to find the
 * total length.  If the last packet is not in the buffer, declare
 * that no message was found.  This might lead to deadlock problems.
 */
			for (p = wait_top; p; p = p->bf_next) {

				if ((p->bf_head.nh_dl_event == event) &&
					(p->bf_head.nh_type == type)) {

					reply->bfr_length +=
						p->bf_head.nh_length;

					if (! (p->bf_head.nh_flags & NMORE))
						break;
				}
			}

			if (! p) {
/*
 * The last packet was not found.  Report that no message was found.
 */
				reply->bfr_reply = 0;
			}
		}
	}

	do_buffer();
	return(0);
}

/*
 *	chkspace
 *
 *	Function:	- checks total amount of buffer space versus maximum
 *			- may shut off user messages and issue rdlvalve()
 *			  calls to all neighbours if maxusr is exceeded
 *			- may shut off to all messages if maxspace is exceeded
 */

static void
chkspace()

{
	if (maxspace < (nspace + MAXNMSGLEN + sizeof(struct buffer))) {
		filter = BFREQ;
	} else {
		filter = BFREQ | BFUSR | BFSYS | BFDLI;
	}

	if (maxusr < nspace) {
		filter &= ~BFUSR;

		if (dlstatus == DLON) {
			valve(DLOFF);
		}
	} else if ((maxusr / 2) > nspace) {

		if (dlstatus == DLOFF) {
			valve(DLON);
		}
	} else {

		if (dlstatus == DLOFF) {
			filter &= ~BFUSR;
		}
	}
}

/*
 *	valve
 *
 *	Function:	- turns neighbour nodes' datalink valves on or off
 *			- gets neighbour node IDs if first call
 *	Accepts:	- datalink valve on or off
 */

static void
valve(dlswitch)

int			dlswitch;

{
	int		i;			/* favourite index */
	struct dlreq	*req;			/* dld request */

	if (! bjones) {
		getsmart();
		njones = getntype(NT_JONES, NT_JONES);

		if (njones < 0) {
			panic(errno);
		}

		if (getnodes(pjones, njones, NT_JONES, NT_JONES) < 0) {
			panic(errno);
		}

		bjones = TRUE;
	}

	req = (struct dlreq *) nhead.nh_data;

	for (i = 0; i < njones; ++i) {
		req->dlq_req = DLQVALVE;
		req->dlq_valve = dlswitch;
		req->dlq_node = getnodeid();

		nhead.nh_dl_event = getroute(pjones[i], NOTEVENT);
		nhead.nh_node = pjones[i];
		nhead.nh_event = EVDLMOM;
		nhead.nh_type = 0;
		nhead.nh_flags = NOBUF;
		nhead.nh_length = 0;

		do_buffer();
	}

	dlstatus = dlswitch;
}

/*
 *	putmsg
 *
 *	Function:	- sends buffered message to the buffer worker
 *	Accepts:	- ptr to network message descriptor
 */

static void
putmsg(nbuf)

struct nmsg		*nbuf;

{
	struct kmsg	kforward;		/* forwarded from daemon */

	kforward.k_event = EVBUFFER;
	kforward.k_type = 0;
	kforward.k_flags = KHOLD;
	kforward.k_length = sizeof(struct nmsg);
	kforward.k_msg = (char *) nbuf;

	if (ksend(&kforward)) {
		panic(errno);
	}

	kforward.k_flags = 0;
	kforward.k_length = nbuf->nh_length;
	kforward.k_msg = nbuf->nh_msg;

	if (ksend(&kforward)) {
		panic(errno);
	}
}
