/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB,RBD
 *
 *	$Log:	castd.c,v $
 * Revision 2.2  92/04/30  14:36:48  trillium
 * Ohio Release
 * 
 *	Function:	- multicast daemon
 *			- An incoming message may be a castd request,
 *			  a multicast or a multireel.  Since castd must always
 *			  be prepared to forward a packet intact, it must
 *			  receive all messages with DFORWARD, so that it gets
 *			  all fields in the header.
 *			- Many different multireels may be in different stages
 *			  of progress through the local node.  Thus, castd
 *			  actually buffers all messages of reels in progress.
 *			- Both header data and message body are cast/reeled.
 *			- User messages are accepted if maxspace >= MAXNMSGLEN.
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 *
 *	Usage:		- castd [<ncasts> [<maxspace>]]
 */

#include <portable.h>
#include <castreq.h>
#include <events.h>
#include <net.h>
#include <shared.h>
#include <terror.h>
#include <errno.h>
#include <typical.h>
#include <t_types.h>
#include <all_hash.h>
#include <all_list.h>

/*
 * global variables
 */
HASH			*ctable;		/* cast table */
struct nmsg		inhead;			/* network message descriptor */
char			inbuf[MAXNMSGLEN];	/* incoming datalink packet */
int4			filter;			/* message type filtering */
int4			maxspace;		/* allowed message space */

/*
 * external variables
 */
extern int		errno;			/* global error code */

/*
 * external functions
 */
extern char		*malloc();
extern int4		getroute2();
extern void		reelit();
extern void		setspace();

/*
 * local functions
 */
static void		ack2parent();
static void		castit();
static void		do_cack();
static void		do_cadd();
static void		do_cask();
static void		do_crem();
static void		do_cstat();
static void		do_cstat2();
static void		reply2client();
static int		sprout();


main(argc, argv)

int			argc;
char			**argv;

{
	int4		ncasts;		/* size of cast table */
	struct cast	*p;		/* favorite pointer */
	struct castreq	*req;		/* castd request */
/*
 * Initialize.
 */
	if (kinit(0)) {
		panic(errno);
	}

	if (rpattach(LOCAL) && (errno != ESERVICE)) {
		panic(errno);
	}
/*
 * Determine run-time parameters: cast table size, message memory space.
 * Set the message filtering mechanism according to the available space.
 */
	ncasts = (argc < 2) ? MAXCAST : atoi4(argv[1]);
	maxspace = (argc < 3) ? MAXSPACE : atoi(argv[2]);
	setspace(INT4_NIL);
/*
 * Allocate cast table.
 */
	if ((ctable = ah_init(ncasts, (int4) sizeof(struct cast),
		NOTNODEID, AHFIXED)) == (HASH *) 0) {

		panic(errno);
	}
/*
 * Wait for a smart router.
 */
	if (getsmart()) {
		panic(errno);
	}
/*
 * Continuously receive and process messages.
 */
	req = (struct castreq *) inhead.nh_data;

	for (;;) {
		inhead.nh_event = EVCASTD;
		inhead.nh_type = filter;
		inhead.nh_flags = DFORWARD;
		inhead.nh_length = MAXNMSGLEN;
		inhead.nh_msg = inbuf;

		if (drecv(&inhead)) {
			panic(errno);
		}
/*
 * Do we have a castd request to process?
 */
		if (inhead.nh_event == EVCASTD) {

			if (req->cq_req == CADD) {
				do_cadd(req, (int4 *) inhead.nh_msg,
					inhead.nh_length / sizeof(int4));
			}

			else if (req->cq_req == CREM) {
				do_crem(req);
			}

			else if ((req->cq_req == CACK) ||
				(req->cq_req == CERR)) {

				do_cack(req);
			}

			else if (req->cq_req == CASK) {
				do_cask(req);
			}

			else if (req->cq_req == CSTAT) {
				do_cstat(req);
			}

			else if (req->cq_req == CSTAT2) {
				do_cstat2(req);
			}

			else {
				reply2client(req->cq_src_node,
					req->cq_src_event, (int4) EBADREQUEST,
					INT4_NIL, (char *) 0,
					INT4_NIL, INT4_NIL);
			}
		}
/*
 * Do a cast or a reel.
 */
		else {
			p = (struct cast *) ah_find(ctable, inhead.nh_node);
			if (p == CNULL) {
				panic(EBADNODE);
			}
			else if (inhead.nh_flags & NREEL) {
				reelit(p);
			}
			else {
				castit(p->c_links);
			}
		}
	}
}

/*
 *	castit
 *
 *	Function:	- cast out messages, demons, etc
 *			- may send message locally as well as to any
 *			  number of links
 *
 *	Accepts:	- output link mask
 */
static void
castit(links)

uint4			links;

{
	int4		i;		/* favourite index */
	int4		temp_node;	/* temporary for nh_node */
/*
 * Check for local node.
 */
	if (links & CLOCAL) {
		temp_node = inhead.nh_node;
		inhead.nh_node = LOCAL;
		inhead.nh_flags &= ~BYBUF;
		inhead.nh_dl_event = inhead.nh_event;

		if (dsend(&inhead)) {
			panic(errno);
		}

		inhead.nh_node = temp_node;
	}

	inhead.nh_flags |= BYBUF;

	for (i = 0; i < MAXCLINK; i++) {

		if (links & 1) {
			inhead.nh_dl_event = dl_event(i);

			if (dsend(&inhead)) {
				panic(errno);
			}
		}

		links >>= 1;
	}
}

/*
 *	do_cadd
 *
 *	Function:	- register a new multicast (CADD)
 *
 *	Accepts:	- cast add request
 *			- ptr to node list
 *			- length of node list
 */
static void
do_cadd(req, pnodes, nnodes)

struct castreq		*req;
int4			*pnodes;
int4			nnodes;

{
	int4 		nbranch;		/* number of branches */
	int4 		nack;			/* number of acks */
	uint4		links;			/* active links at node */
	struct cast	c_entry;		/* cast entry */
	struct cast	*p;			/* favorite pointer */
/*
 * If castid is defined it is either a collision or the registration
 * is in progress (still not completed).
 */
	p = (struct cast *) ah_find(ctable, req->cq_castid);

	if (p != CNULL) {
		if (p->c_reg_nack > 0) {
			errno = EINPROGRESS;
		} else {
			errno = ECOLLIDE;
		}

		ack2parent(req, CNULL, CERR);
		return;
	}
/*
 * This is a new castid.  Sprout new branches as necessary.
 */
	if (sprout(req, pnodes, nnodes, &nbranch, &links)) {
		ack2parent(req, CNULL, CERR);
		return;
	}

	nack = (links & CLOCAL) ? nbranch - 1 : nbranch;
/*
 * Set up the new entry in the cast hash table.
 */
	c_entry.c_castid = req->cq_castid;
	c_entry.c_root_node = req->cq_root_node;
	c_entry.c_parent_node = req->cq_src_node;
	c_entry.c_combine = req->cq_combine;
	c_entry.c_nbranch = nbranch;
	c_entry.c_links = links;
	c_entry.c_reels = (LIST *) 0;
	c_entry.c_reg_event = req->cq_src_event;
	c_entry.c_reg_nack = nack;
	c_entry.c_reg_retcode = CACK;

	if (ah_insert(ctable, (void *) &c_entry)) {
		ack2parent(req, CNULL, CERR);
		return;
	}
/*
 * Do we need to make an immediate response due to no child branches?
 */
	if (nack == INT4_NIL) {
		ack2parent(req, CNULL, CACK);
	}
}

/*
 *	sprout
 *
 *	Function:	- Sprout some new branches
 *			- This functions gets a list of nodes, divides
 *			  them into groups associated with distinct outgoing
 *			  links and then sends them to the neighbours along
 *			  these links in the form of CADD requests.
 *
 *	Accepts: 	- ptr to the incoming CADD request
 *			- ptr to node list
 *			- length of node list
 *			- ptr to number of outgoing branches (value returned)
 *			- ptr to link mask of outgoing branches (value returned)
 *
 *	Returns:	- 0 or ERROR
 */
static int
sprout(inreq, pnodes, nnodes, nbranch, links)

struct castreq		*inreq;
int4			*pnodes;
int4			nnodes;
int4			*nbranch;
uint4			*links;

{
	int4		k;			/* favourite index */
	int4		event;			/* forwarding event */
	int4		outlink;		/* output link number */
	int4		*pbranch[MAXCLINK];	/* categorized node lists */
	int4		lbranch[MAXCLINK];	/* length of each list */
	int		save_errno;		/* saved error code */
	struct castreq	*outreq;		/* outgoing request */
	struct nmsg	outhead;		/* outgoing net msg desc */
/*
 * Allocate space for the divided branches of the spanning tree.
 */
	for (k = 0; k < MAXCLINK; k++) {
		lbranch[k] = 0;
		pbranch[k] = (int4 *) malloc((uint4) nnodes * sizeof(int4));
		if (pbranch[k] == (int4 *) 0) {
			save_errno = errno;
			for (k--; k >= 0; k--) {
				(void) free((char *) pbranch[k]);
			}
			errno = save_errno;
			return(ERROR);
		}
	}
/*
 * Convert node list to local byte ordering.
 */
	ttolcpy((uint4 *) pnodes, (uint4 *) pnodes, nnodes);
	
	*links = 0;
	*nbranch = 0;

	for (k = 0; k < nnodes; k++) {
/*
 * We don't want any castids among cast members.
 */
		if (tiscast(getrtype(pnodes[k]))) {
			for (k = 0; k < MAXCLINK; k++) {
				(void) free((char *) pbranch[k]);
			}
			errno = EINVAL;
			return(ERROR);
		}
/*
 * Get the forwarding event, which indicates the link number.
 */
		event = getroute2(pnodes[k], EVCASTD);

		if (event == NOTEVENT) {
			for (k = 0; k < MAXCLINK; k++) {
				(void) free((char *) pbranch[k]);
			}
			errno = EBADNODE;
			return(ERROR);
		}
		else if (event == EVCASTD) {
			*links |= CLOCAL;
			(*nbranch)++;
		}
		else {
			outlink = dl_link(event);
			pbranch[outlink][lbranch[outlink]] = pnodes[k];
			lbranch[outlink]++;
		}
	}
/*
 * Send each non-empty list on its branch of the spanning tree
 * as a CADD request.
 */
	outreq = (struct castreq *) outhead.nh_data;

	outreq->cq_req = CADD;
	outreq->cq_castid = inreq->cq_castid;
	outreq->cq_root_node = inreq->cq_root_node;
	outreq->cq_src_node = getnodeid();
	outreq->cq_src_event = inreq->cq_src_event;
	outreq->cq_combine = inreq->cq_combine;

	outhead.nh_node = LOCAL;
	outhead.nh_flags = BYBUF;
	outhead.nh_event = EVCASTD;
	outhead.nh_type = 0;

	for (k = 0; k < MAXCLINK; k++) {

		if (lbranch[k] > 0) {
			(*nbranch)++;
			*links |= (1 << k);
/*
 * Convert node sub-list to Trillium byte ordering.
 */
			ltotcpy((uint4 *) pbranch[k],
				(uint4 *) pbranch[k], lbranch[k]);

			outhead.nh_dl_event = dl_event(k);
			outhead.nh_length = lbranch[k] * sizeof(int4);
			outhead.nh_msg = (char *) pbranch[k];

			if (dsend(&outhead)) {
				panic(errno);
			}
		}
	}
/*
 * Free the branch lists.
 */
	for (k = 0; k < MAXCLINK; k++) {
		(void) free((char *) pbranch[k]);
	}

	return(0);
}

/*
 *	do_cack
 *
 *	Function:	- handle two types of acknowledgement requests
 *			  from children: positive or error acks
 *
 *	Accepts:	- ptr to CACK or CERR request
 */
static void
do_cack(req)

struct castreq		*req;

{
	struct cast	*p;			/* favourite ptr */
/*
 * Locate the cast ID in the cast table.
 */
	p = (struct cast *) ah_find(ctable, req->cq_castid);
	if (p == CNULL) {
		return;
	}
/*
 * An error condition wins over an ack.
 */
	if (req->cq_req == CERR) {
		p->c_reg_retcode = CERR;
	}
/*
 * Update and respond to parent if needed.
 */
	if (--(p->c_reg_nack) <= 0) {
		ack2parent(CQNULL, p, p->c_reg_retcode);
	}
}

/*
 *	do_crem
 *
 *	Function:	- remove a cast definition
 *
 *	Accepts:	- ptr to CREM request
 */
static void
do_crem(req)

struct castreq		*req;

{
	struct cast	*p;		/* favorite pointer */
	struct castreel	*pp;		/* another pointer */
	struct castpkt	*p2;		/* yet another pointer */
/*
 * Locate the correct entry in the cast table.
 */
	p = (struct cast *) ah_find(ctable, req->cq_castid);
	if (p == CNULL) {
		return;
	}
/*
 * Cast the CREM request down the spanning tree.
 */
	inhead.nh_node = LOCAL;
	castit(p->c_links & ~CLOCAL);
/*
 * Free all data related to this castid.
 */
	if (p->c_reels) {
/*
 * Loop deleting each entry in the list of reels.
 */
		pp = (struct castreel *) al_top(p->c_reels);
		while (pp) {

			if (pp->cr_pkt) {
/*
 * Loop deleting each entry in the list of packets.
 */
				p2 = (struct castpkt *) al_top(pp->cr_pkt);
				while (p2) {
					if (p2->cp_msg) {
						(void) free(p2->cp_msg);
						setspace(p2->cp_length);
					}
					p2 = (struct castpkt *)
						al_next(pp->cr_pkt, (char *)p2);
				}
/*
 * Free the list of packets.
 */
				(void) al_free(pp->cr_pkt);
			}

			pp = (struct castreel *)
				al_next(p->c_reels, (char *) pp);
		}
/*
 * Free the list of reels.
 */
		(void) al_free(p->c_reels);
	}
/*
 * Delete the castid entry in the hash table.
 */
	(void) ah_delete(ctable, req->cq_castid);
/*
 * Remove castid from the local router.
 */
	if (rremroute(LOCAL, req->cq_castid)) {
		if (errno != EDELETE) {
			panic(errno);
		}
	}
}

/*
 *	do_cask
 *
 *	Function:	- inquire the number of out branches at this link
 *			- handy information for a client that wants to do reel
 *			  combining without the aid of a separate process
 *
 *	Accepts:	- ptr to CASK request
 */
static void
do_cask(req)

struct castreq		*req;

{
	struct cast	*p;		/* favorite pointer */
/*
 * Locate the cast ID in the cast table.
 */
	p = (struct cast *) ah_find(ctable, req->cq_castid);

	if (p == CNULL) {
		reply2client(req->cq_src_node, req->cq_src_event,
			(int4) EBADNODE, INT4_NIL, (char *) 0,
			INT4_NIL, INT4_NIL);
	}
	else {
		reply2client(req->cq_src_node, req->cq_src_event,
			INT4_NIL, p->c_nbranch, (char *) 0, INT4_NIL, INT4_NIL);
	}
}

/*
 *	do_cstat
 *
 *	Function:	- send back castd primary state information
 *
 *	Accepts:	- ptr to CSTAT request
 */
static void
do_cstat(req)

struct castreq		*req;

{
	struct cast	*p;		/* favorite pointer */
	struct cstate	*pp;		/* another pointer */
	int4		i;		/* favorite counter */
	int4		ncast;		/* number of castids */
	int4		length;		/* message length */
/*
 * Return all entries up to the maximum given limit.
 */
	i = length = 0;
	ncast = ah_count(ctable);
	p = (struct cast *) ah_top(ctable);
	pp = (struct cstate *) inbuf;

	while ((p) && (i < req->cq_maxstat)) {
		pp->cs_castid = p->c_castid;
		pp->cs_root_node = p->c_root_node;
		pp->cs_parent_node = p->c_parent_node;
		pp->cs_combine = p->c_combine;
		pp->cs_nbranch = (p->c_reg_nack == 0) ?
			p->c_nbranch : (- p->c_nbranch);
		pp->cs_nreels = (p->c_reels) ? al_count(p->c_reels) : 0;

		pp++;
		i++;
		length += sizeof(struct cstate);
		if ((MAXNMSGLEN - length) < sizeof(struct cstate)) {
/*
 * This packet is full, send it.
 */
			reply2client(req->cq_src_node, req->cq_src_event,
				INT4_NIL, ncast, inbuf, length, DINTMSG);
			length = 0;
			pp = (struct cstate *) inbuf;
		}

		p = (struct cast *) ah_next(ctable, (void *) p);
	}
/*
 * Flush the last packet out, if any.
 */
	if ((length > 0) || (i == 0)) {
		reply2client(req->cq_src_node, req->cq_src_event,
			INT4_NIL, ncast, inbuf, length, DINTMSG);
	}
}

/*
 *	do_cstat2
 *
 *	Function:	- send back secondary state information for castid
 *
 *	Accepts:	- ptr to CSTAT request
 */
static void
do_cstat2(req)

struct castreq		*req;

{
	struct cast	*p;		/* favorite pointer */
	struct cstate2	*pp;		/* another pointer */
	struct castreel	*preel;		/* ptr to reel entry */
	int4		i;		/* favorite counter */
	int4		nreels;		/* # reeled messages */
	int4		length;		/* message length */

	p = (struct cast *) ah_find(ctable, req->cq_castid);
	if ((p == CNULL) || (p->c_reels == (LIST *) 0)) {
/*
 * Castid not defined or list of reels is not defined.
 * First case return EINVAL otherwise just a zero-length message.
 */
		reply2client(req->cq_src_node, req->cq_src_event,
			(p == CNULL) ? EINVAL : INT4_NIL, INT4_NIL,
			(char *) 0, INT4_NIL, INT4_NIL);
		return;
	}
/*
 * Return all entries up to the maximum given limit.
 */
	length = i = 0;
	nreels = al_count(p->c_reels);
	preel = (struct castreel *) al_top(p->c_reels);
	pp = (struct cstate2 *) inbuf;

	while ((preel) && (i < req->cq_maxstat)) {
		pp->cs2_castid = p->c_castid;
		pp->cs2_event = preel->cr_event;
		pp->cs2_type = preel->cr_type;
		pp->cs2_npkt = al_count(preel->cr_pkt);

		i++;
		pp++;
		length += sizeof(struct cstate2);
		if ((MAXNMSGLEN - length) < sizeof(struct cstate2)) {
/*
 * This packet is full, send it.
 */
			reply2client(req->cq_src_node, req->cq_src_event,
				INT4_NIL, nreels, inbuf, length, DINTMSG);
			length = 0;
			pp = (struct cstate2 *) inbuf;
		}

		preel = (struct castreel *) al_next(p->c_reels, (char *) preel);
	}
/*
 * Flush the last packet out, if any.
 */
	if ((length > 0) || (i == 0)) {
		reply2client(req->cq_src_node, req->cq_src_event,
			INT4_NIL, nreels, inbuf, length, DINTMSG);
	}
}

/*
 *	ack2parent
 *
 *	Function:	- send an acknowledgement request to our parent
 *			- requires either ptr to request or ptr to cast
 *
 *	Accepts:	- ptr to request (may be NULL)
 *			- ptr to cast entry (may be NULL)
 *			- request code
 */
static void
ack2parent(preq, pcast, reqcode)

struct castreq		*preq;
struct cast		*pcast;
int4			reqcode;

{
	struct castreq	*outreq;	/* outgoing upstairs request */
	struct nmsg	outhead;	/* outgoing request msg desc */
	int4		castid;		/* cast identifier */
	int4		root_node;	/* root of the cast tree */
	int4		src_node;	/* client nodeid */
	int4		src_event;	/* client event */
/*
 * Grab the information needed.
 */
	if (pcast) {
		castid = pcast->c_castid;
		root_node = pcast->c_root_node;
		src_node = pcast->c_parent_node;
		src_event = pcast->c_reg_event;
	}
	else {
		castid = preq->cq_castid;
		root_node = preq->cq_root_node;
		src_node = preq->cq_src_node;
		src_event = preq->cq_src_event;
	}
/*
 * Register the castid with the router if it is an ack.
 */
	if (reqcode == CACK) {
		if (rsetroute(LOCAL, castid, EVCASTD, EVCASTD, NT_CAST)) {
/*
 * Ok the router let us down, don't expect a beautiful exit, but
 * at least try to notify the top of the tree instead of shouting panic.
 */
			reqcode = CERR;
		}
	}
/*
 * If we are the root reply to rcast() client.
 */
	if (root_node == getnodeid()) {
		reply2client(src_node, src_event,
			(reqcode == CERR) ? errno : INT4_NIL,
			INT4_NIL, (char *) 0, INT4_NIL, INT4_NIL);
	}
/*
 * Send an ack request upstairs.
 */
	else {
		outreq = (struct castreq *) outhead.nh_data;
		outreq->cq_src_node = getnodeid();
		outreq->cq_src_event = src_event;
		outreq->cq_castid = castid;
		outreq->cq_req = reqcode;

		outhead.nh_node = src_node;
		outhead.nh_event = EVCASTD;
		outhead.nh_type = 0;
		outhead.nh_flags = 0;
		outhead.nh_length = 0;
		outhead.nh_msg = (char *) 0;

		if (nsend(&outhead)) {
			panic(errno);
		}
	}
}

/*
 *	reply2client
 *
 *	Function:	- send a reply to a castd client
 *
 *	Accepts:	- client node
 *			- client event
 *			- reply code
 *			- return value (when needed)
 *			- ptr to message
 *			- length of message
 *			- message flags
 */
static void
reply2client(src_node, src_event, replycode, result, mesg, length, flags)

int4			src_node;
int4			src_event;
int4			replycode;
int4			result;
char			*mesg;
int4			length;
int4			flags;

{
	struct castreply
			*reply;		/* castd reply */
	struct nmsg	outhead;	/* outgoing message descriptor */

	reply = (struct castreply *) outhead.nh_data;
	reply->cr_reply = replycode;
	reply->cr_result = result;

	outhead.nh_node = src_node;
	outhead.nh_event = src_event;
	outhead.nh_type = 0;
	outhead.nh_flags = flags;
	outhead.nh_length = length;
	outhead.nh_msg = mesg;

	if (nsend(&outhead)) {
		panic(errno);
	}
}
