/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *      VSD
 *
 *	$Log:	all_hash.c,v $
 * Revision 2.2  92/04/30  14:40:01  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:32:25  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:58  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/03/01  18:01:17  gdburns
 * removed unused variable, full
 * 
 * Revision 2.0  90/02/20  11:56:42  trillium
 * Homo erectus
 * 
 *
 *	Function:	Hashing functions -
 *			1. rd_hash() : to read from the hash table,
 *			2. wr_hash() : to write into hash table,
 *			3. wr_kick() : to write into hash table and kick
 *				       out oldest entry if table is full.
 *			   Call init_table() prior to using any of the
 *			   three functions.	
 *
 *	When can you use these routines :-
 *	-------------------------------
 *	-  A Hash Table (HT) consists of 'num' number of records. The 
 *	   struct of the HT-record has in its first field the identifier
 *	   'id' which is the key used for hashing into the HT. This 'id'
 *	   must be of type INTEGER.
 *	-  No 'id' in the HT must have the value NOTNODEID (see <net.h>).
 *	-  To use wr_kick(), in addition to above condition, the second
 *	   field of the HT-record must be 'count' field of type INTEGER.
 *	   This field is to be initialized and updated by the user of 
 *	   wr_kick().
 *
 *	How to use these routines :-
 *	-------------------------
 *	-  The parameters required are :
 *	   a) id     : value of the key to be read/wriiten in HT
 *	   b) htable : pointer to the HT
 *	   c) num    : number of records in HT
 *	   d) rsize  : sizeof (struct HT-record), even # of bytes.
 *      -  The HT must be first initialized by init_table().
 *         This sets all 'id' fields of HT to NOTNODEID.
 *      -  To hash an 'id' into HT, call wr_hash().
 *         If successful, it returns the index of the HT onto which
 *         the 'id' was hashed and filled.
 *         If unsuccessful, it returns error 'EFULL'. This would
 *	   happen when there are no empty slots in HT. 
 *      -  To read from HT, call rd_hash().
 *         Given an 'id', it returns the index of HT where that 'id'
 *         is stored. If 'id' is not found, it returns error EBADNODE.
 *	-  wr_kick() works like wr_hash() when HT is not full. If HT
 *	   is full, wr_kick() kicks out the 'id' with minimum 'count'
 *	   value and inserts the new 'id' there and returns that index.
 *      -  rd_hash(), wr_hash, wr_kick() can be called in any order 
 *	   once init_table() has been invoked.
 */

#include <net.h>	/* for NOTNODEID 		    		*/
#include <terror.h>	/* for EFULL, EBADNODE				*/
#include <typical.h>	/* for ERROR					*/

extern int errno;	/* for error codes				*/

/*----------------------------------------------------------------------*/
/*
 * init_table()
 *
 * Function :		Initializes all id's in the HT to NOTNODEID 
 *			and sets no. of filled slots in HT to 0.
 */
void 
init_table (htable, num, rsize)

void  *htable;		/* pointer to HT				*/
int   num;		/* no. of records in HT each of size rsize	*/
int   rsize;    	/* size of records in HT			*/

{
  int 	i;
  int	*hentry;

  for (i=0; i<num; i++) {
      hentry = (int *) ((char *)htable + (i * rsize));
      (*hentry) = NOTNODEID;
    }
}

/*----------------------------------------------------------------------*/
/*
 * rd_hash()
 *
 * Function :		To read from hash table.
 *         		Given an 'id', it returns the index of HT where
 *			that 'id' is stored. If 'id' is not found, it 
 *			returns error EBADNODE.
 */

int 
rd_hash (id, htable, num, rsize)

int	id;		/* key value to be searched for in HT		*/
void	*htable;	/* pointer to HT				*/
int	num;		/* number of records in HT			*/
int	rsize;		/* size of records in HT			*/

{
  int index, start, run;
  int *hentry;

  if (id == NOTNODEID) return(ERROR);
  index = abs(id) % num;
  start = index;
  run = 1;
  hentry = (int *) ((char *)htable + (index * rsize));
  while ( ((*hentry) != id) && ((*hentry) != NOTNODEID) && run) {
      index = (index+1) % num;
      if (index == start)  run=0;
      hentry = (int *) ((char *)htable + (index * rsize));
    }
  if ( (*hentry)  == id ) {
      return (index);
    }
  else {
      errno = EBADNODE;
      return (ERROR);
    }
}

/*----------------------------------------------------------------------*/
/*
 * wr_hash()
 *
 * Function :		To write an 'id' into hash table.
 *         		If successful, it returns the index of the HT
 *			onto which the 'id' was hashed and filled.
 *			If unsuccessful, it returns error 'EFULL'.
 *			This would happen when there are no empty slots
 *			in HT. 
 */

int 
wr_hash (id, htable, num, rsize)

int	id;        	/* key value to be written into HT		*/
void	*htable;	/* pointer to HT				*/
int	num;		/* number of records in HT			*/
int	rsize;		/* size of records in HT			*/

{
  int index, start, run;
  int *hentry;

  if (id == NOTNODEID) return(ERROR);
  index = abs(id) % num;
  start = index;
  run = 1;
  hentry = (int *) ((char *)htable + (index * rsize));
  while ( ( (*hentry) != NOTNODEID) && run ) {
	if ((*hentry) == id) {
	   return(index);
	}
      index = (index+1) % num;
      if (index == start)   run = 0;
      hentry = (int *) ((char *)htable + (index * rsize));
    }
  if (!run) {
      errno = EFULL;
      return (ERROR);
    }
  else {
      (*hentry) = id;
      return (index);
    }
}

/*----------------------------------------------------------------------*/
/* 
 * wr_kick()
 *
 * Function :		To write an 'id' into hash table.
 *	  		wr_kick() works like wr_hash() when HT is not
 *			full. If HT is full, wr_kick() kicks out the 
 *			'id' with minimum 'count' value and inserts 
 *			the new 'id' there and returns that index.
 */

int 
wr_kick (id, htable, num, rsize)

int	id;		/* 'id' to be inserted into HT			*/
void	*htable;	/* pointer to HT				*/
int	num;		/* number of records in HT			*/
int	rsize;		/* size of records in HT			*/

{
  int	index, i, mincnt;
  int 	*hentry;

  if (id == NOTNODEID) return(ERROR);
  index = wr_hash (id, htable, num, rsize);
  if (index != ERROR) {
	return(index);
  }
  else {
	hentry = ((int *) htable) + 1;	/* hentry = htable[0].count */
	index = 0;
	mincnt = (*hentry);
	for (i=0; i<num; i++) {
		hentry = ((int *) ((char *)htable + (i * rsize))) + 1;
		/* +1 to skip id field and go to count field */
		if ((*hentry) < mincnt) {
			index = i;
			mincnt = (*hentry);
		}
	}
	hentry = (int *) ((char *)htable + (index * rsize));
	(*hentry) = id;
	return(index);
  }
}

/*------------------------------------------------------------------------*/

