/*
 *      OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *      GDB
 *
 *      $Log:	all_list.c,v $
 * Revision 2.2  92/04/30  14:40:04  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/10/20  21:16:06  gdburns
 * initial revision
 * 
 *	Function:	- general purpose dynamic list package
 *			- assumes a fast malloc()
 *			- doubly linked
 *			- The dl_head structure is prepended to
 *			  the list element given by the user.
 *			  The user does not see dl_head.
 *			  Thus, we often have to shift back and
 *			  forth between the user element pointer
 *			  and the dl_head pointer, which is used to
 *			  traverse the list.
 *
 *			- The user must supply a function
 *			  that compares two list elements and returns
 *			  -1 if the first element is less than the second,
 *			  0 if they are equal, and
 *			  1 if the first element is greater than the second.
 *			  Lists are maintained in increasing element order.
 *
 *			- functions that operate on a list:
 *
 *			  al_init - Allocate and initialize a new list.
 *			  al_delete - Delete a list element.
 *			  al_insert - Insert a list element.
 *			  al_append - Insert an element at the end of a list.
 *			  al_free - Blow away a list.
 *			  al_find - Find a list element.
 *			  al_next - Find the next list element.
 *			  al_prev - Find the previous list element.
 *			  al_count - Count the elements in a list.
 */

#include <all_list.h>
#include <terror.h>
#include <typical.h>

/*
 * local macros
 */
#define NULL ((char *) 0)

/*
 * external functions
 */
extern char		*malloc();
extern char		*memcpy();
extern int		free();

/*
 * external variables
 */
extern int		errno;

/*
 *	al_init
 *
 *	Function:	- Initialize a new list descriptor.
 *
 *	Accepts:	- size of user element
 *			- comparison function
 *
 *	Returns:	- ptr to new list descriptor
 */

LIST *
al_init(elemsize, comp)

int4			elemsize;		/* element size */
int4			(*comp)();		/* comparison function */

{
	LIST		*ald;			/* new al descriptor */

	ald = (LIST *) malloc(sizeof(LIST));

	if (ald == (LIST *) 0) {
		return((LIST *) 0);
	}

	ald->al_elemsize = elemsize;
	ald->al_ltop = ALNULL;
	ald->al_lbottom = ALNULL;
	ald->al_comp = comp;

	return(ald);
}

/*
 *	al_delete
 *
 *	Function:	- Delete an element from a list.
 *
 *	Accepts:	- list descriptor ptr
 *			- ptr to existing element
 *
 *	Returns:	- 0 or ERROR
 */

int4
al_delete(ald, old)

LIST			*ald;			/* al descriptor */
char			*old;			/* old list element */

{
	struct al_head	*curr;
	struct al_head	*prev;

	curr = ald->al_ltop;
	prev = ALNULL;

	while ((curr != ALNULL) && (al_body(curr) != old)) {
		prev = curr;
		curr = curr->al_next;
	}

	if (curr == ALNULL) {
		errno = EDELETE;
		return(ERROR);
	}

	if (curr == ald->al_ltop) {
		ald->al_ltop = curr->al_next;
	} else {
		prev->al_next = curr->al_next;
	}

	if (curr == ald->al_lbottom) {
		ald->al_lbottom = prev;
	} else {
		curr->al_next->al_prev = prev;
	}

	(void) free(old - sizeof(struct al_head));
	return(0);
}

/*
 *	al_insert
 *
 *	Function:	- Insert an element into a list.
 *			- The user's new element is copied
 *			  into space allocated internally for a
 *			  new list element.
 *
 *	Accepts:	- list descriptor ptr
 *			- new element ptr
 * 
 *	Returns:	- new list element ptr
 */

char *
al_insert(ald, unew)

LIST			*ald;			/* al descriptor */
char			*unew;			/* user's new element */

{
	struct al_head	*curr;
	struct al_head	*prev;
	struct al_head	*new_head;		/* new internal list element */
	char		*new;			/* user portion of list
						   element */

	new_head = (struct al_head *) malloc((unsigned) (ald->al_elemsize +
		sizeof(struct al_head)));

	if (new_head == ALNULL) {
		return(NULL);
	}

	new = al_body(new_head);
	(void) memcpy(new, unew, ald->al_elemsize);
	curr = ald->al_ltop;
	prev = ALNULL;

	while ((curr != ALNULL) &&
		(((*(ald->al_comp))(new, al_body(curr))) >= 0)) {

		prev = curr;
		curr = curr->al_next;
	}

	new_head->al_next = curr;
	new_head->al_prev = prev;

	if (curr == ald->al_ltop) {
		ald->al_ltop = new_head;
	} else {
		prev->al_next = new_head;
	}

	if (curr == ALNULL) {
		ald->al_lbottom = new_head;
	} else {
		curr->al_prev = new_head;
	}

	return(new);
}

/*
 *	al_append
 *
 *	Function:	- Append a new element to a list.
 *			- Element is added at the end of the list.
 *			- A new list element is allocated internally
 *			  and the user supplied element is copied,
 *			  as with al_insert().
 *
 *	Accepts:	- list descriptor ptr
 *			- new element ptr
 *
 *	Returns:	- new list element ptr
 */

char *
al_append(ald, unew)

LIST			*ald;			/* al descriptor */
char			*unew;			/* user's new element */

{
	struct al_head	*new_head;		/* new internal list element */
	char		*new;			/* user portion of list
						   element */

	new_head = (struct al_head *) malloc((unsigned) (ald->al_elemsize +
		sizeof(struct al_head)));

	if (new_head == ALNULL) {
		return(NULL);
	}

	new = al_body(new_head);
	(void) memcpy(new, unew, ald->al_elemsize);
	
	new_head->al_next = ALNULL;
	new_head->al_prev = ald->al_lbottom;

	if (ald->al_lbottom != ALNULL) {
		ald->al_lbottom->al_next = new_head;
	} else {
		ald->al_ltop = new_head;
	}

	ald->al_lbottom = new_head;
	return(new);
}

/*
 *	al_free
 *
 *	Function:	- Free all list elements and list descriptor.
 *			- Effectively, blow away the list.
 *
 *	Accepts:	- list descriptor ptr
 *
 *	Returns:	- 0
 */

int4
al_free(ald)

LIST			*ald;			/* al descriptor */

{
	struct al_head	*curr;
	struct al_head	*next;

	curr = ald->al_ltop;

	while (curr != ALNULL) {
		next = curr->al_next;
		free((char *) curr);
		curr = next;
	}

	free((char *) ald);
	return(0);
}

/*
 *	al_find
 *
 *	Function:	- Find a list element that matches the
 *			  user supplied element's key.
 *			- The user supplied comparison function
 *			  is employed to determine equality.
 *
 *	Accepts:	- list descriptor ptr
 *			- search key element ptr
 *
 *	Returns:	- found list element ptr
 */

char *
al_find(ald, key)

LIST			*ald;
char			*key;

{
	struct al_head	*curr;

	curr = ald->al_ltop;

	while ((curr != ALNULL) && ((*(ald->al_comp))(key, al_body(curr)))) {
		curr = curr->al_next;
	}

	return(curr ? al_body(curr) : (char *) 0);
}

/*
 *	al_next
 *
 *	Function:	- find the next element in a list
 *
 *	Accepts:	- list descriptor ptr
 *			- existing list element ptr
 *
 *	Returns:	- next list element ptr
 */

char *
al_next(ald, old)

LIST			*ald;
char			*old;

{
	struct al_head	*curr;

	curr = (struct al_head *) (old - sizeof(struct al_head));
	return(curr->al_next ? al_body(curr->al_next) : (char *) 0);
}

/*
 *	al_prev
 *
 *	Function:	- find the previous element in a list
 *
 *	Accepts:	- list descriptor ptr
 *			- existing list element ptr
 *
 *	Returns:	- previous list element ptr
 */

char *
al_prev(ald, old)

LIST			*ald;
char			*old;

{
	struct al_head	*curr;

	curr = (struct al_head *) (old - sizeof(struct al_head));
	return(al_body(curr->al_prev));
}

/*
 *	al_count
 *
 *	Function:	- Count the number of elements in a list.
 *
 *	Accepts:	- list descriptor ptr
 *
 *	Returns:	- total number of list elements
 */

int4
al_count(ald)

LIST			*ald;

{
	int4		n;
	struct al_head	*curr;

	n = 0;
	curr = ald->al_ltop;

	while (curr != ALNULL) {
		n++;
		curr = curr->al_next;
	}

	return(n);
}
