/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	all_queue.c,v $
 * Revision 2.2  92/04/30  14:40:06  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/10/20  21:17:54  raja
 * Overhaul of interface to fit the all_* standard.
 * 
 * Revision 2.1.1.1  91/06/05  21:36:25  raja
 * Use memcpy() instead of bcopy() for better portability.
 * 
 * Revision 2.1  91/03/20  11:32:26  gdburns
 * Ohio Release
 * 
 * Revision 2.0  90/11/16  15:27:15  raja
 * Package for managing generic queues.
 *
 *	Function:	- generic queue management code
 *			- fully dynamic version
 *				- aq_init(): create a queue
 *				- aq_free(): free a queue
 *				- aq_insert(): insert element
 *				- aq_shove(): insert element by force
 *				- aq_delete(): delete element
 *				- aq_find(): read element
 *				- aq_expand(): increase queue size
 */

#include <memory.h>
#include <malloc.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <all_queue.h>

/*
 * external variables
 */
extern int			errno;


/*
 *	aq_init
 *
 *	Function:	- create a queue
 *
 *	Accepts:	- size of queue
 *			- size of queue element
 *
 *	Returns:	- ptr to queue descriptor or NULL
 */
QUEUE *
aq_init(size, elemsize)

int4			size;
int4			elemsize;

{
	QUEUE		*aqd;
/*
 * Allocate the queue descriptor.
 */
	if ((aqd = (QUEUE *) malloc(sizeof(QUEUE))) == (QUEUE *) 0) {
		return((QUEUE *) 0);
	}

	aqd->aq_maxnelem = size;
	aqd->aq_elemsize = elemsize;
	aqd->aq_nelem = INT4_NIL;
	aqd->aq_first = INT4_NIL;
	aqd->aq_last = INT4_NIL;
/*
 * Allocate the queue array.
 */
	aqd->aq_queue = (void *) malloc((uint4) size * elemsize);
	if (aqd->aq_queue == (void *) 0) {
		free((char *) aqd);
		return((QUEUE *) 0);
	}

	return(aqd);
}

/*
 *	aq_free
 *
 *	Function:	- free a queue
 *			- queue must have been created with aq_init()
 *
 *	Accepts:	- ptr to queue descriptor
 */
void
aq_free(aqd)

QUEUE			*aqd;

{
	if (aqd) {
		if (aqd->aq_queue) {
			free((char *) aqd->aq_queue);
		}
		free((char *) aqd);
	}
}

/*
 *	aq_insert
 *
 *	Function:	- insert an element at the end of the queue
 *
 *	Accepts:	- ptr to queue descriptor
 *			- ptr to the element
 *
 *	Returns:	- 0 or ERROR
 */
int
aq_insert(aqd, elem)

QUEUE			*aqd;
void			*elem;

{
	char		*p;		/* favorite pointer */
/*
 * Check if the queue is full.
 */
	if (aqd->aq_nelem >= aqd->aq_maxnelem) {
		errno = EFULL;
		return(ERROR);
	}
/*
 * Enqueue the new element.
 */
	aqd->aq_nelem++;
	p = ((char *) aqd->aq_queue) + (aqd->aq_last * aqd->aq_elemsize);
	(void) memcpy(p, (char *) elem, aqd->aq_elemsize);
	aqd->aq_last = (aqd->aq_last + 1) % aqd->aq_maxnelem;

	return(0);
}

/*
 *	aq_shove
 *
 *	Function:	- insert an element at the end of the queue
 *			- if the queue is full double its size
 *
 *	Accepts:	- ptr to queue descriptor
 *			- ptr to the element
 *
 *	Returns:	- 0 or ERROR
 */
int
aq_shove(aqd, elem)

QUEUE			*aqd;
void			*elem;

{
	if (aq_insert(aqd, elem)) {
/*
 * The queue is full, expand it by doubling its size.
 */
		if (aq_expand(aqd, 2 * aqd->aq_maxnelem)) {
			return(ERROR);
		}
/*
 * Insert the element.
 */
		if (aq_insert(aqd, elem)) {
			errno = EIMPOSSIBLE;
			return(ERROR);
		}
	}

	return(0);
}

/*
 *	aq_delete
 *
 *	Function:	- delete an element from the top of the queue
 *
 *	Accepts:	- ptr to queue descriptor
 *
 *	Returns:	- 0 or ERROR
 */
int
aq_delete(aqd)

QUEUE			*aqd;

{
/*
 * Check if the queue if empty.
 */
	if (aqd->aq_nelem <= INT4_NIL) {
		errno = EDELETE;
		return(ERROR);
	}
/*
 * Dequeue an element.
 */
	aqd->aq_nelem--;
	aqd->aq_first = (aqd->aq_first + 1) % aqd->aq_maxnelem;

	return(0);
}

/*
 *	aq_find
 *
 *	Function:	- locate the first element in the queue
 *
 *	Accepts:	- ptr to queue descriptor
 *
 *	Returns:	- ptr to element or NULL
 */
void *
aq_find(aqd)

QUEUE			*aqd;

{
/*
 * Check if the queue is empty.
 */
	if (aqd->aq_nelem <= INT4_NIL) {
		return((void *) 0);
	}
	else {
		return((void *) (((char *) aqd->aq_queue) +
			(aqd->aq_first * aqd->aq_elemsize)));
	} 
}

/*
 *	aq_expand
 *
 *	Function:	- expand the size of the queue
 *			- new size must be bigger than old size
 *
 *	Accepts:	- ptr to queue descriptor
 *			- new queue size
 *
 *	Returns:	- 0 or ERROR
 */
int
aq_expand(aqd, newsize)

QUEUE			*aqd;
int4			newsize;

{
	void		*new_q;		/* new queue array */
	int4		size;

	if (newsize < aqd->aq_maxnelem) {
		errno = EINVAL;
		return(ERROR);
	}
	else if (newsize == aqd->aq_maxnelem) {
		return(0);
	}
/*
 * Create a new queue array of a bigger size.
 */
	new_q = (void *) malloc((uint4) newsize * aqd->aq_elemsize);
	if (new_q == (void *) 0) {
		return(ERROR);
	}

	if (aqd->aq_nelem == INT4_NIL) {
/*
 * Do nothing, no elements to copy.
 */
	}
	else if (aqd->aq_first + aqd->aq_nelem <= aqd->aq_maxnelem) {
/*
 * The current queue does not wrap around the array boundary.
 * A single memory copy is enough.
 */
		(void) memcpy((char *) new_q, ((char *) aqd->aq_queue) +
			(aqd->aq_first * aqd->aq_elemsize),
			(aqd->aq_nelem * aqd->aq_elemsize));
	}
	else {
/*
 * The current queue wraps around the array boundary.
 * Two memory copies are required for the two separate chunks.
 */
		size = (aqd->aq_maxnelem - aqd->aq_first) * aqd->aq_elemsize;
		(void) memcpy((char *) new_q, ((char *) aqd->aq_queue) +
			(aqd->aq_first * aqd->aq_elemsize), size);
		(void) memcpy((char *) new_q + size, (char *) aqd->aq_queue,
			aqd->aq_last * aqd->aq_elemsize);
	}
/*
 * Finish the clean up of the queue descriptor.
 */
	if (aqd->aq_queue) {
		free((char *) aqd->aq_queue);
	}
	aqd->aq_queue = new_q;
	aqd->aq_first = INT4_NIL;
	aqd->aq_last = aqd->aq_nelem;
	aqd->aq_maxnelem = newsize;

	return(0);
}
