/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	kernel.c,v $
 * Revision 2.2  92/04/30  14:47:08  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:36:26  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:28:56  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/02/21  12:05:34  gdburns
 * Forgot to initialize the last process descriptor to SFREE.
 * 
 * Revision 2.0  90/02/20  11:55:33  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:11:20  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0.1.1  88/01/12  10:52:18  gdburns
 * Fixed find_event to work properly when etable fills with vacant descriptors.
 * 
 * Revision 1.0  87/12/19  15:46:35  trillium
 * Initial revision
 * 
 *
 *	Function:	- simple message rendez-vous server
 *			- kernel of the Trillium OS
 *			- receives requests and replies results through 
 *			  abstraction presented by the routines "recv_req" 
 *			  and "send_reply"
 *			- should only receive requests within the machine
 *			  on which it is running
 *			- Kernel requests are broken into two groups, those
 *			  that manage processes (kattach, kdetach, kdoom,
 *			  kstate) and those that support interprocess 
 *			  communication (ksend, krecv, ktry_recv, ktry_send).
 *			  There is no kcreate kernel request.
 *			  Applications have access to a kcreate subroutine, but
 *			  it does not interact with the kernel.  Instead, new
 *			  processes become kernel processes by "attaching"
 *			  to the kernel.  This leads to the uncommon practice
 *			  of a process selecting its priority and process ID.
 *
 *			  The kernel does not pass messages.  It simply makes
 *			  a rendez-vous between sending and receiving parties,
 *			  similar to that of a dating service.  When the
 *			  rendez-vous is made, the kernel replies to both
 *			  parties and the sender transmits its message directly
 *			  to the receiver.
 *
 *			  A rendez-vous is not made by one party specifying the
 *			  process ID of the other.  Instead, both parties
 *			  select an arbitrary number which they each know
 *			  (and which the programmer chooses) a priori.
 *			  When both a send and receive have been done on a
 *			  particular number, called an event number, the
 *			  highest priority sender is introduced to the highest
 *			  priority receiver through kernel replies.  The
 *			  sender receives from the kernel the communication
 *			  port of the receiver and the message transfer takes
 *			  place directly.
 *
 *			  As of the 2.0 release, this kernel now features
 *			  optional process scheduling.  The message priority
 *			  presented to the kernel in the kattach request is now
 *			  also the process scheduling priority.  If the priority
 *			  is 0, the process is uncontrolled and dispatched when-
 *			  ever it is ready.  Only the highest priority
 *			  controlled process is ever dispatched.  It runs until
 *			  it makes another kernel request, even if a higher
 *			  priority controlled process becomes ready in the
 *			  meantime.
 *
 *			  The 2.0 kernel also features generalized signal
 *			  delivery.  See the kkenya.c module for more details.
 *
 *			- This module contains the kernel initialization,
 *			  mainline and list processing routines.
 */


#include <debug.h>
#include <kernel.h>
#include <kreq.h>
#include <ksignal.h>
#include <typical.h>
#include <terror.h>


/*
 * major kernel data structures
 */
struct kreq		request;		/* kernel request */
struct kreply		reply;			/* kernel reply */

struct kprocess		ptable[PMAX],		/* the process table */
			*pprocess,		/* top of process list */
			*pproc_tail,		/* bottom of process list */
			*pfree,			/* top of free list */
			*pfree_tail,		/* bottom of free list */
			*pactive[2];		/* active controlled
						   processes */

struct kevent		etable[EMAX];		/* the message event table */

int			ncontrolled;		/* # of controlled,
						   READY processes */
			
/*
 * kernel request handlers
 */
extern void		kattach(),		/* declare kernel requests */
			kdetach(),
			kstate(),
			kdoom(),
			ksync();

void			(*do_req[NREQS])() = {	/* request routines */
				kattach,
				kdetach,
				kstate,
				kdoom,
				ksync,
				ksync,
				ksync,
				ksync
			};
/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern void		init_request(),
			recv_req(),
			send_reply();

/*
 * defined global functions
 */
struct kevent		*find_event();			
struct kprocess		*getprocbypid();
void			insert();
void			delete();
void			schedule();
void			dispatch_ctl();
void			dispatch();


/*
 *	main
 *
 *	Function:	- kernel mainline routine
 *			- initializes process and event tables
 *			- interprets request code and call appropriate routine
 *			- loops forever
 */

main()

{
	struct kprocess		*p;		/* favourite generic pointer */
	struct kevent		*e;		/* not my favourite pointer */
	int			i;		/* favourite index variable */

/*
 * Initialize kernel communication interface.
 */
	init_request();
	ncontrolled = 0;
/*
 * Initialize process table.
 */
	for (p = ptable, i = 0; i < (PMAX - 1); ++i, ++p) {
		p->k_pnext[PPROCESS] = p + 1;
		p->k_state = SFREE;
		p->k_priority = -1;
		p->k_event = NOTEVENT;
		p->k_ptarget = NULL;
	}

	pfree = ptable;
	pfree_tail = p;
	p->k_state = SFREE;
	p->k_pnext[PPROCESS] = pprocess = pproc_tail =
		pactive[0] = pactive[1] = NULL;
/*
 * Initialize event table.
 */
	for (e = etable, i = 0; i < EMAX; ++i, ++e) {
		e->ev_event = NOTEVENT;
		e->ev_head = NULL;
		e->ev_tail = NULL;
	}

	db(("kernel:  initialized\n"));
/*
 * Mainline Kernel Loop
 */
	for (;;) { 
		recv_req(); 

		if (request.kq_index != -1) {
/*
 * Existing process making a kernel request.
 */
			p = ptable + request.kq_index;
/*
 * Check for controlled process.
 */
			if (p == pactive[0]) {
				pactive[0] = NULL;
			}

			else if (p == pactive[1]) {
				pactive[1] = NULL;
			}
/*
 * Check for pending signal.
 */
			if (p->k_signal) {
				p->k_errno = EINTR;
				p->k_state = SERROR;
				schedule(p);
			}
/*
 * Process request.
 */
			else {
				(*do_req[request.kq_request - NORMAL])(p);
			}
		}

		else {
/*
 * New process wishes to attach.
 */
			kattach(NULL);
		}
/*
 * Ensure that a controlled process is running, if possible.
 */
		while (ncontrolled && !pactive[0] && !pactive[1]) {
			dispatch_ctl();
		}
	}   /* for */
}


/*
 *	find_event
 *
 *	Function:	- Find the requested event in the table of events.
 *			- Etable is organized as a simple hash table with
 *			  linear search after collision (open addressing).
 *			- If requested event is not found, an empty event
 *			  descriptor will be initialized.
 *			- If no empty event descriptor can be found, an idle
 *			  empty (no blocked processes) is usurped.
 *
 *	Accepts:	- event to be found in event table
 *
 *	Returns:	- pointer to an event table entry
 */


struct kevent
*find_event(search_ev)

int			search_ev;		/* event to be found */

{
	int		i,			/* favourite index variable */
			key_ev,			/* current event number */
			ievent;			/* index in event table */
	struct kevent	*e;			/* fast local pointer */


/*
 * First try looking in the direct hash location.
 */
	ievent = hash(search_ev, EMAX);
	e = etable;

/*
 * ...then do a linear search.
 */
	for (i = EMAX; i; --i) {

		if ((key_ev = e[ievent].ev_event) == search_ev) {
			return(e + ievent);
		}

		else if (key_ev == NOTEVENT) {
			e[ievent].ev_event = search_ev;
			return(e + ievent);
		}

		ievent = rem((ievent + 1), EMAX);
	}

/*
 * The table is full - kick somebody out.
 */
	for (i = EMAX; i; --i) {

		if (!e[ievent].ev_head) {
			e[ievent].ev_event = search_ev;
			return(e + ievent);
		}

		ievent = rem((ievent + 1), EMAX);
	}

	errno = ENOEVDESCRIPTORS;
	return((struct kevent *) 0);
}


/*
 *	insert
 *
 *	Function:	- insert a process descriptor in a list sorted
 *			  by priority field
 *
 *	Accepts:	- pointer to top of list
 *			- pointer to new process descriptor
 *			- linked list index
 */

void
insert(top, bottom, new, ilist)

struct kprocess			**top,		/* top of list */
				**bottom,	/* bottom of list */
				*new;		/* new element in list */
int				ilist;		/* which list? */

{
	struct kprocess		*curr,		/* current chase pointer */
				*prev;		/* previous chase pointer */

/*
 * Take one good guess at where the process might be inserted.
 */
	if ((!(*bottom)) || (new->k_priority <= (*bottom)->k_priority)) {
		curr = NULL;
		prev = *bottom;
		*bottom = new;
	}

	else {
/*
 * Find point in list sorted by priority where new descriptor
 * should be inserted.
 */
		curr = prev = *top;

		while (curr && (new->k_priority <= curr->k_priority)) {
			prev = curr;
			curr = curr->k_pnext[ilist];
		}
	}

	new->k_pnext[ilist] = curr;

	if (curr == *top) {
		*top = new;
	}
		
	else {
		prev->k_pnext[ilist] = new;
	}
}


/*
 *	delete
 *
 *	Function:	- delete a process descriptor from a list
 *			  sorted by the priority field
 *
 *	Accepts:	- pointer to top of descriptor list
 *			- pointer to descriptor in list to be deleted
 *			- linked list index
 *
 *	Returns:	- pointer to the deleted descriptor
 */

void
delete(top, bottom, old, ilist)

struct kprocess			**top,		/* top of list */
				**bottom,	/* bottom of list */
				*old;		/* old element in list */
int				ilist;		/* which list? */

{
	struct kprocess		*curr,		/* current chase pointer */
				*prev;		/* previous chase pointer */

/*
 * Take one good guess at where the process might be.
 */
	if (*top == old) {
		curr = old;
		prev = NULL;
		*top = old->k_pnext[ilist];
	}

	else {
/*
 * Find old descriptor in list.
 */
		curr = prev = *top;

		while (curr != old) {
			prev = curr;
			curr = curr->k_pnext[ilist];
		}

		prev->k_pnext[ilist] = old->k_pnext[ilist];
	}

	if (*bottom == old) {
		*bottom = prev;
	}
}


/*
 *	getprocbypid
 *
 *	Function:	- get process pointer from process identifier
 *			- uses linear search through process list.
 *
 *	Accepts:	- process identifier
 *
 *	Returns:	- process pointer
 */

struct kprocess
*getprocbypid(pid)

int			pid;

{
	struct kprocess	*p;			/* favourite pointer variable */


	for (p = pprocess; p && p->k_pid != pid; p = p->k_pnext[PPROCESS]);

	if (!p) {
		errno = ENOTPROCESS;
	}

	return(p);
}


/*
 *	schedule
 *
 *	Function:	- identifies processes for dispatch
 *			- uncontrolled processes are dispatched immediately
 *			- controlled processes are compared with the highest
 *			  priority READY controlled process, which is updated
 *			  as necessary
 *			- The ncontrolled counter is an optimization that tells
 *			  us whether to bother searching the process list
 *			  for a controlled READY process.
 *
 *	Accepts:	- pointer to a READY process descriptor
 */

void
schedule(pclient)

struct kprocess		*pclient;		/* ptr to client process */

{
	if (pclient->k_priority) {
		++ncontrolled;
	}

	else {
/*
 * The client is an uncontrolled process, ready to execute.
 * However, there are two cases where we do not want to immediately dispatch
 * an uncontrolled process.  When an uncontrolled process syncs with a
 * controlled process, the uncontrolled process becomes effectively
 * controlled until it receives its turn to run based on its priority
 * (which is always 0, lower than all other legitimate controlled processes).
 * Therefore, we wish to bypass this immediate dispatch and let the
 * code in dispatch and dispatch_ctl take care of pclient.
 * These two cases are 1) when the uncontrolled process is in SSYNC state
 * and 2) when the uncontrolled process is in SREADY state after the
 * kdoom request with SIGNOOP.  Both cases are detected when the k_ptarget
 * field of the process descriptor points to a controlled process.
 */
		if (!pclient->k_ptarget || !pclient->k_ptarget->k_priority ||
			(pclient->k_state == SSTATE)) {

			dispatch(pclient);
		}

		else {
			++ncontrolled;
		}
	}
}


/*
 *	dispatch_ctl
 *
 *	Function:	- ensure that a controlled process is dispatched
 *			- for dramatic increases in simplicity, called after
 *			  every kernel request
 */

void
dispatch_ctl()

{
	struct kprocess	*p;			/* favourite pointer variable */

/*
 * Find the highest priority process that is READY.
 */
	for (p = pprocess; p->k_state > SREADY; p = p->k_pnext[PPROCESS]);

	--ncontrolled;
/*
 * Does the dispatched process require a second process to be dispatched?
 * This happens after a message rendez-vous when both synchronizing processes
 * are controlled processes.
 */
	if (p->k_state == SSYNC) {
		pactive[1] = p->k_ptarget;
		p->k_ptarget->k_signal |= sigmask(SIGBACK);
		dispatch(p->k_ptarget);
		--ncontrolled;
	}
/*
 * If a temporarily controlled process is now resuming its life
 * as an uncontrolled process, we don't want to note it as an active
 * controlled process.
 */
	if (p->k_priority) {
		pactive[0] = p;
	}

	dispatch(p);
}


/*
 *	dispatch
 *
 *	Function:	- dispatch a process
 *			- build reply structure and call "send_reply"
 */

void
dispatch(pclient)

struct kprocess		*pclient;

{
	struct kprocess	*p;			/* favourite pointer variable */

	p = pclient->k_ptarget;

	switch (pclient->k_state) {

		case SSYNC:
		reply.krm_port = p->k_port;
		reply.krm_aux = p->k_aux;
		reply.krm_length = pclient->k_size;
		reply.krm_type = p->k_type;
		reply.kr_reply = SUCCESS;
		break;

		case SATTACH:
		reply.kra_index = pclient->k_index;
		reply.kr_reply = SUCCESS;
		break;

		case SSTATE:
/*
 * Verify that the process ID still exists.  Notice that the snapshot of
 * process state is taken just before the requesting client is dispatched.
 * Clear the k_ptarget field in case the client is really an uncontrolled
 * process.  This will allow the process to be immediately dispatched on
 * subsequent kernel requests in dispatch().
 */
		if ((pclient->k_kqs_pid == p->k_pid) && (p->k_state != SFREE)) {
			reply.krs_state = p->k_state;
			reply.krs_priority = p->k_priority;
			reply.krs_event = p->k_event;
			reply.krs_type = p->k_type;
			reply.krs_signal = p->k_signal;
			reply.kr_reply = SUCCESS;
		}

		else {
			reply.kr_reply = ENOTPROCESS;
		}

		pclient->k_ptarget = NULL;
		break;

		case SERROR:
		reply.kr_reply = pclient->k_errno;
		break;

		default:
/*
 * The other cases are READY after kdetach and kdoom.
 */
		pclient->k_ptarget = NULL;
		reply.kr_reply = SUCCESS;
		break;
	}

	reply.kr_signal = pclient->k_signal;
	pclient->k_signal = 0;
	pclient->k_state = SACTIVE;
	send_reply(pclient->k_client);
}
