/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	kipc.c,v $
 * Revision 2.2  92/04/30  14:47:11  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/04  18:29:18  gdburns
 * Always update descriptor event and type.
 * 
 * Revision 2.1.1.1  91/10/09  15:06:23  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/03/20  11:36:28  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/09/27  18:01:06  gdburns
 * Converted to modern portable.h.
 * 
 * Revision 2.0.1.1  90/07/13  19:29:01  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:55:34  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:11:26  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0.1.1  88/01/12  10:53:46  gdburns
 * Added more comments.
 * 
 * Revision 1.0  87/12/19  15:46:37  trillium
 * Initial revision
 * 
 *
 *	Function:	- rendez-vous interprocess communication routines
 *
 *			  The kernel does not pass messages.  It simply makes
 *			  a rendez-vous between sending and receiving parties,
 *			  similar to that of a dating service.  When the
 *			  rendez-vous is made, the kernel replies to both
 *			  parties and the sender transmits its message directly
 *			  to the receiver.
 *
 *			  A rendez-vous is not made by one party specifying the
 *			  process ID of the other.  Instead, both parties
 *			  select an arbitrary number which they each know
 *			  (and which the programmer chooses) a priori.
 *			  When both a send and receive have been done on a
 *			  particular number, called an event number, the
 *			  highest priority sender is introduced to the highest
 *			  priority receiver through kernel replies.  The
 *			  sender receives from the kernel the communication
 *			  port of the receiver and the message transfer takes
 *			  place directly.
 */


#include <debug.h>
#include <kernel.h>
#include <kreq.h>
#include <typical.h>
#include <terror.h>
#include <portable.h>


/*
 * external data structures
 */
extern struct kreq	request;		/* kernel request */
extern struct kreply	reply;			/* kernel reply */

extern int		errno;

/*
 * external functions
 */
extern void		schedule();
extern void		insert();
extern void		delete();


/*
 *	ksync
 *
 *	Function:	- synchronize on an event number
 *			- kernel request
 *			- first parameter is event number
 *			- ksync can encounter one of two cases:
 *			
 *			  1) no one is blocked on event
 *			     - set client to blocked on event and
 *			       enqueue client on send/recv blocked list for 
 *			       event according to request type
 *
 *			  2) one or more processes are blocked on event
 *			     - rendez-vous with highest priority process
 *			       on send/recv blocked list for event according
 *			       to request type
 *
 *			- etable is organized as a simple hash table
 *			- fails if no event descriptors are available
 *			- two IF statements are required to distinguish
 *			  send requests from receive requests;  I wish
 *			  they weren't needed and if I wasn't so busy
 *			  I'd get rid of them somehow.
 *
 *	Accepts:	- pointer to requesting process descriptor
 */

void ksync(pclient)

struct kprocess			*pclient;	/* pointer to client */

{
	struct kprocess		**pblkd,	/* top of blocked process
						   list */
				**pblkd_tail,	/* bottom of blocked process
						   list */
				*p;		/* favourite pointer variable */
	struct kevent		*pevent,	/* pointer to event */
				*find_event();	/* find event in etable */
	int			ftry,		/* non-blocking request? */
				blocked;	/* blocked state */


	db(("kernel: ksync: client=%d, req=%d, event=%d, type=%x, length=%d\n", 
		pclient->k_pid, request.kq_request,
		request.kqm_event, request.kqm_type, request.kqm_length));
/*
 * Save client's auxiliary port.  The purpose of this is to facilitate
 * multiple virtual circuits.
 */
	pclient->k_aux = request.kqm_aux;
/*
 * Find the event in the event table, or allocate a new event.
 */
	pevent = find_event(request.kqm_event);

	if (!pevent) {
		pclient->k_state = SERROR;
		pclient->k_errno = errno;
		schedule(pclient);
		return;
	}

	pblkd = &(pevent->ev_head);
	pblkd_tail = &(pevent->ev_tail);
/*
 * Determine ksync characteristics.
 */
	ftry = request.kq_request & 1;
	blocked = (request.kq_request < KTRY_RECV) ? SSEND_BLKD : SRECV_BLKD;
/*
 * Search for a synchronizing process with a matching type.
 */
	for (p = *pblkd; p && ((p->k_state == blocked) ||
			(p->k_type && request.kqm_type &&
				!(p->k_type & request.kqm_type)));
		p = p->k_pnext[PBLKD]);

	pclient->k_type = request.kqm_type;
	pclient->k_event = request.kqm_event;
		
	if (!p) {
/*
 * We've NOT found a synchronizing process, so was the request blocking?
 */
		if (!ftry) {
/*
 * Sure thing, block the client.
 */
			pclient->k_size = request.kqm_length;
			insert(pblkd, pblkd_tail, pclient, PBLKD);
			pclient->k_state = blocked;
		}

		else {
/*
 * No way, let the client run again.
 */
			pclient->k_errno = EWOULDBLOCK;
			pclient->k_state = SERROR;
			schedule(pclient);
		}
	}

	else {
/*
 * There exists at least one blocked process which synchronizes.
 * Return the minimum length.
 */
		p->k_size = pclient->k_size =
			(p->k_size > request.kqm_length) ?
			request.kqm_length : p->k_size;
		pclient->k_ptarget = p;
		pclient->k_state = SSYNC;
		p->k_ptarget = pclient;
		p->k_state = SSYNC;
/*
 * Remove blocked process from blocked list.
 */
		delete(pblkd, pblkd_tail, p, PBLKD);
		schedule(pclient);
		schedule(p);
	}
}
