/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	kkenya.c,v $
 * Revision 2.2  92/04/30  14:47:13  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/04/08  20:03:07  gdburns
 * Rewrite kill file when client detaches.
 * 
 * Revision 2.1.1.2  92/02/23  22:38:19  gdburns
 * Record client process identifiers.
 * 
 * Revision 2.1.1.1  91/11/13  17:43:00  gdburns
 * Remove unportable pointer arithmetic.
 * 
 * Revision 2.1  91/03/20  11:36:29  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/01/14  15:30:11  gdburns
 * Fixed process state bug in doom routine.
 * 
 * Revision 2.0.1.2  90/07/13  19:29:02  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/05/15  14:24:17  gdburns
 * Correctly set pclient->k_index in case of re-attachment
 * in kattach().
 * 
 * Revision 2.0  90/02/20  11:55:35  trillium
 * Homo erectus
 *
 *	Function:	- process management routines (kenya)
 *
 *			- The kernel performs a very limited role in
 *			  process management.  It does not actually
 *			  create processes or kill them, but instead
 *			  attaches and detaches them by adding or deleting
 *			  descriptors to the active process list.  The
 *			  primary effect of limited kernel authority is
 *			  that processes can only be controlled while
 *			  they are making a kernel request.  An "active"
 *			  process is completely beyond the kernel's control.
 *
 *			  For example, the kdoom request can only deliver
 *			  signals immediately to blocked or ready processes.
 *			  Active processes will not receive their signal
 *			  until their next kernel request which will fail
 *			  with EINTR and the proper signal delivered.
 *			  An active process in an infinite loop devoid of
 *			  kernel request can receive signals and therefore
 *			  cannot be terminated.
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <debug.h>
#include <kernel.h>
#include <kreq.h>
#include <typical.h>
#include <terror.h>

/*
 * external functions
 */
extern struct kprocess	*getprocbypid();	/* find process by PID */
extern int		kkillrewrite();		/* rewrite kill file */
extern int		kkillwrite();		/* note PID for killing */
extern void		insert(),		/* insert process into list */
			delete(),		/* delete process from list */
			schedule(),		/* schedule processes */
			dispatch(),		/* give control to a process */
			send_reply();		/* kernel I/O module */

/*
 * external variables
 */
extern struct kreq	request;		/* kernel request */
extern struct kreply	reply;			/* kernel reply */
extern struct kprocess	*pfree,			/* top of free list */
			*pfree_tail,		/* bottom of free list */
			*pprocess,		/* top of process list */
			*pproc_tail;		/* bottom of process list */
extern int		errno;			/* global error number */
extern struct kprocess	ptable[];		/* the process table */

/*
 * local functions
 */
static int		unblock();		/* unblock process from event */
			
/*
 *	kattach
 *
 *	Function:	- attach yourself to the kernel
 *			- kernel request
 *			- fails if process is already attached or
 *			  kernel is out of process descriptors
 *
 *	Accepts:	- ptr to process descriptor, NULL if first attach
 */

void kattach(pclient)

struct kprocess	*pclient;

{
	int		i;		/* favourite index */

	db(("kernel: kattach: client=%d, port=%d, priority=%d\n",
		request.kqa_pid, request.kqa_port, request.kqa_priority));
/*
 * Is the client attempting to change priority by reattaching?
 */
	if (pclient) {
		delete(&pprocess, &pproc_tail, pclient, PPROCESS);
		pclient->k_index = request.kq_index;
	}

	else {
/*
 * Do we have any process descriptors available?
 */
		if (pfree == NULL) {
			reply.kr_reply = ENOKDESCRIPTORS;
			reply.kr_signal = 0;
			send_reply(request.kqa_client);
			return;
		}
/*
 * Proceed with attaching the new process.
 */
		pclient = pfree;
		delete(&pfree, &pfree_tail, pfree, PPROCESS);
		pclient->k_pid = request.kqa_pid;
		pclient->k_port = request.kqa_port;
		pclient->k_client = request.kqa_client;
		pclient->k_signal = 0;
		pclient->k_event = NOTEVENT;
/*
 * Find the actual table index of the process descriptor.  Use BFI.
 */
		for (i = 0; pclient != &ptable[i]; ++i) {
		}

		pclient->k_index = i;
	}

	pclient->k_state = SATTACH;
	pclient->k_priority = request.kqa_priority;
	insert(&pprocess, &pproc_tail, pclient, PPROCESS);
	schedule(pclient);
	(void) kkillwrite(pclient->k_pid);
}

/*
 *	kdetach
 *
 *	Function:	- detach a process from the kernel
 *			- kernel request
 *			- fails if process is not attached
 *
 *	Accepts:	- pointer to client's process descriptor
 */

void kdetach(pclient)

struct kprocess			*pclient;		/* ptr to client proc */

{
	struct kprocess		*pdetach;		/* ptr to target proc */


	db(("kernel: kdetach: client=%d, target=%d\n",
		pclient->k_pid, request.kqu_pid));
/*
 * Identify the target process.
 */
	pdetach = (request.kqu_pid) ? getprocbypid(request.kqu_pid) : pclient;
/*
 * Find target process and detach it.
 * The option of detaching another process is offered as a clean-up
 * mechanism only.  Detaching an active, healthy process will probably
 * blow up real good.
 */
	if (pdetach) {
		(void) unblock(pdetach);
		pdetach->k_state = SREADY;
		dispatch(pdetach);
		delete(&pprocess, &pproc_tail, pdetach, PPROCESS);
		insert(&pfree, &pfree_tail, pdetach, PPROCESS);
		pclient->k_state = SREADY;
		pdetach->k_state = SFREE;

		if (pclient != pdetach) {
			schedule(pclient);
		}

		(void) kkillrewrite();
	}

	else {
		pclient->k_errno = errno;
		pclient->k_state = SERROR;
		schedule(pclient);
	}

}


/*
 *	kdoom
 *
 *	Function:	- deliver a signal to the specified process 
 *			- kernel request
 *
 *	Accepts:	- pointer to client's process descriptor
 */

void kdoom(pclient)

struct kprocess			*pclient;	/* pointer to client */

{
	struct kprocess		*pdoomed;	/* pointer to doomed process */


	db(("kernel: kdoom: client=%d, target=%d, signal=%d\n",
		pclient->k_pid, request.kqd_pid, request.kqd_signal));
/*
 * Identify the target process.
 */
	pdoomed = (request.kqd_pid) ? getprocbypid(request.kqd_pid) : pclient;
/*
 * Find target process and doom it.
 */
	if (pdoomed) {
		pdoomed->k_signal |= request.kqd_signal;

		if (pdoomed == pclient) {
			pclient->k_state = SREADY;
		}

		else if (unblock(pdoomed)) {
			pdoomed->k_errno = EINTR;
			pdoomed->k_state = SERROR;
			schedule(pdoomed);
		}
	}

	else {
		pclient->k_errno = errno;
		pclient->k_state = SERROR;
	}

	schedule(pclient);
}


/*
 *	unblock
 *
 *	Function:	- unblock a process from an event
 *
 *	Accepts:	- pointer to process descriptor to be unblocked
 *
 *	Returns:	- flag set if the process was blocked
 */

static int unblock(pblocked)

struct kprocess			*pblocked;	/* pointer to blocked client */

{
	struct kevent		*pevent,	/* pointer to an event */
				*find_event();	/* find event in etable */


	if ((pblocked->k_state == SSEND_BLKD) ||
		(pblocked->k_state == SRECV_BLKD)) {

		pevent = find_event(pblocked->k_event);
		delete(&(pevent->ev_head), &(pevent->ev_tail), pblocked, PBLKD);
		return(TRUE);
	}

	else {
		return(FALSE);
	}
}


/*
 *	kstate
 *
 *	Function:	- get the state of a specified process
 *			- kernel request
 *			- replies to client with fields from the specified
 *			  process's descriptor:
 *
 *			  state, priority, blocked event, signals pending
 *
 *	Accepts:	- pointer to requesting process descriptor
 */

void kstate(pclient)

struct kprocess			*pclient;	/* pointer to client */

{
	struct kprocess		*pstate;	/* pointer to state process */


	db(("kernel: kstate: client=%d, target=%d\n",
		pclient->k_pid, request.kqs_pid));
/*
 * Identify target process.
 */
	pstate = (request.kqs_pid) ? getprocbypid(request.kqs_pid) : pclient;

	if (pstate) {
		pclient->k_ptarget = pstate;
		pclient->k_state = SSTATE;
		pclient->k_kqs_pid = pstate->k_pid;
	}

	else {
		pclient->k_errno = errno;
		pclient->k_state = SERROR;
	}

	schedule(pclient);
}
