/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	ksignal.c,v $
 * Revision 2.2  92/04/30  14:47:38  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:37:05  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:27:45  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:42  trillium
 * Homo erectus
 * 
 *
 *	Function:	- signal package
 *			- Any kernel request may return one or more signals.
 *			  A signal is indicated by one bit in a word.  When a
 *			  signal is turned on, it is vectored through a table of *			  signal handlers.  By default most signals are ignored. *			  Some cause the process to terminate and a few may have *			  meanings particular to specific Trillium services,
 *			  such as the network layer.  Most signals may be
 *			  redirected by the user to user supplied handling
 *			  routines.
 *			- for compatiblity with Unix, the signals start at 1
 */


#include <events.h>
#include <kio.h>
#include <kmsg.h>
#include <terror.h>
#include <typical.h>


/*
 * external variables
 */
extern struct kio_t	_kio;			/* Kernel Input/Output block */
extern int		errno;			/* global error variable */

/*
 * external functions
 */
extern int		kdoom();		/* deliver a signal */

/*
 * local functions
 */
static int		sigign();		/* ignore signal */
static int		sigback();		/* go back to kernel */
static int		sigexit();		/* terminate process */
static int		sigstop();		/* stop process */

/*
 * defined global functions
 */
void			_follow_signal();	/* follow set signals */
void			_siginit();		/* initialize signal module */
int			kpause();		/* wait for a signal */
int			(*ksignal())();		/* redirect signals */
int			ksigsetmask();		/* install entire block mask */
int			ksigblock();		/* set selected block bits */
int			ksigsetretry();		/* install entire retry mask */
int			ksigretry();		/* set selected retry bits */
			  

/*
 *	_follow_signal
 *
 *	Function:	- follow signals that may be set in signal word
 */

void _follow_signal()

{
	int		i;		/* favourite index variable */
	int		sigbit;		/* signal bit */


	if (_kio.ki_signal) {
/*
 * Check every possible signal.
 */
		for (i = 1; i <= TNSIG; ++i) {
/*
 * If the signal is turned on, call the handling routine.
 */
			sigbit = sigmask(i);

			if ((_kio.ki_signal & sigbit) &&
				!(_kio.ki_sigmask & sigbit)) {

				_kio.ki_signal &= ~sigbit;
				(*(_kio.ki_sig_handlers[i - 1]))(i, 0, 0);
			}
		}
	}
}


/*
 *	ksignal
 *
 *	Function:	- simple software signal director
 *			- redirects a signal to a user supplied handling
 *			  routine
 *
 *	Accepts:	- signal number
 *			- handler function
 *
 *	Returns:	- previous handler function
 */

int
(*(ksignal(signo, sigfunc)))()

int			signo;
int			(*sigfunc)();

{
	int		(*oldsigfunc)();


	if ((signo < 1) || (signo > TNSIG)) {
		errno = EINVAL;
		return(BADSIG);
	}

	if (sigmask(signo) & SIG_PRIV) {
		errno = EINVAL;
		return(BADSIG);
	}

	--signo;
	oldsigfunc = _kio.ki_sig_handlers[signo];

	if (sigfunc == SIG_DFL) {
		_kio.ki_sig_handlers[signo] = _kio.ki_sig_default[signo];
	}

	else if (sigfunc == SIG_IGN) {
		_kio.ki_sig_handlers[signo] = sigign;
	}

	else {
		_kio.ki_sig_handlers[signo] = sigfunc;
	}

	return(oldsigfunc);
}


/*
 *	_siginit
 *
 *	Function:	- initialize signal handling table
 */

void _siginit()

{
	int		i;			/* favourite index variable */

/*
 * Retry all signals except SIGRELEASE, SIGBACK and SIGNOOP.
 * Don't block any signals.
 * No signals are currently pending.
 */
	_kio.ki_sigretry = ~(sigmask(SIGRELEASE) | sigmask(SIGBACK) |
			sigmask(SIGNOOP));
	_kio.ki_sigmask = 0;
	_kio.ki_signal = 0;
/*
 * Ignore most signals...
 */
	for (i = 0; i < TNSIG; ++i) {
		_kio.ki_sig_handlers[i] = sigign;
	}
/*
 * ...except for a few special cases.
 */
	_kio.ki_sig_handlers[SIGBACK - 1] = sigback;
	_kio.ki_sig_handlers[SIGUDIE - 1] = sigexit;
	_kio.ki_sig_handlers[SIGARREST - 1] = sigstop;
/*
 * Now copy the current handling table to the default table.
 */
	for (i = 0; i < TNSIG; ++i) {
		_kio.ki_sig_default[i] = _kio.ki_sig_handlers[i];
	}
}


/*
 *	sigign
 *
 *	Function:	- ignore a signal
 */

static int sigign() {}


/*
 *	sigback
 *
 *	Function:	- go back into the kernel by signalling myself
 *			  with the no-operation signal
 */

static int sigback()

{
	(void) kdoom(0, SIGNOOP);
}


/*
 *	sigexit
 *
 *	Function:	- terminate the process
 */

static int sigexit()

{
	kexit(1);
}


/*
 *	sigstop
 *
 *	Function:	- stop or pause the process
 *			- uses the EVSTOP or EVPAUSE event to indefinitely
 *			  hang until a future signal is received
 *
 *	Accepts:	- signal that caused sigstop to be called
 */

static int sigstop(sig)

int			sig;

{
	struct kmsg	forever;		/* local message descriptor */
	char		just_in_case;		/* unused message buffer */


	forever.k_event = (sig == SIGARREST) ? EVSTOP : EVPAUSE;
	forever.k_type = 0;
	forever.k_flags = 0;
	forever.k_length = sizeof(just_in_case);
	forever.k_msg = &just_in_case;
	(void) krecv(&forever);
}


/*
 *	kpause
 *
 *	Function:	- wait for a signal to occur
 *			- implemented by blocking on an event
 *			- must disable all retry flags while pausing
 *
 *	Returns:	- ERROR
 */

int
kpause()

{
	int		save_retry;

	save_retry = _kio.ki_sigretry;
	_kio.ki_sigretry = 0;
	(void) sigstop(0);
	_kio.ki_sigretry = save_retry;
	errno = EINTR;
	return(ERROR);
}


/*
 *	ksigsetmask
 *	ksigblock
 *	ksigsetretry
 *	ksigretry
 *
 *	Function:	- operate on signal inhibiting mask and kernel request
 *			  retry flags
 *	Returns:	- old value of mask or flags
 */

int ksigsetmask(mask)

int			mask;

{
	int		oldmask;		/* previous signal mask */

	oldmask = _kio.ki_sigmask;
	mask &= ~SIG_PRIV;
	_kio.ki_sigmask = mask;
/*
 * Some pending signals may now be unblocked.
 */
	if (_kio.ki_signal) {
		_follow_signal();
	}

	return(oldmask);
}

int ksigblock(mask)

int			mask;

{
	int		oldmask;		/* previous signal mask */

	oldmask = _kio.ki_sigmask;
	mask &= ~SIG_PRIV;
	_kio.ki_sigmask |= mask;
	return(oldmask);
}

int ksigsetretry(flags)

int			flags;

{
	int		oldflags;		/* previous retry flags */

	oldflags = _kio.ki_sigretry;
	flags &= ~(SIG_PRIV | SIGRELEASE);
	_kio.ki_sigretry = flags;
	return(oldflags);
}

int ksigretry(flags)

int			flags;			/* new retry flags */

{
	int		oldflags;		/* previous retry flags */


	oldflags = _kio.ki_sigretry;
	flags &= ~(SIG_PRIV | SIGRELEASE);
	_kio.ki_sigretry |= flags;
	return(oldflags);
}
