/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	ktrace.c,v $
 * Revision 2.2  92/04/30  14:47:43  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/01/10  17:39:15  gdburns
 * Add TID, the process's trace identifier.
 * 
 * Revision 2.1  91/10/03  16:36:56  gdburns
 * initial version
 * 
 *	Function:	- runtime trace package
 *			- an internal package, hence does not check or
 *			  return many error conditions
 */

#include <kio.h>
#include <ktrace.h>
#include <net.h>

/*
 * external variables
 */
extern struct kio_t	_kio;			/* Kernel Input/Output block */
extern int		errno;			/* global error variable */
/*
 * external functions
 */
extern int		atkexit();		/* register cleanup func */
extern int		(*ksignal())();		/* redirect signals */
extern int		rtrattach();
extern int		rtrdetach();
extern int		rtrstore();
extern char		*malloc();
extern char		*memcpy();
/*
 * local functions
 */
static void		tralloc();		/* allocate trace buffers */
/*
 * global functions
 */
void			ktrcflow();		/* cflow trace */
void			ktrdflow();		/* dflow trace */
void			ktrexit();		/* trace cleanup */
void			ktrinit();		/* initialize traces */
void			ktrstart();		/* start trace */
void			ktrstop();		/* end trace */
void			ktrunload();		/* unload traces to traced */

/*
 *	ktrinit
 *
 *	Function:	- initializes trace system
 */
void
ktrinit()

{
	int		i;			/* favourite index */
/*
 * Initialize trace buffer header array.
 * Its size limits the total size of trace data per process
 * since each individual trace buffer is MAXNMSGLEN long.
 */
	_kio.ki_trcurr = 0;
	_kio.ki_trhead = (struct trhead *) malloc((unsigned)
		(TRHMAX * sizeof(struct trhead)));

	if (! _kio.ki_trhead) {
		kexit(errno);
	}

	for (i = 0; i < TRHMAX; ++i) {
		_kio.ki_trhead[i].trh_length = 0;
		_kio.ki_trhead[i].trh_buf = (char *) 0;
	}
/*
 * Vector SIGTRACE to unload trace data to the local daemon.
 * Register a cleanup function.
 */
	(void) ksignal(SIGTRACE, ktrunload);
	(void) atkexit(ktrexit, (char *) 0);
/*
 * Allocate the first few trace buffers.
 */
	tralloc(0);
/*
 * Attach to traced.
 */
	if (rtrattach(LOCAL) < 0) {
		kexit(errno);
	}
/*
 * Kick things off with a trace start record.
 */
	ktrstart(getnodeid());
}

/*
 *	tralloc
 *
 *	Functions:	- allocates memory for the next few trace buffers
 *			- next TRHQUAN buffers are allocated
 *			- assumes TRHQUAN evenly divides TRHMAX
 *
 *	Accepts:	- index of first trace buffer ptr
 */
static void
tralloc(idx)

int			idx;

{
	int		i;
	int		imax;
	char		*p;

	p = malloc((unsigned) (MAXNMSGLEN * TRHQUAN));
	if (! p) kexit(errno);

	imax = idx + TRHQUAN;

	for (i = idx; i < imax; ++i) {
		_kio.ki_trhead[i].trh_buf = p;
		p += MAXNMSGLEN;
	}
}

/*
 *	ktrunload
 *
 *	Function:	- unloads all used trace buffers to the local
 *			  trace daemon
 */
void
ktrunload()

{
	int		i;

	for (i = 0; i < TRHMAX; ++i) {

		if (_kio.ki_trhead[i].trh_length > 0) {

			if (rtrstore(LOCAL, _kio.ki_trhead[i].trh_buf,
				_kio.ki_trhead[i].trh_length) < 0) {

				kexit(errno);
			}

			_kio.ki_trhead[i].trh_length = 0;
		}
	}
}

/*
 *	ktrpoint
 *
 *	Function:	- determines where the next trace can be stored
 *			- calls tralloc, if necessary
 *			- may permanently turn off tracing if all buffers fill
 *
 *	Accepts:	- size of next trace
 *
 *	Returns:	- ptr to next trace location
 */
char *
ktrpoint(size)

int			size;

{
	struct trhead	*p;

	p = _kio.ki_trhead + _kio.ki_trcurr;
/*
 * If the current trace buffer is filled, move to the next one.
 */
	if ((p->trh_length + size) > MAXNMSGLEN) {

		_kio.ki_trcurr += 1;
		p++;
/*
 * If the next trace buffer is unallocated, allocate it plus a few more.
 */
		if (p->trh_buf == (char *) 0) {
			tralloc(_kio.ki_trcurr);
		}
/*
 * If the next trace buffer is the last, store an end trace & shut off tracing.
 * This means that the last buffer will only hold two traces, the
 * price we pay for having very long dflow traces.
 */
		if (_kio.ki_trcurr >= (TRHMAX - 1)) {
			ktrstop();
			_kio.ki_rtf &= ~RTF_TRACE;
		}
	}
/*
 * Increment the trace buffer length before passing the address to be stored.
 */
	p->trh_length += size;
	return(p->trh_buf + (p->trh_length - size));
}

/*
 *	ktrexit
 *
 *	Function:	- cleanup trace system prior to process exit
 */
void
ktrexit()

{
	int		i;
/*
 * Store a trace end.
 */
	ktrstop();
/*
 * Unload all traces to the local daemon.
 */
	ktrunload();
/*
 * Detach from the local traced.
 */
	(void) rtrdetach(LOCAL);
/*
 * Free all trace buffers.
 */
	for (i = 0; i < TRHMAX; ++i) {

		if (_kio.ki_trhead[i].trh_buf != (char *) 0) {
			(void) free(_kio.ki_trhead[i].trh_buf);
		}
	}
/*
 * Free the trace header array.
 */
	(void) free((char *) _kio.ki_trhead);
}

/*
 *	ktrstart
 *
 *	Function:	- enables tracing and stores a start trace
 *	Accepts:	- trace ID
 */
void
ktrstart(tid)

int4			tid;

{
	struct trstart	*p;

	if (! (_kio.ki_rtf & RTF_TRACE)) {
		return;
	}

	p = (struct trstart *) ktrpoint(sizeof(struct trstart));
	p->tr1_head.trt_type = TRSTART;
	p->tr1_head.trt_stamp = ldtimer();
	p->tr1_head.trt_tid = tid;
	p->tr1_node = getnodeid();
	p->tr1_pid = getpid();

	_kio.ki_tid = tid;
	_kio.ki_rtf |= RTF_TRSWITCH;
}

/*
 *	ktrstop
 *
 *	Function:	- stores an end trace and disables tracing
 */
void
ktrstop()

{
	struct trend	*p;

	if ((_kio.ki_rtf & RTF_TRON) != RTF_TRON) {
		return;
	}

	p = (struct trend *) ktrpoint(sizeof(struct trend));
	p->tr2_head.trt_type = TREND;
	p->tr2_head.trt_stamp = ldtimer();
	p->tr2_head.trt_tid = _kio.ki_tid;

	_kio.ki_rtf &= ~RTF_TRSWITCH;
}

/*
 *	ktrcflow
 *
 *	Function:	- stores a control flow trace
 *
 *	Accepts:	- identifier for control flow
 *			- arbitrary data item, usually progression counter
 */
void
ktrcflow(type, data)

int4			type;
int4			data;

{
	struct trcflow	*p;

	if ((_kio.ki_rtf & RTF_TRON) != RTF_TRON) {
		return;
	}

	p = (struct trcflow *) ktrpoint(sizeof(struct trcflow));
	p->tr3_head.trt_type = TRCFLOW;
	p->tr3_head.trt_stamp = ldtimer();
	p->tr3_head.trt_tid = _kio.ki_tid;
	p->tr3_type = type;
	p->tr3_data = data;
}

/*
 *	ktrdflow
 *
 *	Function:	- stores a data flow trace
 *
 *	Accepts:	- identifier for data flow
 *			- size of arbitrary data in words
 *			- arbitrary data array ptr
 */
void
ktrdflow(type, size, data)

int4			type;
int4			size;
int4			*data;

{
	struct trdflow	*p;

	if ((_kio.ki_rtf & RTF_TRON) != RTF_TRON) {
		return;
	}

	if (size > (MAXNMSGLEN - sizeof(struct trend))) {
		return;
	}

	p = (struct trdflow *)
		ktrpoint(sizeof(struct trdflow) + ((size - 1) * 4));
	p->tr4_head.trt_type = TRDFLOW;
	p->tr4_head.trt_stamp = ldtimer();
	p->tr4_head.trt_tid = _kio.ki_tid;
	p->tr4_type = type;
	p->tr4_size = size;
	(void) memcpy((char *) p->tr4_data, (char *) data, size * 4);
}
