/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1989 Cornell University
 *
 *	JRB/GDB
 *
 *	$Log:	rmalloc.c,v $
 * Revision 2.2  92/04/30  14:53:19  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/11/10  23:35:19  gdburns
 * Minor overhaul fixes several problems.
 * 
 * Revision 2.1.1.1  91/06/03  16:36:22  raja
 * Corrected more signal masking in ksigblock().
 * 
 * Revision 2.1  91/03/20  11:38:42  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:29:45  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:11:40  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:55:57  trillium
 * Homo erectus
 * 
 * Revision 1.1.1.2  88/12/21  14:53:22  fielding
 * Change old rmalloc to use correct arguments for new rmalloc
 * 
 * Revision 1.1.1.1  88/07/07  13:55:10  braner
 * checks for amount==0,
 * calls _memrec if successful
 * added _memrec and _free
 * rfree calls _free to delete entry
 * 
 * Revision 1.1  88/06/05  01:13:49  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:35  trillium
 * Initial revision
 * 
 *
 *	Function:	- itb memory allocater runtime interface
 *			- memory is allocated in regions
 */


#include <net.h>
#include <mareq.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <terror.h>


/*
 * external variables
 */
extern int			errno;		/* global error variable */

/*
 * external functions
 */
extern int			atkexit();

/*
 * global functions
 */
char				*rmalloc();
int				rfree();
int				rfreeall();
int				lfreeall();


/*
 *	rmalloc
 *
 *	Function:	- allocate memory from a given region
 *			- works remotely
 *
 *	Accepts:	- amount of memory to allocate
 *			- region from which to allocate
 *			- target node relative address
 *			- address alignment value
 *
 *	Returns:	- pointer to requested memory
 *			- null pointer is an error if errno is non-zero
 */

char *
rmalloc (node, amount, region, align)

unsigned int			amount;		/* amount to allocate */
int				region,		/* allocate in this region */
				node,		/* target node address */
				align;		/* memory alignment */

{
	struct nmsg		nhead;		/* network message desc. */
	struct mareply		*reply;		/* malloc reply */
	struct mareq		*request;	/* malloc request */
	int			mask;


	if (amount == 0) {
		errno = EINVAL;
		return ((char *) 0);
	}
	
 	request = (struct mareq *) nhead.nh_data;
	reply = (struct mareply *) nhead.nh_data;	
	
	request->maq_src_node = (node == LOCAL) ? node : getnodeid();
	request->maq_src_event = -getpid();
	request->maq_region = region;
	request->maq_amount = amount;
	request->maq_align = align;
	request->maq_request = MAQALLOC;
	request->maq_pid = getpid();
	request->maq_node = request->maq_src_node;

	nhead.nh_node = node;
	nhead.nh_event = EVMALLOCD;
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return((char *) 0);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (nrecv(&nhead)) {
		return((char *) 0);
	}

	(void) ksigsetmask(mask);

	if (reply->mar_reply == 0) {

		if (node == LOCAL) {

			if (atkexit(lfreeall, 0)) {
				return((char *) 0);
			}
		}

		return(reply->mar_address);
	}

	else {
		errno = reply->mar_reply;
		return((char *) 0);
	}
}


/*
 *	rfree.c
 *
 *	Function:	- itb memory deallocater runtime interface
 *			- free memory starting from a given address
 *			- works remotely
 *
 *	Accepts:	- address of block to be deallocated
 *
 *	Returns:	- 0 or ERROR if address does not point to
 *			  valid allocated block
 */

int
rfree (node, address)

char				*address;	/* address to allocate */
int				node;		/* target node address */

{
	struct nmsg		nhead;		/* network message desc. */
	struct mareq		*request;	/* malloc request */
	struct mareply		*reply;		/* malloc reply */
	int			mask;


	if (address == 0) {
		errno = EINVAL;
		return(ERROR);
	}

 	request = (struct mareq *) nhead.nh_data;
	reply = (struct mareply *) nhead.nh_data;	

	request->maq_src_node = (node == LOCAL) ? node : getnodeid();
	request->maq_src_event = -getpid();
	request->maq_address = (int) address;
	request->maq_request = MAQFREE;
	request->maq_pid = getpid();
	request->maq_node = request->maq_src_node;

	nhead.nh_node = node;
	nhead.nh_event = EVMALLOCD;
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead) == ERROR) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (nrecv(&nhead) == ERROR) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	if (reply->mar_reply == 0) {
		return(0);
	}

	else {
		errno = reply->mar_reply;
		return(ERROR);
	}
}


/*
 *	rfreeall
 *
 *	Function:	- itb memory deallocater runtime interface
 *			- free all memory allocated to given pid and my nodeid
 *			- works remotely
 *
 *	Accepts:	- target node
 *			- target pid
 *
 *	Returns:	- 0 or ERROR
 */

int
rfreeall (node, pid)
 
int                             node;		/* target node address */
int				pid;		/* pid to be freed */
 
{
        struct nmsg             nhead;		/* network message desc. */
        struct mareq            *request;	/* malloc request */
        struct mareply          *reply;		/* malloc reply */
	int			mask;
 

 	request = (struct mareq *) nhead.nh_data;
	reply = (struct mareply *) nhead.nh_data;	
	
        request->maq_src_node = (node == LOCAL) ? node : getnodeid();
        request->maq_src_event = -getpid();
        request->maq_request = MAQFREEALL;
        request->maq_pid = pid;
        request->maq_node = request->maq_src_node;
 
        nhead.nh_node = node;
        nhead.nh_event = EVMALLOCD;
        nhead.nh_type = 0;
        nhead.nh_flags = NOBUF;
        nhead.nh_length = 0;
        nhead.nh_msg = (char *) 0;
 
        if (nsend(&nhead)) {
                return(ERROR);
        }
 
	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
        nhead.nh_event = -getpid();
 
        if (nrecv(&nhead)) {
                return(ERROR);
        }

	(void) ksigsetmask(mask);
 
        if (reply->mar_reply == 0) {
                return(0);
        }
 
        else {
                errno = reply->mar_reply;
                return(ERROR);
        }
}


/*
 *	lfreeall
 *
 *	Function:	- local call to rfreeall()
 */

int
lfreeall()

{
	return(rfreeall(LOCAL, getpid()));
}
