/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	cboot.c,v $
 * Revision 2.2  92/04/30  14:54:13  trillium
 * Ohio Release
 * 
 *	Function:	- full blown single program loader
 *			- speaks to moses on a neighbour itb node
 *			- -m <fd> writes the next available address
 *			  after the booted program to filed desc, <fd>
 *			- Sboot reads back the output from cboot -m.
 *			- load module is always stripped, symbol and
 *			  relocation tables are never loaded
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 *
 *			- memory layout of cboot, ignoring -t and -w
 *			  possibilities
 *
 *			  -------------------------	-m output
 *			  | argv strings          | 
 *			  -------------------------
 *			  | argv pointers         |
 *			  -------------------------
 *			  | PCB			  |
 *			  | pointer to PCB	  |
 *			  | return address	  |
 *			  | stack		  |
 *			  -------------------------
 *			  | static data/bss       |
 *			  -------------------------
 *			  | header/text           |
 *			  -------------------------	-l <address>
 */

#include <stdio.h>
#include <boot.h>
#include <bootd.h>
#include <bootreq.h>
#include <fcntl.h>
#include <kio.h>
#include <machine/maconfig.h>
#include <moses.h>
#include <net.h>
#include <pcb.h>
#include <t.out.h>
#include <t_types.h>
#include <typical.h>
#include <unistd.h>

#undef NULL
#define	NULL		((char *) 0)	

/*
 * external functions
 */
extern char		*findpath();
extern char		*getparam();
extern char		*malloc();
extern char		**pack_argv();
extern int		do_args();
extern int		get_2nd_let();
extern int		read();
extern int		write();
extern int		OpenLink();	/* open device */
extern int		CloseLink();	/* close device */
extern int		WriteLink();	/* write to device */
extern void		addarg();	/* add arg to arg vector */
extern void		followed();
extern void		validopts();
extern void		toutpatch();
extern void		toutrev();
extern void		twoletter();
extern long		lseek();

/*
 * local functions
 */
static void		ewrite();
static void		help();
static void		load();
static void		moses_exec();
static void		setpaths();	/* initialize search paths */
static void		writew();
static char		*eread();

/*
 * local variables
 */
static char		*me;		/* name I was exec'd by */
static char		**p_prog;	/* bootstrap search paths */
static int		fl_verbose;	/* Be verbose? */
static int		n_prog;		/* count of p_prog */
static int		(*xwrite)();	/* appropriate write function */

static char		*usage = "usage: %s [<options>] <program> <device>\n";

/*
 * external variables
 */
extern int		errno;
extern int		argtailc;	/* # args after -- */
extern char		**argtailv;	/* args after -- */


main(argc, argv)

int			argc;
char			*argv[];

{
	struct exec	*tout;		/* executable info header */
	struct exec	rtout;		/* remote executable info header */
	char		*ifname;	/* input filename */
	char		*ofname;	/* output filename */
	char		*ifullname;	/* input full pathname */

	char		**pargv;	/* ptr to packed arg vectors */
	char		*parg;		/* a temporary when relocating pargv */
	char		*pdata;		/* data segment buffer */
	char		*preloc;	/* relocate table buffer */
	char		*psyms;		/* symbol table buffer */
	char		*ptext;		/* text segment buffer */
	char		*p;		/* favourite ptr */

	int		ifd;		/* input file descriptor */
	int		ofd;		/* output file descriptor */
	int		mfd;		/* highwater file descriptor */

	int		argpt;		/* loadpt of single argument */
	int		argvpt;		/* loadpt of argv vector */
	int		codept;		/* loadpt of code */
	int		datapt;		/* loadpt of data */
	int		entrypt;	/* entrypt of code */
	int		loadpt;		/* primary loadpt */
	int		spawnpt;	/* loadpt of spawn table entry */
	int		stackpt;	/* loadpt of stack */
	int		waterpt;	/* new high water mark */

	int		avsize;		/* total argument vector size */
	int		stksize;	/* stack size */
	int		wp;		/* workspace address */
	int		i;		/* favourite index */
	int		ispawn;		/* index in spawn table */
	int		priority;	/* process priority */

	struct pcb_t	pcb;		/* Process Control Block */
	struct btprog	spawn;		/* spawn table entry */

	me = argv[0];
/*
 * Initialize option parser.
 */
	validopts("hvL");
	followed("iklmopstw");
	twoletter("b:cr");

	if (do_args(&argc, argv) < 0) {
		fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	if (opt_taken('l')) {
		intparam('l', &loadpt);
	} else {
		fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('w')) {
		intparam('w', &stackpt);
	} else {
		stackpt = 0;
	}

	if (opt_taken('s')) {
		intparam('s', &stksize);
	} else {
		fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('k')) {
		intparam('k', &priority);
	} else {
		priority = PRIORITY;
	}

	fl_verbose = opt_taken('v');

	--argc;
	++argv;
	setpaths();
	ifname = NULL;
	ofname = NULL;
/*
 * Check for a single file argument with no decidability between input
 * and output.
 */
	if (!opt_taken('o') && !opt_taken('i') && (argc == 1)) {
		fprintf(stderr, usage, me);
		exit(1);
	}
/*
 * Establish input file name.
 */
	if (opt_taken('i')) {
		ifname = getparam('i');
	} else if (argc > 0) {
		ifname = argv[0];
		--argc;
		++argv;
	}
/*
 * Establish output file name.
 */
	if (opt_taken('o')) {
		ofname = getparam('o');
	} else if (argc > 0) {
		ofname = argv[0];
		--argc;
		++argv;
	}

	if (argc > 0) {
		fprintf(stderr, usage, me);
		exit(1);
	}
/*
 * Establish input file descriptor.
 */
	if (ifname == NULL) {
		ifd = 0;
		ifname = "stdin";
	} else {
	    	if ((ifullname = findpath(ifname, p_prog, R_OK | X_OK))
				== NULL) {
			fprintf(stderr, "%s: can't find %s\n", me, ifname);
			exit(1);
		}

		ifname = ifullname;
		ifd = open(ifname, O_RDONLY, 0);

		if (ifd < 0) {
			perror(ifname);
			exit(errno);
		}
	}

	if (fl_verbose) fprintf(stderr, "%s: input is %s\n", me, ifname);
/*
 * Read in the executable info structure.
 */
	tout = (struct exec *) eread(ifd, (unsigned long) sizeof(struct exec));
/*
 * Verify magic number of a valid executable file.
 */
	if ((tout->a_magic != TO_MAGIC_TN_X) &&
		(tout->a_magic != TO_MAGIC_T4_X) &&
		(tout->a_magic != TO_MAGIC_T8_X)) {

		fprintf(stderr, "%s: %s: bad magic number\n", me, ifname);
		exit(1);
	}
/*
 * Establish output file descriptor.
 */
	if (ofname == NULL) {
		xwrite = write;
		ofd = 1;
		ofname = "stdout";
	} else {

		if (opt_taken('L')) {
			xwrite = WriteLink;
#ifdef PARSY
/*
 * Replace '/' with ' ' in Parsytec link name.
 * Thanks to the person who decided to embed blanks in a single identifier.
 */
			for (p = ofname; *p; ++p) if (*p == '/') *p = ' ';
#endif
/*
 * Open the device.
 */
			ofd = OpenLink(ofname);

			if (ofd < 0) {
				fprintf(stderr, "%s: cannot open %s\n",
					me, ofname);
				exit(1);
			}
		}

		else {
			xwrite = write;
			ofd = open(ofname, O_WRONLY | O_CREAT | O_APPEND, 0644);

			if (ofd < 0) {
				fprintf(stderr, "%s: cannot open %s\n",
					me, ofname);
				perror(ofname);
				exit(errno);
			}
		}
	}

	if (fl_verbose) fprintf(stderr, "%s: output is %s\n", me, ofname);
/*
 * Locate remote load points.
 */
	if (opt_taken('t')) {
		datapt = loadpt;
		intparam('t', &codept);
		codept += sizeof(struct exec);
	} else {
		codept = loadpt + sizeof(struct exec);
		datapt = codept + tout->a_text;
	}

	if (stackpt) {
		argvpt = datapt + tout->a_data + tout->a_bss;
	} else {
		stackpt = datapt + tout->a_data + tout->a_bss;
		argvpt = stackpt + stksize;
	}

	wp = stackpt + stksize;
	entrypt = codept + tout->a_entry;

	if (fl_verbose) {
		fprintf(stderr, "%s: text  %x\n", me, codept);
		fprintf(stderr, "%s: entry %x\n", me, entrypt);
		fprintf(stderr, "%s: data  %x\n", me, datapt);
		fprintf(stderr, "%s: stack %x\n", me, stackpt);
		fprintf(stderr, "%s: argv  %x\n", me, argvpt);
	}
/*
 * Read everything beyond header.
 */
	if (! opt_taken('t')) {
		ptext = eread(ifd, tout->a_text);
	} else {
		if (lseek(ifd, (long) tout->a_text, SEEK_CUR) < 0) {
			exit(errno);
		}
	}

	pdata = eread(ifd, tout->a_data);
	preloc = eread(ifd, tout->a_drsize);
	psyms = eread(ifd, tout->a_syms);
/*
 * Patch the data segment.
 */
	toutpatch(tout, preloc, pdata, codept, datapt);
/*
 * Reverse byte ordering in header, reloc and symbol tables.
 */
	bcopy((char *) tout, (char *) &rtout, sizeof(struct exec));

	if (ltot(TO_MAGIC_T8) != TO_MAGIC_T8) {
		toutrev(tout, &rtout, preloc, psyms);
	}
/*
 * Load everything.
 */
	if (! opt_taken('t')) {
		load(ofd, (char *) &rtout, loadpt,
			(unsigned long) sizeof(rtout));
		load(ofd, ptext, codept, tout->a_text);
		(void) free(ptext);
	}

	load(ofd, pdata, datapt, tout->a_data);
	(void) free(pdata);
/*
 * Load argv ptrs and strings.
 */
	pargv = pack_argv(argtailc, argtailv, &avsize);
	argpt = argvpt + ((argtailc + 1) * sizeof(char *));

	for (i = 0; i < argtailc; ++i) {
		parg = pargv[i];
		pargv[i] = (char *) ltot(argpt);
		argpt += strlen(parg) + 1;
	}
		
	load(ofd, (char *) pargv, argvpt, (unsigned long) avsize);
	(void) free((char *) pargv);
	avsize = (avsize + (sizeof(int) - 1)) & ~(sizeof(int) - 1);
/*
 * If -m, toss the final highwater mark out on the given file
 * descriptor so that sboot can read it.
 */ 
	waterpt = argvpt + avsize;
	if (fl_verbose) fprintf(stderr, "%s: water %x\n", me, waterpt);

	if (opt_taken('m')) {
		intparam('m', &mfd);
		write(mfd, (char *) &waterpt, sizeof(int));
	}
/*
 * Initialize stack frame.
 */
	pcb.pc_header = (struct exec *) ltot(codept - sizeof(struct exec));
	pcb.pc_stack = (char *) ltot(stackpt);
	pcb.pc_pid = ltot((wp + 1) & ~MININT);
	pcb.pc_stksize = ltot(stksize);
	pcb.pc_nodeid = ltot(NOTNODEID);
	pcb.pc_argc = ltot(argtailc);
	pcb.pc_argv = (char **) ltot(argvpt);
	pcb.pc_rtf = ltot(RTF_SYSGEN);
	pcb.pc_stdin = ltot(0);
	pcb.pc_stdout = ltot(1);
	pcb.pc_stderr = ltot(2);

	wp -= sizeof(pcb);
	load(ofd, (char *) &pcb, wp, (unsigned long) sizeof(pcb));
	writew(ofd, wp - sizeof(int), wp);
	wp -= 2 * sizeof(int);
	writew(ofd, wp, datapt);
	wp -= sizeof(int);
/*
 * Load entry in spawn table, if requested.
 */
	if (opt_taken('p')) {
		intparam('p', &ispawn);

		if (ispawn >= BTPMAX) {
			fprintf(stderr, "%s: spawn table index too high\n");
			exit(1);
		}

		spawnpt = BOOTSPAWN + (ispawn * sizeof(struct btprog));

		spawn.btp_load = (void *) ltot(codept - sizeof(struct exec));
		spawn.btp_data = (void *) ltot(datapt);
		spawn.btp_stack = (void *) ltot(stackpt);
		spawn.btp_argv = (char **) ltot(argvpt);
		spawn.btp_stksize = ltot(stksize);
		spawn.btp_argc = ltot(argtailc);
		spawn.btp_avsize = ltot(avsize);
		spawn.btp_flags = ltot((opt_taken('t') ? BTSHARE : 0));
		spawn.btp_priority = ltot(priority);
		spawn.btp_tag = 0;

		writew(ofd, BOOTNSPAWN, ispawn + 1);
		load(ofd, (char *) &spawn, spawnpt,
			(unsigned long) sizeof(spawn));
	}
/*
 * Execute program.
 */
	if (priority) {
		wp += 1;
	}
		
	if (opt_taken('b')) {
		moses_exec(ofd, entrypt, wp, (get_2nd_let('b') == 'r')
			? M_RUNP : M_CALL);
	}

	(void) close(ifd);

	if (xwrite == write) {
		(void) close(ofd);
	} else {
		(void) CloseLink(ofd);
	}

	return(0);
}

/*
 *	setpaths
 *
 *	Function:	- initializes search paths
 */

static void
setpaths()

{
/*
 * input files
 */
	n_prog = 0;
	p_prog = (char **) 0;
	addarg(&n_prog, &p_prog, "");
	addarg(&n_prog, &p_prog, "$TROLLIUSHOME");
	addarg(&n_prog, &p_prog, DEFPPROGS);
}
 
/*
 *        help
 *
 *        Function:     - prints helpful information on the cboot command
 */
 
static void
help()

{
	printf("\nSynopsis:\tcboot [options] <program> <device>\n");
	printf("\nDescription:\tBoot program on neighbour ITB node.\n");
	printf("\nOptions:        -h\t\tPrint this message.\n");
	printf("\t\t-bc\t\tBoot by transferring control.\n");
	printf("\t\t-br\t\tBoot by spawning a process.\n");
	printf("\t\t-i <program>\tSet executable input file.\n");
	printf("\t\t-k <priority>\tSet process priority.\n");
	printf("\t\t-l <address>\tBegin loading at <address>.\n");
	printf("\t\t-m <fd>\t\tWrite high water mark to pipe.\n");
	printf("\t\t-o <device>\tSet output device or file.\n");
	printf("\t\t-p <index>\tLoad one entry in spawn table.\n");
	printf("\t\t-s <size>\tSet stack size.\n");
	printf("\t\t-v\t\tBe verbose.\n");
	printf("\t\t-w <address>\tPlace stack at <address>.\n");
	printf("\t\t-t <address>\tText segment is already at <address>.\n");
	printf("\t\t-L\t\tOutput is a link device.\n");
	printf("\nExample:\tcboot -l 2800 -L -s 200 -br prog1 /dev/link0\n");
	printf("\t\t\t\"Boot `prog1' over `/dev/link0' at address 2800\n");
	printf("\t\t\t\with a stack of 200 bytes.\"\n\n");
}

/*
 *	load
 *
 *	Function:	- moses load operation
 *
 *	Accepts:	- file descriptor
 *			- remote load address
 *			- buffer pointer
 *			- buffer length
 */

static void
load(fd, buffer, loadpt, bufsize)

int			fd;
int			loadpt;
unsigned long		bufsize;
char			*buffer;

{
	char		cmd;		/* moses command */

	if (bufsize > 0) {
		cmd = M_LOAD;
		(*xwrite)(fd, &cmd, sizeof(char), 0);
		bufsize = ltot((int) bufsize);
		(*xwrite)(fd, (char *) &bufsize, sizeof(int), 0);
		loadpt = ltot(loadpt);
		(*xwrite)(fd, (char *) &loadpt, sizeof(int), 0);
		(*xwrite)(fd, buffer, ttol((int) bufsize), 0);
	}
}

/*
 *	writew
 *
 *	Function:	- moses write word operation
 *
 *	Accepts:	- device file descriptor
 *			- word address
 *			- data word
 */

static void
writew(fd, addr, data)

int			fd;
int			addr;
int			data;

{
	char		cmd;		/* moses command */

	cmd = M_WRIT;
	(*xwrite)(fd, &cmd, sizeof(char), 0);
	addr = ltot(addr);
	(*xwrite)(fd, (char *) &addr, sizeof(int), 0);
	data = ltot(data);
	(*xwrite)(fd, (char *) &data, sizeof(int), 0);
}

/*
 *	moses_exec
 *
 *	Function:	- moses call or runp operation
 *
 *	Accepts:	- device file descriptor
 * 			- instruction pointer
 *			- workspace pointer
 *			- call or runp
 */

static void
moses_exec(fd, ip, wp, cmd)

int			fd;
int			ip;
int			wp;
char			cmd;

{
	(*xwrite)(fd, &cmd, sizeof(char), 0);
	ip = ltot(ip);
	(*xwrite)(fd, (char *) &ip, sizeof(int), 0);
	wp = ltot(wp);
	(*xwrite)(fd, (char *) &wp, sizeof(int), 0);
}

/*
 *	eread
 *
 *	Function:	- error proof read operation
 *			- makes the read system call and checks for all errors
 *			- eof is an error
 *			- allocates buffer
 *
 *	Accepts:	- file descriptor open for read
 *			- size of data
 *
 *	Returns:	- buffer ptr
 *			- may call error or perror and never return
 */

static char *
eread(fd, bufsize)

int			fd;
unsigned long		bufsize;

{
	char		*buffer;	/* pointer to data */
	int		num_read = 0;	/* return code from system call */

	if (bufsize == 0) {
		return(NULL);
	}

	buffer = malloc((unsigned int) bufsize);

	if (buffer == NULL) {
		perror(me);
		exit(errno);
	}

	do {
		bufsize -= num_read;
		buffer += num_read;

		if ((num_read = read(fd, buffer, (int) bufsize)) < 0) {
			perror(me);
			exit(errno);
		}

		else if (num_read == 0) {
			fprintf(stderr, "%s: premature EOF\n", me);
			exit(1);
		}

	} while (num_read < bufsize);

	return(buffer);
}
