/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	map.c,v $
 * Revision 2.2  92/04/30  14:56:33  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/01/31  18:19:53  gdburns
 * Look at the environment for the boot schema filename.
 * 
 * Revision 2.1.1.2  91/12/04  17:15:26  raja
 * Added link weight option to route2 algorithm.
 * 
 * Revision 2.1.1.1  91/11/26  17:46:25  raja
 * Big overhaul.  Added second route table.
 * 
 *	Function:	- route generation tool (route table "compiler")
 *			- generate a route file for the multicomputer network
 *			  described in the given Boot Schema (bnail file)
 *			- full connectivity: O(N^2) routes generated
 *			- near-optimal routes for message passing
 *			- optimal routes for multicast/multireel
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <net.h>
#include <nail.h>
#include <routeio.h>
#include <debug.h>
#include <castreq.h>
#include <all_hash.h>
#include <map.h>

/*
 * external variables
 */
extern FILE		*yyin;		/* defined in lex as input file */
extern int		errno;		/* for global error number */

/*
 * global variables
 */
char			*me;		/* tool name */
int			fl_debug;	/* debug option flag */
int			fl_stats;	/* statistics flag */
int			fl_verbose;	/* verbose option flag */
int			fl_weight;	/* use link weights in route2 flag */

struct cmdesc		*cmatrix;	/* connectivity matrix descriptor */
struct bschema		nail;		/* bnail file information */

/*
 * static variables
 */
static char		*f_bnail;	/* boot schema filename */
static char		*full_bnail;	/* boot schema expanded filename */
static char		**p_bnail;	/* boot schema pathname prefix */
static char	 	rtfname[256];	/* route file filename */
static int4		extra;		/* extra rtable entries for castids */
static int		fl_write;	/* write route file flag */

static char *usage =
	"usage: %s [-dhsvwVW] [-i <extra>] [-c <bnail>] [-p <route>]\n";
 
/* 
 * external functions
 */
extern void		exit();
extern void		followed();
extern void		validopts();
extern char		*findpath();
extern char		*getenv();
extern char		*getparam();
extern int		addarg();
extern int		do_args();
extern int		opt_taken();
extern int		parse_nail();
extern int4		intparam();
extern int4		stoi4();

/* 
 * static functions
 */
static void		map_help();
static void		map_stats();
static void		prt_defaults();
static void		set_defaults();
static int		map_create();
static int		map_rtfill();
static int		map_wrtfile();

/*
 * Main program. 
 */
int
main(argc, argv)  

int			argc;
char			*argv[];

{
/*
 * Assign the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("dhsvwVW");
	followed("cip");

	if ((do_args(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
		fprintf(stderr, usage, me); 
		exit(ERROR);
	}
/* 
 * Print the help menu if needed.
 */ 
	if (opt_taken('h')) { 
		map_help();
		exit(0);
	}
/*
 * Set the mode of operation.
 */
	set_defaults();

	if (fl_debug) {
		prt_defaults();
	}
/* 
 * Locate the boot schema file.  
 */
	if ((full_bnail = findpath(f_bnail, p_bnail, R_OK)) == NULL) {
		fprintf(stderr, "%s: cannot locate bnail file \"%s\"\n",
			me, f_bnail);
		exit(ERROR);
	}
/*
 * Open the boot schema file.
 */
	DBUG("%s: opening the boot schema file: %s\n", me, full_bnail);
	if ((yyin = fopen(full_bnail, "r")) == NULL) {
		perror(me);
		exit(errno);
	}
/*
 * Parse the boot schema.
 */
	if (parse_nail(&nail)) {
		fprintf(stderr, "%s: cannot parse the NaIL syntax\n", me);
		exit(ERROR);
	}
/*
 * Close and pretty print the boot schema.
 */
	(void) fclose(yyin);

	if ((opt_taken('V')) || (fl_debug)) {
		pretty_bnail(&nail);
		if (opt_taken('V')) {
			exit(0);
		}
	}
/*
 * Determine the host node of the network.
 * It will be the root of the spanning tree.
 */
	if (find_host(&nail)) {
		fprintf(stderr, "%s: cannot find the network host\n", me);
		exit(ERROR);
	}
	DBUG("%s: found hostnode %d\n", me, nail.bs_hostnode->tn_nodeid);
/*       
 * Generate the network spanning tree.
 */              
	DBUG("%s: generating the spanning tree ...\n", me);
	if (make_spantree(&nail)) {
		fprintf(stderr, "%s: cannot generate spanning tree\n", me);
		exit(ERROR);
	}
	if (fl_debug) {
		print_spantree(&nail);
	}

	if (map_create()) {
		DBUG("%s: unable to create the route file\n", me);
		exit(ERROR);
	}

	exit(0);
}
 
/*
 * 	map_help
 *
 *	Function:	- print the help menu
 */
static void 
map_help()

{
	printf("\nSynopsis:\t%s [options]\n", me);
	printf("\nDescription:\tGenerate a route file from a Boot Schema.\n");
	printf("\nOptions:\t-h\t\tthis message\n");
	printf("\t\t-d\t\tturn on debugging\n");
	printf("\t\t-s\t\tprint route statistics\n");
	printf("\t\t-v\t\tverbose mode\n");
	printf("\t\t-w\t\tdo not create route file\n");
	printf("\t\t-V\t\tpretty print the bnail file\n");
	printf("\t\t-W\t\tignore link weights for route 2\n");
	printf("\t\t-c bnail\tuse \"bnail\" as the Boot Schema file\n");
	printf("\t\t-i extra\tadd \"extra\" entries to each route table\n");
	printf("\t\t-p route\tuse \"route\" as output route file\n");
}

/*
 *	set_defaults
 *
 *      Function:	- set the values of key variables controlling
 *			  the operation of the tool
 */
static void 
set_defaults()

{
        int		n_bnail;
        char		*p;
/* 
 * Pathnames for the boot schema file.
 */
	n_bnail = 0;
	p_bnail = (char **) 0;
	addarg(&n_bnail, &p_bnail, "");				/* absolute */
	addarg(&n_bnail, &p_bnail, "$TROLLIUSHOME/boot");	/* env */
	addarg(&n_bnail, &p_bnail, DEFPBNAIL);			/* default */
/*
 * Pathnames for the files.
 */
	if (opt_taken('c')) {
      		f_bnail = getparam('c');
	} else if (f_bnail = getenv("TROLLIUSBNAIL")) {
	} else {
		f_bnail = DEFFBNAIL;
	}

	fl_stats = opt_taken('s');
	fl_write = !opt_taken('w');
	fl_verbose = opt_taken('v');
	fl_weight = !opt_taken('W');
  
	if (opt_taken('d')) {
		fl_debug = TRUE;
		fl_verbose = TRUE;
		fl_stats = TRUE;
	} else {
		fl_debug = FALSE;
	}

	if (opt_taken('i')) {
		extra = stoi4(getparam('i'));
	} else {
		extra = MAXCAST;
	}

	if (opt_taken('p')) {
		(void) strcpy(rtfname, getparam('p'));
	} else {
		(void) strcpy(rtfname, DEFFROUTE);
		if (p = strchr(rtfname, '.')) {
			*p = '\0';
		}
		p = strrchr(f_bnail, STRDIR);
		if (p = strchr((p) ? p + 1 : f_bnail, '.')) {
			(void)strcat(rtfname, ".");
			(void)strcat(rtfname, p + 1);
		}
	}
}

/*
 *	prt_defaults
 *
 *	Function:	- print the values of key variables
 *			- mostly used for debugging
 */
static void 
prt_defaults()

{
	printf("%s: default settings :\n", me);
	printf("%s: boot file (input): %s\n", me, f_bnail);
	printf("%s: route file (output): %s\n", me, rtfname);
	printf("%s: debug flag: %d\n", me, fl_debug);
	printf("%s: stats flag: %d\n", me, fl_stats);
	printf("%s: verbose flag: %d\n", me, fl_verbose);
	printf("%s: weight flag: %d\n", me, fl_weight);
	printf("%s: write flag: %d\n", me, fl_write);
	printf("%s: extra entries: %d\n", me, extra);
}

/*
 *	map_create
 *
 *	Function:	- create a full route file
 *			- root of the "intelligence" behind the map tool
 *
 *	Returns:	- 0 or ERROR
 */
static int
map_create()

{
/*
 * Create and initialize the connectivity matrix.
 */
	if ((cmatrix = cm_init(&nail, extra)) == (struct cmdesc *) 0) {
		DBUG("%s (map_create): cannot create matrix\n", me);
		return(ERROR);
	}
/*
 * Compute the first route table.  This route table is used for generic
 * message passing, in which case the link weights are taken into account.
 */
	if (map_route1()) {
		DBUG("%s (map_create): cannot generate first routes\n", me);
		cm_free(cmatrix);
		return(ERROR);
	}
/*
 * Compute the second route table.  This route table is used for multicasting
 * and multireeling and requires the routes to be along a min-hop spanning
 * tree with no reconvergent paths.  Link weights are not used.
 */
	if (map_route2()) {
		DBUG("%s (map_create): cannot generate second routes\n", me);
		cm_free(cmatrix);
		return(ERROR);
	}
/*
 * Write the route table file.
 */
	if (fl_write && map_wrtfile()) {
		DBUG("%s (map_create): cannot write route file\n", me);
		cm_free(cmatrix);
		return(ERROR);
	}
/*
 * Write out the statistics.
 */
	if (fl_stats) {
		map_stats();
	}
/*
 * Cleanup and exit.
 */
	cm_free(cmatrix);
	return(0);
}

/*
 *	map_wrtfile
 *
 *	Function:	- create and write the route file
 *
 *	Returns:	- 0 or ERROR
 */
static int
map_wrtfile()

{
	RTIOFD		*rfd;		/* route file descriptor */
	int4		nodeid;		/* target nodeid */
	int		i;		/* favorite counter */
/*
 * Create a route file descriptor.
 */
	if ((rfd = rtio_init(rtfname, &nail, extra)) == (RTIOFD *) 0) {
		DBUG("%s (rtio_init): cannot create route file desc.\n", me);
		return(ERROR);
	}
/*
 * Create the header and the directory of the route file.
 */
	VERBOSE("creating the route file ...\n");
	if (rtio_header(rfd) || rtio_dir(rfd)) {
		rtio_cleanup(rfd);
		return(ERROR);
	}
/*
 * Loop through all nodes, in the required order, creating their
 * route tables and writing them to the route file.
 */
	i = 0;
	rtio_topnode(rfd);
	while ((nodeid = rtio_thisnode(rfd)) != NOTNODEID) {
/*
 * Beautify the output in the pure verbose mode.
 */
		VERBOSE("[%d] ", nodeid);
		if (++i >= 8) {
			VERBOSE("\n");
			i = 0;
		}
/*
 * Initialize the internal route table with the nodeids and node types.
 */
		if (rtio_rtinit(rfd)) {
			rtio_cleanup(rfd);
			return(ERROR);
		}
/*
 * Fill in this node's route table.
 */
		if (map_rtfill(rfd, nodeid)) {
			rtio_cleanup(rfd);
			return(ERROR);
		}
/*
 * Write the node's route table in the route file.
 */
		if (rtio_rtwrite(rfd, nodeid)) {
			rtio_cleanup(rfd);
			return(ERROR);
		}
	}

	if (i != 0) {
		VERBOSE("\n");
	}
	VERBOSE("route file created\n");
	rtio_cleanup(rfd);
	return(0);
}

/*
 *	map_rtfill
 *
 *	Function:	- fill in the route file descriptor's internal
 *			  route table the target node's route table
 *
 *	Accepts:	- ptr route file descriptor
 *			- target nodeid
 *
 *	Returns:	- 0 or ERROR
 */
static int
map_rtfill(rfd, nodeid)

RTIOFD			*rfd;
int4			nodeid;

{
	struct t_node	*tnode;		/* ptr to TNODE */
	struct cmelem	*pcm;		/* ptr to matrix entry */
	struct route	*prt;		/* ptr to route entry */
/*
 * Loop over all the nodes in the system.
 */
	tnode = (struct t_node *) ah_top(nail.bs_tnhd);
	while (tnode) {
/*
 * Locate the correct entry in the connectivity matrix.
 */
		pcm = cm_find(cmatrix, nodeid, tnode->tn_nodeid);
		if (pcm == (struct cmelem *) 0) {
			fprintf(stderr, "%s (map_wrtfile): no (%d, %d) entry\n",
				me, nodeid, tnode->tn_nodeid);
			return(ERROR);
		}
/*
 * Locate the correct route table entry.
 */
		if ((prt = (struct route *) ahs_find(&(rfd->rd_route),
			tnode->tn_nodeid)) == (struct route *) 0) {

			fprintf(stderr,
				"%s (map_wrtfile): no node %d route entry\n",
				me, tnode->tn_nodeid);
			return(ERROR);
		}
/*
 * Store the route table information.
 */
		prt->r_event = pcm->ce_event;
		prt->r_event2 = pcm->ce_event2;
		if (pcm->ce_nodetype != NOTNODETYPE) {
			prt->r_nodetype = pcm->ce_nodetype;
		}

		tnode = (struct t_node *) ah_next(nail.bs_tnhd, (void *) tnode);
	}
 
	return(0);
}

/*
 *	map_stats
 *
 *	Function:	- print out the routing statistics gathered
 */
static void
map_stats()

{
	struct cmstats	*p;		/* favorite pointer */

	p = &(cmatrix->cd_stats);
/*
 * Print statistics for the first and second route.
 */
	printf("\n%49s\n", "Routing Statistics");
	printf("Number of nodes: %d\n", ah_count(nail.bs_tnhd));
	printf("Number of links: %d\n\n", p->cs_nlink);

	printf("%28s%13s%13s%13s\n", "Min", "Max", "Ave", "Std Dev");
	printf("%-19s  %7d   %10d   %10.3f   %10.5f\n", "R1 link usage",
		p->cs_r1minlu, p->cs_r1maxlu, p->cs_r1avelu, p->cs_r1stdlu);
	printf("%-19s  %7d   %10d   %10.3f   %10.5f\n", "R1 link weight",
		p->cs_r1minlw, p->cs_r1maxlw, p->cs_r1avelw, p->cs_r1stdlw);
	if (fl_weight) {
		printf("%-19s  %7d   %10d   %10.3f   %10.5f\n",
			"R2 broadcast weight", p->cs_r2minbh,
			p->cs_r2maxbh, p->cs_r2avebh, p->cs_r2stdbh);
	} else {
		printf("%-19s  %7d   %10d   %10.3f   %10.5f\n",
			"R2 broadcast hops", p->cs_r2minbh,
			p->cs_r2maxbh, p->cs_r2avebh, p->cs_r2stdbh);
	}
}
