/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	map_route1.c,v $
 * Revision 2.2  92/04/30  14:56:41  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/11/26  18:10:39  raja
 * initial revision
 * 
 *	Function:	- generate the first route table for the map tool
 *			- used by calling map_route1()
 *			- the algorithm proceeds as follows:
 *				- assign direct neighbour (1-hop) routes
 *				- assign TROUTE tuples
 *				- generate host-to-all connections
 *				- generate routes in TORDER priority
 *				- generate remaining all-to-all routes
 */

#include <stdio.h>
#include <math.h>
#include <portable.h>
#include <typical.h>
#include <debug.h>
#include <net.h>
#include <events.h>
#include <all_hash.h>
#include <all_list.h>
#include <all_queue.h>
#include <nail.h>
#include <map.h>
#include <routeio.h>

/*
 * external variables
 */
extern struct bschema	nail;		/* network information */
extern struct cmdesc	*cmatrix;	/* connectivity matrix descriptor */
extern int		fl_debug;	/* debug mode flag */
extern int		fl_stats;	/* statistics flag */
extern int		fl_verbose;	/* verbose mode flag */
extern char		*me;		/* tool name */

/*
 * static variables
 */
static HASH		*navhd;		/* graph navigation hash desc. */
static QUEUE		*navqd;		/* graph navigation queue desc. */
static int		fl_tohost;	/* flag to-host path */
static int		fl_fromhost;	/* flag from-host path */

/*
 * external functions
 */
extern int		boot_cmp();
extern int		is_link();
extern int		on_span();
extern int4		cm_linkcomp();

/*
 * static functions
 */
static int		r1_all();
static int		r1_backward();
static int		r1_dolink();
static int		r1_findroute();
static int		r1_host2all();
static int		r1_init();
static int		r1_stats();
static int		r1_tlink();
static int		r1_torder();
static int		r1_troute();
static int4		r1_forward();
static void		r1_free();
static void		r1_reset();

/*
 *	map_route1
 *
 *	Function:	- generate the first route table for each node
 *			- routes are used for generic message passing
 *			- link weights used to pick the least busy path
 *
 *	Returns:	- 0 or ERROR
 */
int
map_route1()

{
	VERBOSE("generating route 1 ...\n");
/*
 * Initialize the graph navigation structures.
 */
	if (r1_init()) {
		return(ERROR);
	}
/*
 * Fill in the direct neighbour paths.  These are read in from the
 * TLINK tuples in the bnail file.
 */
	if (r1_tlink()) {
		r1_free();
		return(ERROR);
	}
/*
 * Fill in the routes specified in the TROUTE tuples in the bnail file.
 */
	if (r1_troute()) {
		r1_free();
		return(ERROR);
	}
/*
 * Give priority to the routes to/from host by generating them first.
 */
	if (r1_host2all()) {
		r1_free();
		return(ERROR);
	}
/*
 * Next in priority are the routes specified in the TROUTE tuples.
 * Generate these route in the order they appear in.
 */
	if (r1_torder()) {
		r1_free();
		return(ERROR);
	}
/*
 * Generate all the remaining routes in the system.
 */
	if (r1_all()) {
		r1_free();
		return(ERROR);
	}
/*
 * Print out the statistics.
 */
	if (fl_stats && r1_stats()) {
		r1_free();
		return(ERROR);
	}

	r1_free();
	return(0);
}

/*
 *	r1_tlink
 *
 *	Function:	- fill the direct neighbour paths
 *			- read them from the TLINK tuples
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_tlink()

{
	struct t_link	*tlink;			/* ptr to TLINK */
	struct cmelem	*p;			/* favorite pointer */
	struct cmlink	link;			/* link/cost element */

	if (nail.bs_ntlink > INT4_NIL) {
		VERBOSE("\tdirect neighbours ...\n");
	}
/*
 * Loop over all the TLINK tuples.
 */
	for (tlink = nail.bs_tlink; tlink; tlink = tlink->tl_next) {
/*
 * Locate the entry in the connectivity matrix.
 * Declare the path between the source and destination to be done.
 */
		p = cm_find(cmatrix, (tlink->tl_srcnode)->tn_nodeid, 
			(tlink->tl_destnode)->tn_nodeid);
		if (p == (struct cmelem *) 0) {
			fprintf(stderr,
				"%s (r1_tlink): no (%d,%d) entry in matrix\n",
				me, (tlink->tl_srcnode)->tn_nodeid,
				(tlink->tl_destnode)->tn_nodeid);
			return(ERROR);
		}
		p->ce_done = TRUE;
/*
 * If needed, allocate the link/cost list for this entry.
 */
		if ( ! p->ce_cost) {
			if ((p->ce_cost = (LIST *)al_init(sizeof(struct cmlink),
				cm_linkcomp)) == (LIST *) 0) {

				fprintf(stderr,
					"%s (r1_tlink): cannot malloc list\n",
					me);
				return(ERROR);
			}
/*
 * Link that list in the cleanup list.
 */
			if ( ! al_append(cmatrix->cd_cleanup, &(p->ce_cost))) {
				fprintf(stderr,
					"%s (r1_tlink): cannot link cleanup\n",
					me);
				return(ERROR);
			}
		}
/*
 * Check if the maximum allowed number of parallel link has been exceeded.
 */
		if (al_count(p->ce_cost) >= PARLINK) {
			fprintf(stderr,
				"%s (r1_tlink): # links > %d in (%d, %d)\n",
				me, PARLINK, (tlink->tl_srcnode)->tn_nodeid,
				(tlink->tl_destnode)->tn_nodeid);
			return(ERROR);
		}
/*
 * Add the link to the list.
 */
		link.cl_linkno = tlink->tl_srclink;
		link.cl_cost = tlink->tl_weight;
		link.cl_weight = tlink->tl_weight;
		link.cl_use = 1;
		if (al_insert(p->ce_cost, (char *) &link) == NULL) {
			fprintf(stderr, "%s (r1_tlink): cannot add link\n", me);
			return(ERROR);
		}
/*
 * If this is the first link between these two nodes, use it to
 * route messages from the source to the destination node.
 */
		if (al_count(p->ce_cost) == 1) {
			p->ce_event = tlink->tl_srclink;
			p->ce_nodetype = NT_JONES |
				((tlink->tl_destnode)->tn_specs)->ns_type; 
/*
 * Check if this tlink is part of the boot spanning tree.  This
 * means that the source node of the link would have to boot the
 * destination node, in the parallel booting scheme.
 */
			if (on_span(tlink)) {
				p->ce_nodetype |= NT_BOOT; 
			}
		}
	}

	return(0);
}

/*
 *	r1_troute
 *
 *	Function:	- fill the indirect routes forced by the user
 *			- read them from the TROUTE tuples
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_troute()

{
	struct t_node	*tnode;		/* ptr to TNODE */
	struct t_route	*troute;	/* ptr to TROUTE */
	struct cmelem	*p;		/* favorite pointer */

	if (nail.bs_ntroute > INT4_NIL) {
		VERBOSE("\tTROUTE forced routes ...\n");
	}
/*
 * Loop over all TROUTE tuples (i.e. loop over all the nodes in
 * the system and for each, loop over all its TROUTE tuples).
 */
	tnode = (struct t_node *) ah_top(nail.bs_tnhd);
	while (tnode) {
		troute = (tnode->tn_specs)->ns_troute;
		for ( ; troute; troute = troute->tr_next) {
/*
 * Locate the entry in the connectivity matrix.
 */
			if ((p = cm_find(cmatrix, tnode->tn_nodeid,
				troute->tr_destnode)) == (struct cmelem *) 0) {

				fprintf(stderr,
					"%s (r1_troute): no (%d,%d) entry\n",
					me, tnode->tn_nodeid,
					troute->tr_destnode);
				return(ERROR);
			}

			if ((p->ce_event != NOTEVENT) &&
				(p->ce_event != troute->tr_destlink)) {
/*
 * There is a collision on route #1.  Ignore the TROUTE entry.
 */
			VERBOSE("Warning: TROUTE <%d, %d, %d, %d> ignored\n",
				tnode->tn_nodeid, troute->tr_destnode,
				troute->tr_destlink, troute->tr_destlink2);
			}
			else {
/*
 * There is no collision between the route entry and the TROUTE tuple.
 * Set the event to the user entry if the given link number exists.
 */
				if (is_link(p, troute->tr_destlink)) {
					p->ce_event = troute->tr_destlink;
				}
				else {
					fprintf(stderr, 
					"%s (r1_troute): link %d not defined\n",
						me, troute->tr_destlink);
					return(ERROR);
				}
			}
		}

		tnode = (struct t_node *) ah_next(nail.bs_tnhd, (void *) tnode);
	}

	return(0);
}

/*
 *	r1_host2all
 *
 *	Function:	- generate the shortest routes to/from the host
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_host2all()

{
	int4		hostid;			/* host nodeid */
	int4		weight;			/* path weight */
	struct t_node	*tnode;			/* ptr to TNODE */

	if (ah_count(nail.bs_tnhd) > (int4) 2) {
		VERBOSE("\thost connections ...\n");
	}
/*
 * Check if the host node is defined.
 */
	if (nail.bs_hostnode) {
		hostid = (nail.bs_hostnode)->tn_nodeid;
	}
	else {
		fprintf(stderr, "%s (r1_host2all): host not defined\n", me);
		return(ERROR);
	}
	weight = nail.bs_tweight.tw_host;
/*
 * Generate all the routes having the host as the destination node.
 * Loop over all the nodes in the system, skipping the host node.
 */
	tnode = (struct t_node *) ah_top(nail.bs_tnhd);
	while(tnode) {
		if (tnode->tn_nodeid != hostid) {
			if (r1_findroute(tnode->tn_nodeid, hostid, weight)) {
				DBUG("%s (r1_host2all): no route (%d, %d)\n",
					me, tnode->tn_nodeid, hostid);
				return(ERROR);
			}
		}

		tnode = (struct t_node *) ah_next(nail.bs_tnhd, (void *)tnode);
	}
/*
 * Generate all the routes having the host as the source node.
 * Loop over all the nodes in the system, skipping the host node.
 */
	tnode = (struct t_node *) ah_top(nail.bs_tnhd);
	while(tnode) {
		if (tnode->tn_nodeid != hostid) {
			if (r1_findroute(hostid, tnode->tn_nodeid, weight)) {
				DBUG("%s (r1_host2all): no route (%d, %d)\n",
					me, hostid, tnode->tn_nodeid);
				return(ERROR);
			}
		}

		tnode = (struct t_node *) ah_next(nail.bs_tnhd, (void *)tnode);
	}

	return(0);
}

/* 
 *	r1_torder
 *
 *	Function:	- generate the shortest route for the paths
 *			  specified in the TORDER tuples
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_torder()

{
	struct t_order	*torder;		/* ptr to TORDER */

	if (nail.bs_ntorder > INT4_NIL) {
		VERBOSE("\tTORDER routes ...\n");
	}
/*
 * Loop over the TORDER tuples.
 */
	for (torder = nail.bs_torder; torder; torder = torder->to_next) {
		if (r1_findroute(torder->to_srcnode, torder->to_destnode,
			torder->to_weight)) {

			DBUG("%s (r1_torder): no route (%d, %d)\n", me,
				torder->to_srcnode, torder->to_destnode);
			return(ERROR);
		}
	}

	return(0);
}

/*
 *	r1_all
 *
 *	Function:	- generate all remaining N*N routes in the system
 *			- guarantees full communication connectivity
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_all()

{
	struct t_node	*src;		/* ptr to source TNODE */
	struct t_node	*dest;		/* ptr to destination TNODE */
	HASH		*tnhd;		/* TNODE hash table descriptor */
	int4		weight;

	if (ah_count(nail.bs_tnhd) > (int4) 2) {
		VERBOSE("\tfull connectivity ...\n"); 
	}

	tnhd = nail.bs_tnhd;
	weight = nail.bs_tweight.tw_mcast;
/*
 * Loop over all nodes, connecting each node to all other nodes.
 * No way around it, it time to generate the N*N paths.
 * Skip the path if the source and destination are the same node.
 */
	src = (struct t_node *) ah_top(tnhd);
	while (src) {
		dest = (struct t_node *) ah_top(tnhd);
		while (dest) {
			if (src != dest) {
				if (r1_findroute(src->tn_nodeid,
					dest->tn_nodeid, weight)) {

					DBUG("%s (r1_all): no route (%d, %d)\n",
						me, src->tn_nodeid,
						dest->tn_nodeid);
					return(ERROR);
				}
			}

			dest = (struct t_node *) ah_next(tnhd, (void *) dest);
		}

		src = (struct t_node *) ah_next(tnhd, (void *) src);
	}

	return(0);
}

/*
 *	r1_findroute
 *
 *	Function:	- generate the shortest route between the given
 *			  source and destination nodes
 *			- update the link usage/weights accordingly
 *
 *	Accepts:	- source nodeid
 *			- destination nodeid
 *			- link weight increment
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_findroute(srcid, destid, weight)

int4			srcid,
			destid,
			weight;

{
	struct cmelem	*p;		/* favorite pointer */
	int4		total_weight;	/* path weight */
/*
 * Locate the proper entry in the connectivity matrix.
 */
	if ((p = cm_find(cmatrix, srcid, destid)) == (struct cmelem *) 0) {
		fprintf(stderr, "%s (r1_findroute): no (%d, %d) matrix entry\n",
			me, srcid, destid);
		return(ERROR);
	}
/*
 * If the path has already been done just return.
 */
	if (p->ce_done == TRUE) {
		return(0);
	}
/*
 * Now get to work and generate this path.  In the forward propagation
 * phase, navigate the network searching for the path.  In the back
 * propagation phase, fill the route tables and update the link weights.
 */
	total_weight = r1_forward(srcid, destid);
	if (total_weight == (int4) ERROR) {
		DBUG("%s (r1_findroute): no path from %d to %d\n",
			me, srcid, destid);
		return(ERROR);
	}

	DBUG("%s (r1_findroute): path %d --> %d found, weight: %d\n",
		me, srcid, destid, total_weight);
	if (r1_backward(srcid, destid, weight)) {
		DBUG("%s (r1_findroute): cannot back propagate (%d, %d)\n",
			me, srcid, destid);
		return(ERROR);
	}

	return(0);
}

/*
 *	r1_init
 *
 *	Function:	- initialize the data structures required to
 *			  navigate the network generating route 1.
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_init()

{
	struct t_node	*tnode;		/* ptr to TNODE */
	struct navelem	entry;		/* navigation table entry */
/*
 * Initialiaze the navigation hash table.
 */
	navhd = ah_init(ah_size(nail.bs_tnhd),
		(int4) sizeof(struct navelem), NOTNODEID, AHFIXED);
	if (navhd == (HASH *) 0) {
		fprintf(stderr, "%s (r1_init): cannot malloc hash table\n", me);
		return(ERROR);
	}

	tnode = (struct t_node *) ah_top(nail.bs_tnhd);
	while (tnode) {
		entry.ne_nodeid = tnode->tn_nodeid;
		entry.ne_tnode = tnode;
		if (ah_insert(navhd, (void *) &entry)) {
			fprintf(stderr, "%s (r1_init): cannot insert node %d\n",
				me, tnode->tn_nodeid);
			return(ERROR);
		}

		tnode = (struct t_node *) ah_next(nail.bs_tnhd, (void *) tnode);
	}
/*
 * Initialize the navigation queue.
 */
	navqd = aq_init((int4) 1, (int4) sizeof(struct navelem *));
	if (navqd == (QUEUE *) 0) {
		fprintf(stderr, "%s (r1_init): cannot malloc queue\n", me);
		return(ERROR);
	}

	return(0);
}

/*
 *	r1_reset
 *
 *	Function:	- reset the contents of the data structures
 *			  used for graph navigation
 */
static void
r1_reset()

{
	struct navelem	*p;		/* favorite pointer */
/*
 * Reset the contents of the navigation hash table.
 */
	if (navhd) {
		p = (struct navelem *) ah_top(navhd);
		while (p) {
			p->ne_mincost = INT4_MAX;
			p->ne_prevnode = (struct navelem *) 0;
			p->ne_prevlink = (struct cmlink *) 0;

			p = (struct navelem *) ah_next(navhd, (void *) p);
		}
	}
}

/*
 *	r1_free
 *
 *	Function:	- free the data structures using in graph navigation
 */
static void
r1_free()

{
	if (navhd) {
		ah_free(navhd);
	}

	if (navqd) {
		aq_free(navqd);
	}
}

/*
 *	r1_forward
 *
 *	Function:	- navigate the network graph from the source 
 *			  node trying to reach the destination node along
 *			  the path with minimum weight
 *			- this is a modified single-source-all-pairs
 *			  shortest route algorithm that takes into account
 *			  routing restrictions at intermediate nodes
 *
 *	Accepts:	- source nodeid
 *			- destination nodeid
 *
 *	Returns:	- weight of path found or ERROR
 */
static int4
r1_forward(srcid, destid)

int4			srcid,
			destid;

{
	struct navelem	*p;		/* favorite pointer */
	struct navelem	**pp;		/* another pointer */
	struct cmelem	*cmentry;	/* ptr to connectivity entry */
	struct link	*plink;		/* link pointer */
/*
 * Reset the navigation data structures.
 */
	r1_reset();
/*
 * If either source or destination node is the host, remember
 * that this will require special handling later on.
 * Basically, when going to/from the host, all intermediate nodes
 * should follow a strict ordering via-a-vis the boot order.
 */
	fl_tohost = fl_fromhost = FALSE;
	if (srcid == (nail.bs_hostnode)->tn_nodeid) {
		fl_fromhost = TRUE;
	}
	else if (destid == (nail.bs_hostnode)->tn_nodeid) {
		fl_tohost = TRUE;
	}
/*
 * Insert the source node entry in the navigation queue.
 * Set the cost at the source node to zero.
 */
	p = (struct navelem *) ah_find(navhd, srcid);
	if (p == (struct navelem *) 0) {
		fprintf(stderr, "%s (r1_forward): node %d not in nav. table\n",
			me, srcid);
		return((int4) ERROR);
	}
	if (aq_shove(navqd, (void *) &p)) {
		fprintf(stderr, "%s (r1_forward): nav. enqueue error\n", me);
		return((int4) ERROR);
	}
	p->ne_mincost = INT4_NIL;
/*
 * Loop till the navigation queue is empty.
 */
	while (aq_count(navqd) > INT4_NIL) {
/*
 * Remove the top element from the queue.  This is the current node.
 */
		pp = (struct navelem **) aq_find(navqd);
		if (pp == (struct navelem **) 0) {
			fprintf(stderr,
				"%s (r1_forward): empty nav. queue\n", me);
			return((int4) ERROR);
		}
		p = *pp;
		if (aq_delete(navqd)) {
			fprintf(stderr,
				"%s (r1_forward): cannot nav. dequeue\n", me);
			return((int4) ERROR);
		}
/*
 * Check if the route from the current node to the destination is defined.
 */
		cmentry = cm_find(cmatrix, p->ne_nodeid, destid);
		if (cmentry == (struct cmelem *) 0) {
			fprintf(stderr,
				"%s (r1_forward): no (%d, %d) matrix entry\n",
				me, p->ne_nodeid, destid);
			return((int4) ERROR);
		}
		if (cmentry->ce_event != NOTEVENT) {
/*
 * Route from current node to destination node has been defined.
 * We have to use it and not bother trying to go through other
 * neighbours of the current node (routing restriction).
 * Locate the TLINK tuple for the link used.
 */
			plink = ((p->ne_tnode)->tn_specs)->ns_lkout;
			for ( ; plink; plink = plink->lk_next) {
				if ((plink->lk_link)->tl_srclink ==
					cmentry->ce_event) {
					break;
				}
			}
			if (plink == (struct link *) 0) {
				fprintf(stderr,
				"%s (r1_forward): no link %d from node %d\n",
					me, cmentry->ce_event, p->ne_nodeid);
				return((int4) ERROR);
			}
			
			if (r1_dolink(p, plink->lk_link)) {
				return((int4) ERROR);
			}
		}
		else {
/*
 * Route from current node to destination node is not defined.
 * We have the flexibility of trying to go in all directions.
 * Loop over all the neighbours of the current node.
 */
			plink = ((p->ne_tnode)->tn_specs)->ns_lkout;
			for (; plink; plink = plink->lk_next) {
				if (r1_dolink(p, plink->lk_link)) {
					return((int4) ERROR);
				}
			}
		}
	}
/*
 * Return the total weight of the path from source to destination.
 */
	p = (struct navelem *) ah_find(navhd, destid);
	if (p == (struct navelem *) 0) {
		fprintf(stderr, "%s (r1_forward): node %d not in nav. table\n",
			me, destid);
		return((int4) ERROR);
	}

	if (p->ne_mincost == INT4_MAX) {
		fprintf(stderr, "%s (r1_forward): no path (%d, %d) found\n",
			me, srcid, destid);
		return((int4) ERROR);
	}
	else {
		return(p->ne_mincost);
	}
}

/*
 *	r1_dolink
 *
 *	Function:	- check if the given link minimizes the path
 *			  weight reaching the next node
 *			- if this link is a better choice update the
 *			  navigation table and queue otherwise do nothing
 *			- in the case of routing to/from the host the
 *			  path has to follow a strict boot ordering
 *
 *	Accepts:	- ptr to source node
 *			- ptr to TLINK to use
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_dolink(src, tlink)

struct navelem		*src;
struct t_link		*tlink;

{
	int4		new_weight;	/* new destination node weight */
	int4		dest_node;	/* destination nodeid */
	struct cmelem	*p;		/* favorite pointer */
	struct navelem	*dest;		/* ptr to destination node */
	struct cmlink	link;		/* matrix link entry */
	struct cmlink	*plink;		/* ptr to matrix link entry */
/*
 * If this path is to the host node, check if the source node is booted
 * after the destination node (i.e. appears later in the boot order).
 * If this is not the case, just skip the current node.
 */
	dest_node = (tlink->tl_destnode)->tn_nodeid;
	if (fl_tohost) {
		if (boot_cmp(src->ne_nodeid, dest_node) < 0) {
			return(0);
		}
	}
/*
 * If this path is from the host node, check if the source node is booted
 * before the destination node (i.e. appears earlier in the boot order).
 * If this is not the case, just skip the current node.
 */
	if (fl_fromhost) {
		if (boot_cmp(src->ne_nodeid, dest_node) > 0) {
			return(0);
		}
	}
/*
 * Locate the destination node in the navigation hash table.
 */
	dest = (struct navelem *) ah_find(navhd, dest_node);
	if (dest == (struct navelem *) 0) {
		fprintf(stderr, "%s (r1_dolink): node %d not in nav. table\n",
			me, dest_node);
		return(ERROR);
	}
/*
 * Locate the link entry in the connectivity matrix.
 */
	p = cm_find(cmatrix, src->ne_nodeid, dest->ne_nodeid);
	if (p == (struct cmelem *) 0) {
		fprintf(stderr, "%s (r1_dolink): no (%d, %d) in matrix\n",
			me, src->ne_nodeid, dest->ne_nodeid);
		return(ERROR);
	}
	link.cl_linkno = tlink->tl_srclink;
	plink = (struct cmlink *) al_find(p->ce_cost, (char *) &link);
	if (plink == (struct cmlink *) 0) {
		fprintf(stderr, "%s (r1_dolink): no link %d in matrix\n",
			me, link.cl_linkno);
		return(ERROR);
	}
/*
 * Compute the new weight for the destination node.
 */
	new_weight = src->ne_mincost + plink->cl_cost + nail.bs_tweight.tw_hop;
	if (new_weight < dest->ne_mincost) {
/*
 * The new path to the destination node through the source node is shorter
 * than the one previously used.  Modify the entry of the destination node
 * to use this path and insert that node in the navigation queue.
 */
		dest->ne_mincost = new_weight;
		dest->ne_prevnode = src;
		dest->ne_prevlink = plink;

		if (aq_shove(navqd, (void *) &dest)) {
			fprintf(stderr,
				"%s (r1_dolink): cannot nav. enqueue node %d\n",
				me, dest->ne_nodeid);
			return(ERROR);
		}
	}

	return(0);
}

/*
 *	r1_backward
 *
 *	Function:	- navigate backwards the path found during the
 *			  forward propagation step (r1_forward)
 *			- going from the destination to the source, 
 *			  modify the route tables and update the link weights
 *
 *	Accepts:	- source nodeid
 *			- destination nodeid
 *			- path weight increment
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_backward(srcid, destid, weight)

int4			srcid,
			destid,
			weight;

{
	struct navelem	*p;		/* favorite pointer */
	struct navelem	*pp;		/* another pointer */
	struct cmelem	*pcm;		/* ptr matrix entry */
/*
 * Locate the destination node entry in the navigation hash table.
 */
	p = (struct navelem *) ah_find(navhd, destid);
	if (p == (struct navelem *) 0) {
		fprintf(stderr, "%s (r1_backward): node %d not in nav. table\n",
			me, destid);
		return(ERROR);
	}
/*
 * Loop over the path found, from the destination to the source.
 */
	pp = (struct navelem *) 0;
	for ( ; p; p = p->ne_prevnode) {
		DBUG("%d", p->ne_nodeid);
/*
 * Set the current node's route table event for the destination node.
 * Skip the destination node's entry since it's initialized to NOTEVENT.
 */
		if (p->ne_nodeid != destid) {
			pcm = cm_find(cmatrix, p->ne_nodeid, destid);
			if (pcm == (struct cmelem *) 0) {
				fprintf(stderr,
				"%s (r1_backward): no (%d , %d) in matrix\n",
					me, p->ne_nodeid, destid);
				return(ERROR);
			}

			pcm->ce_event = (pp->ne_prevlink)->cl_linkno;
		}
/*
 * Update the cost and usage of the link used to reach the current node.
 */
		if (p->ne_prevnode) {
			DBUG(" <- ");
			(p->ne_prevlink)->cl_cost += weight;
			(p->ne_prevlink)->cl_use += 1;
		}

		pp = p;
	}
/*
 * Declare the current path from source to destination to be done.
 */
	if ((pcm = cm_find(cmatrix, srcid, destid)) == (struct cmelem *) 0) {
		fprintf(stderr, "%s (r1_backward): no (%d , %d) in matrix\n",
			me, srcid, destid);
		return(ERROR);
	}
	pcm->ce_done = TRUE;

	DBUG("\n");
	return(0);
}

/*
 *	r1_stats
 *
 *	Function:	- compute link weight and usage statistics
 *			- compute the means, standard deviations,
 *			  minima and maxima
 *
 *	Returns:	- 0 or ERROR
 */
static int
r1_stats()

{
	struct cmstats	*st;		/* ptr to stats info */
	struct t_link	*tlink;		/* ptr TLINK info */
	struct cmelem	*pcm;		/* ptr matrix entry */
	struct cmlink	lk;		/* link cost/usage element */
	struct cmlink	*p;		/* favorite pointer */
	int4		srcid;		/* link source node */
	int4		destid;		/* link destination node */
	int4		sum_weight;	/* sum of link weights */
	int4		sum2_weight;	/* sum of link weights squared */
	int4		sum_usage;	/* sum of link usages */
	int4		sum2_usage;	/* sum of link usages squared */
/*
 * Initialize.
 */
	st = &(cmatrix->cd_stats);
	st->cs_nlink = sum_weight = sum2_weight = INT4_NIL;
	sum_usage = sum2_usage = INT4_NIL;
	st->cs_r1minlw = st->cs_r1minlu = INT4_MAX;
	st->cs_r1maxlw = st->cs_r1maxlu = INT4_NIL;
/*
 * Loop over all the TLINKs in the system.
 */
	for (tlink = nail.bs_tlink; tlink; tlink = tlink->tl_next) {
/*
 * Locate the current link entry in the connectivity matrix.
 */
		srcid = (tlink->tl_srcnode)->tn_nodeid;
		destid = (tlink->tl_destnode)->tn_nodeid;
		pcm = cm_find(cmatrix, srcid, destid);
		if (pcm == (struct cmelem *) 0) {
			fprintf(stderr, "%s (r1_stats): no (%d,%d) in matrix\n",
				me, srcid, destid);
			return(ERROR);
		}
/*
 * Locate the link cost/usage entry.
 */
		if (pcm->ce_cost == (LIST *) 0) {
			fprintf(stderr,
				"%s (r1_stats): empty cost list at (%d, %d)\n",
				me, srcid, destid);
			return(ERROR);
		}
		lk.cl_linkno = tlink->tl_srclink;
		p = (struct cmlink *) al_find(pcm->ce_cost, (char *) &lk);
		if (p == (struct cmlink *) 0) {
			fprintf(stderr,
				"%s (r1_stats): no link %d at (%d, %d)\n",
				me, srcid, destid);
			return(ERROR);
		}
/*
 * Process this link information.
 */
		DBUG("%s: link %d -> %d, linkno %d\tweight: %d\tusage: %d\n",
			me, srcid, destid, p->cl_linkno, p->cl_cost, p->cl_use);
		st->cs_nlink++;
		sum_weight += p->cl_cost;
		sum2_weight += (p->cl_cost * p->cl_cost);
		sum_usage += p->cl_use;
		sum2_usage += (p->cl_use * p->cl_use);
		st->cs_r1minlw = min(st->cs_r1minlw, p->cl_cost);
		st->cs_r1maxlw = max(st->cs_r1maxlw, p->cl_cost);
		st->cs_r1minlu = min(st->cs_r1minlu, p->cl_use);
		st->cs_r1maxlu = max(st->cs_r1maxlu, p->cl_use);
	}
/*
 * Compute the results of this statistical exercise.
 */
	if (st->cs_nlink > INT4_NIL) {
		st->cs_r1avelw = ((float) sum_weight) / st->cs_nlink;
		st->cs_r1avelu = ((float) sum_usage) / st->cs_nlink;
		st->cs_r1stdlw = sqrt( (((double) sum2_weight) / st->cs_nlink) -
			(st->cs_r1avelw * st->cs_r1avelw) );
		st->cs_r1stdlu = sqrt( (((double) sum2_usage) / st->cs_nlink) -
			(st->cs_r1avelu * st->cs_r1avelu) );
	}
	else {
		st->cs_r1minlw = st->cs_r1minlu = INT4_NIL;
		st->cs_r1avelw = st->cs_r1avelu = (float) 0;
		st->cs_r1stdlw = st->cs_r1stdlu = (double) 0;
	}

	return(0);
}
