/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	map_route2.c,v $
 * Revision 2.2  92/04/30  14:56:54  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/12/04  17:20:02  raja
 * Added link weight option to route2 algorithm.
 * 
 * Revision 2.1  91/11/26  18:11:57  raja
 * initial revision
 * 
 *	Function:	- generate the second route table for the map tool
 *			- used by calling map_route2()
 *			- the algorithm generates minimum-hop spanning trees
 *			  rooted at each node and uses these tress to compute
 *			  the routing table
 *			- these routes are used during multicasts/multireels
 */

#include <stdio.h>
#include <math.h>
#include <portable.h>
#include <typical.h>
#include <debug.h>
#include <net.h>
#include <events.h>
#include <all_hash.h>
#include <all_list.h>
#include <all_queue.h>
#include <nail.h>
#include <map.h>
#include <routeio.h>

/*
 * external variables
 */
extern struct bschema	nail;		/* network information */
extern struct cmdesc	*cmatrix;	/* connectivity matrix descriptor */
extern int		fl_debug;	/* debug mode flag */
extern int		fl_stats;	/* statistics flag */
extern int		fl_verbose;	/* verbose mode flag */
extern int		fl_weight;	/* use link weights flag */
extern char		*me;		/* tool name */

/*
 * static variables
 */
static HASH		*treehd;	/* navigation tree hash desc. */
static QUEUE		*navqd;		/* navigation queue desc. */
static int4		sum_hops;	/* sum of broadcast hops */
static int4		sum2_hops;	/* sum of hops squared */
static int4		min_hops;	/* minimum broadcast hops */
static int4		max_hops;	/* maximum broadcast hops */

/*
 * static functions
 */
static int		r2_dolink();
static int		r2_init();
static int		r2_stats();
static int		r2_sptree();
static int4		r2_backtrace();
static int4		r2_extract();
static void		r2_free();
static void		r2_reset();

/*
 *	map_route2
 *
 *	Function:	- generate the second route table for each node
 *			- routes are used for multicasting/multireeling
 *			- routes are overlayed onto min-hop spanning trees
 *
 *	Returns:	- 0 or ERROR
 */
int
map_route2()

{
	struct stelem	*node;		/* ptr spanning tree entry */
	int4		bhops;		/* tree broadcast hops */

	VERBOSE("generating route 2 ...\n");
/*
 * Initialize the spanning tree building structures.
 */
	if (r2_init()) {
		return(ERROR);
	}
/*
 * Loop over all nodes in the system.
 */
	node = (struct stelem *) ah_top(treehd);
	while (node) {
/*
 * Reset the tree structure.
 */
		r2_reset();
/*
 * Build the spanning tree rooted at the current node.
 */
		if (r2_sptree(node)) {
			return(ERROR);
		}
/*
 * Extract the routing information from the spanning tree and 
 * fill the route table entries accordingly.
 * Update the statistics counters.
 */
		if ((bhops = r2_extract()) < INT4_NIL) {
			return(ERROR);
		}
		else if (fl_stats) {
			sum_hops += bhops;
			sum2_hops += (bhops * bhops);
			min_hops = min(min_hops, bhops);
			max_hops = max(max_hops, bhops);
		}

		node = (struct stelem *) ah_next(treehd, (void *) node);
	}
/*
 * Compute the statistics.
 */
	if (fl_stats && r2_stats()) {
		r2_free();
		return(ERROR);
	}

	r2_free();
	return(0);
}

/*
 *	r2_init
 *
 *	Function:	- initialize the data structures required to
 *			  navigate the network generating route 2.
 *
 *	Returns:	- 0 or ERROR
 */
static int
r2_init()

{
	struct t_node	*tnode;		/* ptr to TNODE */
	struct stelem	entry;		/* spanning tree entry */
/*
 * Initialize the hash table used for building the spanning tree.
 */
	treehd = ah_init(ah_size(nail.bs_tnhd),
		(int4) sizeof(struct stelem), NOTNODEID, AHFIXED);
	if (treehd == (HASH *) 0) {
		fprintf(stderr, "%s (r2_init): cannot malloc hash table\n", me);
		return(ERROR);
	}

	tnode = (struct t_node *) ah_top(nail.bs_tnhd);
	while (tnode) {
		entry.st_nodeid = tnode->tn_nodeid;
		entry.st_tnode = tnode;
		if (ah_insert(treehd, (void *) &entry)) {
			fprintf(stderr, "%s (r2_init): cannot insert node %d\n",
				me, tnode->tn_nodeid);
			return(ERROR);
		}

		tnode = (struct t_node *) ah_next(nail.bs_tnhd, (void *) tnode);
	}
/*
 * Initialize the navigation queue.
 */
	navqd = aq_init((int4) 1, (int4) sizeof(struct stelem *));
	if (navqd == (QUEUE *) 0) {
		fprintf(stderr, "%s (r2_init): cannot malloc queue\n", me);
		return(ERROR);
	}
/*
 * Clear the statistics adders.
 */
	sum_hops = sum2_hops = max_hops = INT4_NIL;
	min_hops = INT4_MAX;

	return(0);
}

/*
 *	r2_reset
 *
 *	Function:	- reset the contents of the data structures
 *			  used for graph navigation
 */
static void
r2_reset()

{
	struct stelem	*p;		/* favorite pointer */
/*
 * Reset the contents of the tree hash table.
 */
	if (treehd) {
		p = (struct stelem *) ah_top(treehd);
		while (p) {
			p->st_mindist = INT4_MAX;
			p->st_upnode = (struct stelem *) 0;
			p->st_uplink = (struct t_link *) 0;
			p->st_children = INT4_NIL;

			p = (struct stelem *) ah_next(treehd, (void *) p);
		}
	}
}

/*
 *	r2_free
 *
 *	Function:	- free the data structures using in graph navigation
 */
static void
r2_free()

{
	if (treehd) {
		ah_free(treehd);
	}

	if (navqd) {
		aq_free(navqd);
	}
}

/*
 *	r2_sptree
 *
 *	Function:	- build a minimum-hop spanning tree rooted at
 *			  the given node
 *
 *	Accepts:	- ptr to root tree entry
 *
 *	Returns:	- 0 or ERROR
 */
static int
r2_sptree(root)

struct stelem		*root;

{
	struct stelem	*node;		/* current tree node */
	struct stelem	**p;		/* favorite pointer */
	struct link	*plk;		/* ptr list of links */

	DBUG("%s (r2_sptree): root node %d\n", me, root->st_nodeid);
/*
 * Initialize and store the root node in the queue.
 */
	root->st_mindist = INT4_NIL;
	if (aq_shove(navqd, (void *) &root)) {
		fprintf(stderr, "%s (r2_sptree): nav. enqueue error\n", me);
		return(ERROR);
	}
/*
 * Loop till the navigation queue is empty.
 */
	while (aq_count(navqd) > INT4_NIL) {
/*
 * Remove the top node in the navigation queue.  This is the current node.
 */
		p = (struct stelem **) aq_find(navqd);
		if (p == (struct stelem **) 0) {
			fprintf(stderr, "%s (r2_sptree): empty queue\n", me);
			return(ERROR);
		}
		node = *p;
		if (aq_delete(navqd)) {
			fprintf(stderr, "%s (r2_sptree): dequeue error\n", me);
			return(ERROR);
		}
/*
 * Loop over the current node's outgoing links doing a breadth-first search.
 */
		plk = ((node->st_tnode)->tn_specs)->ns_lkout;
		for ( ; plk; plk = plk->lk_next) {
			if (r2_dolink(node, plk->lk_link)) {
				return(ERROR);
			}
		}
	}

	return(0);
}

/*
 *	r2_dolink
 *
 *	Function:	- check if the given link can be used in building
 *			  the spanning tree from the given node
 *			- if used, add its end node to the navigation queue
 *			  and link the end node in the tree
 *
 *	Accepts:	- ptr current tree entry (source node)
 *			- ptr to TLINK to use
 *
 *	Returns:	- 0 or ERROR
 */
static int
r2_dolink(srcnode, tlink)

struct stelem		*srcnode;
struct t_link		*tlink;

{
	struct stelem	*destnode;		/* ptr destination node */
	int4		new_dist;		/* new dest. node distance */
/*
 * Find the destination node tree entry.
 */
	if ((destnode = (struct stelem *) ah_find(treehd,
		(tlink->tl_destnode)->tn_nodeid)) == (struct stelem *) 0) {

		fprintf(stderr, "%s (r2_dolink): node %d not in tree\n",
			me, (tlink->tl_destnode)->tn_nodeid);
		return(ERROR);
	}

	if (fl_weight) {
/*
 * We are using the link weights to generate the shortest spanning tree.
 * Check if this link yields a shorter route to the destination node.
 * If it does not, just return.  Equal route are reconsidered, in order
 * to force all trees to look the same no matter which node is the root.
 * This algorithm is slower but gives better quality routes.
 */
		new_dist = srcnode->st_mindist + tlink->tl_weight;
		if (new_dist > destnode->st_mindist) {
			return(0);
		}
		else {
			destnode->st_mindist = new_dist;
		}
	}
	else {
/*
 * We are generating minimum hop routes without using the link weights.
 * This is a faster algorithm but gives a slightly degraded quality.
 * If the destination node has been visited no further processing is needed.
 */
		if (destnode->st_mindist == INT4_NIL) {
			return(0);
		}
		else {
			destnode->st_mindist = INT4_NIL;
		}
	}
/*
 * Update the number of children for the current and previous
 * source (parent) nodes (if any).
 */
	srcnode->st_children++;
	if (destnode->st_upnode) {
		(destnode->st_upnode)->st_children--;
	}
/*
 * Link the destination node to the tree and add it to the navigation queue.
 */
	destnode->st_upnode = srcnode;
	destnode->st_uplink = tlink;

	if (aq_shove(navqd, (void *) &destnode)) {
		fprintf(stderr, "%s (r2_dolink): cannot enqueue node %d\n",
			me, destnode->st_nodeid);
		return(ERROR);
	}

	return(0);
}

/*
 *	r2_extract
 *
 *	Function:	- extract the routing information from the
 *			  spanning tree and fill the route table entries
 *
 *	Returns:	- broadcast hops or ERROR
 */
static int4
r2_extract()

{
	int4		max_hops;	/* max tree hops */
	int4		hops;		/* tree branch hops */
	struct stelem	*p;		/* favorite pointer */
/*
 * Loop over the nodes in the tree, locating the tree leaves.
 */
	max_hops = INT4_NIL;
	p = (struct stelem *) ah_top(treehd);
	while (p) {
		if (p->st_children == INT4_NIL) {
/*
 * Found a leave node.  Trace the path from it to the root, filling
 * the routing information along the way.
 */
			if ((hops = r2_backtrace(p)) < INT4_NIL) {
				return((int4) ERROR);
			}
			else {
				max_hops = max(max_hops, hops);
			}
		}
		else if (p->st_children < INT4_NIL) {
			fprintf(stderr,
				"%s (r2_extract): node %d has %d children\n",
				me, p->st_nodeid, p->st_children);
			return((int4) ERROR);
		}

		p = (struct stelem *) ah_next(treehd, (void *) p);
	}

	return(max_hops);
}

/*
 *	r2_backtrace
 *
 *	Function:	- navigate from the given node to the root of
 *			  the tree, filling the proper routing information
 *
 *	Accepts:	- ptr to tree entry
 *
 *	Returns:	- path weight or ERROR
 */
static int4
r2_backtrace(node)

struct stelem		*node;

{
	struct stelem	*cur;		/* ptr current node */
	struct stelem	*p;		/* favorite pointer */
	struct cmelem	*route;		/* route table entry */
	struct t_link	*curtlink;	/* current TLINK used */
	int4		curid;		/* current nodeid */
	int4		hops;		/* path hops weight */

	DBUG("%d", node->st_nodeid);
/*
 * Loop over all the nodes upstream from the given leaf node.
 */
	hops = INT4_NIL;
	curtlink = node->st_uplink;
	for (cur = node->st_upnode; cur; cur = cur->st_upnode) {
		if (fl_weight) {
			hops += curtlink->tl_weight;
		} else {
			hops++;
		}
		curid = cur->st_nodeid;
		DBUG(" <- %d [%d]", curid, curtlink->tl_srclink);
/*
 * For each current node, fill its routing information for nodes
 * downstream from it, going from the leaf node up (i.e. only the
 * downstream nodes that are in the tree branch being navigated).
 */
		for (p = node; p != cur; p = p->st_upnode) {

			route = cm_find(cmatrix, curid, p->st_nodeid);
			if (route == (struct cmelem *) 0) {
				fprintf(stderr,
					"%s (r2_backtrace): no (%d, %d)\n",
					me, curid, p->st_nodeid);
				return((int4) ERROR);
			}
/*
 * Fill the routing information, after doing a sanity check.
 */
			if ((route->ce_event2 != NOTEVENT) &&
				(route->ce_event2 != curtlink->tl_srclink)) {

				fprintf(stderr,
		"%s (r2_backtrace): algorithm bug: (%d, %d) event2: %d <- %d\n",
					me, curid, p->st_nodeid,
					route->ce_event2, curtlink->tl_srclink);
				return((int4) ERROR);
			}
			else {
				route->ce_event2 = curtlink->tl_srclink;
			}
		}

		curtlink = cur->st_uplink;
	}

	DBUG("\n");
	return(hops);
}

/*
 *	r2_stats
 *
 *	Function:	- compute broadcast path statistics
 *			- compute the means, standard deviations,
 *			  minima and maxima
 *
 *	Returns:	- 0 or ERROR
 */
static int
r2_stats()

{
	struct cmstats	*st;		/* ptr to stats info */
	int4		nnodes;		/* number of nodes */
/*
 * Compute the results of this statistical exercise.
 */
	st = &(cmatrix->cd_stats);
	nnodes = ah_count(nail.bs_tnhd);
	if (nnodes > INT4_NIL) {
		st->cs_r2minbh = min_hops;
		st->cs_r2maxbh = max_hops;
		st->cs_r2avebh = ((float) sum_hops) / nnodes;
		st->cs_r2stdbh = sqrt( (((double) sum2_hops) / nnodes) -
			(st->cs_r2avebh * st->cs_r2avebh) );
	}
	else {
		st->cs_r2minbh = INT4_NIL;
		st->cs_r2maxbh = INT4_NIL;
		st->cs_r2avebh = (float) 0;
		st->cs_r2stdbh = (double) 0;
	}

	return(0);
}
