/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	rasm.c,v $
 * Revision 2.2  92/04/30  15:14:33  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/05  16:35:33  gdburns
 * Look at the environment for the boot schema filename.
 * 
 * Revision 2.1.1.1  91/11/26  14:45:39  raja
 * New design/implementation including support for a second route table.
 * 
 *	Function:	- route table "assembler"
 *			- generate a route file for the routes explicitly
 *			  specified in the given boot schema (bnail file)
 *			  TLINK and TROUTE tuples only
 *			- no connectivity checks, no "intelligence"
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <malloc.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <net.h>
#include <events.h>
#include <nail.h>
#include <routeio.h>
#include <debug.h>
#include <castreq.h>
#include <all_hash.h>

/*
 * external variables
 */
extern FILE		*yyin;		/* defined in lex as input file */
extern int		errno;		/* for global error number */
 
/*
 * global variables
 */
char			*me;		/* tool name */
int			fl_verbose;	/* verbose option flag */
int			fl_debug;	/* debug option flag */
 
/*
 * static local variables
 */
static char		rtfname[256];	/* route file filename */
static char		*f_bnail;	/* boot schema filename */
static char		*full_bnail;	/* expanded boot schema filename */
static char		**p_bnail;	/* boot schema pathname prefix */
static int4		extra;		/* extra rtable entries for castids */
static struct bschema	nail;		/* bnail file information */

static char *usage =
	"usage: %s [-dhvV] [-i <extra>] [-c <bnail>] [-p <route>]\n";

/* 
 * external functions
 */
extern void		exit();
extern void		followed();
extern void		validopts();
extern char		*findpath();
extern char		*getenv();
extern char		*getparam();
extern int		addarg();
extern int		do_args();
extern int		on_span();
extern int		opt_taken();
extern int		parse_nail();
extern int4		intparam();
extern int4		stoi4();
 
/*
 * static functions
 */
static void		help_rasm();
static void		prt_defaults();
static void		set_defaults();
static int		rasm_create();
static int		rasm_tlink();
static int		rasm_troute();


main(argc, argv)  

int			argc;
char			*argv[];

{
/*
 * Assign the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("dhvV");
	followed("cip");

	if ((do_args(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, usage, me);
		} else {
			terror(me);
		}
		exit(ERROR);
	}
/*        
 * Print the help menu if needed.
 */
	if (opt_taken('h')) {
		help_rasm();
		exit(0);
	}
/*
 * Set the mode of operation.
 */
	set_defaults();

	if (fl_debug) {
		prt_defaults();
	}
/*
 * Locate the Boot Schema file.
 */
	if ((full_bnail = findpath(f_bnail, p_bnail, R_OK)) == NULL) {
		fprintf(stderr, "%s: cannot locate bnail file \"%s\"\n",
			me, f_bnail);
		exit(ERROR);
	}
/*
 * Open the boot schema file.
 */
	DBUG("opening the boot schema file: ...\n", full_bnail);
	if ((yyin = fopen(full_bnail, "r")) == NULL) {
		perror(me);
		exit(errno);
	}
/*
 * Parse the boot schema.
 */
	if (parse_nail(&nail)) {
		fprintf(stderr, "%s: cannot parse the NaIL syntax\n", me);
		exit(ERROR);
	}
/*
 * Close and pretty print the boot schema.
 */
	(void) fclose(yyin);

	if ((opt_taken('V')) || (fl_debug)) {
		pretty_bnail(&nail);
		if (opt_taken('V')) {
			exit(0);
		}
        }
/*
 * Determine the host node of the network.
 * It will be the root of the spanning tree.
 */
	if (find_host(&nail)) {
		fprintf(stderr, "%s: cannot find the network host\n", me);
		exit(ERROR);
	}
	DBUG("%s: found hostnode %d\n", me, nail.bs_hostnode->tn_nodeid);
/*
 * Generate the network spanning tree.
 */
	DBUG("%s: generating the spanning tree ...\n", me);
	if (make_spantree(&nail)) {
		fprintf(stderr, "%s: cannot generate spanning tree\n", me);
		exit(ERROR);
	}
	if (fl_debug) {
		print_spantree(&nail);
	}

	if (rasm_create()) {
		DBUG("%s: unable to create the route file\n", me);
		exit(ERROR);
	}

	VERBOSE("\nroute file created\n");
	exit(0);
}
 
/*
 *	help_rasm
 *
 *	Function:	- print the help menu
 */
static void 
help_rasm()

{
	printf("\nSynopsis:\t%s [options]\n", me);
	printf("\nDescription:\tAssemble a route file from a Boot Schema.\n");
	printf("\nOptions:\t-h\t\tthis message\n");
	printf("\t\t-v\t\tverbose mode\n");
	printf("\t\t-d\t\tturn on debugging\n");
	printf("\t\t-V\t\tpretty print the bnail file\n");
	printf("\t\t-c bnail\tuse \"bnail\" as the Boot Schema file\n");
	printf("\t\t-i extra\tadd \"extra\" entries to each route table\n");
	printf("\t\t-p route\tuse \"route\" as the output route file\n");
}

/*
 *	set_defaults
 *
 *	Function: 	- set the values of key variables controlling
 *			  the operation of the tool
 */
static void 
set_defaults()

{
	int		n_bnail;
	char		*p;
/*
 * Pathnames for the boot schema file.
 */
	n_bnail = 0;
	p_bnail = (char **) 0;
	addarg(&n_bnail, &p_bnail, "");				/* absolute */
	addarg(&n_bnail, &p_bnail, "$TROLLIUSHOME/boot");	/* env */
	addarg(&n_bnail, &p_bnail, DEFPBNAIL);			/* default  */
/*
 * Pathnames for the files.
 */
	if (opt_taken('c')) {
		f_bnail = getparam('c');
	} else if (f_bnail = getenv("TROLLIUSBNAIL")) {
        } else {
		f_bnail = DEFFBNAIL;
	}
 
	fl_verbose = opt_taken('v');

	if (opt_taken('d')) {
		fl_debug = TRUE;
		fl_verbose = TRUE;
	} else {
		fl_debug = FALSE;
	}
 
	if (opt_taken('i')) {
		extra = stoi4(getparam('i'));
	} else {
		extra = MAXCAST;
	}
 
	if (opt_taken('p')) {
		(void) strcpy(rtfname, getparam('p'));
	} else {
		(void) strcpy(rtfname, DEFFROUTE);
		if (p = strchr(rtfname, '.')) {
			*p = '\0';
		}
		p = strrchr(f_bnail, STRDIR);
		if (p = strchr((p) ? p + 1 : f_bnail, '.')) {
			(void)strcat(rtfname, ".");
			(void)strcat(rtfname, p + 1);
		}
	}
}

/*
 *	prt_defaults
 *
 *	Function:	- print the values of key variables
 *                      - used for debugging
 */
static void 
prt_defaults()

{
	printf("\nDefault settings\n\n");
	printf("The boot file (input): %s\n", f_bnail);
	printf("The route file (output): %s\n", rtfname);
	printf("debug flag: %d\n", fl_debug);
	printf("verbose flag: %d\n", fl_verbose);
	printf("extra entries: %d\n", extra);
}

/*
 *	rasm_create
 *
 *	Function:	- create a "rasm level" route file
 *
 *	Returns:	- 0 or ERROR
 */
static int
rasm_create()

{
	RTIOFD		*rfd;		/* route file descriptor */
	int4		nodeid;		/* target nodeid */
	int		i;		/* favorite counter */
/*
 * Create a route file descriptor.
 */
	if ((rfd = rtio_init(rtfname, &nail, extra)) == (RTIOFD *) 0) {
		DBUG("%s (rtio_init): cannot create route file desc.\n", me);
		return(ERROR);
	}
/*
 * Create the header and the directory of the route file.
 */
	VERBOSE("creating the route file ...\n");
	if (rtio_header(rfd) || rtio_dir(rfd)) {
		rtio_cleanup(rfd);
		return(ERROR);
	}
/*
 * Loop through all nodes, in the required order, creating their
 * route tables and writing them to the route file.
 */
	i = 0;
	rtio_topnode(rfd);
	while ((nodeid = rtio_thisnode(rfd)) != NOTNODEID) {
/*
 * Beautify the output in the pure verbose mode.
 */
		VERBOSE("[%d] ", nodeid);
		if (fl_verbose && ! fl_debug) {
			if (++i >= 8) {
				printf("\n");
				i = 0;
			}
		}
/*
 * Initialize the internal route table with the nodeids and node types.
 */
		if (rtio_rtinit(rfd)) {
			rtio_cleanup(rfd);
			return(ERROR);
		}
/*
 * Handle any relevant TLINK and TROUTE tuples for this node's route table.
 * TLINK tuples have a higher priority than the TROUTE tuples.
 */
		if (rasm_tlink(rfd, nodeid) || rasm_troute(rfd, nodeid)) {
			rtio_cleanup(rfd);
			return(ERROR);
		}
/*
 * Write the node's route table in the route file.
 */
		if (rtio_rtwrite(rfd, nodeid)) {
			rtio_cleanup(rfd);
			return(ERROR);
		}
	}

	rtio_cleanup(rfd);
	return(0);
}

/*
 *	rasm_tlink
 *
 *	Function:	- enter the TLINK information for the route
 *			  table of the given nodeid
 *
 *	Accepts:	- ptr to route file descriptor
 *			- target nodeid
 *
 *	Returns:	- 0 or ERROR
 */
static int
rasm_tlink(rfd, nodeid)

RTIOFD			*rfd;
int4			nodeid;

{
	struct t_link	*tlink;		/* ptr to TLINK information */
	struct route	*p;		/* favorite pointer */

	DBUG("\t\tdoing the TLINKs ...\n");
 
	for (tlink = (rfd->rd_net)->bs_tlink; tlink; tlink = tlink->tl_next) {
		if (nodeid != (tlink->tl_srcnode)->tn_nodeid) {
			continue;
		}
/*
 * Found a TLINK headed by the target nodeid.
 */
		p = (struct route *) ahs_find(&(rfd->rd_route),
			(tlink->tl_destnode)->tn_nodeid);
		if (p == (struct route *) 0) {
			fprintf(stderr,
				"%s (rasm_tlink): node %d not in route table\n",
				me, (tlink->tl_destnode)->tn_nodeid);
			return(ERROR);
		}
/*
 * Store the route table information at the current index.
 */
		p->r_event = tlink->tl_srclink;
		p->r_event2 = tlink->tl_srclink;
		p->r_nodetype |= NT_JONES;
		if (on_span(tlink)) {
			p->r_nodetype |= NT_BOOT;
		}
	}
 
	return(0);
}

/* 
 *	rasm_troute
 *
 *	Function:	- enter the TROUTE information for the route
 *			  table of the given nodeid
 *
 *	Accepts:	- ptr to route file descriptor
 *			- target nodeid
 *
 *	Returns:	- 0 or ERROR
 */
static int
rasm_troute(rfd, nodeid)

RTIOFD			*rfd;
int4			nodeid;

{
	struct t_route	*troute;		/* ptr to TROUTE info */
	struct t_node	*tnode;			/* ptr to TNODE info */
	struct route	*p;			/* favorite pointer */

	DBUG("\t\tdoing the TROUTEs ...\n");
/*
 * Locate the target node's TNODE in the network information.
 */
	tnode = (struct t_node *) ahs_find((rfd->rd_net)->bs_tnhd, nodeid);
	if (tnode == (struct t_node *) 0) {
		fprintf(stderr, "%s (rasm_troute): no node %d in the network\n",
			me, nodeid);
		return(ERROR);
	}
/*
 * Loop over the TROUTE tuples of the target node.
 */
	troute = (tnode->tn_specs)->ns_troute;
	for ( ; troute; troute = troute->tr_next) {

		if ((p = (struct route *) ahs_find(&(rfd->rd_route),
			troute->tr_destnode)) == (struct route *) 0) {

			fprintf(stderr,
				"%s (rasm_troute): n%d not in route table\n",
				me, troute->tr_destnode);
			return(ERROR);
		}
/*
 * Store the route table information at the current entry only if the entry
 * is still empty otherwise the previous information takes precedence.
 */
		if  ((p->r_event == NOTEVENT) && (p->r_event2 == NOTEVENT)) {
			p->r_event = troute->tr_destlink;
			p->r_event2 = troute->tr_destlink2;
		}
		else if ((p->r_event != troute->tr_destlink) ||
			(p->r_event2 != troute->tr_destlink2)) {

			VERBOSE("Warning: TROUTE <%d, %d, %d, %d> ignored\n",
				nodeid, troute->tr_destnode,
				troute->tr_destlink, troute->tr_destlink2);
		}
	}

	return(0);
}
