/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	sboot.c,v $
 * Revision 2.2  92/04/30  15:15:12  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  16:02:07  gdburns
 * major rewrite
 * 
 *	Function:	- boots ITB operating system on neighbour
 *			  of OTB node
 *			- toolkit layer utility
 */

#include <boot.h>
#include <bootreq.h>
#include <conf.h>
#include <debug.h>
#include <machine/machine.h>
#include <machine/maconfig.h>
#include <stdio.h>
#include <sys/errno.h>
#include <terror.h>
#include <typical.h>
#include <unistd.h>

/*
 * external variables
 */
extern int		errno;
extern char		**exttopv;	/* -T will override conf topology */
extern int		exttopc;

/*
 * external functions
 */
extern void		split();
extern void		cnfparse();
extern void		exclusive();
extern void		followed();
extern void		validopts();
extern int		addarg();
extern int		cnfexec();
extern int		do_args();
extern int		unlink();
extern int4		cnfshare();
extern char		*findpath();
extern char		*getparam();
extern char		*rindex();
extern char		*strcpy();
extern char		*strcat();
extern CONF		*cnffind();

/*
 * local functions
 */
static void		help();
static void		sbexec();
static void		setdefaults();
static CONF		*config();

/*
 * local variables
 */
static char		*f_conf;		/* process schema file */
static char		*f_device;		/* link device */
static char		*f_frozen;		/* frozen boot file */
static char		*f_strap;		/* bootstrap program */
static char		*me;			/* who am I? */
static char		**p_conf;		/* conf pathname prefixes */
static char		**p_prog;		/* program pathname prefixes */
static char		*t_cboot;		/* 1 program boot tool */
static char		*t_export;		/* export tool */
static char		*t_reset;		/* reset tool */
static char		*t_thaw;		/* thaw tool */

static int		fl_debug;		/* debugging option */
static int		fl_pretend;		/* pretend option */
static int		fl_verbose;		/* verbose option */
static int		n_prog;			/* # of p_prog prefixes */
static int		n_conf;			/* # of p_conf prefixes */

static char
	*usage = " [-bdhrvxzNRVZ] [-c <conf>] [-f <device>] [-i <frozen>] [-T <topology>]";

main(argc, argv)

int			argc;
char			*argv[];

{
	CONF		*cnftop;	/* parsed process schema list */
	CONF		*p;		/* favourite ptr */
	char		*full;		/* full pathname */
	char		*tail;		/* tail of full pathname */
	char		**toolv;	/* tool invocation arg vector */
	int4		i;		/* favourite index */
	int4		loadpt;		/* remote addr of header/text */
	int4		newmark;	/* new high water mark */
	int4		onchipmark;	/* current onchip high water mark */
	int4		watermark;	/* current high water mark */
	int		n;		/* favourite counter */
	int		toolc;		/* # toolv args */
	int		mfd[2];		/* pipe to cboot child */
	char		anum[16];	/* ascii representation of number */

	me = argv[0];
	cnftop = CNFNULL;
/*
 * Initialize option parser.
 */
	validopts("bdhrvxzNRVZ");
	followed("cfiT");
	exclusive("zZ");

	if ((do_args(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		fprintf(stderr, "usage: %s%s\n", me, usage);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	setdefaults();

	if (!opt_taken('z') && !opt_taken('b')) {
		cnftop = config();

		if (! cnftop) {
			fprintf(stderr, "%s: nothing to do in %s\n", me,
				f_conf);
			exit(0);
		}

		if (opt_taken('V')) {
			cnfprint(cnftop);
			exit(0);
		}
	}
/*
 * Remove any existing frozen file, if freezing.
 */
	if (opt_taken('Z')) {

		if ((unlink(f_frozen) < 0) && (errno != ENOENT)) {
			fprintf(stderr, "%s: can't remove %s\n", me, f_frozen);
			perror(me);
			exit(errno);
		}
	}
/*
 * Execute the reset program.
 */
	if (!opt_taken('r') && !opt_taken('Z')) {
		VERBOSE("reset ...\n");
		toolc = 0;
		toolv = (char **) 0;
		addarg(&toolc, &toolv, t_reset);
		addarg(&toolc, &toolv, f_device);

		sbexec(toolv);
	}
/*
 * Execute the export program.
 */
	if (! opt_taken('x') && !opt_taken('Z')) {
/*
 * Find the bootstrap program.
 */
		full = findpath(f_strap, p_prog, R_OK | X_OK);

		if (! full) {
			fprintf(stderr, "%s: can't find strap file - %s\n",
					me, f_strap);
			exit(1);
		} else {
			f_strap = full;
			DBUG("%s: f_strap = \"%s\"\n", me, f_strap);
		}

		VERBOSE("export ...\n");
		toolc = 0;
		toolv = (char **) 0;
		addarg(&toolc, &toolv, t_export);
		addarg(&toolc, &toolv, f_strap);
		addarg(&toolc, &toolv, f_device);

		sbexec(toolv);
	}

	if (opt_taken('b')) {
		return(0);
	}
/*
 * Boot with the frozen file.
 */
	if (opt_taken('z')) {
/*
 * Find the frozen file.
 */
		full = findpath(f_frozen, p_conf, R_OK);

		if (! full) {
			fprintf(stderr, "%s: can't find frozen file - %s\n",
					me, f_frozen);
			exit(1);
		} else {
			f_frozen = full;
			DBUG("%s: f_frozen = \"%s\"\n", me, f_frozen);
		}

		VERBOSE("frozen boot ...\n");
		toolc = 0;
		toolv = (char **) 0;
		addarg(&toolc, &toolv, t_thaw);
		addarg(&toolc, &toolv, f_frozen);
		addarg(&toolc, &toolv, f_device);

		sbexec(toolv);
	}
/*
 * Boot with the process schema.
 */
	else {
		VERBOSE("boot ...\n");
		watermark = DEFMEMSTART;
		onchipmark = DEFONCHIPSTART;
		n = 0;
		p = cnftop;

		if (pipe(mfd) < 0) {
			perror(me);
			exit(errno);
		}
/*
 * Loop through all the programs in the parsed config file.
 */
		while (p) {
			toolc = 0;
			toolv = (char **) 0;
			addarg(&toolc, &toolv, t_cboot);
			addarg(&toolc, &toolv, p->cnf_argv[0]);

			if (opt_taken('Z')) {
				addarg(&toolc, &toolv, f_frozen);
			} else {
				addarg(&toolc, &toolv, f_device);
			}

			if ((p->cnf_next == CNFNULL) && !opt_taken('R')) {
				addarg(&toolc, &toolv, "-bc");
			} else {
				addarg(&toolc, &toolv, "-br");
			}

			sprintf(anum, "0x%X", watermark);
			addarg(&toolc, &toolv, "-l");
			addarg(&toolc, &toolv, anum);

			sprintf(anum, "0x%X", p->cnf_stksize);
			addarg(&toolc, &toolv, "-s");
			addarg(&toolc, &toolv, anum);

			if (p->cnf_flags & BTONCHIPSTK) {
				sprintf(anum, "0x%X", onchipmark);
				addarg(&toolc, &toolv, "-w");
				addarg(&toolc, &toolv, anum);
				onchipmark += p->cnf_stksize;
			}
/*
 * Look for possible text segment sharing and if found, use the
 * shared module's text ptr.
 */
			p->cnf_loadpt = watermark;

			if (p->cnf_flags & BTSHARE) {
				loadpt = cnfshare(p, cnftop);

				if (loadpt) {
					p->cnf_loadpt = loadpt;
					sprintf(anum, "0x%X", loadpt);
					addarg(&toolc, &toolv, "-t");
					addarg(&toolc, &toolv, anum);
				}
			}

			if (! opt_taken('Z')) {
				addarg(&toolc, &toolv, "-L");
			}

			sprintf(anum, "%d", mfd[1]);
			addarg(&toolc, &toolv, "-m");
			addarg(&toolc, &toolv, anum);

			sprintf(anum, "%d", p->cnf_priority);
			addarg(&toolc, &toolv, "-k");
			addarg(&toolc, &toolv, anum);

			sprintf(anum, "%d", n);
			addarg(&toolc, &toolv, "-p");
			addarg(&toolc, &toolv, anum);

			addarg(&toolc, &toolv, "--");

			for (i = 0; i < p->cnf_argc; i++) {
				addarg(&toolc, &toolv, p->cnf_argv[i]);
			}

			if (fl_debug) {
				printf("sboot:");

				for (i = 0; i < toolc; ++i) {
					printf(" %s", toolv[i]);
				}

				printf("\n");
			}

			sbexec(toolv);

			if (!fl_pretend) {
				if (read(mfd[0], (char *) &newmark,
						sizeof(int4)) < sizeof(int4)) {
					perror(me);
					exit(1);
				}
			}

	    		n++;

			if (fl_verbose) {
				tail = rindex(p->cnf_argv[0], '/');
				tail = (tail) ? tail + 1 : p->cnf_argv[0];
				printf("[%2d] %6x %s", n, watermark & ~MININT,
					tail);

				for (i = 1; i < p->cnf_argc; i++) {
		    			printf(" %s", p->cnf_argv[i]);
				}

				printf("\n");
			}

			if (!fl_pretend) {
				watermark = newmark;
			}

			p = p->cnf_next;
/*
 * Check if watermark limits have been exceeded.
 */
			if ((watermark > MAIN_START) ||
					(onchipmark > ONCHIP_START)) {
				fprintf(stderr,
					"%s: high water limit exceeded\n", me);
				exit(EHIGHWATER);
			}
		}
	}

	DBUG("%s: final watermark = %x\n", me, watermark);

	return(0);
}

/*
 *	config
 *
 *	Function:	- reads in the process schema
 *			- may print error messages
 *			- may exit()
 *
 *	Returns:	- ptr to process schema
 */

static CONF *
config()

{
	CONF		*cnftop;	/* top of schema program list */
	CONF		*p;		/* favourite ptr */
	FILE		*fp_conf;	/* schema open file ptr */
	char		*full;		/* schema full pathname */
/*
 * Locate the conf file.
 */
	full = findpath(f_conf, p_conf, R_OK);

	if (! full) {
		fprintf(stderr, "%s: can't find conf file - %s\n", me, f_conf);
		exit(1);
	} else {
		f_conf = full;
	}
/*
 * Dump debugging information.
 */
	DBUG("%s: f_conf = \"%s\"\n", me, f_conf);
/*
 * Open the conf file.
 */
	fp_conf = fopen(f_conf, "r");

	if (! fp_conf) {
		fprintf(stderr, "%s: can't open conf file - %s\n", me, f_conf);
		perror(me);
		exit(errno);
	}
/*
 * Allow -T to override conf file topology variable.
 */
	if (opt_taken('T')) {
		exttopv = (char **) 0;
		exttopc = 0;
		split(getparam('T'), &exttopv, &exttopc);
	}
/*
 * Parse the conf file.
 */
	cnfparse(fp_conf, &cnftop);
/*
 * Find programs with sufficient permissions.
 */
	if ((p = cnffind(p_prog, cnftop)) != CNFNULL) {
		fprintf(stderr, "%s: can't find %s with permissions\n",
			me, p->cnf_argv[0]);
		exit(1);
	}

	if (fl_debug) {

		for (p = cnftop; p; p = p->cnf_next) {
			printf("%s: found %s\n", me, p->cnf_argv[0]);
		}
	}

	return(cnftop);
}

/*
 *	setdefaults
 *
 *	Function:	- sets default files and paths
 */

static void
setdefaults()

{
/*
 * prefix paths
 */
	n_conf = 0;
	p_conf = (char **) 0;
	addarg(&n_conf, &p_conf, "");
	addarg(&n_conf, &p_conf, "$TROLLIUSHOME/boot");
	addarg(&n_conf, &p_conf, DEFPCONFIG);

	n_prog = 0;
	p_prog = (char **) 0;
	addarg(&n_prog, &p_prog, "");
	addarg(&n_prog, &p_prog, "$TROLLIUSHOME/itb/bin");
	addarg(&n_prog, &p_prog, DEFPPROGS);
/*
 * flags
 */
	fl_debug = opt_taken('d');;
	fl_verbose = opt_taken('v');
	fl_pretend = opt_taken('N');
/*
 * files
 */
	if (opt_taken('f')) {
		f_device = getparam('f');
	} else {
		f_device = DEFFDEVICE;
	}

	if (opt_taken('c')) {
		f_conf = getparam('c');
	} else {
		f_conf = DEFFCONFIGS;
	}

	if (opt_taken('i')) {
		f_frozen = getparam('i');
	} else {
		f_frozen = DEFFFROZENS;
	}

	f_strap = DEFFSTRAP;
/*
 * tools
 */
	t_reset = DEFTRESETS;
	t_export = DEFTEXPORT;
	t_cboot = DEFTCBOOT;
	t_thaw = DEFTTHAW;
}

/*
 *	help
 *
 *	Function:	- prints helpful information on the sboot command
 */

static void
help()

{
    printf("\nSynopsis:	sboot [options]\n");
    printf("\nDescription:	Boot a neighbour ITB node.\n");
    printf("\nOptions:	-h		Print this message.\n");
    printf("\t\t-d		Print debug information.\n");
    printf("\t\t-v		Be verbose.\n");
    printf("\t\t-b		Skip the boot step.\n");
    printf("\t\t-r		Skip the reset step.\n");
    printf("\t\t-x		Skip the export step.\n");
    printf("\t\t-z		Perform a frozen boot.\n");
    printf("\t\t-N		Pretend to sboot.\n");
    printf("\t\t-R		Retain bootstrap control over link.\n");
    printf("\t\t-V		Pretty print the conf file.\n");
    printf("\t\t-Z		Create a frozen file.\n");
    printf("\t\t-c <conf>	Use <conf> as the process schema.\n");
    printf("\t\t-f <device>	Use <device> for the output device.\n");
    printf("\t\t-i <frozen>	Use <frozen> for the frozen file.\n");
    printf("\t\t-T <topo>	Set $topology variable.\n\n");
}

/*
 *	sbexec
 *
 *	Function:	- executes tool and checks errors
 *			- may call exit()
 *
 *	Accepts:	- arg vector for tool invocation
 */

static void
sbexec(toolv)

char			**toolv;

{
	int		status;		/* exit status of child process */

	if (!fl_pretend) {
		status = cnfexec(toolv);

		if (status < 0) {
			fprintf(stderr, "%s: can't exec %s\n", me, toolv[0]);
			perror(me);
			exit(errno);
		} else if (status > 0) {
			fprintf(stderr, "%s: %s exited with code %d\n",
					me, toolv[0], status);
			exit(status);
		}
	}
}
