/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	face.c,v $
 * Revision 2.2  92/04/30  15:15:45  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/01/31  18:20:46  gdburns
 * Upgrade to suit the overhauled boot subsystem.
 * 
 * Revision 2.1  91/03/20  13:03:16  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  91/03/11  01:48:29  gdburns
 * Added recursive boot option.
 * 
 * Revision 2.0  90/11/16  16:25:52  raja
 * Overhaul of the topology layer tools.
 * 
 *	Function:	- spread system interface
 *			- native O.S. specific (Unix(TM) version)
 *			- each command forks a child process and wait
 *			- the tserver call is a stub
 *
 *	Accepts:	- tool name for error message
 *			- various command specific arguments
 *
 *	Returns:	- 0 or error code
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <terror.h>
#include <typical.h>
#include <debug.h>
#include <nail.h>

/*
 * external variables
 */
extern char		*me;			/* program name */
extern char		*t_hboot;		/* hboot tool */
extern char		*t_sboot;		/* sboot tool */
extern char		*t_vboot;		/* vboot tool */
extern char		*t_tkill;		/* tkill tool */
extern char		*t_tserv;		/* tserver tool */
extern int		errno;
extern int		fl_debug;		/* debug mode */
extern int		fl_frozen;		/* use frozen files */
extern int		fl_pretend;		/* pretend mode */
extern int		fl_recurse;		/* recursive booting */
extern int		fl_verbose;		/* verbose mode */

/* 
 * external functions
 */
extern void		execvp();
extern int		fork();
extern int		wait();

 
/*
 *	exec_tkill
 *
 *	Function:	- execute the tkill command
 *			- terminate Trollius on the host node
 *
 *	Returns:	- 0 or ERROR
 */
int
exec_tkill()
{
	int		n_cmd;
	char		**v_cmd;

	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_tkill);

	if (fl_debug) {
		printfv(n_cmd, v_cmd);
	}
	VERBOSE("tkill ...\n");

	if (! fl_pretend) {
		if (exec_tool(t_tkill, v_cmd)) {
			return(ERROR);
		}
	}
	return(0);
}
 
/*
 *	exec_sboot
 *
 *	Function:	- execute the sboot command
 *			- boot a device connected node
 *
 *	Accepts:	- nodeid to boot
 *			- device from which to boot
 *
 *	Returns:	- 0 or ERROR
 */
int
exec_sboot(nodeid, device)
 
int4			nodeid;
char			*device;
 
{
	int		n_cmd;
	char		**v_cmd;

	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_sboot);
	addarg(&n_cmd, &v_cmd, "-f");
	addarg(&n_cmd, &v_cmd, device);
	if (fl_verbose) {
		addarg(&n_cmd, &v_cmd, "-v");
	}
	if (fl_frozen) {
		addarg(&n_cmd, &v_cmd, "-z");
	}

	if (fl_debug) {
		printfv(n_cmd, v_cmd);
	}
	VERBOSE("sboot n%d ...\n", nodeid);

	if (! fl_pretend) {
		if (exec_tool(t_sboot, v_cmd)) {
			return(ERROR);
		}
	}
	return(0);
}
 
/*
 *	exec_hboot
 *
 *	Function:	- execute the hboot command
 *			- boot Trollius on the host node
 *
 *	Accepts:	- host nodeid
 *			- topology information
 *
 *	Returns:	- 0 or ERROR
 */
int
exec_hboot(hostid, topology)

int4			hostid;
char			*topology;

{
	int		n_cmd;
	char		**v_cmd;

	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_hboot);
	addarg(&n_cmd, &v_cmd, "-T");
	addarg(&n_cmd, &v_cmd, topology);
	if (fl_verbose) {
		addarg(&n_cmd, &v_cmd, "-v");
	}

	if (fl_debug) {
		printfv(n_cmd, v_cmd);
	}
	VERBOSE("hboot n%d ...\n", hostid);

	if (! fl_pretend) {
		if (exec_tool(t_hboot, v_cmd)) {
			return(ERROR);
		}
	}
	return(0);
}

/*
 *	exec_stage
 *
 *	Function:	- execute the vboot command
 *			- set up the staging nodes
 *
 *	Accepts:	- array of nodedis
 *			- # of nodes in array
 *			- host nodeid
 *			- device from which to boot
 *
 *	Returns:	- 0 or ERROR
 */
int
exec_stage(nodeids, nnodes, hostid, device)

int4			*nodeids;
int4			nnodes;
int4			hostid;
char			*device;

{
	int		n_cmd;
	int4		i;			/* favorite counter */
	char		**v_cmd;
	char		tmp_string[80];

	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_vboot);
	if (fl_verbose) {
		addarg(&n_cmd, &v_cmd, "-v");
	}
/*
 * Check the four cases of recursive and frozen staging.
 */
	addarg(&n_cmd, &v_cmd, "-ax");
	if (fl_frozen) {
		addarg(&n_cmd, &v_cmd, "-az");
		if (fl_recurse) {
			addarg(&n_cmd, &v_cmd, "-ez");
		}
	}
	else {
		addarg(&n_cmd, &v_cmd, "-al");
		if (fl_recurse) {
			addarg(&n_cmd, &v_cmd, "-el");
		}
	}

	for (i = 0; i < nnodes; i++) {
		sprintf(tmp_string, "n%#x", nodeids[i]);
		addarg(&n_cmd, &v_cmd, tmp_string);
	}
/*
 * Make lint happy because these variables are only used in PC-T.
 */
	hostid = hostid;
	device = device;

	if (fl_debug) {
		printfv(n_cmd, v_cmd);
	}
	VERBOSE("vboot");
	for (i = 0; i < nnodes; i++) {
		VERBOSE(" n%d", nodeids[i]);
	}
	VERBOSE(" ...\n");
	
	if (! fl_pretend) {
		if (exec_tool(t_vboot, v_cmd)) {
			return(ERROR);
		}
	}
	return(0);
}

/*
 *	exec_vboot
 *
 *	Function:	- execute the vboot command
 *			- boot a target node from a staging node
 *
 *	Accepts:	- staging nodeid
 *			- target nodeid
 *			- host nodeid
 *			- output link
 *			- device from which to boot
 *
 *	Returns:	- 0 or ERROR
 */
int
exec_vboot(stageid, targetid, hostid, outlink, device)

int4			stageid,
			targetid,
			hostid,
			outlink;
char			*device;

{
	int		n_cmd;
	char		**v_cmd;
	char		tmp_string[80];

	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_vboot);
	if (fl_verbose) {
		addarg(&n_cmd, &v_cmd, "-v");
	}
	addarg(&n_cmd, &v_cmd, "-bx");
	addarg(&n_cmd, &v_cmd, "-sx");
	if (fl_frozen) {
		addarg(&n_cmd, &v_cmd, "-bz");
		addarg(&n_cmd, &v_cmd, "-sz");
	}
	else {
		addarg(&n_cmd, &v_cmd, "-bl");
		addarg(&n_cmd, &v_cmd, "-sl");
	}
/*
 * Add the target nodeid to the command.
 */
	addarg(&n_cmd, &v_cmd, "-n");
	sprintf(tmp_string, "%#x", targetid);
	addarg(&n_cmd, &v_cmd, tmp_string);
/*
 * Add the output link number to the command.
 */
	addarg(&n_cmd, &v_cmd, "-L");
	sprintf(tmp_string, "%d", outlink);
	addarg(&n_cmd, &v_cmd, tmp_string);
/*
 * Add the staging nodeid to the command.
 */
	sprintf(tmp_string, "n%#x", stageid);
	addarg(&n_cmd, &v_cmd, tmp_string);
/*
 * Make lint happy because these variables are only used in PC-T.
 */
	hostid = hostid;
	device = device;

	if (fl_debug) {
		printfv(n_cmd, v_cmd);
	}
	VERBOSE("vboot n%d ...\n", targetid);

	if (! fl_pretend) {
		if (exec_tool(t_vboot, v_cmd)) {
			return(ERROR);
		}
	}

	return(0);
}

/*
 *	exec_tserv
 *
 *	Function:	- execute the tserver command
 *			- this is a stub in Unix(tm)
 *
 *	Accepts:	- host nodeid
 *			- number of itb nodes
 *			- device from which to boot
 *
 *	Returns:	- 0 or ERROR
 */
int
exec_tserv(hostid, device, nitb)

int4			hostid,
			nitb;
char			*device;

{
/*
 * Make lint happy, use all variables.
 */
	hostid = hostid;
	nitb = nitb;
	device = device;
	return(0);
}

/*
 *	exec_tool
 *
 *	Function:	- execute the given command line
 *			- fork a child process
 *			- wait for completion of child process
 *
 *	Accepts:	- command name
 *			- command line (argv format)
 *
 *	Returns:	- 0 or child error code
 */
int
exec_tool(tool, command)

char 			*tool;
char			**command;

{
	union wait	status;		/* status returned after forking */
	int pid;			/* used when forking */

	if ((pid = fork()) < 0) {	/* error */
		perror(me);
		exit(errno);
	}
	else if (pid == 0) {		/* child */
		execvp(tool, command);
		perror(tool);
		exit(errno);
	}
	else {				/* parent */
		while ((wait(&status) != 0) && (! WIFEXITED(status)));
		DBUG("%s: status: %d\n", tool, (int)status.w_retcode);
	}
	return((int)status.w_retcode);
}
